/*****************************************************************************
F s e a r c h
==============================================================================
fsearch.c                                                    Complete C Source
==============================================================================

This C source file contains the complete program source of the project.

==============================================================================

This UNIX/LINUX console application sends a FSEARCH command to a WIBU-BOX/ST
and reads the response from the BOX.

==============================================================================

This program is written in ANSI C.

==============================================================================

Copyright  2000-2005 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 54, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 1.00 of 00Aug14
Project .. Implementation Kit BOX ST/fsearch
Author ... Rainer Opitz (ro), Markus Neumann (mn), (ha)
System ... UNIX/LINUX

Change Report
00Aug18 1.00 ro: initial release
01Sep14 3.13 mn: bugfix: LED-Box is blinking 5times (Version Jump)
01Sep14 3.10 mn: Some honor to other OSs: Sun; Thanks to Bruce Fitzsimons

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>   // UNIX/Linux specific file desciptor manipulation
#include <sys/io.h>
#include <termios.h> // UNIX/Linux specific i/o parameter settings (struct termios)
#include <unistd.h>  // contains usleep() function for timeout in RecvDirect()

#include <string.h>

#ifndef CRTSCTS
  #define CRTSCTS 0 /* for AIX 5 */
#endif

// ******* timeout constants while accessing the serial port - OS-DEPENDENT used with the timing function usleep()
#define NUMBER_OF_READS 30000 // times max. try to read from box:RecvDirect().

// Additional Changes are possible and or nessecary at points that marked with

/*** I/O-ADAPTATION { ***/
//!Ch:mn CHANGE baudrate ONLY here: B4800, B9600, B19200 [B38400 or B57600]
#define WK_BAUDRATE B9600
/*** I/O-ADAPTATION } ***/

// special commands of the Box /ST (send directly without uu-encoding)
#define CMDST_TERMINATE     0x23     // BOX /ST-Command '#'

#ifndef FALSE
  #define FALSE               0
#endif

#ifndef TRUE
  #define TRUE                1
#endif

// Offset for uu-decoding
#define OFFSET              0x30

#define FSEARCH             "\"\"'40%"

// function prototypes
void PrintUsage (char*);
void UUStatusDecode (unsigned char*, unsigned char*);
void PrintSerialNumber (unsigned char*);
void PrintCopyright (char*);

/*****************************************************************************
m a i n        
==============================================================================

This is the main function. It checks the number of arguments, opens the port
and sends the FSEARCH command to the specified port. All received bytes are
echoed and the serial number and the status are decoded.

Parameters
  iNumberOfArg  number of command line arguments including the program call
  apcArguments  command line arguments

Return
  The function returns 0 if successful, otherwise 1.

*****************************************************************************/

int main (int iNumberOfArg, char *apcArguments[])
{
  char cByte = 0;                            // Container for the r/w-Bytes
  int  iNumberRead = 0;                      // 
  int  iStatusByte = 0;                      // Counter of received Bytes
  unsigned char aucStatusUU[10];             // Received String
  unsigned char aucFsearchCommand[]=FSEARCH; // Command String

  int            iHandlePort = 0;
  int            cCount = 0;                 // Breakout counter
  struct termios tioNew;
  struct termios tioOld;

  // print version number and copyright
  PrintCopyright(apcArguments[0]);

  // check number of arguments
  if (iNumberOfArg > 2) {
    fprintf(stdout, "Too many arguments.\n");
    PrintUsage(apcArguments[0]);
    return (1);
  }	// if

  if (iNumberOfArg < 2) {
    fprintf(stdout, "No port specified.\n");
    PrintUsage(apcArguments[0]);
    return (1);
  }	// if

  // iNumberOfArg == 2

  /*
   * Hard coded AccessCOM()
   */

  /*** I/O-ADAPTATION { ***/
  // open port for reading and writing
  iHandlePort = open(apcArguments[1],  O_RDWR | O_NOCTTY | O_NDELAY);
  if (iHandlePort == -1) {
    fprintf(stdout, "Can't open port %s.\n", apcArguments[1]);
    PrintUsage(apcArguments[0]);
    return (1);
  }	// if

  // set port parameters (9600bps, 8 databits, 2 stopbits, no parity)
  tcgetattr (iHandlePort, &tioOld);
  #if !defined(sun) //!N:mn
    memset (&tioNew, 0, sizeof(tioNew));
  #else //!N:mn
    tcgetattr (iHandlePort, &tioNew);//!N:mn
  #endif //!N:mn

  // Set raw in- & output flow control, no parity, 8 databits, 2 stopbits.
  // The following settings are the same as cfmakeraw (&tioNew) on
  // LINUX/UNIX systems.
  tioNew.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | IXON);
  tioNew.c_oflag &= ~OPOST;
  tioNew.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
  tioNew.c_cflag &= ~(CSIZE | PARENB);  // **** no parity checking
  tioNew.c_cflag |= CS8;  // **** 8 databits

  // These settings are some complements for the WIBU-BOX/ST.
  tioNew.c_cflag &= ~CRTSCTS;
  tioNew.c_cflag |= CLOCAL;
  tioNew.c_cflag |= CREAD;
  tioNew.c_cflag |= CSTOPB;  // **** 2 stopbits

  // set the baudrate for in- & output
  cfsetospeed (&tioNew, WK_BAUDRATE);
  cfsetispeed (&tioNew, WK_BAUDRATE);

  // set now all settings to the port immediately
  tcsetattr (iHandlePort, TCSANOW, &tioNew);

  /*** I/O-ADAPTATION } ***/

  iNumberRead = 1;

  /*** I/O-ADAPTATION { ***/
  // send FSEARCH to the BOX/ST
  write(iHandlePort, aucFsearchCommand, 6);
  /*** I/O-ADAPTATION } ***/
    
  // wait for the reply, send the echo of every received byte
  cByte = 0;
  iStatusByte = 0;
  while (cByte != CMDST_TERMINATE) {
    // read one byte from the box
    // simulate timeout
    iNumberRead = 0;
    cByte = 1;
    // delay until one byte is receceived
    cCount = 0;
    do {

      /*** I/O-ADAPTATION { ***/
      // now, receive one byte
        iNumberRead = (int) read(iHandlePort, &cByte, 1);//:mn
      /*** I/O-ADAPTATION } ***/

      cCount++;
    } while ((iNumberRead <= 0 || cByte == 0) && (cCount < NUMBER_OF_READS));

    // no byte read --> timeout, box doesn't reply
    if (iNumberRead == 0) {
        break;
    } // if

    if (iStatusByte == 0) {
      // BOX/ST sent "acknowledged" --> wait until the user hits a key
      fprintf(stdout, "Press ENTER to continue ...\n");
      getchar();
    } // if

    /*** I/O-ADAPTATION { ***/
    // send the echo
    write(iHandlePort, &cByte, 1);
    /*** I/O-ADAPTATION } ***/

    if (iStatusByte == 10) {
        fprintf(stdout, "Error: Too many bytes received from WIBU-BOX.\n");
        exit(1);
    } // if
    aucStatusUU[iStatusByte] = cByte;
    iStatusByte++;
  } // while

  if (iNumberRead == 0) {
      fprintf(stdout, "WIBU-BOX doesn't reply.\n\n");
  } else {
      PrintSerialNumber (aucStatusUU);
  } // if

  /*
   * Hard coded ReleaseCOM()
   */
  // reset the port settings of the chosen serial port, wait until all chars
  // reached; then change mode !

  /*** I/O-ADAPTATION { ***/
  //!Ch:mn tcsetattr (iHandlePort, TCSANOW, &tioOld);
  tcsetattr (iHandlePort, TCSADRAIN, &tioOld);//!N,E:mn LED-box blinks 5times

  // close file and port
  close(iHandlePort);
  /*** I/O-ADAPTATION } ***/

  return (0);

} // main()


/*****************************************************************************
P r i n t U s a g e
==============================================================================

Displays the correct program call with parameter list.

Parameters
  pcFilename  Filename of the exe-file

*****************************************************************************/

void PrintUsage (char *pcFilename)
{
  fprintf(stdout, "Use: %s <portspec>\n", pcFilename);
  fprintf(stdout, "   <portspec>   specification of the port where the WIBU-BOX/ST is connected\n");
  fprintf(stdout, "                i.e.  %s /dev/ttyS0\n", pcFilename);

} // PrintUsage()


/*****************************************************************************
U U S t a t u s D e c o d e
==============================================================================

This function decodes the uu-encoded status bytes.

Parameters
  pucInput   array of 8 uu-encoded status bytes
  pucOutput  array of 6 decoded status bytes

*****************************************************************************/

void UUStatusDecode (unsigned char *pucInput, unsigned char *pucOutput)
{
  // convert uu-code into bytes (4 uu-coded bytes -> 3 bytes)
  pucOutput[0] = ((pucInput[0] - OFFSET) & 0x3F)
               + (((pucInput[1] - OFFSET) & 0x03) << 6);
  pucOutput[1] = (((pucInput[1] - OFFSET) & 0x3C) >> 2)
               + (((pucInput[2] - OFFSET) & 0x0F) << 4);
  pucOutput[2] = (((pucInput[2] - OFFSET) & 0x30) >> 4)
               + (((pucInput[3] - OFFSET) & 0x3F) << 2);
  pucOutput[3] = ((pucInput[4] - OFFSET) & 0x3F)
               + (((pucInput[5] - OFFSET) & 0x03) << 6);
  pucOutput[4] = (((pucInput[5] - OFFSET) & 0x3C) >> 2)
               + (((pucInput[6] - OFFSET) & 0x0F) << 4);
  pucOutput[5] = (((pucInput[6] - OFFSET) & 0x30) >> 4)
               + (((pucInput[7] - OFFSET) & 0x3F) << 2);
} // UUStatusDecode()


/*****************************************************************************
P r i n t S e r i a l N u m b e r
==============================================================================

This function displays the serial number which the BOX sent.

Parameters
  pucStatusUU  10 bytes, received from the BOX

*****************************************************************************/

void PrintSerialNumber (unsigned char *pucStatusUU)
{
  unsigned char aucStatusByte[6];
  unsigned long ulSerialNumber;

  // pucStatusUU[0] == 0x21 (acknowledge)
  // pucStatusUU[1-8] == serial number + status, uu-encoded
  // pucStatusUU[9] == 0x23 (terminate)
  UUStatusDecode(pucStatusUU + 1, aucStatusByte);

  ulSerialNumber = ((unsigned long) aucStatusByte[0]) +
                   ((unsigned long) aucStatusByte[1] << 8) +
                   ((unsigned long) aucStatusByte[2] << 16) +
                   ((unsigned long) aucStatusByte[3] << 24);

  fprintf(stdout, "Serial Number: %d-%ld\n", aucStatusByte[4], ulSerialNumber);

  if (aucStatusByte[5]) {
    fprintf(stdout, "No further WIBU-BOX follows.\n\n");
  } else {
    fprintf(stdout, "A further WIBU-BOX follows and can be read by another NSEARCH command.\n\n");
  } // if

} // PrintSerialNumber()


/*****************************************************************************
P r i n t C o p y r i g h t
==============================================================================

This function displays the version and the copyright.

Parameters
  pcFilename  Filename of the exe-file

*****************************************************************************/

void PrintCopyright (char *pcFilename)
{
  fprintf(stdout, "%s - WIBU-KEY Test Tool For WIBU-BOX/ST.\n", pcFilename);
  fprintf(stdout, "Version 1.00 of 00-Aug-18 for UNIX/LINUX.\n");
  fprintf(stdout, "Copyright (C) 2000 by WIBU-SYSTEMS AG. All rights reserved.\n\n");
} // PrintCopyright()

// ================
// end of fsearch.c
// ================
