/**
 * @mainpage
 *
 * @author   Heiko Abraham (ha)
 * @date     2008 Sep 25
 * @version  5.0
 *
 * wkloop is a simple test program for WIBU-KEY and 
 * demonstrates the usage of the WIBU-KEY API.
 * The test will access to a WIBU-BOX entry 10:13 and make 
 * a infinity loop with wait and encrypt/decrypt a random 
 * string by using WkbCrypt2().
 * 
 * All sources are written in ANSI C.
 * 
 * ---
 * 
 * Copyright  2000 - 2008 by WIBU-SYSTEMS AG
 * 
 *          Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
 *          
 *          Website: http://www.wibu.com, Email: info@wibu.com
 *          
 *          Phone: +49-721-93172-0, Fax: +49-721-93172-22
 *          
 *          All rights reserved.
 */



#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include <wk.h>

#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>




#ifndef VERSION
  #define VERSION          4    //!< major version number that required by this test
#endif

#ifndef SUBVERSION
  #define SUBVERSION       0    //!< minor version number that required by this test
#endif

#define DRV_VERSION        2    //!< major driver version number (required)
#define DRV_SUBVERSION    31    //!< minor driver version number (required)

#define FIRM_CODE         10    //!< the FirmCode that used by this test 
#define USER_CODE         13    //!< the UserCode that used by this test

#define DEBUG 1                 //!< DEBUG=1 if debug output to stdout is required

#define CB_MAX_DATA 60          //!< max data length for WkbCrypt2() 

// structure that holds all important WIBU-KEY informations
typedef struct {
  LONG lFirmCode;       //!< FirmCode to use
  LONG lUserCode;       //!< UserCode to use
  ULONG ulSelection;    //!< SelectionCode to use
  WKBSELECT wkbsel;     //!< Selection informations
  HWKBENTRY hwkbe;      //!< handle of opened WIBU-BOX entry
  HWKBSYSTEM hwkbsys;   //!< handle of used subsystem
  INT nVersion;         //!< Version number as integer
  char szVersion[10];  //!< Version number as string
  BOOL fSysDrv;         //!< flag "using kernel driver"
  BOOL fWkDos;          //!< flag "using resident WkDos driver"
  BOOL fCheckRunning;   //!< flag "CheckBox is running"
} WIBUKEYDATA;


WIBUKEYDATA wibudata;     //!< global variables to hold all important WIBU-KEY informations
WIBUKEYDATA *pwk = NULL;  //!< pointer to wibudata

int nExit = 0;

/**
 * @brief A little helper function to return user defined error messages.
 *
 * GetWkbError return to a given error code the corresponending 
 * error text as a const string. 
 *
 * @warning WkbGetLastError() ist not Thread safe in WkRt-Lin version =< 4.10.
 */
const char *GetWkbError (int err) {
  if (err == WKB_NO_ERROR)
    return "No error occurred";
  if (err == WKB_ERROR_INVALID_HANDLE)
    return "An invalid handle was specified";
  if (err == WKB_ERROR_NO_MORE_HANDLES)
    return "No more handles are available";
  if (err == WKB_ERROR_INVALID_PORT)
    return "The specified port is invalid";
  if (err == WKB_ERROR_INVALID_PARAMETER)
    return "The specified parameter is invalid";
  if (err == WKB_ERROR_BUFFER_OVERFLOW)
    return "The data area passed to the WIBU-KEY driver is too small";
  if (err == WKB_ERROR_LOCK_VIOLATION)
    return "A function of the WIBU-KEY driver is locked by another driver call";
  if (err == WKB_ERROR_NOT_READY)
    return "The accessed WIBU-BOX is not ready (timeout)";
  if (err == WKB_ERROR_NOT_CONNECTED)
    return "No WIBU-BOX is connected at the specified port";
  if (err == WKB_ERROR_INVALID_FUNCTION)
    return "A not existing WIBU-KEY API function is called";
  if (err == WKB_ERROR_NOT_CRYPTIC)
    return "The calling program is not encrypted by WKCRYPT.EXE";
  if (err == WKB_ERROR_OPTION_NOT_SUPPORTED)
    return "The specified option is not supported by WIBU-KEY driver function";
  if (err == WKB_ERROR_MEMORY_TOO_SMALL)
    return "There is not enough memory available in WIBU-KEY driver";
  if (err == WKB_ERROR_INVALID_VERSION)
    return "The used WIBU-BOX has an invalid version";
  if (err == WKB_ERROR_INVALID_LEVEL)
    return "A WIBU-KEY driver function is called in a non-supported level";
  if (err == WKB_ERROR_CRC_VERIFY)
    return "An error occurred in the WIBU-BOX data communication (CRC)";
  if (err == WKB_ERROR_NOT_SELECTED)
    return "The accessed WIBU-BOX is not selected for encryption";
  if (err == WKB_ERROR_ENTRY_NOT_FOUND)
    return "A suitable WIBU-BOX entry was not found";
  if (err == WKB_ERROR_SERIAL_NOT_FOUND)
    return "A WIBU-BOX with the specified Serial Number was not found";
  if (err == WKB_ERROR_ENTRY_NOT_CHANGED)
    return "The WIBU-BOX entry was not changed by the programming operation";
  if (err == WKB_ERROR_INVALID_ENTRY)
    return "The specified WIBU-BOX entry is invalid";
  if (err == WKB_ERROR_BAD_SIGNATURE)
    return "The specified WIBU-BOX programming signature is wrong";
  if (err == WKB_ERROR_POWER_TOO_SMALL)
    return "The supplied power is too small for WIBU-BOX programming task";
  if (err == WKB_ERROR_ALREADY_SELECTED)
    return "The used WIBU-BOX is already selected by another process";
  if (err == WKB_ERROR_BOXLIMIT_IS_ZERO)
    return "The Limit Counter of the used WIBU-BOX reached 0";
  if (err == WKB_ERROR_PROGRAM_VERIFY)
    return "The verification of a WIBU-BOX entry change failed";
  if (err == WKB_ERROR_PORT_IS_BUSY)
    return "An accessed WIBU-BOX port is already used by another process";
  if (err == WKB_ERROR_INVALID_SELECTION)
    return "The specified Selection Code is invalid";
  if (err == WKB_ERROR_ENTRY_ALREADY_USED)
    return "The used WIBU-BOX entry is already used by another instance";
  if (err == WKB_ERROR_OUT_OF_DATE)
    return "The used WIBU-BOX entry has an passed Expiration Date";
  if (err == WKB_ERROR_BAD_ADDRESS)
    return "A specified parameter points to an invalid address";
  if (err == WKB_ERROR_INVALID_COMMAND)
    return "The specified command code is not supported";
  if (err == WKB_ERROR_INVALID_SUBSYSTEM)
    return "The specified WIBU-KEY subsystem is not available";
  if (err == WKB_ERROR_DATA_TOO_COMPLEX)
    return "The specified data is too large for local operation";
  if (err == WKB_ERROR_EXCLUSIVE_MODE_CONFLICT)
    return "An exclusive access conflicts with another exclusive access";
  if (err == WKB_ERROR_NO_USER_SET)
    return "An identification of the user is not stored in the WIBU-KEY profile";
  if (err == WKB_ERROR_IN_ACTION)
    return "The specified WIBU-KEY driver action hat not completed yet";
  if (err == WKB_ERROR_UNDEFINED_ACTION)
    return "The specified WIBU-KEY driver action is undefined";
  if (err == WKB_ERROR_TOO_MANY_ACTIONS)
    return "There are too many WIBU-KEY driver action";
  if (err == WKB_ERROR_SUBSYSTEM_IN_ACTION)
    return "The WIBU-KEY subsystem is waiting for termination a concurrent action";
  if (err == WKB_ERROR_NO_NET_SYSTEM)
    return "The WIBU-KEY network system is not available";
  if (err == WKB_ERROR_NET_SEND_FAULT)
    return "A WIBU-KEY network communication was stopped with a sending fault";
  if (err == WKB_ERROR_NET_RECEIVE_FAULT)
    return "A WIBU-KEY network communication was stopped with a receiving fault";
  if (err == WKB_ERROR_NET_NOT_READY)
    return "The WIBU-KEY network is not ready";
  if (err == WKB_ERROR_NET_IS_BUSY)
    return "The WIBU-KEY network is busy";
  if (err == WKB_ERROR_BAD_NET_SYSTEM)
    return "The WIBU-KEY network is currupt";
  if (err == WKB_ERROR_NET_LIMIT_IS_ZERO)
    return "The user maximum of the WIBU-KEY network is reached";
  if (err == WKB_ERROR_NET_USER_CANCELLED)
    return "The current user was cancelled in the WIBU-KEY network";
  if (err == WKB_ERROR_NET_DATA_TOO_COMPLEX)
    return "The specified data is too large for a WIBU-KEY network transfer";
  if (err == WKB_ERROR_SERVER_NOT_FOUND)
    return "The specified server was not found in the WIBU-KEY network";
  if (err == WKB_ERROR_NO_KRNL_DRIVER)
    return "The WIBU-KEY kernel driver is not available";
  if (err == WKB_ERROR_KRNL_COMMUNICATION)
    return "An error occurred in the communication with the WIBU-KEY kernel driver";
  return "unknown error code";
}



/**
 * @brief Helper function to open a box and initialize all data structures
 *
 *  Initialize global structure for the WIBU-KEY.
 *
 * @param pwkdata  points to the global WIBU-KEY data
 *
 * @post pwkdata contain a vaild handle, if it was possible to open a box.
 *
 * @return  This function returns TRUE if all actions were completed. If any error
 *          occurs FALSE will be returned.
 */
BOOL OpenBox (WIBUKEYDATA * pwkdata) {
  ULONG ulLowVersion = 0;
  ULONG ulHighVersion = 0;
  ULONG flCtrl = 0;
  WKBSTATUS wkbst;

  pwkdata->ulSelection = 0UL;
  pwkdata->nVersion = 0;
  memset (&(pwkdata->wkbsel), 0, sizeof ((pwkdata->wkbsel)));
  
  if (wkbst.usFlags & WKB_ST_KRNL) {
    pwkdata->fSysDrv = TRUE;
  } else {
    pwkdata->fSysDrv = FALSE;
    if (wkbst.usFlags & WKB_ST_WKDOS) {
      pwkdata->fWkDos = TRUE;
    } else {
      pwkdata->fWkDos = FALSE;
    } // if
  } // if

  pwkdata->hwkbe = NULL;
  pwkdata->hwkbsys = NULL;
  pwkdata->fCheckRunning = FALSE;

  // search & open WIBU-BOX entry of accessed WIBU-KEY subsystem
  // NOTE: If you use WKB_VERSION2 it will not be possible to access
  // version 1 WIBU-BOXes (the silver ones)
  flCtrl = WKB_STDCTRL | WKB_OPEN_FIND | WKB_VERSION2;
  pwkdata->hwkbe = WkbOpen2 (HWKB_LOCAL,	// handle to local system
           flCtrl,	// control flags (use version 2, find entry)
           NULL,	// not used for net access
           pwkdata->lFirmCode,	// Firm Code
           pwkdata->lUserCode,	// User Code
           NULL	// no optional data
    );

  if (!pwkdata->hwkbe) {
    printf ("error(%i) : WkbOpen2() - %s\n", 
      WkbGetLastError (), GetWkbError (WkbGetLastError ()) );
    return FALSE;
  } // if
  return TRUE;
} // OpenBox()



/**
 * @brief Make a random encryption and decryption for a open box 
 *
 * @pre   pwkdata must contain a vaild handle to a WIBU-KEY subsystem
 *
 * Crypt a random string and compare the result. If result valid, then return 0.
 *
 * @param pwkdata points to the global WIBU-KEY data
 */
int WkhCrypt2 (WIBUKEYDATA * pwkdata) {
  char szTest[CB_MAX_DATA];
  char szCrypt[CB_MAX_DATA];
  char szDecrypt[CB_MAX_DATA];

  BOOL fValid = FALSE;
  UINT cbCrypt = 0;
  UINT cbData = 0;
  ULONG flCtrl = 0;
  ULONG ulSC = 0;
  UINT i = 0;
  WKBAREA wkbar;

  pwk = pwkdata;

  if (pwkdata->fCheckRunning) {
    #if DEBUG
      printf ("we already run a check\n");
    #else
      printf ("E");
    #endif
    return 0;
  } // if 
  memset (szTest, 0, CB_MAX_DATA);
  memset (szCrypt, 0, CB_MAX_DATA);
  memset (szDecrypt, 0, CB_MAX_DATA);

  // reopen WIBU-Box again, if not already open
  if (pwkdata->hwkbe == NULL) {
    #if DEBUG
      printf ("handle lost, reopen box\n");
    #else
      printf("E");
    #endif
    WkbClose2 (pwkdata->hwkbe);
    flCtrl = WKB_STDCTRL | WKB_OPEN_FIND | WKB_VERSION2;
    pwkdata->hwkbe = WkbOpen2 (HWKB_LOCAL,
             flCtrl,
             NULL,
             pwkdata->lFirmCode,
             pwkdata->lUserCode, NULL);
    if (!pwkdata->hwkbe) {
      #if DEBUG
        printf ("error by calling WkbOpen2, (%i - %s)\n",WkbGetLastError (), GetWkbError(WkbGetLastError ())  );
      #else
      printf("E");
      #endif
      return 1;
    } // if 
  } // if(reopen)

  // generate a random selection code
  // AND set counter (bit 31) to zero
  ulSC = (ULONG) random () & 0x0FFFF;

  // get a random test-stream
  for (i = 0; i < CB_MAX_DATA; i++) {
    szTest[i] = (char) (65 + (random () % 24));
  } // for 

  pwkdata->fCheckRunning = TRUE;

  wkbar.ulCrc = 0;
  wkbar.ulReserved = 0;
  wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
  wkbar.flSelCtrl = WKB_STDCTRL | WKB_SEL_DIRECT;
  wkbar.ulSelectCode = ulSC;
  wkbar.pbGapMap = NULL;
  wkbar.pvCtrl = NULL;
  cbData = CB_MAX_DATA;

  // crypt  test-stream
  memcpy (szCrypt, szTest, CB_MAX_DATA);
  fValid = WkbCrypt2 (pwkdata->hwkbe,
          WKB_STDCTRL | WKB_CRYPT_CTRL,
          szCrypt, &wkbar, cbData, &cbCrypt);
  if (!fValid) {
  #if DEBUG
    printf ("not valid crypt (err %i - %s) (handle %i)\n",
      WkbGetLastError (), GetWkbError(WkbGetLastError ()), (int) pwkdata->hwkbe);
  #else
    printf("E");
  #endif
    WkbClose2 (pwkdata->hwkbe);
    pwkdata->hwkbe = NULL;
    pwkdata->fCheckRunning = FALSE;
    return 2;
  } // if 

  // decrypt received test-stream
  memcpy (szDecrypt, szCrypt, CB_MAX_DATA);
  fValid = WkbCrypt2 (pwkdata->hwkbe,
          WKB_STDCTRL | WKB_CRYPT_CTRL,
          szDecrypt, &wkbar, cbData, &cbCrypt);
  if (!fValid) {
  #if DEBUG
    printf ("not valid decrypt (err %i - %s) (handle %i)\n",
      WkbGetLastError (), GetWkbError(WkbGetLastError ()), (int) pwkdata->hwkbe);
  #else
    printf("E");
  #endif
    WkbClose2 (pwkdata->hwkbe);
    pwkdata->hwkbe = NULL;
    pwkdata->fCheckRunning = FALSE;
    return 3;
  } // if
  pwkdata->fCheckRunning = FALSE;

  // check if both test-stream are equal
  for (i = 0; i < CB_MAX_DATA; i++) {
    if (szDecrypt[i] != szTest[i]) {
    #if DEBUG
      printf ("string not equal\n");
    #else
      printf("E");
    #endif
      return 1;
    }
  } // for
  return 0;
}

/** 
 * @brief Close access to WIBU-BOX.
 *
 * If pwkdata contain a valid handle to a WIBU-KEY sub system the this 
 * function will close the subsystem access. 
 *
 * @param pwkdata  points to the global WIBU-KEY data
 *
 * @return This function returns TRUE if all actions were completed. If any error
 *         occurs FALSE will be returned.
 *
 * @post Handle in pwkdata will be set to NULL.
 *
 * @warning This function handles no WkLAN access.
 *
 */
BOOL CloseBox (WIBUKEYDATA * pwkdata) {
  if (pwkdata->hwkbe != NULL) {
    // close opened WIBU-BOX entry
    if (WkbClose2 (pwkdata->hwkbe) == 0) {
      printf ("error(%i) : WkbClose2() - %s\n", 
        WkbGetLastError (),
        GetWkbError (WkbGetLastError ()) );
    } // if 
  } // if
  return TRUE;
} // CloseBox()



/**
 * @brief The function callback for SysV-signals.
 *
 * Catch system signals and make cleanup, the exit application.
 *
 */
void WkhCatcher (int i) {
  // make to be exit
  
  nExit = 1;
} // WkhCatcher


/**
 * @brief the program main function 
 *
 *  @return This function return '0' if the test was vaild, else '1'.
 */
int main (int argc, char *argv[]) {
  char *appname = argv[0];
  int r = 0;
  int i = 0;
  int nOpt = 0;
  
  int nFirmCode = FIRM_CODE;
  int nUserCode = USER_CODE;
  
  // setup sysv-signals
  signal (SIGINT, WkhCatcher);
  signal (SIGTERM, WkhCatcher);

  // output some informations 
  printf ("Simple Demonstration of the WIBU-KEY API under LINUX/UNIX with C.\n");
  printf ("%s - Simple software protection.\n", appname);
  printf ("Copyright (c) 1996-2008 WIBU-SYSTEMS AG\n\n" );
  
  // get command line options
  while((nOpt=getopt(argc, argv, "u:f:h")) != -1 )
    switch(nOpt) {
      case 'u':
        // User-Code
        nUserCode = atoi(optarg);
        break;
      case 'f':
        // Firm-Code
        nFirmCode = atoi(optarg);
        break;
      case 'h':
      default:
        printf("usage:\n");
        printf("  %s [-f <firmcode>] [-u <usercode>] [-h]\n", argv[0]);
        return 0;
    } // switch
  
  printf ("Use FirmCode=%i UserCode=%i\n", nFirmCode, nUserCode);
  
  wibudata.lFirmCode = nFirmCode;
  wibudata.lUserCode = nUserCode;
  
  // initialize <wibudata>-structure, check version, etc.
  if (!OpenBox (&wibudata)) {
    printf ("No access to WIBU-BOX!\n");
    return (1);
  } // if

  // display access & version information
  if (wibudata.fSysDrv) {
    printf ("Access to WIBU-BOXes via kernel driver.\n");
  } // if

  if (wibudata.fWkDos) {
    printf ("Access to WIBU-BOXes via resident WkDos driver.\n");
  } // if

  printf ("output:\n .  wait\n c  make a encryption/decryption\n E  error received\n\n");
  printf ("press ctrl+c for exit\n");


  while ( nExit == 0) {
    r = (int) random () % 10;
    for (i = 0; i < (r + 1); i++) {
      usleep (100 * 1000);
      fprintf (stdout, ".");
      fflush (stdout);
    } // for

    fprintf (stdout, "c");
    fflush (stdout);

    WkhCrypt2 (&wibudata);
  } // while
  return 0;
} // main
