/**
 * @mainpage
 *
 * @author   Heiko Abraham (ha)
 * @date     2005 Feb 02
 * @version  5.0
 *
 * wkscan is a simple test program for WIBU-KEY and 
 * demonstrates the usage of the WIBU-KEY API.
 * This test will scan for all local connected WIBU-BOXes
 * and list the entry of each WIBU-BOX.
 *
 * All sources are written in ANSI C.
 *
 * ---
 *
 * Copyright  2000 - 2005 by WIBU-SYSTEMS AG
 * 
 * Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
 *          
 * Website: http://www.wibu.com, Email: info@wibu.com
 *          
 * Phone: +49-721-93172-0, Fax: +49-721-93172-22
 *          
 * All rights reserved.
 */




#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include "wk.h"





#define CB_MAX_PORTNAME 32



// structure that holds all important WIBU-KEY informations
typedef struct {
  LONG lFirmCode;             //!< FirmCode to use
  LONG lUserCode;             //!< UserCode to use
  WKBSELECT wkbsel;           //!< Selection informations
  HWKBENTRY hwkbe;            //!< handle of opened WIBU-BOX entry
  HWKBSYSTEM hwkbsys;         //!< handle of used subsystem
  INT nVersion;               //!< Version number as integer
} WIBUKEYDATA;

WIBUKEYDATA wibudata;         //!< global variables to hold all important WIBU-KEY informations
WIBUKEYDATA *pwkdata = NULL;  //!< pointer to wibudata

USHORT flCtrl = 0;
WKBSTATUS wkbst;
INT idError = 0;



/**
 * @brief The function callback for SysV-signals.
 *
 * Catch system signals and make cleanup, the exit application.
 *
 */
void WkhCatcher (int i) {
  // add clean up code here 
  // i.e. WkbClose2()

  // display massage 
  printf ("(II) catch crtl+c, exit application\n");

  // exit application 
  exit (1);
} // WkhCatcher 



/**
 * @brief init global pwkdata
 *
 * Initialize global structure for the WIBU-KEY.
 * 
 * @post  pwkdata (global var) points to the global WIBU-KEY data
 * 
 * @return This function returns '0' if all actions were completed and
 *         otherwise greater.
 *
 */
int WkhInit () {
  ULONG ulVersion = 0, ulSubVersion = 0;
  USHORT usMinDrvVersion = 0;
  char szVerBuf[200];

  // init data-struc
  pwkdata = &wibudata;
  memset (pwkdata, 0, sizeof (WIBUKEYDATA));

  // determe the API version
  pwkdata->nVersion = WkbGetVersion3 (HWKB_LOCAL);

  // calculate the min version
  ulVersion = pwkdata->nVersion / 256;
  ulSubVersion = pwkdata->nVersion % 256;

  printf("Local driver version:\n");
  printf ("  WkbVersion3 %lu.%s%lu\n",
    ulVersion, (ulSubVersion < 10) ? "0" : "", ulSubVersion);
  
  if (0 == WkbQueryInfo2(HWKB_LOCAL, WKB_QI_VERTEXT, szVerBuf, sizeof(szVerBuf)  )) {
    printf ("Error(%i) - WkbQueryInfo2()\n", WkbGetLastError() );
    return 1;
  } else {
    printf ("  WkbQueryInfo2() %s (local driver)\n", szVerBuf );
  } // if

  return 0;
}





/**
 * @brief list serial number and box contents of SRL1 local WIBU-BOX.
 *
 * Result a list of the entries of local connected WIBU-BOXes.
 *
 * @pre a valid init pwkdata-structure
 *
 * @return Function return '1' if the list operation was successful else '0'.
 *
 */
int ListBoxes () {

  typedef struct {
    // room for 5 serial number entries
    WKBBOXLIST wkbbxl;	// first serial number
    WKBSERIAL awkbxe[5];	// ten other serial numbers
  } PORTLIST;

  PORTLIST prtlist;
  WKBBOXDATA wkbbxd;
  BOOL fValid = FALSE;
  INT nMaxBox = 0;
  INT i = 0;
  INT j = 0;
  UINT uType = 0;
  INT nEntry = 0;
  char szPortName[CB_MAX_PORTNAME];
  INT nReturn = 0;
  UINT nDay = 0, nMonth = 0, nYear = 0;
  printf ("Analysing WIBU-KEY ports ...\n");

  szPortName[0] = 0; // start of listing
  memset(szPortName, 0x0, sizeof(szPortName) );
  memset(&prtlist, 0x0, sizeof(prtlist) );

    
    nMaxBox = 1;
    // list WIBU-BOXes at this port
    fValid = WkbListPort2 (HWKB_LOCAL, WKB_STDCTRL,
               /* szPortName */ "SRL1" , &prtlist,
               sizeof (prtlist));
    if (fValid && (prtlist.wkbbxl.cBox > 0) ) {
      // one or more WIBU-BOXes found at szPortName: list first WIBU-BOX
      printf ("\n    %i WIBU-BOX(es) found at %s: ",
        prtlist.wkbbxl.cBox, szPortName);

      // list serial numbers at this port
      nMaxBox = prtlist.wkbbxl.cBox - 1;
      if (nMaxBox > 10) {
        nMaxBox = 10;
      } // if 

      for (i = 0; i <= nMaxBox; i++) {
        printf ("\n    Contents of WIBU-BOX #%i (%u-%08lu):\r\n", i + 1, prtlist.wkbbxl.awkbsr[i].usMask, prtlist.wkbbxl.awkbsr[i].ulSerial);
        fValid = WkbListBox2 (HWKB_LOCAL,	// HWKB_LOCAL, 
                  WKB_STDCTRL,
                  /* szPortName */ "SRL1",
                  prtlist.wkbbxl.awkbsr +
                  i, &wkbbxd,
                  sizeof (wkbbxd));
        if (!fValid)  {
          printf ("(EE) error : WkbListBox2 (err %i)\n", WkbGetLastError ());
          return FALSE;
        }


        // list contents of WIBU-BOX
        for (j = 0; j < (INT) wkbbxd.cEntry; j++) {
          WKBBOXENTRY *pwkbbxe = wkbbxd.awkbbxe + j;
          uType = (UINT) pwkbbxe->flCtrl & WKB_BE_TYPEMASK;
          nEntry = (INT) pwkbbxe->flCtrl & WKB_BE_INDEXMASK;
          switch (uType) {
            case WKB_BE_FCUC:
              printf ("%2d: FC/UC:  %8lu:%-8lu", nEntry, pwkbbxe->ulFirmCode, pwkbbxe->ulUserCode);
              break;
            case WKB_BE_MASTER:
              printf ("%2d: Master: %8lu:%-8lu", nEntry, pwkbbxe->ulFirmCode, pwkbbxe->ulUserCode);
              break;
            case WKB_BE_DATA:
              printf ("%2d: User-D: %8lu:%-lu", nEntry, pwkbbxe->ulData2, pwkbbxe->ulData1);
            default:
              break;
          }	// switch


          if (pwkbbxe->flCtrl & WKB_BE_ADDDATA)
            printf (" Added Data: %lu-%lu", pwkbbxe->ulData2, pwkbbxe->ulData1);
          else if (pwkbbxe->
             flCtrl & WKB_BE_ADDLIMIT)
            printf (" Added Limit Counter: %lu", pwkbbxe->ulData1);
          else if (pwkbbxe->
             flCtrl & WKB_BE_ADDEXPDATE)
          {
            nDay = (UINT) pwkbbxe->ulData1 & 0xff;
            nMonth = (UINT) (pwkbbxe->ulData1 & 0xff00) / 0x100;
            nYear = (UINT) ((pwkbbxe->ulData1 & 0xfff0000) / 0x10000);
            if (pwkbbxe->ulData2 == 0xffff)
            {
              printf (" Expiration Date: %2d.%2d.%4d (is expired)", nDay, nMonth, nYear);
            }
            else
            {
              printf (" Expiration Date: %2d.%2d.%4d (%i days left)", nDay, nMonth, nYear, (int) pwkbbxe->ulData2);
            }	// if
          }	// if
          printf ("\r\n");
        }	// for
      }	// for(i...nMaxBox)
    }		// if
  return 1;
}				// ListBoxes()



/**
 * @brief the program main function 
 *
 * This is the main program function, it will init the driver and list all 
 * detected local WIBU-BOXes. 
 *
 * @return This function return '0'.
 *
 * @warning Only WIBU-BOXes will be list, that ports a vaild. I.e. the enviroment
 *          WKCOM and WKLPT define valid ports.
 */
int main (int argc, char *argv[]) {

  // setup sysv-signals
  signal (SIGINT, WkhCatcher);

  // init DRVC and get Drv-Version
  if ( WkhInit () == 0) {
    // scan local ports and list content 
    ListBoxes ();
  } // if

  return 0;
} // main
