/**
 * @mainpage
 *
 * @author   Heiko Abraham (ha)
 * @date     2010 Jun 20
 * @version  5.0
 *
 * wkscan is a simple test program for WibuKey and 
 * demonstrates the usage of the WibuKey API.
 * This test will scan for all local connected WibuBoxes
 * and list the entry of each WibuBox.
 *
 * All sources are written in ANSI C.
 *
 * ---
 *
 * Copyright  2000 - 2010 by WIBU-SYSTEMS AG
 * 
 * Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
 *          
 * Website: http://www.wibu.com, Email: info@wibu.com
 *          
 * Phone: +49-721-93172-0, Fax: +49-721-93172-22
 *          
 * All rights reserved.
 */


#include "wktest.h"

#define INDCRYPT 1 
#define LANACC   1

INT nNoError = 0;
INT nNoTests = 0;


/**
 * @brief internal 
 */
static void PrintHex(char *szBuf, int cbLen) 
{
  int k = 0;
  for (k = 0;k<cbLen; k++) {
    printf("0x%02x ", szBuf[k] & 0xFF);
  }
  printf("\n");
}


/**
 * @brief Print test result message
 *
 * If @a nStatus is not Zero, than we assume an error is occur and 
 * we print an error message 
 */
static void PrintMessage(int nStatus, char *szTest, char *szMessage)
{
  if ( (nStatus != 0) && (nStatus != -1) ) {
    nNoError++;
  }
  
  if (nStatus != -1) {
    nNoTests++;
  }
  
  printf("%s %s %s\n",
    (nStatus == 0) ? "***" : 
    (nStatus == -1) ? "   " : "!!!",
    (NULL == szTest) ? "(null)" : szTest,
    (NULL == szMessage) ? "" : szMessage );
}


/**
 * @brief internal 
 */
static void TestWkbVersion(void)
{
  INT nRet = 0;
  INT nMajor = 0;
  INT nMinor = 0;
  char szVersion[16];
  
  szVersion[0] = 0;
  
  ULONG ulVersion = 0, ulSubVersion = 0;
  USHORT usMinDrvVersion = 0;

  // check WkbVersion2
  nRet = WkbGetVersion2 (HWKB_LOCAL);
  if ( nRet > 0 ) {
    nMajor = nRet & 0xFF;
    nMinor = (nRet & 0xFF00) / 0xFF;
    snprintf(szVersion, sizeof(szVersion), "%i.%02i", 
      nMajor, nMinor);
    szVersion[sizeof(szVersion)-1] = 0;
  }
  PrintMessage( ((nRet == 0) || (nMajor != 2)) ? 1 : 0,
    "WkbGetVersion2()", szVersion);


  // check WkbVersion3
  nRet = WkbGetVersion3(HWKB_LOCAL);
  if ( nRet > 0 ) {
    nMajor = (nRet & 0xFF00) / 0xFF;
    nMinor = nRet & 0xFF;
    snprintf(szVersion, sizeof(szVersion), "%i.%02i", 
      nMajor, nMinor);
    szVersion[sizeof(szVersion)-1] = 0;
  }
  
  PrintMessage( ((nRet == 0) || (nMajor < 3)) ? 1 : 0,
    "WkbGetVersion3()", szVersion);
  
  WKBVERSION wkver[2];
  nRet = WkbQueryInfo2(HWKB_LOCAL, WKB_QI_VERVALUE | WKB_QI_DRIVER, &wkver, sizeof(wkver));
  snprintf(szVersion, sizeof(szVersion), "%i.%02i.%03i.%03i", 
    wkver[0].usVersion, wkver[0].usSubVersion,
    wkver[0].usBuild, wkver[0].usCount);
  szVersion[sizeof(szVersion)-1] = 0;
  
  PrintMessage( !((nRet == 16) && (wkver[0].usVersion == nMajor) && (wkver[0].usSubVersion == nMinor)),
    "WkbQueryInfo2(WKB_QI_VERVALUE)", szVersion);
}


/**
 * @brief internal 
 */
static void TestErrorCode(void)
{
  INT nRet = 0;
  int i = 0;
  int nSuc = 0;
  
  nRet = WkbGetLastError();
  if ( (nRet < WKB_NO_ERROR) || (nRet > WKB_ERROR_KRNL_COMMUNICATION) ) {
    nSuc = 1;
  }
  PrintMessage( nSuc, "WkbGetLastError()", NULL);
  
  nSuc = 0;
  for (i=WKB_NO_ERROR;i<WKB_ERROR_KRNL_COMMUNICATION;i++) {
    WkbSetLastError(i);
    nRet = WkbGetLastError();
    if (nRet != i ) {
      nSuc = 1;
      break;
    }
  }
  
  PrintMessage( nSuc, "WkbSetLastError()", NULL);
}


/**
 * @brief internal 
 */
void TestWkbOpen(void)
{
  INT nRet = 0;
  int nSuc = 0;
  HWKBENTRY h;
  HWKBENTRY h2;
  ULONG flCtrl = 0;
  char *pszPort;
  LONG lFirmCode = 0;
  LONG lUserCode = 0;
  void *pvCtrl = NULL;
  char szErrorCode[32];
  WKBOPENENTRY wkbOpenEntry;
  
  szErrorCode[sizeof(szErrorCode)-1] = 0;
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    // WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,STD,FC:UC)", NULL);
  
  if ( (HANDLE)NULL != h ) {
    nRet = WkbClose2(h);
    PrintMessage( (nRet == 1) ? 0 : 1, "WkbClose2()", NULL);
  }
  nRet = WkbClose2(h);
  PrintMessage( (nRet == 1) ? 1 : 0, "WkbClose2(invalid)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_FIND;
  pszPort = NULL;
  lFirmCode = 1;
  lUserCode = 1;
  nSuc = 0;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  }
  PrintMessage( nSuc, "WkbOpen2(Local, Invalid FC:UC)", NULL);
  
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_RED;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,RED,FC:UC)", NULL);
  
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION2;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,WKB2,FC:UC)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_FIND | WKB_VERSION3;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,WKB3,FC:UC)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_FIND | WKB_VERSION5;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,WKB5,FC:UC)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_MULTIFIND;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = 0;
  lUserCode = 0;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  }
  PrintMessage( nSuc, "WkbOpen2(Local, Invalid FC:UC)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_MULTIFIND;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = 11;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  }
  PrintMessage( nSuc, "WkbOpen2(Local, Invalid FC)", NULL);
  
  // test (invalid)
  flCtrl = WKB_OPEN_MULTIFIND;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = FIRM_CODE;
  lUserCode = 1;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  }
  PrintMessage( nSuc, "WkbOpen2(Local, Invalid UC)", NULL);
  
  // test 
  flCtrl = WKB_OPEN_MULTIFIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_MULTIFIND,STD,FC:UC)", NULL);
  
  
  // test 
  flCtrl = WKB_OPEN_MULTIFIND | WKB_VERSION_RED;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_MULTIFIND,RED,FC:UC)", NULL);
  
  
  // test 
  flCtrl = WKB_OPEN_MULTIFIND | WKB_VERSION2;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_MULTIFIND,WKB2,FC:UC)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_MULTIFIND | WKB_VERSION3;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_MULTIFIND,WKB3,FC:UC)", NULL);
  
  
  // test (invalid)
  flCtrl = WKB_OPEN_MULTIFIND | WKB_VERSION5;
  pszPort = NULL;
  nSuc = 0;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 1;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_MULTIFIND,WKB5,FC:UC)", NULL);
  
  
  
  
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD | WKB_OPEN_NONETLIMIT;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    h2 = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
    if ( (HANDLE)NULL != h2 ) {
      nSuc = 0;
      WkbClose2(h2);
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,NONETLIMIT,STD,FC:UC)", NULL);
  
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD | WKB_OPEN_STATIONSHARE;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    h2 = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
    if ( (HANDLE)NULL != h2 ) {
      nSuc = 0;
      WkbClose2(h2);
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,STATIONSHARE,STD,FC:UC)", NULL);
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD | WKB_OPEN_GLOBAL;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    h2 = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
    if ( (HANDLE)NULL != h2 ) {
      nSuc = 1;
      WkbClose2(h2);
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,GLOBAL,STD,FC:UC)", NULL);

  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD | WKB_OPEN_EXCLUSIVE;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    h2 = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
    if ( (HANDLE)NULL != h2 ) {
      nSuc = 1;
      WkbClose2(h2);
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,EXCLUSIVE,STD,FC:UC)", NULL);
  
  
  // todo - test for "WKB_OPEN_ENTRY"
  
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  // pszPort = "SRL1|SRL2|SRL3|SRL4|LPT1|LPT2|LPT3|USB";
  pszPort = "SRL1|SRL2|SRL3|SRL4|USB";
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbOpen2(Local,WKB_OPEN_FIND,SRL*|LPT*|USB,FC:UC)", NULL);
}


/**
 * @brief internal 
 */
static void TestWkbCrypt(void) 
{
  INT nRet = 0;
  int nSuc = 0;
  HWKBENTRY h;
  ULONG flCtrl = 0;
  ULONG flCtrlCrypt = 0;
  char *pszPort = NULL;
  LONG lFirmCode = 0;
  LONG lUserCode = 0;
  UINT cbCrypt = 0;
  UINT cbData = 0;
  void *pvCtrl = NULL;
  char szErrorCode[32];
  char szOrign[10] = {0x57,0x4a,0x54,0x52,0x48,0x4b,0x4d,0x4a,0x4e,0x43};
  char szPlain[10];
  char szCrypt[10];
  WKBAREA wkbar;
  BYTE abInit[] = "mySecretVector";
  
  szErrorCode[sizeof(szErrorCode)-1] = 0;
  
  
  // test 10:13
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123456 & 0x0FFFFFFF;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0x5c,0x52,0x86,0xe2,0x28,0xc0,0xab,0x53,0x57,0xbd};
      if ( 0 == memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:13,Sel1)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));


  
  // test 10:13
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123457 & 0x0FFFFFFF;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0x5c,0x52,0x86,0xe2,0x28,0xc0,0xab,0x53,0x57,0xbd};
      if ( 0 != memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:13,Sel2)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));

  
  

  // test 10:13
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = USER_CODE;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_STDCTRL | WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 1234567 & 0x0FFFFFFF;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0x32,0x0c,0x29,0x80,0x3e,0x9c,0x3a,0xd8,0x50,0x72};
      if ( 0 == memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:13;Sel2)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));
  
  
  
  // test 10:14
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = 14;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_STDCTRL | WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123456 & 0x0FFFFFFF;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0x72,0x4c,0x1a,0x8c,0x87,0x15,0xae,0x33,0x91,0xc0};
      if ( 0 == memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:14;Sel1)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));
  
  
  // test 10:14
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = 14;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_STDCTRL | WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123456 & 0x0FFFFFFF;
    wkbar.ulSelectCode |= WKB_SC_BOXLIMIT;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0x0e,0x65,0x77,0xbb,0x0e,0x7a,0x88,0xac,0x00,0xcf};
      if ( 0 == memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:14;Sel2)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));
  
    // test 10:15
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = 15;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_STDCTRL | WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123456 & 0x0FFFFFFF;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0x18,0xd5,0x7c,0xe9,0x5d,0x65,0x39,0x14,0xf7,0xfb};
      if ( 0 == memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:15,Sel1)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));

  
  // test 10:15
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = 15;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_STDCTRL | WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123456 & 0x0FFFFFFF;
    wkbar.ulSelectCode |= WKB_SC_EXPDATE;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( (nRet == 1) && (cbCrypt == cbData) ) {
      char szCheck[10] = {0xf8,0x85,0xe0,0xc1,0x2d,0xe1,0xbb,0x40,0x6e,0xec};
      if ( 0 == memcmp(szCheck, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:15,Sel2)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));

  
  // test 10:16
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  pszPort = NULL;
  nSuc = 1;
  lFirmCode = FIRM_CODE;
  lUserCode = 16;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  if ( (HANDLE)NULL != h ) {
    // we use direct encryption
    memset(&wkbar, 0x0, sizeof(wkbar));
    flCtrlCrypt = WKB_STDCTRL | WKB_CRYPT_CTRL;
    wkbar.ulCrc = 0;
    wkbar.ulReserved = 0;
    wkbar.flCtrl = WKB_STDCTRL | WKB_AREA_SELECT;
    wkbar.flSelCtrl = WKB_SEL_DIRECT;
    wkbar.ulSelectCode = 123456 & 0x0FFFFFFF;
    wkbar.ulSelectCode |= WKB_SC_EXPDATE;
    wkbar.pbGapMap = NULL;
    wkbar.pvCtrl = NULL;
    cbData = (INT)sizeof(szOrign);
    memcpy (szCrypt, szOrign, sizeof(szOrign));
    nRet = WkbCrypt2(h,flCtrlCrypt, szCrypt, &wkbar, cbData, &cbCrypt );
    if ( WkbGetLastError() == WKB_ERROR_OUT_OF_DATE ) {
      // must failed, if date is newer that the expired date (expected)
      if ( 0 == memcmp(szOrign, szCrypt, sizeof(szCrypt))) {
        nSuc = 0;
      }
    }
    WkbClose2(h);
  } 
  PrintMessage( nSuc, "WkbCrypt2(10:16,Sel2)", NULL);
  printf("    Plain: "); PrintHex(szOrign, (int)sizeof(szOrign));
  printf("    Crypt: "); PrintHex(szCrypt, (int)sizeof(szCrypt));

}





/**
 * @brief internal 
 */
void TestEnumPorts(void)
{
  BOOL fValid = FALSE;
  int nSuc = 0;
  int i = 0;
  int nPortWithBoxes = 0;
  char szPortName[256];
  UINT cbLen = (UINT)sizeof(szPortName);
  
  cbLen = (UINT)(sizeof(szPortName) -1);
  memset(szPortName, 0x0, sizeof(szPortName));
  fValid = WkbEnumPorts2 (HWKB_LOCAL, 
        WKB_EPT_LIST, szPortName, cbLen);
  PrintMessage( (fValid==0) ? 1:0, "WkbEnumPorts2(WKB_EPT_LIST)", szPortName);
  
  memset(szPortName, 0x0, sizeof(szPortName));
  for (;;) {
    fValid = WkbEnumPorts2 (HWKB_LOCAL, WKB_EPT_NEXT, szPortName, cbLen );
    if ( fValid ) {
      PrintMessage( (fValid==0) ? 1:0, "WkbEnumPorts2(WKB_EPT_NEXT)", szPortName);
      
      INT nVal = 0;
      typedef struct {
        WKBBOXLIST wkbbxl;
        WKBSERIAL awkbxe[20];
      } PORTLIST;
      PORTLIST prtlist;
      nVal = WkbListPort2 (HWKB_LOCAL, WKB_STDCTRL, szPortName,
               &prtlist, sizeof (prtlist));
      if (nVal && (prtlist.wkbbxl.cBox > 0) ) {
        // one or more boxes connected
        nPortWithBoxes++;
      }
      PrintMessage( (nVal==0) ? 1:0, "WkbListPort2(WKB_STDCTRL)", szPortName);
    }
    
    if (!fValid) {
      // assume that all ports read: exit loop
      // GetLastError() should return WKB_ERROR_INVALID_PORT
      if (WkbGetLastError () != WKB_ERROR_INVALID_PORT) {
        nSuc = 1;
      }
      break;
    }
    i++;
  } // for 
  PrintMessage( ((i==0) || (i>128)) ? 1 : 0, "WkbEnumPorts2(WKB_EPT_NEXT)", NULL);
  PrintMessage( (nPortWithBoxes==0) ? 1 : 0, "WkbListPort2(CheckForBoxes)", NULL);
}



/**
 * @brief The function callback for SysV-signals.
 *
 * Catch system signals and make cleanup, the exit application.
 *
 */
void WkhCatcher (int i) {
  // add clean up code here 

  // exit application 
  exit(1);
} // WkhCatcher 




/**
 * @brief list serial number and box contents of each local WIBU-BOX.
 *
 * Result a list of the entries of all local connected WIBU-BOXes.
 *
 * @pre a valid init pwkdata-structure
 *
 * @return Function return '1' if the list operation was successful else '0'.
 *
 */
int ListBoxes (void) {

  typedef struct {
    // room for 5 serial number entries
    WKBBOXLIST wkbbxl;	// first serial number
    WKBSERIAL awkbxe[20];	// ten other serial numbers
  } PORTLIST;

  PORTLIST prtlist;
  WKBBOXDATA wkbbxd;
  BOOL fValid = FALSE;
  INT nMaxBox = 0;
  INT i = 0;
  INT j = 0;
  UINT uType = 0;
  INT nEntry = 0;
  char szPortName[CB_MAX_PORTNAME];
  INT nReturn = 0;
  UINT nDay = 0, nMonth = 0, nYear = 0;

  szPortName[0] = 0;	// start of listing
  memset(szPortName, 0x0, sizeof(szPortName) );
  memset(&prtlist, 0x0, sizeof(prtlist) );

  // NOTE: We have no outside knowledge how many ports are available and
  // how many WIBU-BOXes are attached to these ports. Therefore we use
  // and endless loop which will be exited when no more ports are available.
  for (;;) {
    
    fValid = WkbEnumPorts2 (HWKB_LOCAL, WKB_EPT_NEXT,	// flag "get next WIBU-BOX port"
          szPortName,	// buffer for portname
          CB_MAX_PORTNAME	// size of portname buffer
      );
    if (!fValid) {
      // assume that all ports read: exit loop
      // GetLastError() should return WKB_ERROR_INVALID_PORT
      nReturn = WkbGetLastError ();
      if (nReturn != WKB_ERROR_INVALID_PORT) {
        printf ("(EE) error : WkbEnumPorts2 (err %i)\n", WkbGetLastError ());
        return 0;
      }	// if

      // NOTE: here we are leaving the loop!
      break;

    }		// if
    
    nMaxBox = 1;
    // list WIBU-BOXes at this port
    fValid = WkbListPort2 (HWKB_LOCAL, WKB_STDCTRL,
               szPortName /* "USB" */ , &prtlist,
               sizeof (prtlist));
    if (fValid && (prtlist.wkbbxl.cBox > 0) ) {
      // one or more WIBU-BOXes found at szPortName: list first WIBU-BOX
      PrintMessage( (fValid==0) ? 1 : 0, "WkbListPort2(WKB_STDCTRL)", szPortName);
      printf ("\n    %i WIBU-BOX(es) found at %s: ",
        prtlist.wkbbxl.cBox, szPortName);

      // list serial numbers at this port
      nMaxBox = prtlist.wkbbxl.cBox - 1;
      if (nMaxBox > 10) {
        nMaxBox = 10;
      } // if 

      for (i = 0; i <= nMaxBox; i++) {
        printf ("\n    Contents of WIBU-BOX #%i (%u-%08lu):\r\n", i + 1, prtlist.wkbbxl.awkbsr[i].usMask, prtlist.wkbbxl.awkbsr[i].ulSerial);
        fValid = WkbListBox2 (HWKB_LOCAL,
                  WKB_STDCTRL,
                  szPortName  ,
                  prtlist.wkbbxl.awkbsr +
                  i, &wkbbxd,
                  sizeof (wkbbxd));
        if (!fValid)  {
          PrintMessage( (fValid==0) ? 1 : 0, "WkbListBox2(WKB_STDCTRL)", NULL);
          return FALSE;
        }


        // list contents of WIBU-BOX
        for (j = 0; j < (INT) wkbbxd.cEntry; j++) {
          WKBBOXENTRY *pwkbbxe =
            wkbbxd.awkbbxe + j;
          uType = (UINT) pwkbbxe->
            flCtrl & WKB_BE_TYPEMASK;
          nEntry = (INT) pwkbbxe->
            flCtrl & WKB_BE_INDEXMASK;
          switch (uType) {
            case WKB_BE_FCUC:
              printf ("%2d: FC/UC:  %8lu:%-8lu", nEntry, pwkbbxe->ulFirmCode, pwkbbxe->ulUserCode);
              break;
            case WKB_BE_MASTER:
              printf ("%2d: Master: %8lu:%-8lu", nEntry, pwkbbxe->ulFirmCode, pwkbbxe->ulUserCode);
              break;
            case WKB_BE_DATA:
              printf ("%2d: User-D: %8lu:%-lu", nEntry, pwkbbxe->ulData2, pwkbbxe->ulData1);
            default:
              break;
          }	// switch


          if (pwkbbxe->flCtrl & WKB_BE_ADDDATA)
            printf (" Added Data: %lu-%lu", pwkbbxe->ulData2, pwkbbxe->ulData1);
          else if (pwkbbxe->
             flCtrl & WKB_BE_ADDLIMIT)
            printf (" Added Limit Counter: %lu", pwkbbxe->ulData1);
          else if (pwkbbxe->
             flCtrl & WKB_BE_ADDEXPDATE)
          {
            nDay = (UINT) pwkbbxe->
              ulData1 & 0xff;
            nMonth = (UINT) (pwkbbxe->
                 ulData1 &
                 0xff00) /
              0x100;
            nYear = (UINT) ((pwkbbxe->
                 ulData1 &
                 0xfff0000) /
                0x10000);
            if (pwkbbxe->ulData2 ==
                0xffff)
            {
              printf (" Expiration Date: %2d.%2d.%4d (is expired)", nDay, nMonth, nYear);
            }
            else
            {
              printf (" Expiration Date: %2d.%2d.%4d (%i days left)", nDay, nMonth, nYear, (int) pwkbbxe->ulData2);
            }	// if
          }	// if
          printf ("\r\n");
        }	// for
        PrintMessage( (fValid==0) ? 1 : 0, "WkbListBox2(WKB_STDCTRL)", szPortName);
      }	// for(i...nMaxBox)
    }		// if
  }			// for(;;)
  return 1;
} // ListBoxes()



/**
 * @brief internal 
 */
void TestQueryInfo2(void)
{
  INT nRet = 0;
  int i = 0;
  ULONG flCtrl = 0;
  char szText[256];
  WKBVERSION wkbVersion;
  UINT cbLen = (UINT)(sizeof(szText)-1);
  
  memset(szText, 0x0, sizeof(szText));
  memset(&wkbVersion, 0x0, sizeof(wkbVersion));
  
  for(i=0;i<127;i++) {
    flCtrl = WKB_QI_ERRTEXT;
    WkbSetLastError(i);
    nRet = WkbQueryInfo2(HWKB_LOCAL, flCtrl, &szText, cbLen);
    
    PrintMessage( (nRet==0) ? 1 : 0, "WkbQueryInfo2(WKB_QI_ERRTEXT)", szText);
  }
  
  flCtrl = WKB_QI_VERVALUE;
  nRet = WkbQueryInfo2(HWKB_LOCAL, flCtrl, &wkbVersion, sizeof(wkbVersion));
  PrintMessage( ((nRet==0) || (wkbVersion.usVersion==0) || (wkbVersion.usVersion>10) ) ? 1 : 0, 
    "WkbQueryInfo2(WKB_QI_VERVALUE)", szText);
  printf("    Version: %i.%02i.%03i.%03i\n", 
    (int)wkbVersion.usVersion, (int)wkbVersion.usSubVersion,
    (int)wkbVersion.usBuild, (int)wkbVersion.usCount );
  
  
  flCtrl = WKB_QI_VERTEXT;
  nRet = WkbQueryInfo2(HWKB_LOCAL, flCtrl, &szText, cbLen);
  PrintMessage( (nRet==0) ? 1 : 0,  "WkbQueryInfo2(WKB_QI_VERTEXT)", szText);
}


/**
 * @brief internal 
 */
void TestQueryEntry2(void)
{
  INT nRet;
  int nSuc = 0;
  HWKBENTRY h;
  ULONG flCtrl = 0;
  char *pszPort = NULL;
  LONG lFirmCode = FIRM_CODE;
  LONG lUserCode = USER_CODE;
  void *pvCtrl = NULL;
  char szErrorCode[32];
  
  WKBOPENENTRY wkbOpenEntry;
  WKBBOXENTRY wksBoxEntry;
  
  szErrorCode[sizeof(szErrorCode)-1] = 0;
  memset(&wkbOpenEntry, 0x0, sizeof(wkbOpenEntry));
  memset(&wksBoxEntry, 0x0, sizeof(wksBoxEntry));
  
  // test 
  flCtrl = WKB_OPEN_FIND | WKB_VERSION_STD;
  nSuc = 1;
  h = WkbOpen2(HWKB_LOCAL, flCtrl, pszPort, lFirmCode, lUserCode, pvCtrl);
  snprintf(szErrorCode, sizeof(szErrorCode) -1, " - error %i", WkbGetLastError() );
  
  PrintMessage( ((HANDLE)NULL != h) ? 0 : 1, "WkbOpen2(Local,WKB_OPEN_FIND,STD,FC:UC)", NULL);
  if ( (HANDLE)NULL != h ) {
    nSuc = 0;
    
    flCtrl = WKB_QE_OPEN;
    nRet = WkbQueryEntry2(h, flCtrl, &wkbOpenEntry, sizeof(wkbOpenEntry));
    
    PrintMessage( ((nRet == 0) || 
                   (wkbOpenEntry.ulFirmCode != FIRM_CODE) || 
                   (wkbOpenEntry.ulUserCode != USER_CODE) || 
                   (wkbOpenEntry.usVersion > 50)
                    ) ? 1 : 0, 
                  "WkbQueryEntry2(WKB_QE_OPEN)", NULL);
    
    
    flCtrl = WKB_QE_BOX;
    nRet = WkbQueryEntry2(h, flCtrl, &wksBoxEntry, sizeof(wksBoxEntry));
    
    PrintMessage( ((nRet == 0) || 
                   (wksBoxEntry.ulFirmCode != FIRM_CODE) || 
                   (wksBoxEntry.ulUserCode != USER_CODE) ||
                   (wksBoxEntry.ulData1 != 0) /* no extended data on 10:13 */
                    ) ? 1 : 0, 
                  "WkbQueryEntry2(WKB_QE_BOX)", NULL);
    
    WkbClose2(h);
  } 
}


/**
 * @brief internal 
 */
void TestExtendedMemory(void)
{
  typedef struct {
    WKBBOXLIST wkbbxl;
    WKBSERIAL  awkbxe[20];
  } PORTLIST;
  PORTLIST prtlist;
  INT fValid = 0;
  WKBBOXENTRY hwkboxentry;
  WKBBOXCONFIG hwkboxconfig;
  WKBEXTMEM hwkextmem;
  UINT cbDest = 0;
  ULONG flCtrl = 0;
  int nRandValue = 12 + ( (int)random() % 54);
  int i = 0,j = 0;
  char *port = "USB";
  memset( &hwkboxentry,  0, sizeof(hwkboxentry) );
  memset( &hwkboxconfig, 0, sizeof(hwkboxconfig) );
  
  fValid = WkbListPort2( HWKB_LOCAL, WKB_STDCTRL,  port,  
     &prtlist, sizeof(prtlist) 	);
  PrintMessage( (fValid==0) ? 1 : 0, "WkbListPort2(WKB_STDCTRL, USB)", NULL);
  if( !fValid) {
    return ;
  }

  flCtrl =  WKB_LIST_CONFIG;
  cbDest = sizeof(hwkboxconfig);
  fValid = WkbListBox2( HWKB_LOCAL, flCtrl, port, &prtlist.wkbbxl.awkbsr[0] , &hwkboxconfig, cbDest) ; 
  PrintMessage( (fValid == 0) ? 1 : 0, "WkbListBox2(WKB_LIST_CONFIG)", NULL);
  if( !fValid) {
    return ;
  }

  if ( hwkboxconfig.usExtMemSize > 0) {
    printf("    Found WibuBox %i-%u with %i Pages ExtMem\n",
      (int)prtlist.wkbbxl.awkbsr[0].usMask,
      prtlist.wkbbxl.awkbsr[0].ulSerial,
      hwkboxconfig.usExtMemSize );
    
      j = 0; // read out only first page 
      memset( &hwkextmem,    0, sizeof(hwkextmem) );
      flCtrl =  WKB_LIST_EXTMEM;
      cbDest = sizeof(hwkextmem);
      hwkextmem.fsCtrl = WKB_XM_USER;  // handle unprotected mem-page
      hwkextmem.iPage = (USHORT)j;
      fValid = WkbListBox2( HWKB_LOCAL, flCtrl, port, &prtlist.wkbbxl.awkbsr[0], &hwkextmem, cbDest) ; 
      PrintMessage( (fValid == 0) ? 1 : 0, "WkbListBox2(WKB_LIST_EXTMEM)", NULL);
      if( !fValid) {
       return ;
      }
  
      if(1 > 0 ) {
        printf("    0x%04x : ", j*32 );
        for(i=0;i<16;i++) {
          printf("%02x ", hwkextmem.abData[i] );
        }
        printf("\n    0x%04x : ", j*32 + i );
        for(i=16;i<32;i++) {
          printf("%02x ", hwkextmem.abData[i] );
        }
        printf("\n");
      }
      
      // write ext mem.
      j = 0; // write only first page 
      flCtrl = WKB_PGM_EXTMEM; 
      memset( &hwkextmem,    0, sizeof(hwkextmem) );
      cbDest = sizeof(hwkextmem);
      hwkextmem.fsCtrl = WKB_XM_USER;  // handle unprotected mem-page
      hwkextmem.iPage = (USHORT)j;
      for (i=0;i<32;i++) 
        hwkextmem.abData[i] = (BYTE)( 32 + (i + j*32) % nRandValue );
      fValid = WkbProgram2( HWKB_LOCAL, flCtrl, port, &prtlist.wkbbxl.awkbsr[0], &hwkextmem );
      PrintMessage( (fValid == 0) ? 1 : 0, "WkbProgram2(WKB_XM_USER)", NULL);
      if( !fValid) {
        return ;
      }
      
      // read again 
      j = 0; // read out only first page 
      memset( &hwkextmem,    0, sizeof(hwkextmem) );
      flCtrl =  WKB_LIST_EXTMEM;
      cbDest = sizeof(hwkextmem);
      hwkextmem.fsCtrl = WKB_XM_USER;  // handle unprotected mem-page
      hwkextmem.iPage = (USHORT)j;
      fValid = WkbListBox2( HWKB_LOCAL, flCtrl, port, &prtlist.wkbbxl.awkbsr[0], &hwkextmem, cbDest) ; 
      PrintMessage( (fValid == 0) ? 1 : 0, "WkbListBox2(WKB_LIST_EXTMEM)", NULL);
      if( !fValid) {
       return ;
      }
  
      if(1 > 0 ) {
        printf("    0x%04x : ", j*32 );
        for(i=0;i<16;i++) {
          printf("%02x ", hwkextmem.abData[i] );
        }
        printf("\n    0x%04x : ", j*32 + i );
        for(i=16;i<32;i++) {
          printf("%02x ", hwkextmem.abData[i] );
        }
        printf("\n");
      }

      
  } // if(nBoxMemSize)
}


/**
 * @brief internal 
 */
static void TestTypeSize(void)
{
  PrintMessage( !(sizeof(char) == 1) , "sizeof(char) == 1", NULL);
  PrintMessage( !(sizeof(short) == 2) , "sizeof(short) == 2", NULL);
  PrintMessage( !(sizeof(int) == 4) , "sizeof(int) == 4", NULL);
  
  PrintMessage( !(sizeof(unsigned char) == 1) , "sizeof(char) == 1", NULL);
  PrintMessage( !(sizeof(unsigned short) == 2) , "sizeof(short) == 2", NULL);
  PrintMessage( !(sizeof(unsigned int) == 4) , "sizeof(int) == 4", NULL);
  
  PrintMessage( !(sizeof(BYTE) == 1) , "sizeof(BYTE) == 1", NULL);
  PrintMessage( !(sizeof(USHORT) == 2) , "sizeof(USHORT) == 2", NULL);
  PrintMessage( !(sizeof(UINT) == 4) , "sizeof(UINT) == 4", NULL);
  PrintMessage( !(sizeof(ULONG) == 4) , "sizeof(ULONG) == 4", NULL);
  // PrintMessage( !(sizeof(HANDLE) == 4) , "sizeof(HANDLE)", NULL);
  
  PrintMessage( !(sizeof(WKBSERIAL) == 8) , "sizeof(WKBSERIAL)", NULL);
  PrintMessage( !(sizeof(WKBBOXLIST) == 16) , "sizeof(WKBBOXLIST)", NULL);
  PrintMessage( !(sizeof(WKBBOXENTRY) == 20) , "sizeof(WKBBOXENTRY)", NULL);
  PrintMessage( !(sizeof(WKBBOXDATA) == 208) , "sizeof(WKBBOXDATA)", NULL);
  PrintMessage( !(sizeof(WKBOPENFIND) == 16) , "sizeof(WKBOPENFIND)", NULL);
  PrintMessage( !(sizeof(WKBEXTMEM) == 44) , "sizeof(WKBEXTMEM)", NULL);
  PrintMessage( !(sizeof(WKBPROGRAM) == 32) , "sizeof(WKBPROGRAM)", NULL);
  PrintMessage( !(sizeof(WKBVERSION) == 16) , "sizeof(WKBVERSION)", NULL);
}



/**
 * @brief the program main function 
 *
 * This is the main program function, it will init the driver and list all 
 * detected local WIBU-BOXes. 
 *
 * @return This function return '0'.
 *
 * @warning Only WIBU-BOXes will be list, that ports a vaild. I.e. the enviroment
 *          WKCOM and WKLPT define valid ports.
 */
int main (int argc, char *argv[]) {

  // setup sysv-signals
  signal (SIGINT, WkhCatcher);
  
  printf("Start WibuKey simple driver test\n");
  printf("Test required one WibuBox/U with follow entries:\n");
  printf("   WibuBox/U version 6 with ext. memory or WibuBox/ST\n");
  printf("   0: FC/UC:        10:13      \n");
  printf("   1: FC/UC:        10:14       Added Limit Counter: 10000\n");
  printf("   2: FC/UC:        10:15       Expiration Date: 31.12.2019\n");
  printf("   3: FC/UC:        10:16       Expiration Date:  1.1.2000 (expired)\n");
  printf("   4: FC/UC:        10:17       Added Data: 1-234567\n");

  TestTypeSize();
  
  TestWkbVersion();
  
  TestErrorCode();
  
  TestQueryInfo2();
  
  TestWkbOpen();
  
  TestWkbCrypt();
  
  TestEnumPorts();
  
  ListBoxes();
  
  TestQueryEntry2();

  TestExtendedMemory();


  // WkbProgram()
  
  printf("\nResults:\n");
  printf("  No. of tests:  %i\n", nNoTests);
  printf("  No. of errors: %i\n", nNoError);

  return 0;
} // main
