/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
bsremul.c                                      Leased code from BSRLIB library
==============================================================================

This C source file implements code and data which is transferred without
significant change from the BSRLIB library. It is currently only used by
the Macintosh and NLM variant of the driver.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1997 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://wwww.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu), Marcus Balb (marb)
System ... DOS 6.20, Windows 95, Windows 3.11, Windows NT 4.00
           Rational System DOS/4G
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
97Aug23 2.32 mabu: initial release
98Mar15 2.41 mabu: adapted to Novell Netware
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#if !defined(O_BINARY)
  #define O_BINARY 0
#endif 

/*****************************************************************************

---- Absolute/relative date calculations -------------------------------------

*****************************************************************************/

static int anLeapDays[] = {
  -1, 30, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365
};

static int anDays[] = {
  -1, 30, 58, 89, 119, 150, 180, 211, 242, 272, 303, 333, 364
};


/*****************************************************************************
BsrFreeMem
==============================================================================

This function destroys a memory block which was allocated previously by
BsrAllocMem.

Parameters
  pPointer  contains the handle of a memory block which was allocated by
            BsrAllocMem.

*****************************************************************************/

void BsrFreeMem(void *pPointer) 
{
  if ( pPointer != NULL ) { 
    free(pPointer); 
    pPointer = NULL; 
  } // if
} // BsrFreeMem() 



/*****************************************************************************
B s r C o n v e r t D a t e C o u n t
==============================================================================

This function converts the specified date count to a year/month/day
specification. The date count is relative to 1st January 1970.

The conversation is limited to the range from 1970 to 2099 because 2100 is
no leap year which is not checked in this function.

Parameters
  uDate  contains the number of days since 1st January 1970.
  pst .. points to the structure which gets the date.

Return
  The function returns TRUE if the date count within the specified range
  and FALSE if not. In the latter case, the 31th Dec 2099 is exceeded.

*****************************************************************************/

int BsrConvertDateCount(UINT uDate, SYSTEMTIME *pst)
{
  int nYear = 0;
  int iMonth = 0;
  int *pnDays = NULL;
  int fLeapYear = FALSE;
  unsigned int uDate2 = uDate;
  // set all days to 0
  memset(pst, 0, sizeof(SYSTEMTIME));
  // first, identify the four-year interval since this makes handling
  // leap-years easy (note that 2000 IS a leap year and 2100 is out-of-range)
  nYear = (unsigned short)uDate2 / (3 * 365 + 366);
  // set <uDate2> to remainder of 4-year-block
  uDate2 -= nYear * (3 * 365 + 366);
  // determine the year of the interval (1972, 1976 .. are leap years!)
  nYear *= 4;
  if (uDate2 >= 365 + 365 + 366) {
    // 1973, 1977, 1981, ...
    uDate2 -= 365 + 365 + 366;
    nYear += 3;
  } else if (uDate2 >= 365 + 365) {
    // 1972, 1976, 1980, ...
    uDate2 -= 365 + 365;
    nYear += 2;
    // this is also a lap year
    fLeapYear = TRUE;
  } else if (uDate2 >= 365) {
    // 1971, 1975, 1979
    uDate2 -= 365;
    nYear += 1;
  } // if
  if (nYear > 2099 - 1970) {
    // year overflow => return FALSE
    return FALSE;
  } // if
  pst->wYear = nYear + 1970;
  // <uDate2> now holds the number of days since the beginning of that year.
  // determine months since January (1 - 12) and day of month (1 - 31)
  pnDays = fLeapYear? anLeapDays : anDays;
  for (iMonth = 1; pnDays[iMonth] < (int)uDate2; iMonth++) {
    ;
  } // for
  pst->wMonth = iMonth;
  // determine day within a month
  pst->wDay = (unsigned short)uDate2 - (unsigned int)pnDays[iMonth - 1];
  // determine day of week in the range 0 up to 6 (0 is sunday)
  // base day 01Jan70 is a Thursday 
  pst->wDayOfWeek = (unsigned short)(uDate + 6 - 2) % 7;
  // return without error
  return TRUE;
} // BsrConvertDateCount()


/*****************************************************************************
B s r G e t C u r r e n t D a t e C o u n t
==============================================================================

This function returns the date count, relative to 1970Jan01, of the current
date.

Return
  The function returns the current time as date count. This value is always
  valid.

Implementation
  The function saves the current date and count in a cache memory to avoid
  that the expensive count calculation is repeated at a specific data.
  Therefore, the function is really fast.

*****************************************************************************/

UINT BSRAPI BsrGetCurrentDateCount(void)
{
  static SYSTEMTIME stCurrent;
  static UINT uCurrentDate;
  SYSTEMTIME st;
  GetLocalTime(&st);
  if (memcmp(&st, &stCurrent, FIELDOFFSET(SYSTEMTIME, wHour))) {
    // date changed: recalculate date offset
    stCurrent = st;
    uCurrentDate = BsrGetDateCount(stCurrent.wYear, stCurrent.wMonth,
      stCurrent.wDay);
    ASSERT(uCurrentDate < UINT_MAX - 2);
  } // if
  // return counter of current date
  return uCurrentDate;
} // BsrGetCurrentDateCount()


/*****************************************************************************
B s r G e t D a t e C o u n t
==============================================================================

This function converts the specified year/month/day specication into a
continous date count which is relative to 1st January 1970. The maximum
supported address is 2099-Dec-31.

Parameters
  nYear   contains the year in the range 1970..2099.
  nMonth  contains the month in the range 1..12.
  nDay .. contains the day in the range 1..31.

Return
  The function returns the continous date count value. If this value is
  near UINT_MAX, the specified parameters were not correct.

Comments
  This function checks also if the specified year/month/day values are valid.
  Following error codes are returned:
    UINT_MAX - 0: wrong year value
    UINT_MAX - 1: wrong month value
    UINT_MAX - 2: wrong day value

*****************************************************************************/

UINT BSRAPI BsrGetDateCount(INT nYear, INT nMonth, INT nDay)
{
  UINT uDate;
  INT *pnDays;
  BOOL fLeapYear;
  if (nYear < 1970 || nYear > 2099) {
    return UINT_MAX - 0;
  } // if
  if (nMonth < 1 || nMonth > 12) {
    return UINT_MAX - 1;
  } // if
  // determine lear year flag
  fLeapYear = (nYear % 4 == 0);
  // set pointer to day array
  pnDays = fLeapYear? anLeapDays : anDays;
  // bring year into range since 1970
  nYear -= 1970;
  // calculate 4-year blocks
  uDate = (nYear / 4) * (365 + 365 + 366 + 365);
  // calculate now rest within 4-year block
  nYear %= 4;
  uDate += 365 * nYear;
  if (nYear == 3) {
    // 1973, 1977, 1981: one day more for leap year
    uDate++;
  } // if
  // month now in range 0..11
  nMonth--;
  // add month value (in dependence of leap year)
  uDate += (UINT)pnDays[nMonth];
  // check day value
  if (nDay < 1 || nDay > pnDays[nMonth + 1] - pnDays[nMonth]) {
    // illegal day => return error
    return UINT_MAX - 2;
  } // if
  // add day value (offset 1 corrected by pnDays-start of "-1")
  uDate += (UINT)nDay;
  // return determined value
  ASSERT(uDate < UINT_MAX - 2);
  return uDate;
} // BsrGetDateCount()




#define lstrcmpi(a, b) strcasecmp((a), (b))

// pointer to allocated profile buffer
static BYTE *pbProfBuf = NULL;

// dummy buffer for non-existing buffer
static BYTE bNoBuffer = 0;

int BSRAPI BsrGetPrivateProfileString(const char *pszSection,
 const char *pszEntry, const char *pszDefault, char *pszDestBuf,
 unsigned int cbDestBuf, const char *pszFileName)
{
  char *pszDest;
  char *pszDestEnd = pszDestBuf + cbDestBuf;
  ASSERT(pszFileName != NULL);
  // simulate full Win32 behaviour of GetPrivateProfileString()
  if (pbProfBuf == NULL) {
    // read data from specified file & compact it (remove spaces/comments)
    ULONG ulSize = 0;
    BYTE *pbBuf = NULL;

    int cbRead = 0;
    int hf;
  #if defined(__GNUC__)
    struct stat mystat;
  #endif 
    // set to dummy buffer
    pbProfBuf = &bNoBuffer;
    hf = open(pszFileName, O_RDONLY | O_BINARY);
    if (hf < 0) {
      goto LOADED;
    } // if
    // determine size of file for buffer
    #if defined(__GNUC__)
      fstat(hf, &mystat);
      ulSize = (ULONG)mystat.st_size;
    #endif

    if ((int)ulSize == -1) {
      ulSize = 0;
    } // if
    
    // bad length => bad file => ignore contents, close file
    if (ulSize > 0xFFFE) {
      ulSize = 0;
    } // if
    if (ulSize > 0) {
      // allocate buffer with this length
      pbBuf = malloc((UINT)ulSize + 1);
      // memory too small: don't read data
      if (pbBuf == NULL) {
        ulSize = 0;
      } // if
    } else {
      goto LOADED;
    }
    // read contents of file into buffer, close file
    cbRead = read(hf, pbBuf, (int)ulSize);
    if (cbRead < 0) {
      // error during read: free memory & set to default
      cbRead = 0;
    } // if
    if (pbBuf != NULL) {
      if (cbRead != 0) {
        // set buffer permanently
        pbProfBuf = pbBuf;
        // set end of file
        pbProfBuf[cbRead] = 0;
      } else {
        // delete allocated buffer (don't use it)
        free(pbBuf);
      } // if
    } // if
    // close file
    close(hf);
    { // compact file for multiple reading
      BYTE *pbSrc = pbProfBuf;
      BYTE *pbDest = pbProfBuf;
      BYTE *pbStart;
      int cch;
      while (*pbSrc != 0) {
        ASSERT(pbSrc >= pbDest);
        // skip leading spaces
        while (*pbSrc == ' ') {
          pbSrc++;
        } // if
        if (*pbSrc == '[') {
          // start of section header: copy it, start with '[' for identifying
          pbStart = pbDest++;
          // leading '[' not stored to avoid (pbSrc > pbDest)!
          pbSrc++;
          do {
            *pbDest++ = *pbSrc++;
          } while (*pbSrc >= ' ' && *pbSrc != ']');
          // set code character for sections
          *pbDest++ = ']';
          // set byte length at start of name
          cch = (int)(pbDest - pbStart - 1);
          if (cch > 255) {
            goto BAD_FILE;
          } // if
          *pbStart = (BYTE)cch;
        } else if (*pbSrc != ';' && *pbSrc != '\r' && *pbSrc != '\n') {
          // valid contents of line: transfer entry name
          // * no topic headline exists => wrong file
          if (pbSrc <= pbDest) {
            goto BAD_FILE;
          } // if
          // reserve byte for length specification
          pbStart = pbDest++;
          do {
            *pbDest++ = *pbSrc++;
          } while(*pbSrc >= ' ' && *pbSrc != '=');
          // remove trailing spaces
          while (*--pbDest == ' ' && pbDest > pbStart);
          pbDest++;
          if (*pbSrc == '=') {
            pbSrc++;
          } // if
          // skip spaces after equation
          while (*pbSrc == ' ') {
            pbSrc++;
          } // if
          // set byte length at start of name
          cch = (int)(pbDest - pbStart - 1);
          if (cch > 255) {
            goto BAD_FILE;
          } // if
          *pbStart = (BYTE)cch;
          // transfer contents of entry
          pbStart = pbDest++;
          do {
            *pbDest++ = *pbSrc++;
          } while(*pbSrc >= ' ');
          // remove trailing spaces
          while (*--pbDest == ' ' && pbDest > pbStart);
          if (pbStart[1] == '\"' && *pbDest == '\"') {
            // remove surrounding quotes
            memmove(pbStart + 1, pbStart + 2, pbDest - pbStart - 2);
            pbDest--; 
          } else {
            // after last character
            pbDest++;
          } // if
          // set byte length at start of name
          cch = (int)(pbDest - pbStart - 1);
          if (cch > 255) {
            goto BAD_FILE;
          } // if
          *pbStart = (BYTE)cch;
        } // if
        // skip remainder until end of line
        while (*pbSrc >= ' ') {
          *pbSrc++;
        } // if
        // search start of next line
        while (*pbSrc != 0 && *pbSrc <= ' ') {
          *pbSrc++;
        } // if
      } // while
      // set end of buffer, reduce to length of compacted buffer
      *pbDest = 0;
      pbProfBuf = (BYTE*) realloc(pbProfBuf, pbDest - pbProfBuf + 1);
      goto LOADED;
    } // block
  BAD_FILE:
    // wrong format of file: don't accept read buffer
    free(pbProfBuf);
    // set to dummy buffer
    pbProfBuf = &bNoBuffer;
  } // block
LOADED:;
  // execute operations with read data
  pszDest = pszDestBuf;
  if (pszSection == NULL) {
    // return all section headers
    BYTE *pbSearch = pbProfBuf;
    while (*pbSearch != 0) {
      // length of next part
      UINT cb = *pbSearch++;
      if (pbSearch[cb - 1] == ']') {
        // section part found: return it
        // * buffer full => exit loop
        if (pszDest + cb >= pszDestEnd) {
          break;
        } // if
        memcpy(pszDest, pbSearch, (size_t)(cb - 1));
        pszDest += cb - 1;
        *pszDest++ = 0;
        pbSearch += cb;
      } else {
        // skip entry name + argument
        pbSearch += cb;
        pbSearch += *pbSearch + 1;
      } // if
    } // while
    // final terminator
    if (pszDest < pszDestEnd) {
      *pszDest++ = 0;
    } // if
    return (int)(pszDest - pszDestBuf);
  } else if (pszEntry == NULL) {
    // return all entry names of specified section
    BYTE *pbSearch = pbProfBuf;
    while (*pbSearch != 0) {
      // length of next part
      UINT cb = *pbSearch++;
      if (pbSearch[cb - 1] == ']') {
        // new section: analyse it
        char szBuf[60];
        memcpy(szBuf, pbSearch, (size_t)(cb - 1));
        szBuf[cb - 1] = 0;
        // to next entry (entry key or next section)
        pbSearch += cb;
        // set pointer to end of header name
        if (lstrcmpi(szBuf, pszSection) == 0) {
          // header matches: set all entries keys into destination
          while (*pbSearch != 0 && pbSearch[*pbSearch] != ']') {
            // transfer found entry key name into destination
            cb = *pbSearch++;
            // destination buffer full => exit loop
            if (pszDest + cb - 1 >= pszDestEnd) {
              break;
            } // if
            memcpy(pszDest, pbSearch, (size_t)cb);
            pszDest += cb;
            *pszDest++ = 0;
            // skip contents of entry
            pbSearch += cb;
            pbSearch += *pbSearch + 1;
          } // while
          // end of destination string, return length
          if (pszDest < pszDestEnd) {
            *pszDest++ = 0;
          } // if
          return (int)(pszDest - pszDestBuf);
        } // if
      } else {
        // key: skip entry name + argument
        pbSearch += cb;
        pbSearch += *pbSearch + 1;
      } // if
    } // while
  } else {
    // search contents of specified key
    BYTE *pbSearch = pbProfBuf;
    while (*pbSearch != 0) {
      // length of next part
      int cb = *pbSearch++;
      if (pbSearch[cb - 1] == ']') {
        // new section: analyse it
        char szBuf[60];
        memcpy(szBuf, pbSearch, (size_t)(cb - 1));
        szBuf[cb - 1] = 0;
        // to next entry (entry key or next section)
        pbSearch += cb;
        // set pointer to end of header name
        if (lstrcmpi(szBuf, pszSection) == 0) {
          // header matches: search entry key which matches
          while (*pbSearch != 0 && pbSearch[*pbSearch] != ']') {
            // check next entry
            cb = *pbSearch++;
            memcpy(szBuf, pbSearch, (size_t)cb);
            szBuf[cb] = 0;
            // to end of entry key field
            pbSearch += cb;
            // read length of entry contents
            cb = *pbSearch++;
            if (lstrcmpi(szBuf, pszEntry) == 0) {
              // section & key matches: transfer into return
              if (pszDest + cb - 1 >= pszDestEnd) {
                // reduce transfer length
                cb = (int)(pszDestEnd - pszDest - 1);
              } // if
              if (cb >= 0) {
                memcpy(pszDest, pbSearch, (size_t)cb);
                pszDest += cb;
                *pszDest = 0;
              } // if
              // end of destination string, return length
              // * the length is not increased by the trailing 0 character
              return (int)(pszDest - pszDestBuf);
            } // if
            // skip entry contents, continue search operation
            pbSearch += cb;
          } // while
        } // if
      } else {
        // key: skip entry name + argument
        pbSearch += cb;
        pbSearch += *pbSearch + 1;
      } // if
    } // while
  } // if
  { // analyse return value
    UINT cbDest = (UINT)(pszDest - pszDestBuf);
    if (cbDest == 0) {
      // set default value
      cbDest = (UINT)_tcslen(pszDefault);
      if (cbDest >= (UINT)cbDestBuf) {
        cbDest = cbDestBuf - 1;
      } // if
      memcpy(pszDestBuf, pszDefault, (size_t)cbDest);
      // terminator
      pszDestBuf[cbDest] = 0;
    } // if
    return cbDest;
  } // block
} // BsrGetPrivateProfileString()

// =====================================
// end of bsremul.c
// =====================================
