/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clcommon.c                                   Common part of the calling driver
==============================================================================

This C source file implements functions which are called from the API modules
to implement the WibuKey-API functions.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992-2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52-54, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All POSIX OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
92Sep13 1.33 mabu: start of development
93Mar18 1.33 mabu: custom source variant released
93Mar21 1.35 mabu: adaption to Windows
93Apr13 1.35 mabu: WkbUniversalCall() included (Module UNICALL.C)
93Jun06 1.35 mabu: conversation to large memory model.
93Jun20 1.35 mabu: adaptation to Win32 environment.
93Jul02 1.35 mabu: first adaptations to WibuBox version 2
93Jul22 1.35 mabu: adapation to DOS
93Aug21 2.00 mabu: greater changes in API of new functions
94Mar08 2.00 mabu: low level driver variants added
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94Aug10 2.00 mabu: adaptation to new WkNet/WkLAN API
94Oct11 2.00 mabu: implementing the look-ahead-function
94Nov11 2.00 mabu: version 2.00 released
95Feb08 2.02 mabu: bug in CreateHandle() function removed
95Feb09 2.05 mabu: extracted from KRNL.C
95Feb27 2.05 mabu: minimum-T management added
95Mar05 2.05 mabu: wrong-Path message extended by current path
95Jun04 2.06 mabu: addng of calling the WKINT interface under DOS
95Jul27 2.06 mabu: checking of SystemDir profile setting
95Aug11 2.07 mabu: critical checking of serial number added
95Sep24 2.08 mabu: single port names don't have index number (ADB1 -> ADB)
95Oct01 2.08 mabu: proper checking if WKINT calls 2.0 driver or higher
95Oct28 2.08 mabu: avoid multiple calling of WkdReleasePort low level release
95Oct31 2.08 mabu: absolute timing mechanism implemented
95Nov08 2.09 mabu: extensions for WibuBox 4 implemented
95Dec04 2.09 mabu: proper port reading implemented (avoids polls)
96Jan29 2.10 mabu: API specific parts moved from CTRL.C
96Feb20 2.10 mabu: WkDos calling scheme replaces WkInt calling scheme
96Mar23 2.11 mabu: function WkdResetStatus() included
96Mar24 2.11 mabu: proper error code management when function returns 0
96Jun23 2.13 mabu: renamed from BSCOMMON.C
96Oct19 2.15 mabu: moving of WkDOS calling to MNCALLS.C
96Nov05 2.20 mabu: DLLTEST mode reduced to tests in debugging mode
96Dec19 2.21 mabu: proper checking of path name length in WkdCheckFilePosition
97Feb11 2.25 mabu: strict checking of API strings in WKWIN[32].DLL included
97Feb11 2.25 mabu: storing back of received structures simplified
97Jun11 2.31 mabu: litte/big endian adaptation for Macintosh
97Jul13 2.31 mabu: simplified indirect encryption management implemented
97Jul26 2.31 mabu: redesign of action management
98Feb07 2.41 mabu: integration of new error codes
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

// define some prototypes
void WkdGetWKBSYSTEM(WKBSYSTEM *pwkbsys);
void WkdGetWKBSERIAL(WKBSERIAL *pwkbsr);
void WkdGetWKBSELECT(WKBSELECT *pwkbsel);
void WkdGetWKBPORTNAME(WKBPORTNAME *pwkbprtn);
void WkdGetWKBPORT(WKBPORT *pwkbprt);
void WkdGetWKBOPENENTRY(WKBOPENENTRY *pwkboe);
void WkdGetWKBBOXENTRY(WKBBOXENTRY *pwkbbxe);
void WkdSetWKBBOXENTRY(WKBBOXENTRY *pwkbbxe);
void WkdGetWKBBOXCONFIG(WKBBOXCONFIG *pwkbbxc);
void WkdSetWKBVERSION (WKBVERSION *pwkbver);


#if ASYNCMODE

/*****************************************************************************

---- Management of Asynchronous States (Action Handling) ---------------------

*****************************************************************************/

// first and last action ID.
// Don't expand to 32 bit to use action ID also in 16 bit return values
enum {
  FIRST_ACTION = -2,
  LAST_ACTION = -32768
};


static ASYNCACTION *pasacList = NULL;
static INT idNextAction = FIRST_ACTION;


/*****************************************************************************
W k d C r e a t e A c t i o n
==============================================================================

This function creates a new action and initializes it.

Globals
  cstWkd*

Return
  The function returns the address of the created action and NULL if the
  action is not created.

*****************************************************************************/

ASYNCACTION * WkdCreateAction(void)
{
  ASYNCACTION *pasac = BsrAllocMem(sizeof(ASYNCACTION));
  if (pasac == NULL) {
    // action not created => return special error
    cstWkd.idError = WKB_ERROR_TOO_MANY_ACTIONS;
    return NULL;
  } // if
  // initialize created action object
  // set all contents to 0
  memset(pasac, 0, sizeof(ASYNCACTION));
  pasac->psys = cstWkd.psys;
  pasac->idAction = idNextAction;
  // because the most functions terminated immediately set ACT_COMPLETE here
  // later analyse a more critical the single operations and set here ACT_RUN
  pasac->idState = ACT_COMPLETE; // ACT_RUN;
  // insert created action in action list (double linked)
  pasac->pNext = pasacList;
  pasac->ppPrev = &pasacList;
  if (pasacList != NULL) {
    pasacList->ppPrev = &pasac->pNext;
  } // if
  pasacList = pasac;
  // determine next action, restart if overrun
  idNextAction--;
  if (idNextAction == -32768) {
    idNextAction = FIRST_ACTION;
  } // if
  // action was created: return address
  return pasac;
} // WkdCreateAction()


/*****************************************************************************
W k d R e m o v e A c t i o n
==============================================================================

This function removes the action which is set in <cstWkd>
as the current action.

Parameters
  pasac  points to the action which is removed.

Globals
  cstWkd

*****************************************************************************/

void WkdRemoveAction(ASYNCACTION *pasac)
{
  ASSERT(pasac != NULL);
  if (pasac->pNext != NULL) {
    // set new predecessor of successor
    pasac->pNext->ppPrev = pasac->ppPrev;
  } // if
  *pasac->ppPrev = pasac->pNext;
  BsrFreeMem(pasac);
} // WkdRemoveAction()


/*****************************************************************************
W k d R e m o v e A l l A c t i o n s O f S u b S y s t e m
==============================================================================

This function removes all actions which are created for the specified
WibuKey subsystem.

Parameters
  psys  points to the subsystem object which actions are removed.

Globals
  cstWkd

*****************************************************************************/

void WkdRemoveAllActionsOfSubSystem(const SYSENTRY *psys)
{
  ASYNCACTION *pasacThis = NULL;
  pasacThis = pasacList;
  while (pasacThis != NULL) {
    ASYNCACTION *pasacNext = pasacThis->pNext;
    if (pasacThis->psys == psys) {
      // same subsystem as specified system => remove action
      WkdRemoveAction(pasacThis);
      // set pointer to next element
      pasacThis = pasacNext;
    } // if
    // select next action in list
    pasacThis = pasacNext;
  } // while
} // WkdRemoveAllActionsOfSubSystem()


/*****************************************************************************
W k d F i n d A c t i o n
==============================================================================

This function finds the action with the specified code and returns a pointer
to the action object.

Parameters
  idAction  contains the action which is searched.

Return
  The function returns the pointer to the found action or NULL if the action
  is not found.

Globals
  cstWkd

*****************************************************************************/

ASYNCACTION * WkdFindAction(INT idAction)
{
  ASYNCACTION *pasacThis;
  for (
      pasacThis = pasacList;
      pasacThis != NULL;
      pasacThis = pasacThis->pNext
    ) {
    if (pasacThis->idAction == idAction) {
      // action found => return its address
      return pasacThis;
    } // if
  } // for
  // action not found => return NULL
  return NULL;
} // WkdFindAction()


/*****************************************************************************
W k d F i n d C a l l e r A c t i o n
==============================================================================

This function finds the action with the specified code and sets it as the
current action into the <cstWkd> structure.

Parameters
  idAction  contains the action which is searched.

Return
  The function returns TRUE if the action is found and FALSE if not. In the
  latter case, <cstWkd.idError> receives the error code
  WKB_ERROR_UNDEFINED_ACTION.

Globals
  cstWkd

*****************************************************************************/

BOOL WkdFindCallerAction(INT idAction)
{
  cstWkd.pasacThis = WkdFindAction(idAction);
  if (cstWkd.pasacThis != NULL) {
    // action found => return TRUE
    return TRUE;
  } else {
    // action not found => return error state
    cstWkd.idError = WKB_ERROR_UNDEFINED_ACTION;
    return FALSE;
  } // if
} // WkdFindCallerAction()

#endif // #if ASYNCMODE


/*****************************************************************************

---- API Call Support Functions ----------------------------------------------

*****************************************************************************/



// fixed declaration of local system
SYSENTRY gsysWkdLocal = {
  STYPE_LOCAL
};


/*****************************************************************************
WkdResetStatus
==============================================================================

This function resets the error code of the current execution and resets the
standard control flags of a function call. The control flags is set to the
<flDtrlDefSet> special value in <cstWkd>.

Globals
  cstWkd*

*****************************************************************************/

void WkdResetStatus(void)
{
  cstWkd.flCtrl = cstWkd.flCtrlDefSet;
  cstWkd.bLevel = 0;
  cstWkd.fUnicodeCall = FALSE;
#if ASYNCMODE
  cstWkd.fAsyncCall = FALSE;
#endif
  // no error detected
  cstWkd.idError = WKB_NO_ERROR;
} // WkdResetStatus()


/*****************************************************************************
W k d I s R e m o t e K e r n e l
==============================================================================

This function returns the state whether a separate kernel driver is used to
access the local WIBU-BOXes.

*****************************************************************************/



/*****************************************************************************
W k d G e t C t r l F l a g s
==============================================================================

This function sets the control flags into the global <cstWkd> structure.
For asynchronous calls, an action object is created.

Parameters
  flCtrl   contains the control flags of the original function call.

Globals
  cstWkd*

Return
  The function returns TRUE if the flag settings are valid and FALSE if an
  error occurred.

*****************************************************************************/

BOOL WkdGetCtrlFlags(ULONG flCtrl)
{
  if ((flCtrl & WKB_LEVELMASK) > WKB_LEVEL1) {
    // invalid level => return error
    cstWkd.idError = WKB_ERROR_INVALID_LEVEL;
    return FALSE;
  } // if

  // outside of WKWIN[32].DLL: WKB_UNICODE not permitted
  if (flCtrl & (WKB_UNICODE | WKB_ASYNCHRONOUS)) {
    cstWkd.idError = WKB_ERROR_OPTION_NOT_SUPPORTED;
    return FALSE;
  } // if

  // store determined level as byte (currently 0 or 1)
  cstWkd.bLevel = HIBYTE((USHORT)flCtrl);

  // save determined flags into global variable
  cstWkd.flCtrl = flCtrl;
  // otherwise: return without error
  return TRUE;
} // WkdGetCtrlFlags()


/*****************************************************************************
W k d G e t S y s t e m
==============================================================================

This function initializes the current function state variable via the
specified specified opened WibuBox entry handle. This function updates also
the specified control flags by using the default flags of the accessed
subsystem.

Parameters
  hwkbsys  contains the handle of the used subsystem.
  flCtrl   contains the control flags of the original function call.

Globals
  cstWkd*

Return
  The function returns TRUE if the subsystem is determined and FALSE if an
  error occurred.

*****************************************************************************/

BOOL WkdGetSystem(HWKBSYSTEM hwkbsys, ULONG flCtrl)
{
  // reset function status
  WkdResetStatus();
  cstWkd.hwkbsys = hwkbsys;
  // analyse type of handle
  if (hwkbsys == HWKB_LOCAL) {
    // local standard system addressed
    cstWkd.psys = &gsysWkdLocal;
  } else {
    // true subsystem: determine pointer & remote handle
    UINT uType = WkdGetHandleType(hwkbsys);
    if (uType != HTYP_SYS) {
      // bad type of handle => return with error
      cstWkd.idError = WKB_ERROR_INVALID_HANDLE;
      return FALSE;
    } // if
    // determine corresponding handle address
    cstWkd.psys = WkdGetHandlePtr(hwkbsys);
    ASSERT(cstWkd.psys != NULL);
    if (!(flCtrl & WKB_EXCLMODE)) {
      // get function standard control flags from standard setting
      // but don't touch flags in <flCtrlExtMask>!
      flCtrl &= ~(STDFNFLAGS & cstWkd.flCtrlDefMask);
      flCtrl |= cstWkd.psys->flStdCtrl & cstWkd.flCtrlDefMask;
    } // if
  } // if
  ASSERT(cstWkd.psys != NULL);
  cstWkd.fuSysType = cstWkd.psys->fuType;
  ASSERT(!(cstWkd.fuSysType & STYPE_WKNET));
  // set control flags in <cstWkd> structure & return status
  return WkdGetCtrlFlags(flCtrl);
} // WkdGetSystem()


/*****************************************************************************
W k d G e t E n t r y S y s t e m
==============================================================================

This function initializes the current function state variable via the
specified specified opened WibuBox entry handle. This function updates also
the specified control flags by using the default flags of the accessed
subsystem.

Parameters
  hwkbe   contains the handle of the opened WibuBox entry.
  flCtrl  contains the control flags of the original function call.

Globals
  cstWkd*

Return
  The function returns TRUE if the subsystem is determined and FALSE if an
  error occurred.

*****************************************************************************/

BOOL WkdGetEntrySystem(HWKBENTRY hwkbe, ULONG flCtrl)
{
  UINT uType = 0;
  HANDLE hwkbsys;
  // validate handle
  uType = WkdGetHandleType(hwkbe);
  if (uType == HTYP_ENTRY) {
    // determine address of WibuBox entry object
    cstWkd.pope = WkdGetHandlePtr(hwkbe);
    ASSERT(cstWkd.pope != NULL);
    // set used system & port number
    hwkbsys = cstWkd.pope->hwkbsys;
    } else

  
  { // bad type of handle => return with error
    // reset function status for proper setting of several flags
    WkdResetStatus();
    cstWkd.idError = WKB_ERROR_INVALID_HANDLE;
    return FALSE;
  } // if
  // determine system from specified system handle
  return WkdGetSystem(hwkbsys, flCtrl);
} // WkdGetEntrySystem()


/*****************************************************************************
W k d R e t u r n
==============================================================================

This function handles the returning of the specified value from a WibuKey
Base API function. If the asynchronous calling mode is active, the return
value and the current error code are stored into the action object and the
action code is returned. If no asynchronous mode is used, the specified value
is returned and the WkbGetLastError() value is updated if an error occurred.

Parameters
  idReturn  contains the original return value of the function.

Globals
  cstWkd*

Return
  The specified return value or the asynchronous action code is returned.

*****************************************************************************/

INT WkdReturn(INT idReturn)
{
  // port may not be accessed longer
  ASSERT(!cstWkd.fAccessed);
#if ASYNCMODE
  if (cstWkd.idError == ASYNCWAIT) {
    // asynchronously called function in waiting mode
    ASSERT(cstWkd.fAsyncCall);
    // return current action code, nothing more to do
    return cstWkd.pasacThis->idAction;
  } else
#endif
  if (cstWkd.idError != WKB_NO_ERROR) {
    #if ASYNCMODE
      // error => remove created action
      if (cstWkd.fAsyncCall) {
        WkdRemoveAction(cstWkd.pasacThis);
      } // if
    #endif
    // set last error code
    gidWkdLastError = cstWkd.idError;
  } else {
    // no error: 0 is returned => reset internal error state
  #if ASYNCMODE
    if (cstWkd.fAsyncCall) {
      // asynchronous execution of already executed action:
      //   save return value into action, return action identification
      cstWkd.pasacThis->idReturn = idReturn;
      cstWkd.pasacThis->idError = cstWkd.idError;
      // return code of action
      idReturn = cstWkd.pasacThis->idAction;
    } // if
  #endif
    if (idReturn == 0) {
      gidWkdLastError = WKB_NO_ERROR;
    } // if
  } // if
  return idReturn;
} // WkdReturn()


/*****************************************************************************
W k d R e t u r n B o o l
==============================================================================

This function handles the returning of the specified value from a WibuKey
Base API function in dependence of the current error state: If no error
occured the return value is TRUE, otherwise FALSE. If the asynchronous calling
mode is active, the return value and the current error code are stored into the
action object and the action code is returned. If no asynchronous mode is
used, the boolean is returned and the WkbGetLastError() value is updated if
an error occurred.

Parameters
  idReturn  contains the original return value of the function.

Globals
  cstWkd*

Return
  The function returns FALSE (0), TRUE (1) or an asynchronous action code.

*****************************************************************************/

INT WkdReturnBool(void)
{
  return WkdReturn(cstWkd.idError == WKB_NO_ERROR);
} // WkdReturnBool()


/*****************************************************************************
W k d R e t u r n C h e c k e d
==============================================================================

This function is identically with WkdReturn() but returns 0 instead of the
specified return value if an error occurred.

Parameters
  idReturn  contains the non-error return value of the function.

Globals
  cstWkd*

Return
  The specified return value, 0 or the asynchronous action code is returned.

*****************************************************************************/

INT WkdReturnChecked(INT idReturn)
{
  if (cstWkd.idError != WKB_NO_ERROR) {
    idReturn = 0;
  } // if
  return WkdReturn(idReturn);
} // WkdReturnChecked()


/*****************************************************************************
W k d R e t u r n L e n g t h
==============================================================================

This function is identically with WkdReturn() but returns 0 instead of the
specified return value if an error except WKB_ERROR_BUFFER_OVERFLOW occurred.

Parameters
  cbReturn  contains the non-error return value of the function.

Globals
  cstWkd*

Return
  The specified return value, 0 or the asynchronous action code is returned.

*****************************************************************************/

INT WkdReturnLength(INT cbReturn)
{
  if (cstWkd.idError != WKB_NO_ERROR &&
      cstWkd.idError != WKB_ERROR_BUFFER_OVERFLOW
    ) {
    cbReturn = 0;
  } // if
  return WkdReturn(cbReturn);
} // WkdReturnLength()


/*****************************************************************************
W k d G e t A p i S t r i n g L e n g t h
==============================================================================

This function checks if the specified address contains a valid string, checks
its length agains a maximum value and returns the length in characters
including the trailing 0-character.

Parameters
  psz ..... points to the string which may be a 0-terminated ASCII or Unicode
            character sequence.
  cchMax .. contains the maximum permitted number of characters including the
            0-terminator.
  fUnicode  contains TRUE if the sting is specified in Unicode and FALSE
            if the string is specified in ASCII.

Globals
  cstWkd*

Return
  The number of characters in the string including the 0-terminator is
  returned. This size can be used to allocate a memory block holding the
  string in ASCII characters. If no string is specified or the string is too
  long, 0 is returned and the WKB_ERROR_BAD_ADDRESS is set.

Comments
* Unicode is only supported in the WKWIN32.DLL. On all other operating
  systems, <fUnicode> must be set to FALSE.
* On Windows, the string is also validated via the IsBadStringPtr()
  Windows API function. This is not done if following situations:
  * The WKWIN32.DLL is used under Win32s or Windows 95 and an Unicode string
    is specified.

*****************************************************************************/

INT WkdGetApiStringLength(const char *psz, UINT cchMax,
  BOOL fUnicode)
{
  UINT cch = 0; // initial: error condition
  ASSERT(cchMax > 0);
  // support outside of Windows: ANSI C standard, no Unicode support
  ASSERT(fUnicode == FALSE);
  cch = (UINT)strlen(psz) + 1;
  // analyse length
  if (cch == 0 || cch > cchMax) {
    // set error code, return 0
    cstWkd.idError = WKB_ERROR_BAD_ADDRESS;
    cch = 0;
  } // if
  // return 0 or determined length (including 0-terminator)
  return cch;
} // WkdGetApiStringLength()


/*****************************************************************************
W k d G e t A p i S t r i n g
==============================================================================

This function transfers a API string specification into the specified
destination address. If the API string is specified in Unicode it is converted
to ASCII.

Parameters
  pszDest   points to a destination buffer which receives always ASCII
            characters. The caller guarantees that the string at <pszSrc>
            can be stored in this buffer.
  pszSrc .. points to a string which may be specified in ASCII or Unicode
            in dependence of the parameter <fUnicode>. This buffer is
            terminated by a 0 character.
  fUnicode  contains TRUE if the string is specified in Unicode and FALSE
            if the string is specified in ASCII.

Comments
  The converting of Unicode to ASCII is only implemented in the WKWIN32.DLL.
  It is implemented by removing the higher byte of an Unicode character. Other
  converting operations are not done.

*****************************************************************************/

void WkdGetApiString(char *pszDest, const char *pszSrc,
  BOOL fUnicode)
{
  // support outside of Windows: ANSI C standard, no Unicode support
  ASSERT(fUnicode == FALSE);
  strcpy(pszDest, pszSrc);
} // WkdGetApiString()


/*****************************************************************************
W k d S e t A p i S t r i n g
==============================================================================

This function transfers a driver internal string to an API string location.
If the API string should be specified in Unicode, it is converted.

Parameters
  pszDest   points to a destination buffer which may be specified in ASCII
            or Unicode in dependence of the parameter <fUnicode>. The caller
            guarantees that the string at <pszSrc> can be stored
            in this buffer.
  pszSrc .. points to a string which is specified always in ASCII.
  cchSrc .. contains the number of characters which are transferred from the
            source without the trailing 0 character.
  fUnicode  contains TRUE if the API string must be specified in Unicode and
            FALSE if the string is specified in ASCII.

Comments
  The converting of ASCII to Unicode is only implemented in the WKWIN32.DLL.
  It is implemented by setting the higher byte of an Unicode character to 0.
  Other converting operations are not done.

*****************************************************************************/

void WkdSetApiString(char *pszDest, const char *pszSrc, INT cchSrc,
  BOOL fUnicode)
{
  // support outside of Windows: ANSI C standard, no Unicode support
  ASSERT(fUnicode == FALSE);
  memcpy(pszDest, pszSrc, (size_t)cchSrc);
  // set explizit trailing 0
  pszDest[cchSrc] = 0;
} // WkdSetApiString()


/*****************************************************************************
W k d G e t S t r u c t A d d r
==============================================================================

This function returns the true address of the specified entry in the structure
It analyses if in the first ULONG entry of the structure the WKB_USEOFFSETS
flag is set. If so, the specified entry value is interpreted as offset value
and calculated to a true parameter value. Otherweise the entry value is
returned unchanged.

Parameters
  pvStruct  points to the begin of the analysed structure
  pvEntry   contains the pointer or offset value of the specified structure
            entry.

Return
  The address specified the true address of the structure entry.

*****************************************************************************/

const void * WkdGetStructAddr(const void *pvStruct,
  const void* pvEntry)
{
  if (*((ULONG*)pvStruct) & WKB_USEOFFSETS) {
    // offsets are specified: calculate true address
    // 0 => return as NULL
    if (pvEntry == NULL) {
      return pvEntry;
    } // if
    return (BYTE*)pvStruct + (ULONG)pvEntry;
  } else {
    // addresses are specified: return unchanged
    return pvEntry;
  } // if
} // WkdGetStructAddr()


/*****************************************************************************
W k d S e t S t r u c t S t r i n g
==============================================================================

This function stores the specified string into the variable part of the
specified structure. The function checks if the specified structure length
is long enough to hold this data. The index of the storing location is also
updated by this function. The function sets also the address or the offset
of the stored string into the specified member field in dependence of the
setting of the WKB_USEOFFSETS flag in <cstWkd.flCtrl>. This function also
assumes that the first entry of the modified structure is <flCtrl> field
which receives the WKB_USEOFFSETS flag if needed.

Parameters
  pvStruct   points to the structure containing the variable part which is
             filled with the string.
  cbStruct   contains the complete length of the structure including the
             variable part. This size is specified from the API caller and
             specifies the maximum structure length.
  pvEntry .. points to the entry in the structure which receives the address
             or the offset of the stored string.
  puOffset   points to the used offset value, relative to the start of the
             structure in <pvStruct>.
  psz ...... points to the string which is stored. This string is always
             specified in ASCII

Globals
  cstWkd.flCtrl, cstWkd.idError

Return
  The function returns TRUE if the string is stored and FALSE if not. The
  offset position is increased always. If the string is not stored, the
  error WKB_ERROR_BUFFER_OVERFLOW is set.

Comments
  The value in <cbStruct> may be less than the fixed part of the structure,
  extremely it may be 0. Thus the setting of the <flCtrl> and the <pvEntry>
  member is checked carefully.

*****************************************************************************/

BOOL WkdSetStructString(void *pvStruct, UINT cbStruct,
  void *pvEntry, UINT *puOffset, const char *psz)
{
  UINT cbPosition = 0;
  UINT cbLength = 0;
  INT cchString = 0;
  INT cbString = 0;
  ASSERT((BYTE*)pvEntry > (BYTE*)pvStruct);
  if (cbStruct >= sizeof(ULONG)) {
    // set valid flags into structure control member
    if (cstWkd.flCtrl & WKB_USEOFFSETS) {
      ASSERT(pvStruct != NULL);
      *(ULONG*)pvStruct |= WKB_USEOFFSETS;
    } // if
    if (cstWkd.fUnicodeCall) {
      ASSERT(pvStruct != NULL);
      *(ULONG*)pvStruct |= WKB_UNICODE;
    } // if
  } // if
  // determine length of string and needed structure length
  cchString = (UINT)strlen(psz);
  cbString = cchString + 1;
  ASSERT(cstWkd.fUnicodeCall == FALSE);
  cbPosition = *puOffset;
  cbLength = cbPosition + cbString;
  ASSERT((BYTE*)pvEntry < (BYTE*)pvStruct + cbPosition);
  // store next offset position independent of data stored or not
  *puOffset = cbLength;
  if (cbLength > cbStruct) {
    // buffer not long enough to hold complete data => return with error
    if ((BYTE*)pvEntry - (BYTE*)pvStruct + sizeof(ULONG*) >= cbStruct) {
      // entry member is valid: store NULL offset/pointer
      // * set "no valid offset"
      *(ULONG*)pvEntry = 0;
    } // if
    cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
    return FALSE;
  } // if
  // copy string at specified address, expand to Unicode if required
  WkdSetApiString((char*)pvStruct + cbPosition, psz, cchString,
    cstWkd.fUnicodeCall);
  // store address or offset into entry
  if (cstWkd.flCtrl & WKB_USEOFFSETS) {
    *(ULONG*)pvEntry = cbPosition;
  } else {
    *(ULONG*)pvEntry = (ULONG)((BYTE*)pvStruct + cbPosition);
  } // if
  return TRUE;
} // WkdSetStructString()


/*****************************************************************************
W k d S t o r e E n u m R e s u l t
==============================================================================

This function saves a result of an WibuKey API enumeration function to the
specified destination address. The 

Parameters
  pszItemList  points to the enumeratino item list. Its format is compatible
               with the WKB_Exx_LIST command: The items are separated by the
               vertical bar ("|").
  pszDest .... points to the destination string. This may also contain the
               start position of the WKB_Exx_NEXT command. This pointer does
               never contain a memory handle address.
  cchDest .... contains the length of the destination buffer in characters.

Globals
  cstWkd.flCtrl, cstWkd.idError

Return
  The function returns the return value of a synchronously called
  WibuKey-API function:
  * If the length in <cchDest> is large enough, the number of stored
    characters is returned without the trailing 0-character.
  * If the length in <cchDest> is not large enough to hold the determined
    string, the number of required characters is returned including the
    trailing 0-character.
  * If the end of the enumeration list is reached or the specified name for
    the WKB_Exx_NEXT command is not found, 0 is returned and <cstWkd.flCtrl>
    contains the error code WKB_ERROR_INVALID_PORT.

Comments
* To return the proper error code when the end of the list is reached, it
  is important, that this error code (WKB_INVALID_PORT_NAME or
  WKB_INVALID_SERVER_NAME) is already stored in <cstWkd.idError> before
  this function is called.
* The command code is stored in <cstWkd.flCtrl> in bits 3 to 7.
  The WKB_Exx_LIST command contains 0x00, the WKB_Exx_NEXT command 0x10 in
  these bits.
* This function supports also Unicode for the Win32 driver.

*****************************************************************************/

INT WkdStoreEnumResult(const char *pszItemList,
  char *pszDest, UINT cchDest)
{
  // return value, initialize with error condition
  size_t cchLength = 0;
  ASSERT(pszItemList != NULL);
  switch (cstWkd.flCtrl & WKB_ESV_MASK) {
    case WKB_ESV_LIST: {
      // determine new string of WkLAN servers (if not cached)
      cchLength = strlen(pszItemList);
      if (cchDest <= cchLength) {
        // destination buffer too small:
        // * return required length (including separator)
        // * one character more for terminator
        cchLength++;
        cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
      } else {
        // write back string (converted to Unicode if required)
        // no error (except WkdSetApiString() produces new error)
        cstWkd.idError = WKB_NO_ERROR;
        WkdSetApiString(pszDest, pszItemList, (UINT)cchLength, cstWkd.fUnicodeCall);
      } // if
      break;
    } // case
    case WKB_ESV_NEXT: {
      char *pszLast = NULL;
      size_t cchLast;
      const char *pszNext;
      cchLast = WkdGetApiStringLength(pszDest, CCH_MAXSTRING,
        cstWkd.fUnicodeCall);
      if (cchLast == 0) {
        // bad string specified => return with error
        goto RETURN_SUBFUNC;
      } // if
      // allocate buffer for specified "last" string
      pszLast = WkdAllocMem(cchLast);
      if (pszLast == NULL) {
        // memory too small => return with error
        goto RETURN_SUBFUNC;
      } // if
      // get string, convert from Unicode if specified
      WkdGetApiString(pszLast, pszDest, cstWkd.fUnicodeCall);
      if (pszLast[0] == 0) {
        // first reading: determine server list, return first server name
        ASSERT(pszItemList != NULL);
        pszNext = pszItemList;
      } else {
        // search specified name
        pszNext = WkdFindEnumName(pszLast, pszItemList);
        if (pszNext == NULL) {
          // specified name not found => return with error
          cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
          goto RETURN_SUBFUNC;
        } // if
        // determine next name
        pszNext += strlen(pszLast);
        if (*pszNext == 0) {
          // end of server list => return with error (is already set)
          ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PORT ||
            cstWkd.idError == WKB_ERROR_SERVER_NOT_FOUND);
          goto RETURN_SUBFUNC;
        } // if
        // set to begin of next name
        pszNext++;
      } // if
      {// search end of next name
        const char *pszEnd = pszNext;
        do {
          pszEnd++;
        } while (*pszEnd != 0 && *pszEnd != '|');
        cchLength = pszEnd - pszNext;
      } // block
      if (cchLength < cchDest) {
        // enough memory in buffer: write new server name + terminator
        // no error (except WkdSetApiString() produces new error)
        cstWkd.idError = WKB_NO_ERROR;
        // convert to Unicode if specified
        WkdSetApiString(pszDest, pszNext, cchLength, cstWkd.fUnicodeCall);
      } else {
        // buffer too small: store number of characters including terminator
        cchLength++;
        cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
      } // if
    RETURN_SUBFUNC:
      // release allocated buffer if any
      BsrFreeMem(pszLast);
      break;
    } // case
    default:
      // specified command does not exist: set error code
      cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
      break;
  } // switch
  // return transferred/required length
  return cchLength;
} // WkdStoreEnumResult()


#if BSR_BIG_ENDIAN || defined(_DEBUG)

/*****************************************************************************

---- Functions to Convert Structures to Little Endian and Big Endian ---------

*****************************************************************************/

/*****************************************************************************
W k d S e t W K B A R E A
==============================================================================

This function converts the members of the WKBAREA structure from the used
endian ordering to the little endian.

Parameters
  pwkbar  points to the WKBAREA structure to be converted.

Comments
  This function does not convert the WKBSELECT structure which is possibly
  addressed by the <pvCtrl> member.

*****************************************************************************/

void WkdSetWKBAREA(WKBAREA *pwkbar)
{
  ULONG flCtrl;
  
#if defined(sun) || defined(SOL_USPARC)
  // fix a allignment issue for sun
  WKBAREA myTemp;
  memcpy(&myTemp, pwkbar, sizeof(WKBAREA) );

  // save system's endian value for later
  flCtrl = myTemp.flCtrl;
  SET_UINT32_LE_INPLACE(&myTemp.flCtrl);
  SET_UINT32_LE_INPLACE(&myTemp.flSelCtrl);
  SET_UINT32_LE_INPLACE(&myTemp.ulSelectCode);
  SET_UINT32_LE_INPLACE(&myTemp.ulCrc);

  // <ulReserved> member not converted
  if (flCtrl & WKB_USEOFFSETS) {
    SET_UINT32_LE_INPLACE((ULONG*)&myTemp.pbGapMap);
    SET_UINT32_LE_INPLACE((ULONG*)&myTemp.pvCtrl);
  } // if

  memcpy(pwkbar, &myTemp, sizeof(WKBAREA) );
#else
  // save system's endian value for later
  flCtrl = pwkbar->flCtrl;
  SET_UINT32_LE_INPLACE(&pwkbar->flCtrl);
  SET_UINT32_LE_INPLACE(&pwkbar->flSelCtrl);
  SET_UINT32_LE_INPLACE(&pwkbar->ulSelectCode);
  SET_UINT32_LE_INPLACE(&pwkbar->ulCrc);

  // <ulReserved> member not converted
  if (flCtrl & WKB_USEOFFSETS) {
    SET_UINT32_LE_INPLACE((ULONG*)&pwkbar->pbGapMap);
    SET_UINT32_LE_INPLACE((ULONG*)&pwkbar->pvCtrl);
  } // if
#endif
} // WkdSetWKBAREA()


/*****************************************************************************
W k d G e t W K B B O X C O N F I G
==============================================================================

This function converts the members of the WKBBOXCONFIG structure from little
endian to the used endian ordering.

Parameters
  pwkbbxc  points to the WKBBOXCONFIG structure to be converted.

*****************************************************************************/

void WkdGetWKBBOXCONFIG(WKBBOXCONFIG *pwkbbxc)
{
  pwkbbxc->flCtrl = GET_UINT32_LE(&pwkbbxc->flCtrl);
  pwkbbxc->usVersion = GET_USHORT_VALUE(&pwkbbxc->usVersion);
  // reserved field <usReserved> not converted
  pwkbbxc->ulProgramCount = GET_UINT32_LE(&pwkbbxc->ulProgramCount);
  pwkbbxc->ulConfig = GET_UINT32_LE(&pwkbbxc->ulConfig);
  if (sizeof(WKBBOXCONFIG) == 20) {
    // change only when new extended WKBBOXCONFIG is used
    pwkbbxc->usAsicBuild = GET_USHORT_VALUE(&pwkbbxc->usAsicBuild);
    pwkbbxc->usExtMemSize = GET_USHORT_VALUE(&pwkbbxc->usExtMemSize);
  } // if
} // WkdGetWKBBOXCONFIG()


/*****************************************************************************
W k d S e t W K B B O X C O N F I G
==============================================================================

This function converts the members of the WKBBOXCONFIG structure from the
used endian ordering to the little endian.

Parameters
  pwkbbxc  points to the WKBBOXCONFIG structure to be converted.

*****************************************************************************/

void WkdSetWKBBOXCONFIG(WKBBOXCONFIG *pwkbbxc)
{
  SET_UINT32_LE_INPLACE(&pwkbbxc->flCtrl);
  SET_USHORT_VALUE_AT(&pwkbbxc->usVersion);
  // reserved field <usReserved> not converted
  SET_UINT32_LE_INPLACE(&pwkbbxc->ulProgramCount);
  SET_UINT32_LE_INPLACE(&pwkbbxc->ulConfig);
  if (sizeof(WKBBOXCONFIG) == 20) {
    // change only when new extended WKBBOXCONFIG is used
    SET_USHORT_VALUE_AT(&pwkbbxc->usAsicBuild);
    SET_USHORT_VALUE_AT(&pwkbbxc->usExtMemSize);
  } // if
} // WkdSetWKBBOXCONFIG()


/*****************************************************************************
W k d G e t W K B B O X D A T A
==============================================================================

This function converts the members of the WKBBOXDATA structure from little
endian to the used endian ordering.

Parameters
  pwkbbxe  points to the WKBBOXDATA structure to be converted.

*****************************************************************************/

void WkdGetWKBBOXDATA(WKBBOXDATA *pwkbbxd)
{
  WKBBOXENTRY *pwkbbxe;
  UINT iEntry;
  // change endian in WKBBOXDATA base structure
  pwkbbxd->flCtrl = GET_UINT32_LE(&pwkbbxd->flCtrl);
  pwkbbxd->usVersion = GET_USHORT_VALUE(&pwkbbxd->usVersion);
  pwkbbxd->cEntry = GET_USHORT_VALUE(&pwkbbxd->cEntry);
  // change now each entry in WKBBOXDATA structure
  pwkbbxe = pwkbbxd->awkbbxe;
  for (iEntry = 0; iEntry < pwkbbxd->cEntry; iEntry++, pwkbbxe++) {
    WkdGetWKBBOXENTRY(pwkbbxe);
  } // for
} // WkdGetWKBBOXDATA()


/*****************************************************************************
W k d G e t W K B B O X E N T R Y
==============================================================================

This function converts the members of the WKBBOXENTRY structure from little
endian to the used endian ordering.

Parameters
  pwkbbxe  points to the WKBBOXENTRY structure to be converted.

*****************************************************************************/

void WkdGetWKBBOXENTRY(WKBBOXENTRY *pwkbbxe)
{
  pwkbbxe->flCtrl = GET_UINT32_LE(&pwkbbxe->flCtrl);
  pwkbbxe->ulFirmCode = GET_UINT32_LE(&pwkbbxe->ulFirmCode);
  pwkbbxe->ulUserCode = GET_UINT32_LE(&pwkbbxe->ulUserCode);
  pwkbbxe->ulData1 = GET_UINT32_LE(&pwkbbxe->ulData1);
  pwkbbxe->ulData2 = GET_UINT32_LE(&pwkbbxe->ulData2);
} // WkdGetWKBBOXENTRY()


/*****************************************************************************
W k d S e t W K B B O X E N T R Y
==============================================================================

This function converts the members of the WKBBOXENTRY structure from the
used endian ordering to the little endian.

Parameters
  pwkbbxe  points to the WKBBOXENTRY structure to be converted.

*****************************************************************************/

void WkdSetWKBBOXENTRY(WKBBOXENTRY *pwkbbxe)
{
  SET_UINT32_LE_INPLACE(&pwkbbxe->flCtrl);
  SET_UINT32_LE_INPLACE(&pwkbbxe->ulFirmCode);
  SET_UINT32_LE_INPLACE(&pwkbbxe->ulUserCode);
  SET_UINT32_LE_INPLACE(&pwkbbxe->ulData1);
  SET_UINT32_LE_INPLACE(&pwkbbxe->ulData2);
} // WkdSetWKBBOXENTRY()


/*****************************************************************************
W k d G e t W K B B O X L I S T
==============================================================================

This function converts the members of the WKBBOXLIST structure from little
endian to the used endian ordering.

Parameters
  pwkbbxl  points to the WKBBOXLIST structure to be converted.
  cbDest   contains the maximum number of bytes which are converted. The
           minimum is the size of the constant part of the structure.

*****************************************************************************/

void WkdGetWKBBOXLIST(WKBBOXLIST *pwkbbxl, UINT cbDest)
{
  WKBSERIAL *pwkbsr = NULL;
  UINT iBox = 0;
  UINT cBox = 0;
  ASSERT(cbDest >= FIELDOFFSET(WKBBOXLIST, awkbsr));
  pwkbbxl->flCtrl = GET_UINT32_LE(&pwkbbxl->flCtrl);
  pwkbbxl->cBox = GET_USHORT_VALUE(&pwkbbxl->cBox);
  // <usReserved> member not converted
  // determine number of stored WibuBox serial numbers in buffer
  cBox = (cbDest - sizeof(WKBBOXLIST)) / sizeof(WKBSERIAL) + 1;
  if (cBox > pwkbbxl->cBox) {
    // no so much WibuBox serial numbers read: reduce counter
    cBox = pwkbbxl->cBox;
  } // if
  pwkbsr = pwkbbxl->awkbsr;
  for (iBox = 0; iBox < cBox; iBox++, pwkbsr++) {
    WkdGetWKBSERIAL(pwkbsr);
  } // for
} // WkdGetWKBBOXLIST()


/*****************************************************************************
W k d G e t W K B O P E N E N T R Y
==============================================================================

This function converts the members of the WKBOPENENTRY structure from little
endian to the used endian ordering.

Parameters
  pwkboe  points to the WKBOPENENTRY structure to be converted.

*****************************************************************************/

void WkdGetWKBOPENENTRY(WKBOPENENTRY *pwkboe)
{
  pwkboe->flCtrl = GET_UINT32_LE(&pwkboe->flCtrl);
#if defined(WK32)
  pwkboe->hwkbsys = (HWKBSYSTEM)GET_UINT32_LE(&pwkboe->hwkbsys);
#else
  // only implemented for 32 bits
  ASSERT(0);
#endif
  pwkboe->lPort = GET_UINT32_LE(&pwkboe->lPort);
  if (pwkboe->flCtrl & WKB_USEOFFSETS) {
    pwkboe->pszPort = (char *)(unsigned long)GET_USHORT_VALUE((UINT*)&pwkboe->pszPort);
  } // if
  WkdGetWKBSERIAL(&pwkboe->wkbsr);
  pwkboe->usVersion = GET_USHORT_VALUE(&pwkboe->usVersion);
  pwkboe->usEntry = GET_USHORT_VALUE(&pwkboe->usEntry);
  pwkboe->ulFirmCode = GET_UINT32_LE(&pwkboe->ulFirmCode);
  pwkboe->ulUserCode = GET_UINT32_LE(&pwkboe->ulUserCode);
} // WkdGetWKBOPENENTRY()


/*****************************************************************************
W k d S e t W K B O P E N E N T R Y
==============================================================================

This function converts the members of the WKBOPENENTRY structure from the used
endian ordering to the little endian.

Parameters
  pwkboe  points to the WKBOPENENTRY structure to be converted.

*****************************************************************************/

void WkdSetWKBOPENENTRY(WKBOPENENTRY *pwkboe)
{
  ULONG flCtrl;
  // save system's endian value for later
  flCtrl = pwkboe->flCtrl;
  SET_UINT32_LE_INPLACE(&pwkboe->flCtrl);
#if defined(WK32)
  SET_UINT32_LE_INPLACE((ULONG*)&pwkboe->hwkbsys);
#else
  // only implemented for 32 bits
  ASSERT(0);
#endif
  SET_UINT32_LE_INPLACE(&pwkboe->lPort);
  if (flCtrl & WKB_USEOFFSETS) {
    SET_UINT32_LE_INPLACE((ULONG*)&pwkboe->pszPort);
  } // if
  WkdSetWKBSERIAL(&pwkboe->wkbsr);
  SET_USHORT_VALUE_AT(&pwkboe->usVersion);
  SET_USHORT_VALUE_AT(&pwkboe->usEntry);
  SET_UINT32_LE_INPLACE(&pwkboe->ulFirmCode);
  SET_UINT32_LE_INPLACE(&pwkboe->ulUserCode);
} // WkdSetWKBOPENENTRY()


/*****************************************************************************
W k d G e t W K B P O R T
==============================================================================

This function converts the members of the WKBPORT structure from little
endian to the used endian ordering.

Parameters
  pwkbprt  points to the WKBPORT structure to be converted.

*****************************************************************************/

void WkdGetWKBPORT(WKBPORT *pwkbprt)
{
  pwkbprt->ulAddress = GET_UINT32_LE(&pwkbprt->ulAddress);
  pwkbprt->fsPortCtrl = GET_USHORT_VALUE(&pwkbprt->fsPortCtrl);
  pwkbprt->fsPortCtrl = GET_USHORT_VALUE(&pwkbprt->fsDrvCtrl);
  // <ulReserved1> & <ulReserved2> members not converted
} // WkdGetWKBPORT()


/*****************************************************************************
W k d S e t W K B P O R T N A M E
==============================================================================

This function converts the members of the WKBPORTNAME structure from little
endian to the used endian ordering.

Parameters
  pwkbprtn  points to the WKBPORTNAME structure to be converted.

*****************************************************************************/

void WkdGetWKBPORTNAME(WKBPORTNAME *pwkbprtn)
{
  pwkbprtn->flCtrl = GET_UINT32_LE(&pwkbprtn->flCtrl);
  pwkbprtn->lPort = GET_UINT32_LE(&pwkbprtn->lPort);
  if (pwkbprtn->flCtrl & WKB_USEOFFSETS) {
    pwkbprtn->pszPort = (char*)GET_UINT32_LE(&pwkbprtn->pszPort);
  } // if
} // WkdGetWKBPORTNAME()


/*****************************************************************************
W k d S e t W K B P R O G R A M
==============================================================================

This function converts the members of the WKBPROGRAM structure from the used
endian ordering to the little endian.

Parameters
  pwkbpgm  points to the WKBPROGRAM structure to be converted.

*****************************************************************************/

void WkdSetWKBPROGRAM(WKBPROGRAM *pwkbpgm)
{
  SET_UINT32_LE_INPLACE(&pwkbpgm->flCtrl);
  WkdSetWKBBOXENTRY(&pwkbpgm->wkbbxe);
} // WkdSetWKBPROGRAM()


/*****************************************************************************
WkdSetWKBEXTMEM
==============================================================================

This function converts the members of the WKBEXTMEM structure
from the used endian ordering to the little endian.

Parameters
  pwkbxmw  points to the WKBEXTMEM structure to be converted.

*****************************************************************************/

void WkdSetWKBEXTMEM(WKBEXTMEM *pwkbxmw)
{
  SET_USHORT_VALUE_AT(&pwkbxmw->fsCtrl);
  SET_USHORT_VALUE_AT(&pwkbxmw->iPage);
} // WkdSetWKBEXTMEM()


/*****************************************************************************
W k d G e t W K B S E L E C T
==============================================================================

This function converts the members of the WKBSLECT structure from little
endian to the used endian ordering.

Parameters
  pwkbsel  points to the WKBSELECT structure to be converted.

*****************************************************************************/

void WkdGetWKBSELECT(WKBSELECT *pwkbsel)
{
  pwkbsel->flCtrl = GET_UINT32_LE(&pwkbsel->flCtrl);
  pwkbsel->ulCacheId = GET_UINT32_LE(&pwkbsel->ulCacheId);
  pwkbsel->ulCrc = GET_UINT32_LE(&pwkbsel->ulCrc);
  pwkbsel->cbInit = GET_USHORT_VALUE(&pwkbsel->cbInit);
} // WkdGetWKBSELECT()


/*****************************************************************************
W k d G e t W K B S E R I A L
==============================================================================

This function converts the members of the WKBSERIAL structure from little
endian to the used endian ordering.

Parameters
  pwkbsr  points to the WKBSERIAL structure to be converted.

*****************************************************************************/

void WkdGetWKBSERIAL(WKBSERIAL *pwkbsr)
{
  pwkbsr->ulSerial = GET_UINT32_LE(&pwkbsr->ulSerial);
  pwkbsr->usMask = GET_USHORT_VALUE(&pwkbsr->usMask);
} // WkdGetWKBSERIAL()


/*****************************************************************************
W k d S e t W K B S E R I A L
==============================================================================

This function converts the members of the WKBSERIAL structure from the used
endian ordering to the little endian.

Parameters
  pwkbsr  points to the WKBSERIAL structure to be converted.

*****************************************************************************/

void WkdSetWKBSERIAL(WKBSERIAL *pwkbsr)
{
  SET_UINT32_LE_INPLACE(&pwkbsr->ulSerial);
  SET_USHORT_VALUE_AT(&pwkbsr->usMask);
} // WkdSetWKBSERIAL()


/*****************************************************************************
W k d G e t W K B S Y S T E M
==============================================================================

This function converts the members of the WKBSYSTEM structure from the used
endian ordering to the little endian.

Parameters
  pwkbprtn  points to the WKBSYSTEM structure to be converted.

*****************************************************************************/

void WkdGetWKBSYSTEM(WKBSYSTEM *pwkbsys)
{
  pwkbsys->flCtrl = GET_UINT32_LE(&pwkbsys->flCtrl);
  pwkbsys->idNetUser = GET_USHORT_VALUE(&pwkbsys->idNetUser);
  pwkbsys->cMaxUser = GET_USHORT_VALUE(&pwkbsys->cMaxUser);
  pwkbsys->cFreeUser = GET_USHORT_VALUE(&pwkbsys->cFreeUser);
  pwkbsys->cbMaxData = GET_USHORT_VALUE(&pwkbsys->cbMaxData);
  pwkbsys->cRefreshError = GET_USHORT_VALUE(&pwkbsys->cRefreshError);
  // <usReserved2> member not converted
  if (pwkbsys->flCtrl & WKB_USEOFFSETS) {
    pwkbsys->pszServerName = (char*)GET_UINT32_LE(&pwkbsys->pszServerName);
  } // if
} // WkdGetWKBSYSTEM()


/*****************************************************************************
WkdSetWKBVERSION
==============================================================================

This function converts the members of the WKBVERSION structure from little
endian to the used endian ordering.

Parameters
  pwkbver  points to the WKBVERSION structure to be converted.

*****************************************************************************/

void WkdSetWKBVERSION (WKBVERSION *pwkbver)
{
  SET_USHORT_VALUE_AT(&pwkbver->usVersion);
  SET_USHORT_VALUE_AT(&pwkbver->usSubVersion);
  SET_USHORT_VALUE_AT(&pwkbver->usBuild);
  SET_USHORT_VALUE_AT(&pwkbver->usCount);
  SET_USHORT_VALUE_AT(&pwkbver->usYear);
  SET_UINT32_LE_INPLACE(&pwkbver->ulReserved1);
} // WkdSetWKBVERSION()

#endif // #if BSR_BIG_ENDIAN || defined(_DEBUG)


// ====================================
// end of clcommon.c
// ====================================
