/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clcrypt.c                                        API Encryption Implementation
==============================================================================

This C source file implements following WibuKey API functions:
  WkbSelect2(), WkbCrypt2(), WkbUnSelect2()

Following driver internally used are also implemented:
  WkdDecryptExpDate()
  WkdUnSelectEntry()

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All POSIX OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
92Sep13 1.33 mabu: start of development
93Mar11 1.33 mabu: version 1.33 released
95Feb01 2.02 mabu: redesigned with parts of internal C driver
95Feb09 2.02 mabu: extracted from KRNL.C
95Jun11 2.06 mabu: indirect encryption moved from kernel to redirector
95Oct27 2.08 mabu: diagnosis event creation added.
96Jun23 2.13 mabu: renamed from BSCRYPT.C
96Aug25 2.14 mabu: API pointer parameter checking
96Aug28 2.14 mabu: implementation of CRC checking & calculation
97Feb26 2.30 mabu: handling of expiration dates and multiopen entries
97Jun11 2.31 mabu: litte/big endian management for Macintosh added
97Jul06 2.31 mabu: synchronous WkNet encryption implemented
97Jul12 2.31 mabu: redesign of calling sema to support WKB_AREA_... options
97Jul12 2.31 mabu: monolithic sent of WkbCrypt2/WKBAREA to server
97Jul12 2.31 mabu: support of indirect encryption for WkNet
97Aug25 2.32 mabu: support of WKB_AREA_CODE for Win16 and Win32
97Oct28 2.40 mabu: setting of level value via remote direct encryption
98Feb07 2.41 mabu: integration of new error codes
98Feb20 2.41 mabu: CRC check for indirect encryption
98Jul13 2.51 mabu: bug in WkbCrypt2() with unselected entry removed
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"


#if defined(__APPLE__) || defined(__linux__)
  // semaphore to synchronize the access to the SELECT & CRYPT
  sem_t* gCryptSemaphoreID = NULL;

  #define EnterCryptAccess()  WkdEnterCS(&gCryptSemaphoreID, "WKXX_CRYPT")
  #define LeaveCryptAccess()  WkdLeaveCS(&gCryptSemaphoreID)
#else
  #define EnterCryptAccess()
  #define LeaveCryptAccess()
#endif


// ====================================
// CRC algorithm: move later into WKLIB
// ====================================

const int cTabEntry = 16;

// use constants in array
static UINT auTab[] = {
  0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
  0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400
}; // auTab[]


/*****************************************************************************
W k l G e n e r a t e C r c
==============================================================================

This function calculates the CRC-16 checksum of the specified block and
returns the value.

Parameters
  pbData  points to the data block for which the CRC checksum is calculated.
  cbData  contains the length of the block in number of bytes.

Return
  The calculated checksum is returned.

*****************************************************************************/

UINT WklGenerateCrc(const BYTE *pbData, UINT cbData)
{
  UINT uCrc = 0;

  ASSERT_ARRAY(pbData, cbData);
  for (;cbData != 0; cbData--, pbData++) {
    // calculate CRC for next byte
    INT i = 0;
    BYTE b;
    b = *pbData;
    i = (uCrc ^ b) & 0x000F;
    uCrc = ((uCrc >> 4) & 0x0FFF) ^ auTab[i];
    b >>= 4;
    i = (uCrc ^ b) & 0x000F;
    uCrc = ((uCrc >> 4) & 0x0FFF) ^ auTab[i];
  } // for
  // return calculated CRC
  return uCrc;
} // WklGenerateCrc()




/******************************************************************************
G e t G a p M a p D a t a
===============================================================================

This function copies all data from the source to the specified destination
which bits are set in the specified gap map. The other bytes are skipped.

Parameters
  pbDest .. points to the memory area which contains the non-gap-data.
  pbSrc ... points to the memory area which contains the source data.
  cbSrc ... contains the length of the source memory area in number of bytes.
  pbGapMap  points to the gap map with the skipped byte information.

Return
  The function returns the number of transferred bytes in the destination.

******************************************************************************/

static UINT GetGapMapData(BYTE *pbDest, const BYTE *pbSrc,
  UINT cbSrc, const BYTE *pbGapMap)
{
  UINT cbDest = 0;
  INT  cGap = 0;
  BYTE bGap = 0;

  // activate gap map skipping
  cGap = 8;
  bGap = *pbGapMap++;
  for (cbDest = 0; cbSrc != 0; cbSrc--, pbSrc++) {
    // transfer or skip next byte
    BOOL fValid = bGap & 1;
    // select next gap map bit, reduce counter
    bGap >>= 1;
    cGap--;
    if (cGap == 0) {
      // set gap counter for next byte, get next gap map byte
      cGap = 8;
      bGap = *pbGapMap++;
    } // if
    if (fValid) {
      // valid bit: transfer byte
      *pbDest++ = *pbSrc;
      // one byte more transferred
      cbDest++;
    } // if
  } // for

  return cbDest;
} // GetGapMapData()


/*****************************************************************************
G e t G a p M a p D a t a L e n g t h
==============================================================================

This function determines the length of the the valid data which is controlled
by the specified gap map.

Parameters
  cbSrc ... contains the length of the source memory area in number of bytes.
  pbGapMap  points to the gap map with the skipped byte information.

Return
  The function returns the number of bytes which is valid by the control
  of the gap map.

*****************************************************************************/

static UINT GetGapMapDataLength(INT cbSrc, const BYTE *pbGapMap)
{
  UINT cbDest = 0;
  INT cGap = 0;
  BYTE bGap;

  // activate gap map skipping
  cGap = 8;
  bGap = *pbGapMap++;
  for (cbDest = 0; cbSrc != 0; cbSrc--) {
    // transfer or skip next byte
    BOOL fValid = bGap & 1;
    // select next gap map bit, reduce counter
    bGap >>= 1;
    cGap--;
    if (cGap == 0) {
      // set gap counter for next byte, get next gap map byte
      cGap = 8;
      bGap = *pbGapMap++;
    } // if
    if (fValid) {
      // valid bit: increment counter
      cbDest++;
    } // if
  } // for

  return cbDest;
} // GetGapMapDataLength()


/*****************************************************************************
S e t G a p M a p D a t a
==============================================================================

This function copies all data from the source to the specified destination
and skips all bytes in the destination which bits are not set in the specified
gap map.

Parameters
  pbDest .. points to the memory area which receives the data..
  pbSrc ... points to the memory area which contains the source data without
            gaps.
  cbSrc ... contains the length of the source memory area wihtout the gap
            information in number of bytes.
  pbGapMap  points to the gap map with the skipped byte information.

*****************************************************************************/

static void SetGapMapData(BYTE *pbDest, const BYTE *pbSrc,
  UINT cbSrc, const BYTE *pbGapMap)
{
  UINT cbDest;
  INT cGap = 0;
  BYTE bGap;
  // activate gap map skipping
  cGap = 8;
  bGap = *pbGapMap++;
  for (cbDest = 0; cbSrc != 0; cbDest++, pbDest++) {
    // transfer or skip next byte
    BOOL fValid = bGap & 1;
    // select next gap map bit, reduce counter
    bGap >>= 1;
    cGap--;
    if (cGap == 0) {
      // set gap counter for next byte, get next gap map byte
      cGap = 8;
      bGap = *pbGapMap++;
    } // if
    if (fValid) {
      // valid bit: transfer byte
      *pbDest = *pbSrc++;
      // one byte less in the source
      cbSrc--;
    } // if
  } // for
} // SetGapMapData()



/*****************************************************************************

---- Kernel specific function for Selection and Encryption -------------------

*****************************************************************************/

/*****************************************************************************
U p d a t e O p e n B o x D a t a
==============================================================================

This function rereads the WibuBox contents for the opened WibuBox entry
and determines the used Expiration Date and the used Limit Counter entry.

Parameters
  pope  points to the OPENENTRY structure which is used and modified

The function returns TRUE if the operation was sucessful and FALSE if not.
In the latter case, an error code is stored in <cstWkd.idError>.

*****************************************************************************/

static BOOL UpdateOpenBoxData(OPENENTRY *pope)
{
  INT idLastError = 0;
  BOOL fValid = FALSE;

  // save current error code (if any, is overwritten by WibuBox low select)
  idLastError = cstWkd.idError;
  if (pope->fMultiFind) {
    // ==========================================================
    // multiopen mode: read complete contents of WibuBox & check
    // ==========================================================
    WKBBOXDATA *pwkbbxd = NULL;
    UINT cbData = 0;
    
    // determine WibuBox contents
    ASSERT(cstWkd.fAccessed);
    
    pwkbbxd = WkdGetBoxContents(&pope->wkbsr, pope->fuEntryOpen, &cbData);
    if (pwkbbxd == NULL) {
      // any error => return
      ASSERT(fValid == FALSE);
      goto RETURN;
    } // if
    // update selection of Expiration Date & Limit Counter entry
    WkdSetMultiFindOptions(pwkbbxd, pope);
    // free allocated memory
    BsrFreeMem(pwkbbxd);
    // operation successfully executed
    fValid = TRUE;
  } else {
    // =============================================
    // analyse only Expiration Date for single entry
    // =============================================
    WKBBOXENTRY wkbbxe;
    INT nVersion = 0;
    ASSERT(cstWkd.fAccessed);
    if (!WkdGetOpenBoxEntry(&pope->wkbsr, pope->iEntry, &wkbbxe, &nVersion, FALSE)) {
      // any error => return with error
      ASSERT(fValid == FALSE);
      goto RETURN;
    } // if
    // analyse read entry
    if (wkbbxe.flCtrl & WKB_BE_ADDEXPDATE) {
      // get value & store
      UINT uExpDate = LOUSHORT(wkbbxe.ulData2);
      // calculate absolute date
      if (uExpDate == USHRT_MAX) {
        // out of date: set to 0
        uExpDate = 0;
      } else {
        // otherwise: add relative value to current date
        uExpDate += BsrGetCurrentDateCount();
      } // if
      pope->uExpDate = uExpDate;
    } else {
      // no expiration date set
      pope->uExpDate = NO_EXPDATE;
    } // if
    // operation successfully executed
    fValid = TRUE;
  } // if
RETURN:
  if (idLastError != WKB_NO_ERROR) {
    // function called with any error: set this again
    // otherwise: if any new error => hold this
    cstWkd.idError = idLastError;
  } // if
  // return status of this function

  return fValid;
} // UpdateOpenBoxData()


/*****************************************************************************
S e l e c t D i r e c t K r n l
==============================================================================

This function implements the direct selection in a kernel driver.

Parameters
  ulSelectCode  contains the Selection Code.

Comments
  This function doesn't return a valid state. Instead of, <cstWkd.idError>
  contains WKB_NO_ERROR (0) if the operation occurred successfully or contains
  an error code if the operation failed.

*****************************************************************************/

static void SelectDirectKrnl(ULONG ulSelectCode)
{
  BOOL fSlctVersion2 = FALSE;// flag "use encryption generator version 2 or 3"
  BOOL fBoxRead = FALSE;     // flag "new read of WibuBox contents done"
  BOOL fRetry = FALSE;       // flag "retry selection operation"
  BOOL fBoxVersion1 = FALSE; // flag "WibuBox has version 1"
  DEVOBJECT *pdevobj = NULL; // used device object
  OPENENTRY *pope = cstWkd.pope; // pointer to locally opened entry

  // force level 1 if not already done
  // * this avoids ignoring of Expiration Date
  if (cstWkd.bLevel < 1) {
    cstWkd.bLevel = 1;
  } // if
  // check parameters
  ASSERT_VALUE(pope);
  // flag is true for algorithm 2 and 3
  fSlctVersion2 = (pope->flCtrl & WKB_VERSION_MASK) != WKB_VERSION1;

  // flag is true for WibuBox version 1
  fBoxVersion1 = pope->usBoxVersion == 1;
  if ((ulSelectCode >= 65536) && !fSlctVersion2) {
    // WibuBox generator 1 supports only 2-byte-selection codes
    cstWkd.idError = WKB_ERROR_INVALID_SELECTION;
    goto RETURN;
  } // if

  // WkbOpen2() garantees that WibuBox version does not use selection 2
  ASSERT(!fSlctVersion2 || !fBoxVersion1);
  if ((ulSelectCode & WKB_SC_NETLIMIT) && (pope->flCtrl & WKB_OE_NONETLIMIT)) {
    // WKB_SC_NETLIMIT flag not permitted with WKB_OPEN_NONETLIMIT option
    cstWkd.idError = WKB_ERROR_INVALID_SELECTION;
    goto RETURN;
  } // if

  // check the station share flag
  if (fSlctVersion2) {
    // check SC for StationShare only for version > WKB_VERSION1
    if ((pope->flCtrl & WKB_OPEN_STATIONSHARE) == WKB_OPEN_STATIONSHARE) {
      if ((ulSelectCode & WKB_SC_STATIONSHARE) != WKB_SC_STATIONSHARE) {
        // if WKB_OPEN_STATIONSHARE is set WKB_SC_STATIONSHARE must also be set
        // ==>> if not return with error
        cstWkd.idError = WKB_ERROR_INVALID_SELECTION;
        goto RETURN;
      } // if
    } // if
  } // if

  if (WkdIsBoxSelected(&pope->wkbsr)) {
    // WibuBox selected by another entry => return error
    cstWkd.idError = WKB_ERROR_ALREADY_SELECTED;
    goto RETURN;
  } // if

  // access port of used entry
  if (!WkdSelectPort((INT)cstWkd.pope->lPort, TRUE)) {
    // port cannot be accessed => return with error
    cstWkd.idError =  WKB_ERROR_INVALID_SELECTION;
    goto RETURN;
  } // if

  if (ulSelectCode & WKB_SC_EXPDATE) {
    // Expiration Date must be checked: check if reread is required
    UINT uExpDate = pope->uExpDate;
    if (uExpDate == UNDET_EXPDATE) {
      // expiration date not determined => update it, return if error
      if (!UpdateOpenBoxData(pope)) {
        goto RETURN_KRNL;
      } // if
      // actual WibuBox contents read => no retry
      fBoxRead = TRUE;
      uExpDate = pope->uExpDate;
    } // if
    ASSERT(uExpDate != UNDET_EXPDATE);
    if (uExpDate < BsrGetCurrentDateCount()) {
      // out of date: set specific error code & return
      cstWkd.idError = WKB_ERROR_OUT_OF_DATE;
      goto RETURN_KRNL;
    } // if
  } // if
  do {
    // do loop until suitable WibuBox entry found
    // * only if WKB_OPEN_MULTIENTRY
    struct {
      WKDEV_REQUEST wkdevreq;
      BYTE ab[8 - 1];
    } s;
    BYTE abAnswer[3];
    // no retry expected
    fRetry = FALSE;
    // execute via class driver
    // * search corresponding WibuBox
    pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(&pope->wkbsr);
    if (pdevobj == NULL) {
      // device not found => return with error
      goto RETURN_KRNL;
    } // if
    memset(&s, 0, sizeof(s));
    s.wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
    s.wkdevreq.usIndex = pope->iEntry;
    if (fBoxVersion1) {
      // create SELECT1 command
      s.wkdevreq.usCmd = BOXCMD_SELECT1;
      s.wkdevreq.cbRequest = 2;
      s.wkdevreq.cbAnswer = 1;
      SET_USHORT_VALUE(s.wkdevreq.abRequest, (USHORT)ulSelectCode);
    } else {
      // create SELECT2 command
      s.wkdevreq.usCmd = BOXCMD_SELECT2;
      s.wkdevreq.cbRequest = fSlctVersion2 ? 8 : 6;
      s.wkdevreq.cbAnswer = fSlctVersion2 ? sizeof(abAnswer) :
        sizeof(abAnswer) - 1;
      // set the control flag (= WKB_VERSIONX)
      s.wkdevreq.abRequest[0] = LBYTE32(pope->flCtrl & WKB_VERSION_MASK);
      // set the user code to match a master entry
      s.wkdevreq.abRequest[1] = LBYTE32(pope->ulUserCode);
      s.wkdevreq.abRequest[2] = MBYTE32(pope->ulUserCode);
      s.wkdevreq.abRequest[3] = NBYTE32(pope->ulUserCode);
      // set the selection code
      SET_UINT32_LE(s.wkdevreq.abRequest + 4, ulSelectCode);
    } // if
    ASSERT_VALUE(cstWkd.pclsdrvSel);
  #if !BSR16
    ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
  #endif
    if (!cstWkd.pclsdrvSel->pfClDrvRequest(&s.wkdevreq, abAnswer, NULL)) {
      // request not properly executed => return with error
      goto RETURN_KRNL;
    } // if
    // analyse error, if any => set into <cstWkd.idError>
    if (abAnswer[0] != 0) {
      WkdConvertStatusIntoError(abAnswer[0]);
    } else if (!fBoxVersion1) {
      // WibuBox version 2: check also 2nd and 3d status byte
      if (abAnswer[1] != 0) {
        WkdConvertStatusIntoError(abAnswer[1]);
      } else if (fSlctVersion2 && abAnswer[2] != 0) {
        WkdConvertStatusIntoError(abAnswer[2]);
      } // if
    } // if
    // check error status of executed command
    if (cstWkd.idError == WKB_ERROR_BOXLIMIT_IS_ZERO) {
      // special handling for Limit Counter is zero
      if (pope->fMultiFind && !fBoxRead) {
        // reread WibuBox contents, use another Limit Counter
        SHORT iLastSelEntry = pope->iEntry;
        WkdReleasePort();
        if (UpdateOpenBoxData(pope) && pope->iEntry != iLastSelEntry) {
          // new Limit Counter candidate => retry operation
          fRetry = TRUE;
          continue;
        } // if
      } // if
    } else if (cstWkd.idError != WKB_NO_ERROR) {
      // any other error => return with error
      goto RETURN_KRNL;
    } // if
  } while (fRetry);

  // now this entry is selected
  cstWkd.pope->flCtrl |= WKB_OE_SELECTED;
  // return from function: release selected WIBU-BOX port
RETURN_KRNL:
  WkdReleasePort();
RETURN:
  return;
} // SelectDirectKrnl()


/*****************************************************************************
C r y p t D i r e c t K r n l
==============================================================================

This function encrypts or decrypts the specified WibuBox data in the kernel
driver.

Parameters
  pbDest   points to the data buffer which receives the encrypted
           or decrypted data. If this pointer is NULL, the data is not
           stored. Otherwise it is assumed that this pointer addresses
           always writable memory.
  pbSrc .. points to the data area which contents is encrypted or
           decrypted.
  cbData   contains the length of the area to be encrypted.
  pcbDest  points to a UINT variable which receives the number of
           encrypted bytes. If this pointer is NULL, no number is stored.

Comments
  This function does not return a valid state. Instead of, <cstWkd.idError>
  contains WKB_NO_ERROR (0) if the operation occurred successfully or contains
  an error code if the operation failed.
  If <pcbDest> is specified, always a value is stored, also in any case of
  error.

*****************************************************************************/

static void CryptDirectKrnl(BYTE *pbDest, BYTE *pbSrc, UINT cbData,
  UINT *pcbDest)
{
  DEVOBJECT *pdevobj = NULL;
  WKDEV_REQUEST *pwkdevreq = NULL;
  OPENENTRY *pope = NULL; // pointer to used open entry
  UINT cbCopy = 0;        // rest of data to encrypt/transfer
  UINT cbEncrypted = 0;   // number of properly encrypted bytes
  BYTE *pabAnswer = NULL; // pointer to answer data, shared with request data

  pope = cstWkd.pope;
  // access port of current WibuBox entry
  if (!WkdSelectPort((INT)cstWkd.pope->lPort, TRUE)) {
    // port cannot be accessed => return with error
    return;
  } // if
  // validate that entry is logically selected
  if (!(pope->flCtrl & WKB_OE_SELECTED)) {
    // not selected => set error code & return
    cstWkd.idError = WKB_ERROR_NOT_SELECTED;
    goto EXIT_CRYPT;
  } // if
  // execute encryption via WibuBox
  // * maximum length of one encryption part is 256 bytes
  cbCopy = cbData;
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
  // execute via class driver
  // * search corresponding WibuBox
  ASSERT_VALUE(cstWkd.pclsdrvSel);
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
  
  pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(&pope->wkbsr);
  if (pdevobj == NULL) {
    // device not found => return with error
    goto EXIT_CRYPT;
  } // if
  ASSERT_VALUE(pdevobj);
  // allocate memory for request data block
  pwkdevreq = WkdAllocMem(FIELDOFFSET(WKDEV_REQUEST, abRequest) +
    cstWkd.pclsdrvSel->cbMaxCrypt + 1);
  if (pwkdevreq == NULL) {
    // memory too small => return with error
    goto EXIT_CRYPT;
  } // if
  // set shared pointer to request data
  pabAnswer = pwkdevreq->abRequest;
  for (;;) {
    UINT cbBlock = 0;
    REQOPTIONS reqopt;
    // get length of rest block
    cbBlock = cbData;
    if (cbBlock > cstWkd.pclsdrvSel->cbMaxCrypt) {
      // limit encryption length to maximum supported block length
      cbBlock = cstWkd.pclsdrvSel->cbMaxCrypt;
    } // if
    // init base part of structure
    memset(pwkdevreq, 0, (size_t)FIELDOFFSET(WKDEV_REQUEST, abRequest));
    pwkdevreq->hwkdev = pdevobj->wkdevobj.hwkdev;
    pwkdevreq->usCmd = BOXCMD_CRYPT;
    pwkdevreq->usIndex = pope->iEntry;
    pwkdevreq->cbRequest = cbBlock + 1;
    pwkdevreq->cbAnswer = cbBlock + 1;
    pwkdevreq->abRequest[0] = (BYTE)cbBlock;
    memcpy(pwkdevreq->abRequest + 1, pbSrc, (size_t)cbBlock);
    pbSrc += cbBlock;
    ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
    // initalize number of encrypted bytes for error state to 0
    reqopt.Crypt.cbEncrypted = 0;
    if (!cstWkd.pclsdrvSel->pfClDrvRequest(pwkdevreq, pabAnswer, &reqopt)) {
      // request not properly executed => return with error
      goto EXIT_CRYPT;
    } // if
    // analyse error, if any => set into <cstWkd.idError>,
    if (pabAnswer[0] != 0) {
      WkdConvertStatusIntoError(pabAnswer[0]);
      // reduce block size to properly transferred bytes (0 is default)
      cbBlock = reqopt.Crypt.cbEncrypted;
    } // if
    if (pbDest != NULL) {
      // store back received data
      memcpy(pbDest, pabAnswer + 1, (size_t)cbBlock);
      pbDest += cbBlock;
    } // if
    cbEncrypted += cbBlock;
    cbData -= cbBlock;
    if (cbData == 0 || cstWkd.idError != WKB_NO_ERROR) {
      // all data encrypted or error => exit loop
      goto EXIT_CRYPT;
    } // if
  } // for
EXIT_CRYPT:
  // release port if not already done
  WkdReleasePort();
  // store number of transferred bytes
  if (pcbDest != NULL) {
    *pcbDest = cbEncrypted;
  } // if
  // free allocated memory
  BsrFreeMem(pwkdevreq);
} // CryptDirectKrnl()


/*****************************************************************************

---- Intermediate Level Support Functions ------------------------------------

*****************************************************************************/

/*****************************************************************************
S e l e c t D i r e c t
==============================================================================

This function implements the direct selection of a WibuBox entry. In the
calling driver, a datagram is sent to the kernel driver or sent.

In the kernel driver, the SelectDirectKrnl() function is called instead.

Parameters
  ulSelectCode  contains the Selection Code.

Comments
  This function does not return a valid state. Instead of, <cstWkd.idError>
  contains WKB_NO_ERROR (0) if the operation occurred successfully or
  contains an error code if the operation failed.

  Please note that the direct selection does not support any member in the
  optional WKBSELECT structure. That's why this parameter is not transferred
  to this function.

*****************************************************************************/

static void SelectDirect(ULONG ulSelectCode)
{



  {// execute direct selection in kernel
    SelectDirectKrnl(ulSelectCode);
  } // block/if
} // SelectDirect()


/*****************************************************************************
W k d U n S e l e c t E n t r y
==============================================================================

This function unselects the current selected entry. If the entry is
indirectly encrypted, the encryption object is destroyed. Otherwise, it is
assumed that the entry is selected for direct encryption. In this case, the
redirector sends the UNSELECT command to the kernel and the kernel driver
resets the <fSelected> flag in the open entry structure.

Comments
  This function doesn't return a valid state. Instead of, <cstWkd.idError>
  contains WKB_NO_ERROR (0) if the operation occurred successfully or contains
  an error code if the operation failed.

*****************************************************************************/

void WkdUnSelectEntry(void)
{

  {//// execute local driver operations
    ASSERT(cstWkd.pope != NULL);
    // entry no longer directly encrypted
    cstWkd.pope->flCtrl &= ~WKB_OE_SELECTED;
  } // if/block
} // WkdUnSelectEntry()


/*****************************************************************************
W k d S t o r e C r y p t D a t a
==============================================================================

This function stores the encrypted data bytes from a buffer to the
destination. It handles different memory options (code or data), the
WKB_AREA_NONE option which inhibits any storing operation and the ignoring
of bytes which are stored in a gap table.

Parameters
  pbDest .... points to a buffer which receives the read data.
  pbSrc ..... points to a buffer which contains the data to copy.
  cbSrc ..... contains the number of data at <pbSrc> which are encrypted
              and should be transferred to <pbDest>, possibly between gap-map-
              controlled skipping data.
  cbDest .... contains the number of data in the destination. If a gap map
              is determined, this number specifies how many data are
              transferred.
  fuDestCtrl  contains the flags which control the storing of the encrypted
              data. These are one of the WKB_AREA_MASK flags, WKB_AREA_NONE
              or WKB_CRYPT_BUFFER (ignored).
  pbGapMap .. points to a gap table which controls the bytes at <pbDest>
              which are overwritten or skipped. May be NULL.
  idError ... contains the error code which was determined during the direct
              encryption. WKB_NO_ERROR (0) means that the data are read
              correctly.

Return
  The function returns the number of destination data, including the gap map
  based skipped data.

Comments
  It is assumed that <pbDest> can hold <cbSrc> bytes, also after skipping
  some bytes by analyzing the information in the <pbGapTable>.

*****************************************************************************/

UINT WkdStoreCryptData(BYTE *pbDest, const BYTE *pbSrc, UINT cbSrc,
  UINT cbDest, UINT fuDestCtrl, const BYTE *pbGapMap, INT idError)
{
  UINT uMemType = 0;
  UINT cbData = 0;
  if (fuDestCtrl & WKB_AREA_NONE) {
    // nothing to stored: do nothing & return destination length
    return pbGapMap != NULL? cbDest : cbSrc;
  } // if
  // determine memory type
  uMemType = fuDestCtrl & WKB_AREA_MASK;
  if (pbGapMap != NULL) {
    if (idError != WKB_NO_ERROR) {
      // gap map and any error: don't transfer data, return 0 bytes
      cbData = 0;
    } else {
      // store back data & skip gap bytes
      SetGapMapData(pbDest, pbSrc, cbSrc, pbGapMap);
      // return destination length (all valid)
      cbData = cbDest;
    } // if
  } else {
    // store back source data stream directly
    memcpy(pbDest, pbSrc, (size_t)cbSrc);
    // return truely transferred data
    ASSERT(cbSrc == cbDest || idError != WKB_NO_ERROR);
    cbData = cbSrc;
  } // if
  // return number of transferred data
  return cbData;
} // WkdStoreCryptData()



/*****************************************************************************
C r y p t D i r e c t
==============================================================================

This function encrypts or decrypts the specified WibuBox data by the
direct encryption.

In a calling driver, a datagram is sent to the kernel driver or server.
In a kernel driver, the CryptDirectKrnl() function is called instead. If
a WKBAREA structure is specified and the option WKB_AREA_SELECT is set, also
the SelectDirectKrnl() and WkbUnSelectLocal() functions are called before
resp. after the encryption.

Parameters
  pbDest .... points to the data buffer which receives the encrypted data.
  pbSrc ..... points to the data area which contents is encrypted.
  cbData .... contains the length of the area to be encrypted.
  pwkbar .... points to a valid WKBAREA control structure or is NULL.
              The WKB_AREA_MASK and WKB_AREA_NONE flags are ignored in
              <pwkbar->flCtrl> because they are stored in <fuDestCtrl>..
  fuDestCtrl  contains the flags which control the storing of the encrypted
              data. These are one of the WKB_AREA_MASK flags, WKB_AREA_NONE
              or WKB_CRYPT_BUFFER.
  pcbDest ... points to a UINT variable which receives the number of
              encrypted bytes. If this pointer is NULL, no number is stored.

Comments
  This function is never called for an indirection selection/encryption.
  That's why the <pvCtrl> member in an optinal WKBAREA structure must be
  set to NULL and the <flSelCtrl> member must be set to 0.

  The gap table for a direct encryption is handled always in a buffer:
  * In dependence of the gap bits, only bytes which are candiates for
    an encryption are moved into the local buffer.
  * This buffer (reduced in the length) is encrypted directly why sending
    bytes to the WibuBox or is sent to the remove driver.
  * When the encrypted data are received, they are stored in dependence
    of the gap bits into the destination buffer.

  This function does not return a valid state. Instead of, <cstWkd.idError>
  contains WKB_NO_ERROR (0) if the operation occurred successfully or contains
  an error code if the operation failed.

*****************************************************************************/

static void CryptDirect(BYTE *pbDest, BYTE *pbSrc, UINT cbData,
  WKBAREA *pwkbar, UINT fuDestCtrl, UINT *pcbDest)
{
  UINT fuAreaCtrl = 0; // WKBAREA control (not WKB_AREA_MASK or WKB_AREA_NONE)
  const BYTE *pbGapMap = NULL;
  if (pwkbar != NULL) {
    // determine area control flags & gap map address
    fuAreaCtrl = (UINT)pwkbar->flCtrl;
    pbGapMap = WkdGetStructAddr(pwkbar, pwkbar->pbGapMap);
  } // if
  {//// execute local driver operations (only direct encryption)
    BOOL fBuffer = FALSE;      // buffer used for transfer
    BOOL fTransfer = TRUE;     // flag "transfer data truly to destination"
    BYTE *pabBuffer = NULL;    // buffer for temporary encryption
    BYTE *pbTempSrc = pbSrc;   // pointer for original/buffered destination
    BYTE *pbTempDest = pbDest; // pointer for original/buffered source
    UINT cbTransfer = cbData;  // true encrypted data (between gaps)
    UINT cbEncrypted = 0;      // number of encrypted bytes
    UINT uNewCrc = 0;          // CRC of source data (valid if WKB_AREA_CALCCRC set)
    if (fuDestCtrl & WKB_AREA_NONE) {
      // inhibit data transfer to destination
      fTransfer = FALSE;
    } // if
    if (fuAreaCtrl & WKB_AREA_SELECT) {
      // select WibuBox entry for direct encryption at kernel level
      WkdUnSelectEntry();
      // only direct encryption is permitted here
      ASSERT((pwkbar->flSelCtrl & WKB_SEL_MASK) == WKB_SEL_DIRECT);
      SelectDirectKrnl(pwkbar->ulSelectCode);
      if (cstWkd.idError != WKB_NO_ERROR) {
        // any error => abort
        return;
      } // if
    } // if
    if (pbGapMap) {
      // determine length of valid data between gaps
      cbTransfer = GetGapMapDataLength(cbData, pbGapMap);
    } // if
    // determine buffer state:
    // * valid if explictly set, CRC check/calculation or gap map
    if (fuDestCtrl & WKB_CRYPT_BUFFER ||
        fuAreaCtrl & (WKB_AREA_CHKCRC|WKB_AREA_CALCCRC) ||
        pbGapMap != NULL
      ) {
      // allocate buffer with <cbTransfer> bytes, maximum size is 4096 bytes
      fBuffer = TRUE;
      if (cbTransfer > 4096) {
        // * return special error code (since version V2.51)
        cstWkd.idError = WKB_ERROR_DATA_TOO_COMPLEX;
        return;
      } // if
      if (cbTransfer != 0) {
        // required length not 0: create buffer
        pabBuffer = WkdAllocMem(cbTransfer);
        if (!pabBuffer) {
          // memory not allocated => return with error
          return;
        } // if
      } // if
      // preliminary destination is temporary buffer
      pbTempDest = pabBuffer;
      if (pbGapMap) {
        // transfer source data to buffer (remove gap-data)
        GetGapMapData(pabBuffer, pbSrc, cbData, pbGapMap);
        pbTempSrc = pabBuffer;
      } // if
    } // if
    if (fuAreaCtrl & (WKB_AREA_CALCCRC | WKB_AREA_CHKCRC)) {
      // calculate source CRC value
      uNewCrc = WklGenerateCrc(pbTempSrc, cbTransfer);
    } // if
    // execute direct encryption/decryption at kernel level
    CryptDirectKrnl(fTransfer? pbTempDest : NULL,
      pbTempSrc, cbTransfer, &cbEncrypted);
    // analyse encrypted data
    if (cstWkd.idError == WKB_NO_ERROR &&
        fuAreaCtrl & WKB_AREA_CHKCRC && fTransfer
      ) {
      // compare CRC of encrypted/decrypted result with stored value
      ASSERT(pwkbar != NULL);
      if (pwkbar->ulCrc != WklGenerateCrc(pbTempDest, cbEncrypted)) {
        // difference in CRC value: set error code
        cstWkd.idError = WKB_ERROR_CRC_VERIFY;
        cbEncrypted = 0;
      } // if
    } // if
    if (fuAreaCtrl & WKB_AREA_CALCCRC) {
      // save calculated CRC value into WKBAREA structure
      ASSERT(pwkbar != NULL);
      pwkbar->ulCrc = uNewCrc;
    } // if
    if (fBuffer) {
      // store back data from temporary buffer to true destination
      cbEncrypted = WkdStoreCryptData(pbDest, pbTempDest, cbEncrypted, cbData,
        fuDestCtrl, pbGapMap, cstWkd.idError);
    } // if
  // EXIT_ENCRYPT:
    if (fBuffer) {
      // free memory of buffer (pabBuffer may be NULL)
      if (pabBuffer != NULL) {
          BsrFreeMem(pabBuffer);
      } // if
    } // if
    if (fuAreaCtrl & WKB_AREA_SELECT) {
      // unselect used WibuBox entry
      INT idError = 0;
      // save current error code (don't change by UnSelectLocal())
      idError = cstWkd.idError;
      // always unselect of WibuBox entry
      WkdUnSelectEntry();
      // return state of encryption/decryption
      cstWkd.idError = idError;
    } // if
    // save number of valid saved bytes
    if (pcbDest != NULL) {
      *pcbDest = cbEncrypted;
    } // if
  } // if/block
} // CryptDirect()




/*****************************************************************************
W k b S e l e c t 2
==============================================================================

This function selects a opened WibuBox entry for encryption or decryption
with the specified selection data.

Parameters
  hwkbe ....... contains the handle of an opened WibuBox entry.
  flCtrl ...... contains flags which control the execution of the function.
                Following flags are currently supported:
                  WKB_SEL_DIRECT   (0x0000)  direct encryption/decryption
                  WKB_SEL_KNUTH20  (0x0010)  indirect Knuth encrypt/decryption
                  WKB_SEL_DEC_FEAL (0x0020)  indirect original FEAL decryption
                  WKB_SEL_ENC_FEAL (0x0030)  indirect original FEAL encryption
                  WKB_SEL_DEC_PERM (0x0040)  indirect permutation decryption.
                  WKB_SEL_ENC_PERM (0x0050)  indirect permutation encryption.
                  WKB_SEL_SYMFEAL  (0x0060)  indirect symmetric FEAL encr/decr
  ulSelectCode  contains the selection code. If the WibuBox entry was opened
                for WibuKey 1, the code must be in range 0..65535. Otherwise
                the complete 32 bit range can be set. Bit 30 and 31 are used
                for special purposes:
                  WKB_SC_STATIONSHARE (bit 28) checks the station share
                                  option.
                  WKB_SC_EXPDATE  (bit 29) checks the expiration date.
                  WKB_SC_NETLIMIT (bit 30) touches the user limitation in a
                                  WkNet or WkLAN subsystem.
                  WKB_SC_BOXLIMIT (bit 31) reduces the limit counter of the
                                  entry by 1 if such an counter exists.
  pvCtrl ...... points to optional control data for the selection. This
                parameter is used only for the selection of an indirect
                encryption or decryption and must be NULL for WKB_SEL_DIRECT.
                This data is stored in a WKBSELECT structure which must not
                be stored as a constant aggregate because some members are
                updated by the WkbSelect2() call.

Return
   The function returns TRUE if the operation was sucessful and FALSE if an
   error occured. In this case a subsequent call of the function
   WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbSelect2(HWKBENTRY hwkbe, ULONG flCtrl,
  ULONG ulSelectCode, void *pvCtrl)
{

  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pvCtrl)) {
    goto RETURN;
  } // if
  if (WkdGetEntrySystem(hwkbe, flCtrl)) {
    // unselect entry if any selection from earlier
    WkdUnSelectEntry();
    // execute kernel of operation
    if ((flCtrl & WKB_SEL_MASK) == WKB_SEL_DIRECT) {
      // execute direct selection
      if (pvCtrl != NULL) {
        // control structure not permitted with WKB_SEL_DIRECT
        cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
        goto RETURN;
      } // if
      SelectDirect(ulSelectCode);
    } else {
      // indirect encryption not implemented
      cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
    } // if
  } // if
RETURN:
  return WkdReturnBool();
} // WkbSelect2()


/*****************************************************************************
W k b U n S e l e c t 2
==============================================================================

This function unselects the specified WibuBox entry. This entry can be
selected now by other opened WibuBox entries.

Parameters
 hwkbe  contains the handle of an opened WibuBox entry.

Return
 The function returns TRUE if the operation was sucessful and FALSE if an
 error occured. In this case a subsequent call of the function
 WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbUnSelect2(HWKBENTRY hwkbe)
{

  if (WkdGetEntrySystem(hwkbe, 0)) {
    // execute kernel of operation
    WkdUnSelectEntry();
  } // if
  return WkdReturnBool();
} // WkbUnSelect2()


/*****************************************************************************
W k b C r y p t 2
==============================================================================

This function encrypts or decrypts specified data via the WibuBox. In
dependence of a previous selection of the opened entry via the function
WkbSelect2(), a direct or indirect encryption algorithm is used.

Parameters
  hwkbe .. contains the handle of an opened WibuBox entry.
  flCtrl   contains flags which controls the execution of the operation.
           Following flags are currently supported:
           WKB_CRYPT_CTRL (0x0000) The parameter <pvCtrl> contains a pointer
                          to an initialized structure which contains
                          additional data to control the encryption. If
                          <pvCtrl> is NULL, no such data are specified.
           WKB_CRYPT_COPY (0x0001) The parameter <pvCtrl> contains a pointer
                          to a buffer which contains the source data to be
                          encrypted. <pvCtrl> may not be NULL!
           WKB_CRYPT_BUFFER (0x0002) The encryption is internally buffered.
                          The specification overwrites the default setting of
                          the buffer option. This default setting is
                          WKB_CRYPT_BUFFER for direct and WKB_CRYPT_NOBUFFER
                          for indirect encryption.
           WKB_CRYPT_NOBUFFER (0x0004) The encryption is internally not
                          buffered. The specification overwrites the default
                          setting of the buffer option. This default setting
                          is WKB_CRYPT_BUFFER for direct and
                          WKB_CRYPT_NOBUFFER for indirect encryption.
           Further flags controls the type of data to be encrypted or
           decrypted:
           WKB_AREA_DATA (0x0000) standard data are addressed which can be
                         freely modified.
           WKB_AREA_CODE (0x0001) executable code is addressed which must be
                         modified with care and additional driver operations
                         under some operating systems or programming
                         languages.
           This flags are only valid if the parameter <pvCtrl>
           doesn't point to a control data structure. Otherwise the flags of
           the <flCtrl> member of the WKBAREA structure are used.
  pvDest   points to a buffer which receives the encrypted bytes.
           If in <flCtrl> the flag WKB_CRYPT_CTRL is specified, this pointer
           also addresses the data to be encrypted.
  pvCtrl   points to a buffer which is used for different purposes in
           dependence of the flags in <flCtrl>. If the flag WKB_CRYPT_COPY is
           set, it points to the source data which should be encrypted.
           Otherwise if the flag WKB_CTRLDATA is set, it points to a structure
           which controls the type of encryption. In this case the pointer may
           be NULL. If it is not NULL it points to a WKBAREA structure which
           contains additional control data for the execution of the function.
           This data must not be stored as a constant aggregate because some
           members are updated by the WkbCrypt2() call.
  cbSrc .. contains the number of bytes which should be encrypted.
  pcbDest  points to an integer variable which receives the number of
           encrypted bytes at the destination address. If the specified
           pointer is NULL, no value is stored.

Comments
  The option WKB_CRYPT_BUFFER is set as standard if the previous selection
  call chose the direct encryption. On the other hand, the option
  WKB_CRYPT_NOBUFFER is used if an indirect encryption was chosen. The
  specification of one of the options in the parameter <flCtrl> is normally
  unnecessary: Since a failed partial indirect encryption is never possible
  as result of a hardware problem, the use of WKB_CRYPT_BUFFER is unnecessary
  in this case. For direct encryption, the specification of WKB_CRYPT_NOBUFFER
  is normally uncessary as far as only small sequences are encrypted which can
  be buffered by the driver. Under DOS, the WKB_CRYPT_BUFFER option is limited
  to 64 bytes.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Preliminary Note
  The length of encryption/decryption is currently limited to 4096 bytes.
  The gap map option of the WKBAREA structure is not implemented.
  The CRC checking of the WKBAREA structure is not implemented.

*****************************************************************************/

INT WKAPIENTRY WkbCrypt2(HWKBENTRY hwkbe, ULONG flCtrl, void *pvDest,
  void *pvCtrl, UINT cbSrc, UINT *pcbDest)
{
  INT iRet = 0;


  THREAD_API_LOCK

  EnterCryptAccess();


  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pvDest)) {
    goto EXITFN;
  } // if
  if (!WkdConvertMemHandle((const void**) &pvCtrl)) {
    goto EXITFN;
  } // if
  if (!WkdConvertMemHandle((const void**) &pcbDest)) {
    goto EXITFN;
  } // if
  // analyse type of function call, execute encryption/decryption
  if (WkdGetEntrySystem(hwkbe, flCtrl)) {
    // execute kernel of operation
    BYTE *pvSrc = NULL;
    WKBAREA *pwkbar = NULL;
    BOOL fCopy = FALSE;
    UINT uDestCtrl = 0;
    // preliminary: assume no separate source
    pvSrc = pvDest;
    pwkbar = NULL;
    fCopy = (BOOL) (flCtrl & ((ULONG) WKB_CRYPT_COPY));
    if (fCopy) {
      // set separate pointer to source data
      pvSrc = pvCtrl;
    } else {
      // set pointer to WKBAREA structure
      pwkbar = pvCtrl;
    } // if
    // determine area encryption options
    if (pwkbar != NULL) {
      // verify WKBAREA structure: must be stored at writable area
      //   (member <ulCrc> may be changed ...)
      // get destination control flags from <flCtrl> of WKBAREA structure
      uDestCtrl = (UINT)pwkbar->flCtrl & (WKB_AREA_MASK|WKB_AREA_NONE);
    } else {
      // get (some) area encryption flags from <flCtrl> of WkbCrypt2()
      uDestCtrl = (UINT)flCtrl & (WKB_AREA_NONE|WKB_AREA_MASK);
    } // if
    // analyse specified control flags
    switch (uDestCtrl & WKB_AREA_MASK) {
      case WKB_AREA_DATA:
        // data (read/write memory): nothing to do
        break;
      case WKB_AREA_CODE:
        // code (read only): activate buffering on Windows 16 bits
        break;
      default:
        // other area types not permitted: set error & abort
        cstWkd.idError = WKB_ERROR_OPTION_NOT_SUPPORTED;
        goto EXITFN;
    } // switch
    if ((pwkbar != NULL) && (pwkbar->flCtrl & WKB_AREA_SELECT) &&
        (pwkbar->flSelCtrl & WKB_SEL_MASK) != WKB_SEL_DIRECT) {
      // indirect encryption not implemented
      cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
      goto EXITFN;
    } // if
    {//// direct encryption
      // determine buffering mode from specified flags
      if (flCtrl & WKB_CRYPT_BUFFER) {
        uDestCtrl |= WKB_CRYPT_BUFFER;
      } else if (!(flCtrl & WKB_CRYPT_NOBUFFER) && cbSrc <= 256) {
        // set default buffering option: buffer when < 256 bytes in size
        uDestCtrl |= WKB_CRYPT_BUFFER;
      } // if
      CryptDirect(pvDest, pvSrc, cbSrc, pwkbar, uDestCtrl, pcbDest);
    } // block
  } // if
EXITFN:
  iRet = WkdReturnBool();
  THREAD_API_UNLOCK;

  LeaveCryptAccess();

  return iRet;
} // WkbCrypt2()


// =====================================
// end of clcrypt.c
// =====================================
