/*****************************************************************************
S r c D r v                               Portable Source Code WibuKey Driver
==============================================================================
cllist.c                                   WIBU-BOX Listing API Implementation
==============================================================================

This module implements following WibuKey API functions:
  WkbListPort2(), WkbListBox2()

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 54, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-0-721-93172-0, Fax: +49-0-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
92Sep13 1.33 mabu: start of development
93Mar18 1.33 mabu: custom source variant released
93Mar21 1.35 mabu: adaption to Windows
93Jun06 1.35 mabu: conversation to large memory model.
93Jun20 1.35 mabu: adaptation to Win32 environment.
93Jul02 1.35 mabu: first adaptations to WibuBox version 2
93Jul22 1.35 mabu: adapation to DOS
93Aug21 2.00 mabu: greater changes in API of new functions
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94Aug10 2.00 mabu: adaptation to new WkNet/WkLAN API
94Oct11 2.00 mabu: implementing the look-ahead-function
94Nov11 2.00 mabu: version 2.00 released
95Feb08 2.02 mabu: bug in CreateHandle() function removed
95Feb09 2.02 mabu: extracted from KRNL.C
95Aug10 2.07 mabu: error handling in WkbListPort2 changed
95Aug11 2.07 mabu: critical checking of serial number added
95Sep20 2.08 mabu: proper reading of configuration data on Macintosh
95Nov08 2.09 mabu: extensions for WibuBox 4 implemented
95Dec19 2.09 mabu: proper abort of WkbListBox2/WKB_LIST_CONFIG with BOX/V1
96Jun23 2.13 mabu: renamed from BSLIST.C
97Feb15 2.25 mabu: decrypting of expiration dates
97Jun11 2.31 mabu: litte/big endian management for Macintosh added
98Feb07 2.31 mabu: proper checking of WKB_LPT_SERIAL command
98Mar16 2.41 mabu: begin of adaptation to new driver class model
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"


/*****************************************************************************
W k b L i s t P o r t 2
==============================================================================

This function determines all available WIBU-BOXes at the specified port. The
serial numbers of all detected WIBU-BOXes are stored into the specified buffer
except the buffer is too small.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To list a port of the standard local subsystem the predefined
           handle value HWKB_LOCAL can be used.
  flCtrl   contains flags which control the operation of the function.
             WKB_PORTNUMBER (0x0008) is set to specify a port number instead
                                     of a port name.
  pszPort  points to a string which contains one single port specification.
           If the flag WKB_PORTNUMBER is set, this parameter is not inter-
           preted as pointer but must contain a valid port specification
           number which is compatible with WibuKey version 1.
  pvDest   points to a destination buffer which receives the list information.
           For this version of the function a WKBBOXLIST structure with
           variable length of array awkbsr[] is used.
  cbDest   contains the length of the destination buffer in bytes.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cbDest> is too small to receive all returned
    information, the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cbDest> is large enough, the function returns the
    number of transferred bytes.

*****************************************************************************/

INT WKAPIENTRY WkbListPort2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  const char *pszPort, void *pvDest, UINT cbDest)
{
  UINT cbData = 0;
  INT iRet = 0;


  THREAD_API_LOCK

  // convert handle address to true address if required
  if (cbDest != 0 && !WkdConvertMemHandle((const void**) &pvDest)) {
    // any error => return with error
    goto EXITFN_RET;
  } // if
  // check if specified system is valid, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto EXITFN_RET;
  } // if
  {//// execute driver operations
    switch (flCtrl & WKB_EPT_MASK) {
      case WKB_LPT_SERIAL: {
        UINT cBox = 0; // number of found WIBU-BOXes
        WKBBOXLIST *pwkbbxl = NULL; // pointer to WibuBox table
        WKBSERIAL  *pwkbsr = NULL;  // pointer in serial number table
        // validate port specification & select and access this port
        if (!WkdAccessPortSpec(pszPort)) {
          goto EXITFN_RET;
        } // if
        {// execute via class driver
          UINT iBox = 0;
          DEVOBJECT *pdevobj = NULL;
          ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceList);
          cBox = cstWkd.pclsdrvSel->pfClDrvGetDeviceList(&pdevobj);
          // calculate determined size of destination data
          cbData = FIELDOFFSET(WKBBOXLIST, awkbsr) + cBox * sizeof(WKBSERIAL);
          if (cbDest <= FIELDOFFSET(WKBBOXLIST, awkbsr)) {
            // too less memory to store header => return with required size
            cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
            if (cbDest == FIELDOFFSET(WKBBOXLIST, awkbsr)) {
              // set header information into destination buffer
              // * set pointer to destination buffer
              pwkbbxl = pvDest;
              // no flags set
              pwkbbxl->flCtrl = 0;
              // store number of WIBU-BOXes
              pwkbbxl->cBox = (USHORT)cBox;
            } // if
            goto EXIT_KRNL;
          } // if
          // set header information into destination buffer
          // * set pointer to destination buffer
          pwkbbxl = pvDest;
          // no flags set
          pwkbbxl->flCtrl = 0;
          // store number of WIBU-BOXes
          pwkbbxl->cBox = (USHORT)cBox;
          // unused area set to 0
          pwkbbxl->usReserved = 0;
          if (cbData > cbDest) {
            // * destination buffer not large enough
            // reduce stored number of WibuBox to buffer size
            ASSERT(cbDest > FIELDOFFSET(WKBBOXLIST, awkbsr));
            cBox = (cbDest - FIELDOFFSET(WKBBOXLIST, awkbsr)) /
              sizeof(WKBSERIAL);
            cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
          } // if
          // set destination pointer to first entry
          pwkbsr = pwkbbxl->awkbsr;
          // transfer all serial numbers until end of list
          // * or destination buffer is full
          for (iBox = 0;
              iBox < cBox;
              iBox++, pwkbsr++, pdevobj = pdevobj->pNext
            ) {
            // transfer serial number from WibuBox device object list
            ASSERT_VALUE(pdevobj);
            *pwkbsr = pdevobj->wkbsr;
          } // for
        } // block
      EXIT_KRNL:
        if (cstWkd.idError == WKB_ERROR_NOT_CONNECTED) {
          // this error code is ignored (no WibuBox is not a true error)
          cstWkd.idError = WKB_NO_ERROR;
        } // if
        WkdReleasePort();
        break;
      } // case
      default:
        // other commands are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
    } // switch
  }
EXITFN_RET:
  iRet = WkdReturnLength(cbData);

  THREAD_API_UNLOCK

  return iRet;
} // WkbListPort2()


/*****************************************************************************
W k b L i s t B o x 2
==============================================================================

This function lists the contents of the specicified WibuBox. The returned
data structure contains the version value of the WibuBox and the interpreted
contents of the entry registers of the WibuBox or as a mode variant the
configuration data of the WibuBox.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To list the contents of a WibuBox at the standard local subsystem
           the predefined handle value HWKB_LOCAL can be used.
  flCtrl   contains flags which control the operation of the function.
           Following flags are currently supported:
             WKB_LIST_DATA   (0x0000) returns the complete contents of the
                             entries of the WibuBox.
             WKB_LIST_CONFIG (0x0010) returns the configuration data of the
                             WibuBox.
             WKB_LIST_EXTMEM (0x0020) returns the stored content in the memor
                             the of the WibuBox.
             WKB_PORTNUMBER  (0x0008) is set to specify a port number instead
                             of a port name.
  pszPort  points to a string which contains one single port specification.
           If the flag WKB_PORTNUMBER is set, this parameter is not inter-
           preted as pointer but must contain a valid port specification
           number which is compatible with WibuKey version 1.
  pwbksr   points to a valid serial number which specifies the WibuBox
           which contents is returned.
  pvDest   points to a destination buffer which receives the list information.
           For this version of the function a WKBBOXDATA or a WKBBOXCONFIG
           structure is used.
  cbDest   contains the length of the destination buffer in bytes. This should
           be the length of the WKBBOXDATA or WKBBOXCONFIG data structure.

Return
 The function returns one of the following values:
 * 0 if any error occured. In this case a subsequent call of the function
   WkbGetLastError() returns the corresponding error code.
 * If the length in <cbDest> is too small to receive all returned information,
   the needed buffer length is returned. A subsequent call of
   WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
 * If the length in <cbDest> is large enough, the function returns the number
   of transferred bytes.

*****************************************************************************/

INT WKAPIENTRY WkbListBox2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  const char *pszPort, const WKBSERIAL *pwkbsr, void *pvDest,
  UINT cbDest)
{
  UINT cbData = 0;
  INT iRet = 0;

  
  THREAD_API_LOCK

  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pwkbsr)) {
    // invalid specification => return with error
    goto RETURN;
  } // if
  if (cbDest != 0 && !WkdConvertMemHandle((const void**) &pvDest)) {
    // invalid specification => return with error
    goto RETURN;
  } // if
  // check if specified system is valid, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  {//// execute driver operations
    // pointer to used device object
    DEVOBJECT *pdevobj = NULL;
    struct {
      WKDEV_REQUEST wkdevreq;
      BYTE ab[1 + 2 + 8 - 1];
    } s;
    BYTE *pabSrc = NULL;
    // number of bytes which should be read
    UINT cbRead = 0;
    // number of needed destination bytes
    UINT cbNeedDest = 0;
  #if BSRDEBUG
    cstWkd.pope = NULL;
  #endif

    // validate port specification, select and access this port
    if (!WkdAccessPortSpec(pszPort)) {
      goto RETURN;
    } // if
    // validate given parameters and set working parameters
    switch (cstWkd.flCtrl & WKB_LIST_MASK) {
      case WKB_LIST_DATA:
        cbNeedDest = sizeof(WKBBOXDATA);
        cbRead = 1 + 60;
        // setting request parameter
        s.wkdevreq.usCmd = BOXCMD_LIST;
        // control code for getting the contents of the WibuBox
        s.wkdevreq.usIndex = 0x00;
        s.wkdevreq.cbRequest = 0;
        break;
      case WKB_LIST_CONFIG:
        // check old size of WKBBOXCONFIG !!!
        cbNeedDest = sizeof(WKBBOXCONFIG) - 2 * sizeof(USHORT);
        cbRead = 1 + 4;
        // setting request parameter
        s.wkdevreq.usCmd = BOXCMD_LIST;
        // control code for getting the configuration of the WibuBox
        s.wkdevreq.usIndex = 0x01;
        s.wkdevreq.cbRequest = 0;
        break;
      case WKB_LIST_EXTMEM: {
        // destination buffer contains also some request parameter
        WKBEXTMEM *pwkbxmr = pvDest;
        USHORT idPage = pwkbxmr->iPage << 1;
        // check the version of the box
        UINT32 ulVersion = WkdGetBoxVersion(pwkbsr);
        if (ulVersion < 6 || ulVersion >= 0x80) {
          cstWkd.idError = WKB_ERROR_OPTION_NOT_SUPPORTED;
          goto RETURN;
        } // if
        // set request parameter
        cbNeedDest = sizeof(WKBEXTMEM);
        cbRead = 1 + 1 + 32;
        // setting request parameter (command PROGRAM !!!)
        s.wkdevreq.usCmd = BOXCMD_PROGRAM;
        // index of entry for getting the programmable contents of the WibuBox
        s.wkdevreq.usIndex = 0;
        // send additional request bytes: Page Index + Access Siganture
        s.wkdevreq.cbRequest = 1 + 2 + 8;
        // set Control Code for reading ExtMem storage
        s.wkdevreq.abRequest[0] = 0x07;
        // copy signature
        ASSERT(sizeof(pwkbxmr->abSignature) == 8);
        if (pwkbxmr->fsCtrl & WKB_XM_PROT) {
          // use protected ExtMem access
          idPage |= 1;
          memcpy(s.wkdevreq.abRequest + 1 + 2, pwkbxmr->abSignature, sizeof(pwkbxmr->abSignature));
        } else {
          // use unprotected ExtMem access (WKB_XM_USER)
          memset(s.wkdevreq.abRequest + 1 + 2, 0, sizeof(pwkbxmr->abSignature));
        } // if
        // set page index
        SET_USHORT_VALUE(s.wkdevreq.abRequest + 1, idPage);
        } // case
        break;
      default:
        // other commands are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        goto RETURN;
    } // switch
    // validate size of destination buffer
    if (cbNeedDest > cbDest) {
      // buffer for storing box data too small, return with error
      // * return this value
      cbData = cbNeedDest;
      cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
      goto RETURN;
    } // if
    // parameters are valid: start operation
    // * allocate local read buffer
    pabSrc = WkdAllocMem(cbRead);
    if (!pabSrc) {
      goto RETURN;
    } // if
    // execute via class driver
    // * determine accessed device from serial number
    ASSERT_VALUE(cstWkd.pclsdrvSel);
    ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
    pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(pwkbsr);
    if (pdevobj == NULL) {
      // device not found => return with error
      goto RETURN_KRNL;
    } // if
    s.wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
    s.wkdevreq.cbAnswer = cbRead;
    // read now determined number of bytes from device
    ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
    //
    if (!cstWkd.pclsdrvSel->pfClDrvRequest(&s.wkdevreq, pabSrc, NULL)) {
      // request not properly executed => return with error
      goto RETURN_KRNL;
    } // if
    switch (cstWkd.flCtrl & WKB_LIST_MASK) {
      case WKB_LIST_DATA: {
        // init and fill WKBBOXDATA structure at destination buffer
        INT i = 0;
        BYTE *pbSrc = NULL;
        INT cEntry = 0;
        // pointer to WibuBox data structure/destination buffer
        WKBBOXDATA *pwkbbxd = pvDest;
        // expand packed WibuBox data into WKBBOXDATA structure
        // * set destination store positition
        WKBBOXENTRY *pwkbbxeDest = pwkbbxd->awkbbxe;
        // clear complete structure
        memset(pwkbbxd, 0, sizeof(WKBBOXDATA));
        // store first byte as version value
        pwkbbxd->usVersion = pabSrc[0];
        for (i = 0, pbSrc = pabSrc + 1; i < 10; i++, pbSrc += 3 * 2) {
          // expand WibuBox contents (convert also expiration dates)
          if (WkdSetBoxEntry(pbSrc, pbSrc + 5 * 6, pabSrc[0], i, pwkbbxeDest, FALSE)) {
            // created entry is stored: increment number of entries
            pwkbbxeDest++;
            cEntry++;
          } // if
        } // for
        // store number of valid entries
        pwkbbxd->cEntry = (USHORT)cEntry;
        } // case
        break;
      case WKB_LIST_CONFIG: {
        // init and fill WKBBOXCONFIG structure at destination buffer
        // * set pointer to destination buffer
        WKBBOXCONFIG *pwkbbxc = pvDest;
        // initialize only old length of WKBBOXCONFIG !!!
        memset(pwkbbxc, 0, sizeof(WKBBOXCONFIG) - 2 * sizeof(USHORT));
        pwkbbxc->usVersion = pabSrc[0];
        pwkbbxc->ulProgramCount = GET_USHORT_VALUE(pabSrc + 1);
        pwkbbxc->ulConfig = GET_USHORT_VALUE(pabSrc + 1 + 2);
        if (cbDest >= sizeof(WKBBOXCONFIG)) {
          // new driver version (> 4.0) ==>> add chip info
          WkdGetChipInfo(pwkbsr, &pwkbbxc->usAsicBuild, &pwkbbxc->usExtMemSize);
          cbNeedDest = sizeof(WKBBOXCONFIG);
        } // if
        } // case
        break;
      case WKB_LIST_EXTMEM: {
        // set pointer to destination buffer
        // ==>> output buffer is interpreted as a 32 byte array buffer 
        WKBEXTMEM *pwkbxm = pvDest;
        ASSERT(sizeof(pwkbxm->abData) == 32);
        // analyse error, if any => set into <cstWkd.idError>
        if (pabSrc[0] != 0) {
          WkdConvertStatusIntoError(pabSrc[0]);
          goto RETURN_KRNL;
        } else if (pabSrc[1] != 0) {
          WkdConvertStatusIntoError(pabSrc[1]);
          goto RETURN_KRNL;
        } // if
        // copy received sequence
        memcpy(pwkbxm->abData, pabSrc + 1 + 1, sizeof(pwkbxm->abData));
        } // case
        break;
      default:
        // other commands are not implemented
        ASSERT(0);
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        goto RETURN;
    } // switch

    // return with from function with <idResult>
    cbData = cbNeedDest;
  RETURN_KRNL:
    BsrFreeMem(pabSrc);
    WkdReleasePort();
  }
RETURN:
  iRet = WkdReturnLength(cbData);

  THREAD_API_UNLOCK

  return iRet;
} // WkbListBox2()


// ====================================
// end of cllist.c
// ====================================
