/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clmem.c                                         User Memory API Implementation
==============================================================================

This module implements following WibuKey API functions:
  WkbAllocMem(), WkbFreeMem(), WkbWriteMem(), WkbReadMem()

It implements also following driver internally used functions:
  WkdConvertMemHandle()

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1995 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://wwww.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
95Oct24 2.08 mabu: initial version
95Nov06 2.08 mabu: handle/address detection reduced from 65536 to 2048.
96Feb27 2.10 mabu: WkDos calling implemented
96Jun23 2.13 mabu: renamed from BSMEM.C
97Jul28 2.31 mabu: critical memory parameter check implemented
98Feb06 2.41 mabu: integration of new error codes
06Jan25 5.10 ha:   CodeMerge
06Aug30 5.12 ha:   changes for SharedMemory access - all data stored LittleEndian
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

/*****************************************************************************
W k b A l l o c M e m
==============================================================================

This function allocates a memory block with the specified length in the memory
heap of the driver and returns a handle which is associated with this block.

Parameters
  hwkbsys  contains the handle of a subsystem for which the memory block is
           allocated. If this subsystem is closed by WkbRelease2(), this
           memory block is freed automatically.
  cbAlloc  contains the length of allocated memory block. This must be a value
           between 0 and 65500.

Return
  The function returns a valid handle if the operation was successful and
  FALSE if an error occurred. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Preliminary Limitation
  The <hwkbsys> parameter is ignored. Closing a subsystem does not free
  automatically all memory blocks which are allocated under this subsystem.

*****************************************************************************/

HWKBMEM WKAPIENTRY WkbAllocMem(HWKBSYSTEM hwkbsys, UINT cbAlloc)
{
  HWKBMEM hwkbmem;
  BYTE *pbBuf;

  if (cbAlloc > 65500) {
    // specified length too long: return with error
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    goto RETURN_ERROR;
  } // if
  // allocate two bytes more for store length at begin of buffer
  pbBuf = WkdCreateHandle(HTYP_MEMBLOCK, 
      cbAlloc + sizeof(USHORT), (HANDLE*)&hwkbmem);
  // handle not created: return with error
  if (hwkbmem == (HWKBMEM)NULL) {
    goto RETURN_ERROR;
  } // if
  // store allocated length at begin of buffer, fill remainder with 0 bytes
  *(USHORT*)pbBuf = (USHORT)cbAlloc;
  pbBuf += sizeof(USHORT);
  memset(pbBuf, 0, cbAlloc);
  return hwkbmem;
RETURN_ERROR:
  gidWkdLastError = cstWkd.idError;
  return (HWKBMEM)NULL;
} // WkbAllocMem()


/*****************************************************************************
W k b F r e e M e m
==============================================================================

This function destroys a memory block which was allocated previously by
WkdAllocMem.

Parameters
  hwkbmem  contains the handle of a memory block which was allocated by
           WkbAllocMem.

Return
  The function returns TRUE if the operation was successful and FALSE if an
  error occurred. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

BOOL WKAPIENTRY WkbFreeMem(HWKBMEM hwkbmem)
{
  UINT uType;

  uType = WkdGetHandleType(hwkbmem);
  if (uType != HTYP_MEMBLOCK) {
    // bad type of handle => return with error
    gidWkdLastError = WKB_ERROR_INVALID_HANDLE;
    return FALSE;
  } // if
  // destroy specified handle, return without error
  WkdDestroyHandle(hwkbmem);
  return TRUE;
} // WkbFreeMem()


/*****************************************************************************
W k b W r i t e M e m
==============================================================================

This function transfers data with the specified length from the specified
source address to the specified allocated memory block.

Parameters
  hwkbmem  contains the handle of memory block which was allocated by
           WkbAllocMem.
  pvSrc .. points to the data which are written into the memory block. This
           must be hold at least <cbData> or more bytes. This parameter may
           also contain a memory handle of another created memory block.
  cbData   contains the number of bytes which should be transferred. There are
           never more bytes transferred than the allocated memory block can
           hold.

Return
  The function returns the number of actually transferred bytes. If this value
  is 0, an error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

UINT WKAPIENTRY WkbWriteMem(HWKBMEM hwkbmem,
  const void *pvSrc, UINT cbData)
{
  UINT uType = 0;
  BYTE *pbBuf = NULL;
  UINT cbBuf = 0;
  UINT iError = WKB_NO_ERROR;

  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvSrc)) {
    iError = WKB_ERROR_INVALID_PARAMETER;
    goto RET_ERROR;
  } // if
  uType = WkdGetHandleType(hwkbmem);
  if (uType != HTYP_MEMBLOCK) {
    iError = WKB_ERROR_INVALID_HANDLE;
    goto RET_ERROR;
  } // if
  pbBuf = WkdGetHandlePtr(hwkbmem);
  cbBuf = *(USHORT*)pbBuf;
  pbBuf += sizeof(USHORT);
  // reduce specified length if required
  if (cbData > cbBuf) {
    cbData = cbBuf;
  } // if
  // transfer data from specified address into buffer
  memcpy(pbBuf, pvSrc, (size_t)cbData);
  // return number of transferred bytes
  return cbData;
RET_ERROR:
  gidWkdLastError = WKB_ERROR_INVALID_HANDLE;
  return 0;
} // WkbWriteMem()


/*****************************************************************************
W k b R e a d M e m
==============================================================================

This function transfers data with the specified length from the specified
allocated memory block to the specified memory address.

Parameters
  hwkbmem  contains the handle of a memory block which was allocated by
           WkbAllocMem.
  pvDest   points to a memory buffer which receives the data from the
           memory block. The memory buffer must be hold at least <cbDest> or
           more bytes. This parameter may also contain a memory handle of
           another created memory block.
  cbDest   contains the number of bytes which should be transferred. There are
           never more bytes transferred than the allocated memory block
           contain really.

Return
  The function returns the number of actually transferred bytes. If this value
  is 0, an error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

UINT WKAPIENTRY WkbReadMem(HWKBMEM hwkbmem, void *pvDest, UINT cbDest)
{
  UINT uType;
  BYTE *pbBuf;
  UINT cbBuf;

  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pvDest)) {
    goto RET_ERROR;
  } // if
  uType = WkdGetHandleType(hwkbmem);
  // bad type of handle => return with error
  if (uType != HTYP_MEMBLOCK) {
    goto RET_ERROR;
  } // if
  pbBuf = WkdGetHandlePtr(hwkbmem);
  cbBuf = *(USHORT*)pbBuf;
  pbBuf += sizeof(USHORT);
  // reduce specified length if required
  if (cbDest > cbBuf) {
    cbDest = cbBuf;
  } // if
  // transfer data from buffer to specified address
  memcpy(pvDest, pbBuf, (size_t)cbDest);
  // return number of transferred bytes
  return cbDest;
RET_ERROR:
  gidWkdLastError = WKB_ERROR_INVALID_HANDLE;
  return 0;
} // WkbReadMem()


/*****************************************************************************
W k d C o n v e r t M e m H a n d l e
==============================================================================

This function checks if the pointer at the specified address contains a handle
value. If so, the handle is searched in the handle table and checked for the
correct type (memory buffer). If the handle is valid, it is replaced at the
pointer address by the I/O address of the allocated buffer.

Parameters
  ppv  points to an pointer which could contain a handle value.

Return
  The function returns TRUE if no handle value is specified or if the
  specified handle was a memory handle and could be converted into its
  corresponding memory address. If an undefined handled is specified or the
  handle is no memory buffer handle, FALSE is returned and <cstWkd.idError>
  contains an error code.

*****************************************************************************/

BOOL WkdConvertMemHandle(const void **ppv)
{
  ULONG ulAddr;
  HANDLE h;
  UINT uType;
  BYTE *pbAddr;
  ulAddr = (ULONG) *ppv;   // UNIX
  // higher 21 bits not 0 => assume that valid pointer, nothing to do
  if (ulAddr >= 2048) {
    return TRUE;
  } // if
  h = (HANDLE) (long) LOUSHORT(ulAddr);
  // NULL specified => nothing to do
  if (h == 0) {
    return TRUE;
  } // if
  // assume that pointer: try to convert
  uType = WkdGetHandleType(h);
  if (uType != HTYP_MEMBLOCK) {
    // bad type of handle => return with error
    cstWkd.idError = WKB_ERROR_INVALID_HANDLE;
    return FALSE;
  } // if
  // handle is valid: set pointer at destination to buffer address
  pbAddr = WkdGetHandlePtr(h);
  // skip leading two length bytes
  *(void**)ppv = pbAddr + sizeof(USHORT);
  return TRUE;
} // WkdConvertMemHandle()


// ===================================
// end of clmem.c
// ===================================
