/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clopen.c                                    WIBU-BOX Open Entry API Management
==============================================================================

This module implements the following WibuKey API functions:
  WkbOpen2(), WkbQueryEntry2(), WkbClose2()

Following driver internally used are also implemented:
  WkdClose

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All POSIX OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
92Sep13 1.33 mabu: start of development
93Aug21 2.00 mabu: greater changes in API of new functions
94Mar08 2.00 mabu: low level driver variants added
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94Aug10 2.00 mabu: adaptation to new WkNet/WkLAN API
94Oct11 2.00 mabu: implementing the look-ahead-function
94Nov11 2.00 mabu: version 2.00 released
95Feb08 2.02 mabu: bug in CreateHandle() function removed
95Feb11 2.05 mabu: extracted from KRNL.C
95Jun05 2.06 mabu: implementation of WkbQueryEntry2/WKB_QE_BOX
95Aug10 2.07 mabu: ignoring of version byte reading when serial mask is 1
95Aug11 2.07 mabu: critical checking of serial number added
95Aug16 2.07 mabu: bug in WkbOpen2/WkNet removed
95Oct19 2.08 mabu: WKB_ERROR_BUFFER_OVERFLOW bug in WkbOpen2() removed
95Dec20 2.09 mabu: proper WibuBox command abort in GetOpenBoxEntry()
96Jun23 2.13 mabu: renamed from BSOPEN.C
96Aug25 2.14 mabu: begin of addition of DBGDRV variant extensions
96Oct24 2.20 mabu: implementation of WKB_OPEN_NONETLIMIT option
96Oct24 2.20 mabu: implementation of WKB_OPEN_GLOBAL option
96Nov04 2.20 mabu: support of WKB_ACCCESS_CODE for WkLAN subsystem opens.
96Dec06 2.21 mabu: saving of subsystem Firm/User Code in base system structure
97Feb26 2.30 mabu: implementation of WKB_OPEN_MULTIFIND command
97Feb26 2.30 mabu: implementation of WKB_QE_MULTIENTRY command
97Mar04 2.31 mabu: bug in WKB_OPEN_MULTIFIND (bad used entry) removed
97Jun08 2.31 mabu: adaptation of datagram sending for Macintosh
97Jun11 2.31 mabu: litte/big endian management for Macintosh added
97Jul05 2.31 mabu: proper setting of WKBSERIAL structure in FindBoxEntry()
97Aug03 2.31 mabu: support of WKB_OPEN_MULTIFIND by WkNet
97Oct27 2.31 mabu: support of WKBOPENFIND structure in WkbOpen2
98Feb07 2.41 mabu: integration of new error codes
98Mar16 2.41 mabu: begin of adaptation to new driver class model
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"


/*****************************************************************************
W k d S e t W K B O P E N F I N D
==============================================================================

This function converts the members of the WKBOPENFIND structure from the used
endian ordering to the little endian.

Parameters
  pwkbof  points to the WKBOPENFIND structure to be converted.

*****************************************************************************/

void WkdSetWKBOPENFIND(WKBOPENFIND *pwkbof)
{
  SET_UINT32_LE_INPLACE(&pwkbof->flCtrl);
  SET_UINT32_LE_INPLACE(&pwkbof->ulFirmCode);
  SET_UINT32_LE_INPLACE(&pwkbof->ulEntrySkip);
} // WkdSetWKBOPENFIND()


/*****************************************************************************
F i n d B o x E n t r y
==============================================================================

This function searchs a WibuBox entry with the specified Firm Code and
User Code at the specified port. If an entry is found it is stored into the
specified WibuBox entry. Otherwise an error is returned.


Parameters
  nPort .... contains the index of the port at which the WibuBox is searched.
  pope ..... points to an initialized internal WibuBox entry structure which
             contains all needed information about the entry to be searched.
  pidSpecificError  points to an INT variable which receives specific errors.
             Currently only two specific error codes are set:
             * WKB_ERROR_INVALID_VERSION
             * WKB_ERROR_ENTRY_ALREADY_USED (overwrites other error code).
             * WKB_ERROR_INVALID_PORT No Permission on Device Path (only LINUX)

Globals
  cstWkd

Return
  The function returns TRUE if the slot is valid and FALSE if not. In the
  latter case, <cstWkd.idError> contains an error code.

Comments
  Starting with version 2.07 of this driver, the mask byte of the serial
  number decides if the optional trailing version byte is read. Because of the
  mask byte is 1 only for WibuBoxes version 1, it is assumed that a WibuBox
  with a higher mask byte value than 1 returns ALWAYS the version byte
  properly.

*****************************************************************************/

static BOOL FindBoxEntry(
  INT nPort, OPENENTRY *pope, INT *pidSpecificError)
{
  BOOL fValid = FALSE;
  // try to access specified port
  if (!WkdSelectPort(nPort, TRUE)) {
    // port cannot be accessed => return with error
    cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
    return FALSE;
  } // if
  pope->lPort = cstWkd.nSelPort;
  // master flag not set
  pope->flCtrl &= ~WKB_OE_MASTER;
  // execute via class driver
  ASSERT_VALUE(cstWkd.pclsdrvSel);
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvFindDevice);
  fValid = (cstWkd.pclsdrvSel->pfClDrvFindDevice)(pope);
  if (!fValid) {
    // entry not found => return with error
    ASSERT(cstWkd.idError != WKB_NO_ERROR);
    if (cstWkd.idError == WKB_ERROR_INVALID_VERSION) {
      // bad version error: store specific except entry already used
      // * WKB_ERROR_ENTRY_ALREADY_USED has higher priority
      if (*pidSpecificError != WKB_ERROR_ENTRY_ALREADY_USED) {
        // store current error as specific error
        *pidSpecificError = cstWkd.idError;
      } // if
    } // if
    // return with error
    goto EXIT_FN_ERROR;
  } // if
  // check if entry is globally opened by another process
  if (WkdIsEntryConflict(pope)) {
    // entry globally used by another open process or exclusive conflict
    // * store specific error
    ASSERT(cstWkd.idError == WKB_ERROR_ENTRY_ALREADY_USED ||
      cstWkd.idError == WKB_ERROR_EXCLUSIVE_MODE_CONFLICT);
    *pidSpecificError = cstWkd.idError;
    // return with error
    goto EXIT_FN_ERROR;
  } // if
  // set single bit for MULTIENTRY option
  pope->fuEntryOpen = 1 << pope->iEntry;
  // set expiration date as unknown
  pope->uExpDate = UNDET_EXPDATE;
  // return without error
  return TRUE;

EXIT_FN_ERROR:
  // return with error but release accessed port
  WkdReleasePort();
  return FALSE;
} // FindBoxEntry()


/*****************************************************************************
O p e n B o x E n t r y
==============================================================================

This function opens a WibuBox entry with the specified parameters if the
entry exists and is valid for required operations.

Parameters
  nPort ..... contains the port index (0..127) of the desired port
  pwkbsr .... points to the serial number of the desired WibuBox.
  iEntry .... contains the index (0..9) of the desired index.
  flCtrl .... contains the WkbOpen2() control flags including the
              algorithm version.
  ulUserCode  contains the User Code which is required to check if a Master
              entry matches the specified entry.
  hwkbsys ... contains the handle of the subsystem which is stored into
              Open-WibuBox-entry structure.
  fIsFC0Allowed specifies if the access to FC=0 is allowed

Return
  The function returns the handle of the created WibuBox entry if the
  operation was successful or NULL if the function failed.

*****************************************************************************/

static HWKBENTRY OpenBoxEntry(INT nPort, WKBSERIAL *pwkbsr,
    INT iEntry, ULONG flCtrl, ULONG ulUserCode, HWKBSYSTEM hwkbsys,
    BOOL8 fIsFC0Allowed)
{
  WKBBOXENTRY wkbbxe;
  UINT uType = 0;
  OPENENTRY *pope = NULL;
  INT nVersion = 0;
  ULONG fuVersion = 0;
  HWKBENTRY hwkbe;

  // access specified port
  if (!WkdSelectPort(nPort, TRUE)) {
    // invalid port => return with error
    return NULL;
  } // if
  // determine contents of specified WibuBox open structure
  if (!WkdGetOpenBoxEntry(pwkbsr, iEntry, &wkbbxe, &nVersion, fIsFC0Allowed)) {
    // any error => return with error
    goto EXIT_FN_ERROR;
  } // if
  // set required algorithm version
  fuVersion = flCtrl & WKB_VERSION_MASK;
  if ((fuVersion == WKB_VERSION_STD) || (fuVersion == WKB_VERSION_RED)
      || ((fuVersion >= WKB_VERSION4) && (fuVersion <= WKB_VERSION_MAX))
    ) {
    // check if version of WibuBox is OK
    if (!WkdCheckBoxVersion(&fuVersion, (USHORT) nVersion, pwkbsr)) {
      // invalid version of the found WibuBox => return with error
      cstWkd.idError = WKB_ERROR_INVALID_VERSION;
      return FALSE;
    } // if
  } // if(fuVersion)
  cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
  uType = (UINT)wkbbxe.flCtrl & WKB_BE_TYPEMASK;
  ASSERT(uType != WKB_BE_EMPTY);
  if (uType >= WKB_BE_DATA) {
    // invalid entry => return with error
    goto EXIT_FN_ERROR;
  } // if
  if (uType == WKB_BE_MASTER) {
    if ((ulUserCode & wkbbxe.ulUserCode) != ulUserCode) {
      // User Code doesn't match: return with error
      goto EXIT_FN_ERROR;
    } // if
    // replace Master Code by used User Code in opened entry
    wkbbxe.ulUserCode = ulUserCode;
  } // if

  if ( ( (nVersion == 1) && (fuVersion > WKB_VERSION1) ) ||
       ( (nVersion == 2) && (fuVersion == WKB_VERSION1) ) ) {
    // bad required version: return with error
    cstWkd.idError = WKB_ERROR_INVALID_VERSION;
    goto EXIT_FN_ERROR;
  } // if
  // entry is valid: create handle & initialize it
  pope = WkdCreateHandle(HTYP_ENTRY, sizeof(OPENENTRY), (HANDLE*) &hwkbe);
  if (hwkbe == NULL) {
    // memory too small => return with error
    goto EXIT_FN_ERROR;
  } // if
  // set version & other open-entry status flags
  pope->flCtrl = fuVersion | (flCtrl & WKB_OPEN_FLAGS);
  if (uType == WKB_BE_MASTER) {
    // master entry opened => set flag
    pope->flCtrl |= WKB_OE_MASTER;
  } // if
  pope->hwkbsys = hwkbsys;
  pope->lPort = nPort;
  pope->wkbsr = *pwkbsr;
  pope->usBoxVersion = (USHORT)nVersion;
  pope->iEntry = iEntry;
  pope->ulFirmCode = wkbbxe.ulFirmCode;
  pope->ulUserCode = wkbbxe.ulUserCode;
  // set single bit for MULTIENTRY option
  pope->fuEntryOpen = 1 << pope->iEntry;
  // set expiration date as unknown
  pope->uExpDate = UNDET_EXPDATE;
  // check if new entry conflicts with any other entry
  if (WkdIsEntryConflict(pope)) {
    // entry globally used by another open process or exclusive conflict:
    // * free created entry again
    WkdDestroyHandle(hwkbe);
    ASSERT(cstWkd.idError == WKB_ERROR_ENTRY_ALREADY_USED ||
      cstWkd.idError == WKB_ERROR_EXCLUSIVE_MODE_CONFLICT);
    // return with error
    goto EXIT_FN_ERROR;
  } // if
  // entry now opened: other entries are initially set to 0
  cstWkd.idError = WKB_NO_ERROR;
  // return without error created handle
  return hwkbe;
EXIT_FN_ERROR:
  // return with error but release accessed port
  WkdReleasePort();
  return NULL;
} // OpenBoxEntry()


/*****************************************************************************
W k b O p e n 2
==============================================================================

This function opens a single entry in a WibuBox for encryption or decryption
with subsequent calls of the functions WkbSelect2() and WkbCrypt2().

The function supports two modes. The first mode (WKB_OPEN_FIND) searches
a WibuBox entry with the special Firm and User Code at some or all available
WibuBox ports and returns a handle of the first found entry.
The second mode (WKB_OPEN_ENTRY) opens an partial specified entry which was
found by an earlier call of WkbOpen2() and returned
by a call of WkbQueryEntry2() or which was returned from another function,
for example WkbListBox2().

Parameters
  hwkbsys .. contains the handle of a valid accessed WibuKey subsystem.
             To open a WibuBox at the standard local subsystem the predefined
             handle value HWKB_LOCAL can be used.
  flCtrl ... contains flags which control the opening of the WibuBox.
             Following flag values are supported:
               The first flag group specifies the mode of the function:
                 WKB_OPEN_FIND  (0x0000)
                     searches a WibuBox entry with the specified Firm and
                     User Code.
                 WKB_OPEN_ENTRY (0x0010)
                     opens the specified WibuBox entry.
                 WKB_OPEN_MULTIENTRY (0x0020)
                     opens all WibuBox entries which match the specified
                     Firm and User Code.
               The second group specifies the version of the WibuBox which
               is opened:
                 WKB_VERSION1 (0x0000)  open WibuBox for WibuKey version 1.
                 WKB_VERSION2 (0x0001)  open WibuBox for WibuKey version 2.
                 WKB_VERSION3 (0x0002)  open WibuBox for WibuKey version 3.
                 WKB_VERSION4 (0x0003)  open WibuBox for WibuKey version 4.
                 WKB_VERSION5 (0x0004)  open WibuBox for WibuKey version 5.
               The third group contains additional flags:
                 WKB_PORTNUMBER (0x0008)  use port number and port names.
  pszPort .. points to a string which contains one or more port specifications
             at which WibuBoxes are searched. If this pointer is set to NULL,
             all available WibuKey ports of the specified WibuKey subsystem
             are scanned.
             This parameter is ignored and must be set to NULL if the
             WKB_OPEN_ENTRY mode is specified.
             If the flag WKB_PORTNUMBER is set, this parameter is not inter-
             preted as pointer but must contain a valid port specification
             number which is compatible with WibuKey version 1.
  lFirmCode  contains the Firm Code (range 0..16777215) for the find mode.
             This parameter is ignored and must be set to 0 if a parameter
             is specified at <pvCtrl>. If this mode is not used, a value
             greater 0 must be used; if WKB_ACCESS_CODE (-1) is set, the
             Firm Code of the access entry of the accessed subsystem is
             used which was specified in the corresponding WkbAccess2 call.
  lUserCode  contains the User Code (range 0..16777215) for the find mode.
             This parameter is ignored and must be set to 0 if a parameter
             is specified at <pvCtrl>. If this mode is not used, a value
             greater 0 must be used; if WKB_ACCESS_CODE (-1) is set, the
             User Code of the access entry of the accessed subsystem is
             used which was specified in the corresponding WkbAccess2 call.
  pvCtrl ... points to optional data which are used in serveral commands.
             For this version of the function it points to an initialized
             WKBOPENENTRY or WKBOPENFIND structure.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Comments
  The length of the string which specifies the port names is limited to 512
  characters. Port names can be specified in lower or upper case, the case is
  not significant. All ports a single port group can be selected with a
  trailing star, "LPT*" searches for all available LPT ports, started with
  LPT1. Multiple port names or port group specifications must be separated
  with the vertical bar ("|"), the search order starts the first specifcation.
  For example the specification "COM3|LPT1|LPT*|COM2" searches at first at
  COM3, then at LPT1, then at all LPT ports and then at COM2.

  The flag WKB_VERSION1 can only be used for an entry in
  a WibuBox version 1 or 3.
  The flag WKB_VERSION2 can only be used for an entry in
  a WibuBox version 2, 3, 4, 6.
  The flag WKB_VERSION3 can only be used for an entry in
  a WibuBox version 5, 7.
  The flag WKB_VERSION4 can only be used for an entry in
  a WibuBox version 6.
  The flag WKB_VERSION5 can only be used for an entry in
  a WibuBox version 7.

*****************************************************************************/

HWKBENTRY WKAPIENTRY WkbOpen2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  const char *pszPort, LONG lFirmCode, LONG lUserCode, void *pvCtrl)
{
  UINT uFuncCmd = 0;
  UINT fuVersion = 0;
  INT iRet = 0;
  HWKBENTRY hwkbe = NULL;
  ULONG ulEntrySkip = 0;

  THREAD_API_LOCK

  // convert handle addresses to true addresses if required
  if (!(flCtrl & WKB_PORTNUMBER)) {
    if (!WkdConvertMemHandle((const void**) &pszPort)) {
      goto RETURN;
    } // if
  } // if
  if (!WkdConvertMemHandle((const void**) &pvCtrl)) {
    goto RETURN;
  } // if
  // determine accessed system, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if

  // check parameter values
  cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
  uFuncCmd = (UINT)flCtrl & WKB_OPEN_MASK;
  fuVersion = (UINT)flCtrl & WKB_VERSION_MASK;

  switch (uFuncCmd) {
    case WKB_OPEN_FIND:
    case WKB_OPEN_MULTIFIND:
      if (pvCtrl != NULL) {
        WKBOPENFIND *pwkbof = pvCtrl;
        // assume WKBOPENFIND structure at <pvCtrl> parameter
        // the first 3 parameters must be 0
        if ( (lFirmCode != 0) || (lUserCode != 0) || (fuVersion != 0) || ( (flCtrl & WKB_OPEN_FLAGS) != 0) ) {
          // WKBOPENFIND specified => other parameters must be 0
          ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PARAMETER);
          goto RETURN;
        } // if
        // set now Firm/User Code variables from structure
        lFirmCode = pwkbof->ulFirmCode;
        lUserCode = pwkbof->ulUserCode;
        flCtrl |= pwkbof->flCtrl & WKB_OPEN_FLAGS;
        fuVersion = (UINT)pwkbof->flCtrl & WKB_VERSION_MASK;
        ulEntrySkip = pwkbof->ulEntrySkip;
      } // if
      if (hwkbsys == HWKB_LOCAL) {
        // local system: set parameters 
        // no negative Firm or User Code, Firm Code not 0
        if ( (lFirmCode <= 0) || (lUserCode < 0) ) {
          // FC=0 is only allowed if the flag WKB_CTRL_NOSRVR
          // is set via the function WkbSetPort2 (for internal use only).
          if (!( (lFirmCode == 0) && (cstWkd.flGlobalInternal & DRVC_FC0_ALLOWED))) {
            goto RETURN;
          } // if
        } // if
      } else {
        // true subsystem: WKB_ACCESS_CODE permitted, otherwise no negative
        // Firm or User Code, Firm Code not 0, no control data
        if ( (lFirmCode < -1) || (lUserCode < -1) || (lFirmCode == 0) ) {
          // FC=0 is only allowed if the flag WKB_CTRL_NOSRVR
          // is set via the function WkbSetPort2 (for internal use only).
          if (!( (lFirmCode == 0) && (cstWkd.flGlobalInternal & DRVC_FC0_ALLOWED))) {
            goto RETURN;
          } // if
        } // if
      } // if
      break;
    case WKB_OPEN_ENTRY:
      // Firm and User Code must be 0, additional control data
      if ( (pszPort != NULL) || 
           (lFirmCode != 0) || (lUserCode != 0)  || 
           ((flCtrl & ( WKB_OPEN_FLAGS | WKB_PORTNUMBER | WKB_VERSION_MASK)) != 0) ||
           (pvCtrl == NULL)
        ) {
        // WKBOPENENTRY must be specified, other parameters must be 0
        ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PARAMETER);
        cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
        goto RETURN;
      } // if
      break;
    default:
      // other commands are not supported
      cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
      goto RETURN;
  } // switch
  if (cstWkd.psys->fuFlags & SFLAG_NONETLIMIT) {
    // limited open of subsystem: set WKB_OPEN_NONETLIMIT flag
    flCtrl |= WKB_OPEN_NONETLIMIT;
  } // if
  if (cstWkd.psys->fuFlags & SFLAG_EXCLUSIVE) {
    // exclusive open of subsystem: set WKB_OPEN_EXCLUSIVE flag
    flCtrl |= WKB_OPEN_EXCLUSIVE;
    if (uFuncCmd == WKB_OPEN_ENTRY) {
      WKBOPENENTRY *pwkboe = pvCtrl;
      pwkboe->flCtrl |= WKB_OE_EXCLUSIVE; 
    } // if
  } // if
  if (cstWkd.psys->fuFlags & SFLAG_STATIONSHARE) {
    // station share open of subsystem: set WKB_OPEN_STATIONSHARE flag
    flCtrl |= WKB_OPEN_STATIONSHARE;
    if (uFuncCmd == WKB_OPEN_ENTRY) {
      WKBOPENENTRY *pwkboe = pvCtrl;
      pwkboe->flCtrl |= WKB_OE_STATIONSHARE; 
    } // if
  } // if
  if (hwkbsys != HWKB_LOCAL) {
    // WKB_ACCESS_CODE: get Firm and User Code from local subsystem
    if (lFirmCode == WKB_ACCESS_CODE) {
      lFirmCode = cstWkd.psys->lFirmCode;
    } // if
    if (lUserCode == WKB_ACCESS_CODE) {
      lUserCode = cstWkd.psys->lUserCode;
    } // if
  } // if
  // no error until here
  cstWkd.idError = WKB_NO_ERROR;
  {//// execute driver operations
    OPENENTRY *pope = NULL;
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    ASSERT(cstWkd.fuSysType & STYPE_LOCAL);
    ASSERT(uFuncCmd <= WKB_OPEN_MULTIFIND);
    // check parameters
    switch (uFuncCmd) {
      case WKB_OPEN_FIND:
      case WKB_OPEN_MULTIFIND: {
        BOOL fParse; // flag "in parsing operation"
        BOOL fMultiple = FALSE; // flag "check multiple ports"
        // special error code
        INT idSpecificError = WKB_NO_ERROR;
        // version check of the encryption algorithm
        if ( (fuVersion > WKB_VERSION_MAX) &&
             (fuVersion != WKB_VERSION_STD) && 
             (fuVersion != WKB_VERSION_RED)
          ){
          // printf("---- Open VersionCheck assert, \n");
          ASSERT(cstWkd.idError == WKB_ERROR_ALGORITHM_NOT_SUPPORTED);
          ASSERT(cstWkd.idError == WKB_ERROR_INVALID_VERSION); // C:ha - include also VersionCheck
          cstWkd.idError = WKB_ERROR_INVALID_VERSION;
          goto RETURN;
        } // if
        if ( (hwkbsys != HWKB_LOCAL) &&
             !(cstWkd.flCtrl & WKB_PORTNUMBER) && 
             (pszPort == NULL)
          ) {
          // no port specified and subsystem used
          // * try at first subsystem entry if Firm/User Code match
          if (cstWkd.psys->lFirmCode != lFirmCode) {
            goto EXIT_USE_SUBSYS_ENTRY;
          } // if
          if (!cstWkd.psys->fMaster) {
            // used entry contains User Code: must match exactly
            if (cstWkd.psys->lUserCode != lUserCode) {
              goto EXIT_USE_SUBSYS_ENTRY;
            } // if
          } else {
            // used entry contains Master Code
            // * nothing to do here: check later when WibuBox entry read
          } // if
          // open entry and check WibuBox version and entry type
          hwkbe = OpenBoxEntry(cstWkd.psys->usPort, &cstWkd.psys->wkbsr,
            cstWkd.psys->usEntry, flCtrl, lUserCode, hwkbsys, FALSE);
          if (hwkbe != NULL) {
            // entry found => use this entry, no search operation
            cstWkd.idError = WKB_NO_ERROR;
            pope = WkdGetHandlePtr(hwkbe);
            ASSERT(pope != NULL);
            goto CHECK_FIND;
          } // if
        EXIT_USE_SUBSYS_ENTRY:;
        } // if
        // start search operation via all known ports
        // * create kernel BOX entry structure
        pope = WkdCreateHandle(HTYP_ENTRY, sizeof(OPENENTRY), (HANDLE*) &hwkbe);
        if (hwkbe == NULL) {
          goto RETURN;
        } // if
        // parameters checked: set fix parameters into OPENENTRY structure
        pope->flCtrl = fuVersion | (flCtrl & WKB_OPEN_FLAGS);
        pope->hwkbsys = hwkbsys;
        pope->ulFirmCode = lFirmCode;
        pope->ulUserCode = lUserCode;
        for (;;) {
          // for-loop for several entries of port name specification
          INT nPort;
          fParse = TRUE;
          if (cstWkd.flCtrl & WKB_PORTNUMBER) {
            // local scan with port specification
            nPort = (INT)(long)pszPort;
            if ( ( (nPort < 0) || (nPort > 127) ) && (nPort != 255) ) {
              // bad port value specified
              ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PORT);
              break;
            } // if
          } else if (pszPort == NULL) {
            // scan all available ports
            nPort = 255;
            cstWkd.flCtrl |= WKB_PORTNUMBER;
          } else {
            // interpret contents of port string
            // * list of port names, separated by '|'s
            cstWkd.idError = WKB_ERROR_INVALID_PORT;
            while ( (*pszPort == ' ') || (*pszPort == '\t') ) {
              pszPort++;
            } // if
            if (*pszPort == '*') {
              pszPort++;
              // skip trailing spaces
              while (*pszPort == ' ' || *pszPort == '\t') {
                pszPort++;
            } // if 
              // no end of specification => error
              if (*pszPort != 0) {
                break;
              } // if
              // set mask "get all ports"
              nPort = 255;
            } else {
              // parse specification into <nPort>
              pszPort = WkdParsePortSpec(pszPort, &nPort);
              if (pszPort == NULL) {
                ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PORT);
                break;
              } // if
              if (*pszPort == '|') {
                // parsing location to next specification
                pszPort++;
                // more than one port specified: set flag
                fMultiple = TRUE;
              } else if (*pszPort != 0) {
                // otherwise: end of string required
                ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PORT);
                break;
              } // if
            } // if
          } // if
          // analyse type of port specification, scan port(s)
          // * now no longer parsing errors
          fParse = FALSE;
          if ((nPort & 7) != 7) {
            // scan single port
            // WibuBox found => continue at multi-check
            if (FindBoxEntry(nPort, pope, &idSpecificError)) {
              goto CHECK_FIND;
            } // if
          } else {
            // scan multiple port
            INT iLastPort = 0;
            // set flag "group specified"
            fMultiple = TRUE;
            if (nPort != 255) {
              // scan all ports of one port group
              INT iPort;
              iLastPort = (nPort | 7) - 1;
              for (iPort = nPort & ~7; iPort < iLastPort; iPort++) {
                // WibuBox found => exit loop
                if (FindBoxEntry(iPort, pope, &idSpecificError)) {
                  goto CHECK_FIND;
                } // if
              } // for
            } else {
              // scan all ports
              INT iPort = 0;
              for (iPort = 0; iPort < 126; iPort++) {
                if ((iPort & 7) == 7) {
                  // skip wildcard
                  continue;
                } // if
                if (FindBoxEntry(iPort, pope, &idSpecificError)) {
                  // WibuBox found => exit loop
                  goto CHECK_FIND;
                } // if
              } // for
            } // if
          } // if
          if (cstWkd.flCtrl & WKB_PORTNUMBER || (*pszPort == 0) ) {
            // end of parsing: analyse if any error
            if (cstWkd.idError != WKB_NO_ERROR) {
              WkdDestroyHandle(hwkbe);
              if (fMultiple && !fParse) {
                // multiple WibuBoxes scanned: create sum error
                if (idSpecificError != WKB_NO_ERROR) {
                  // set special error (WKB_ERROR_INVALID_VERSION ...)
                  cstWkd.idError = idSpecificError;
                } else {
                  // set general sum error
                  cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
                } // if
              } // if
            } // if
            // exit now loop
            break;
          } // if
        } // for
        if (cstWkd.idError != WKB_NO_ERROR) {
          // any error => return with error
          goto EXIT_KRNL;
        } // if
      CHECK_FIND:
        // ======================================================
        // analyse found entry: HWKBE is valid, no error detected
        // ======================================================
        ASSERT(cstWkd.idError == WKB_NO_ERROR);
        if ( (uFuncCmd == WKB_OPEN_MULTIFIND) || (ulEntrySkip != 0) ) {
          // expand opened entry to compatible entries in WibuBox
          WKBBOXDATA *pwkbbxd = NULL;
          UINT cb = 0;
          // get all entries for analyzing
          pwkbbxd = WkdGetBoxContents(&pope->wkbsr, 0xFFFF, &cb);
          if (pwkbbxd == NULL) {
            // any error in reading => close again & return with error
            WkdDestroyHandle(hwkbe);
            goto EXIT_KRNL;
          } // if
          if (uFuncCmd == WKB_OPEN_MULTIFIND) {
            // set flag that multifind-open
            pope->fMultiFind = TRUE;
          } // if
          {// analyse entry: find all matching
            UINT fuEntryOpen = 0;
            UINT iEntry = 0;
            WKBBOXENTRY *pwkbbxe = NULL;
            for (iEntry = 0, pwkbbxe = pwkbbxd->awkbbxe;
                iEntry < pwkbbxd->cEntry;
                iEntry++, pwkbbxe++
              ) {
              UINT uEntry = (UINT)pwkbbxe->flCtrl & WKB_BE_INDEXMASK;
              switch (pwkbbxe->flCtrl & WKB_BE_TYPEMASK) {
                case WKB_BE_FCUC:
                  // FC/UC entry: exact match required
                  if ( (pwkbbxe->ulFirmCode == pope->ulFirmCode) &&
                       (pwkbbxe->ulUserCode == pope->ulUserCode)
                    ) {
                    // set entry index into multi-mask
                    break;
                  } // if
                  // entry does not match: check next
                  continue;
                case WKB_BE_MASTER:
                  // master entry: FC must match, UC must be compatible
                  if ( (pwkbbxe->ulFirmCode == pope->ulFirmCode) &&
                       ( (pope->ulUserCode & pwkbbxe->ulUserCode) == pope->ulUserCode)
                    ) {
                    // entry match => set entry index into multi-mask
                    break;
                  } // if
                  // entry does not match: check next
                  continue;
                default:
                  // other entry type: does not match, check next
                  continue;
              } // switch
              // set entry index into multi-open mask
              SET_BIT(fuEntryOpen, uEntry);
            } // for
            if (ulEntrySkip != 0) {
              // entry skipping active: reduce list of found entries
              fuEntryOpen &= ~(UINT)ulEntrySkip;
              if (fuEntryOpen == 0) {
                // result mask empty => WibuBox entry not found
                cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
                WkdDestroyHandle(hwkbe);
                goto EXIT_KRNL;
              } // if
              // found first usable entry and use as index of entry
              // * avoid this when subsystem is used and its entry is valid
              if (hwkbsys == HWKB_LOCAL || IS_BIT(fuEntryOpen, iEntry)) {
                //!N:mabu/V2.41: subsystem dependence changed
                for (iEntry = 0; iEntry < 10; iEntry++) {
                  if (IS_BIT(fuEntryOpen, iEntry)) {
                    // bit set => use this as default entry, exit loop
                    pope->iEntry = iEntry;
                    break;
                  } // if
                } // for
              } // for
              ASSERT(iEntry < DIMOF(pwkbbxd->awkbbxe));
              if (!pope->fMultiFind) {
                // set only one open entry in bit array
                fuEntryOpen = 0;
                SET_BIT(fuEntryOpen, iEntry);
              } // if
            } // if
            // WibuBox returned always first matchable entry
            ASSERT(pope->iEntry != -1);
            // original find index must be part of mask
            ASSERT(IS_BIT(fuEntryOpen, pope->iEntry));
            pope->fuEntryOpen = fuEntryOpen;
            if (pope->fMultiFind) {
              // determine better entries than first found entry
              // * candiates are Limit Counters & Expiration Date
              WkdSetMultiFindOptions(pwkbbxd, pope);
            } // if
          } // block
          // free allocated memory
          BsrFreeMem(pwkbbxd);
          break;
        } // if
        break;
      } // case
      case WKB_OPEN_ENTRY: {
        // open specified entry & return
        WKBOPENENTRY *pwkboe = pvCtrl;
        BOOL8 fIsFC0Allowed = 
            ((cstWkd.flGlobalInternal & DRVC_FC0_ALLOWED) == DRVC_FC0_ALLOWED) ? TRUE : FALSE;
        hwkbe = OpenBoxEntry ((INT)pwkboe->lPort, &pwkboe->wkbsr,
            pwkboe->usEntry, pwkboe->flCtrl, pwkboe->ulUserCode,
            hwkbsys, fIsFC0Allowed);
        if (hwkbe == NULL) {
          // specified entry cannot be opened => return with error
          ASSERT(cstWkd.idError != 0);
          goto EXIT_KRNL;
        } // if
        // entry now opened: other entries are initially set to 0
        cstWkd.idError = WKB_NO_ERROR;
        break;
      } // case
      default:
        ASSERT(0);
        break;
    } // switch
  EXIT_KRNL:
    // release port if accessed
    WkdReleasePort();
  }
RETURN:

  iRet = (INT)WkdReturnChecked((INT)hwkbe);

  THREAD_API_UNLOCK

  return (HWKBENTRY)(long)iRet;
} // WkbOpen2()


/*****************************************************************************

---- WibuBox Entry Management -----------------------------------------------

*****************************************************************************/

/*****************************************************************************
W k b Q u e r y E n t r y 2
==============================================================================

This function queries information about the specified opened entry and
writes it into the the structure at the specified destination address. The
function supports two mode. The first mode (WKB_QE_OPEN) returns the driver
internal stored data about an opened WibuBox entry. The second mode
(WKB_QE_BOX) reads the complete data of the opened WibuBox entry by reading
the specific WibuBox contents. The second mode is useful to determine the
added data or limit counter of a opened WibuBox entry.

Parameters
  hwkbe   contains the handle of an opened WibuBox entry.
  flCtrl  contains flags which controls the execution of the function.
          Following flags are currently supported:
            WKB_QE_OPEN (0x0000)  returns the internal WkbOpen2() data into
                                  a WKBOPENENTRY structure.
            WKB_QE_BOX  (0x0100)  returns the complete contents of the
                                  opened WibuBox entry in a WKBBOXENTRY
                                  structure.
  pvDest  points to a destination buffer which receives the status data.
          For this version of the function the structure WKBOPENENTRY or
          WKBBOXENTRY is used.
  cbDest  contains the length of the destination buffer in bytes. This should
          be the enough to receive a WKBOPENENTRY structure and the name
          string of the port or to receive the WKBBOXENTRY structure.

Comments
  If the destination buffer is long enough the WKB_QE_OPEN mode stores the
  name of the port behind the WKBOPENENTRY structure as a 0 terminated string.
  To address this string from the WKBOPENENTRY structure the WKB_USEOFFSETS
  bit is set in the <flCtrl> member and the <pszPort> member contains the
  offset of the string relative to the start of the structure.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cbDest> is too small to receive all returned
    information, the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cbDest> is large enough, the function returns the number
    of transferred bytes.

*****************************************************************************/

INT WKAPIENTRY WkbQueryEntry2(HWKBENTRY hwkbe, ULONG flCtrl,
  void *pvDest, UINT cbDest)
{
  UINT uFuncCmd = 0;
  UINT cbDestData = 0;
  INT iRet = 0;


  THREAD_API_LOCK


  // convert handle address to true address if required
  if (cbDest != 0 && !WkdConvertMemHandle((const void**) &pvDest)) {
    // any error => goto return
    goto RETURN;
  } // if
  // check parameter values
  cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
  uFuncCmd = (UINT)flCtrl & WKB_QE_MASK;
  if (uFuncCmd > WKB_QE_MULTIENTRY) {
    // invalid command specified => error & return
    cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
    goto RETURN;
  } // if
  // determine accessed system, initialize <cstWkd>
  if (!WkdGetEntrySystem(hwkbe, flCtrl)) {
    goto RETURN;
  } // if
  {// execute kernel operations
    ASSERT(cstWkd.fuSysType & STYPE_LOCAL);
    switch (uFuncCmd) {
      case WKB_QE_OPEN: {
        char szName[20];
        UINT cch = 0;
        WKBOPENENTRY *pwkboe = (WKBOPENENTRY*)pvDest;
        memset(szName, 0x0, sizeof(szName));
        if (!WkdGetPortName(szName, sizeof(szName), &cch)) {
          goto EXIT_KRNL;
        } // if
        // initialize destination length
        cbDestData = (UINT) sizeof(WKBOPENENTRY);
        if (cbDest >= cbDestData) {
          // copy data directly from internal structure to external structure
          pwkboe->flCtrl = cstWkd.pope->flCtrl;
          pwkboe->hwkbsys = cstWkd.hwkbsys;
          pwkboe->usFill = cstWkd.pope->usFill;
          pwkboe->lPort = cstWkd.pope->lPort;
          pwkboe->pszPort = cstWkd.pope->pszPort;
          pwkboe->wkbsr.ulSerial = cstWkd.pope->wkbsr.ulSerial;
          pwkboe->wkbsr.usMask = cstWkd.pope->wkbsr.usMask;
          pwkboe->wkbsr.usReserved = cstWkd.pope->wkbsr.usReserved;
          pwkboe->usVersion = cstWkd.pope->usBoxVersion;
          pwkboe->usEntry = cstWkd.pope->iEntry;
          pwkboe->ulFirmCode = cstWkd.pope->ulFirmCode;
          pwkboe->ulUserCode = cstWkd.pope->ulUserCode;
        } // if
        // add port string at end of structure if possible
        // create WKB_ERROR_BUFFER_OVERFLOW error if not enough memory
        WkdSetStructString(pwkboe, cbDest, &pwkboe->pszPort,
          &cbDestData, szName);
        break;
      } // case
      case WKB_QE_BOX: {
        WKBBOXENTRY *pwkbbxe = (WKBBOXENTRY*)pvDest;
        // WKBOPENENTRY *pwkboe = NULL;
        INT nVersion = 0; // dummy
        cbDestData = sizeof(WKBBOXENTRY);
        if (cbDest < cbDestData) {
          // specified destination too small => return with error
          goto EXIT_KRNL;
        } // if
        // execute list box operation with specified data
        OPENENTRY* pwkboe = (OPENENTRY*)WkdGetHandlePtr(hwkbe);
        // access port of current WibuBox entry
        if (!WkdSelectPort((INT)cstWkd.pope->lPort, TRUE)) {
          // port cannot be accessed => return with error
          goto EXIT_KRNL;
        } // if
        WkdGetOpenBoxEntry(&pwkboe->wkbsr, pwkboe->iEntry,
            pwkbbxe, &nVersion, FALSE);
        } // case
        break;
      case WKB_QE_MULTIENTRY: {
        // return all selected entries
        WKBBOXDATA *pwkbbxd = NULL;
        // access port of current WibuBox entry
        if (!WkdSelectPort((INT)cstWkd.pope->lPort, TRUE)) {
          // port cannot be accessed => return with error
          goto EXIT_KRNL;
        } // if
        pwkbbxd = WkdGetBoxContents(&cstWkd.pope->wkbsr,
          cstWkd.pope->fuEntryOpen, &cbDestData);
        if (pwkbbxd == NULL) {
          // any error in reading => return with error
          goto EXIT_KRNL;
        } // if
        if (cbDest >= cbDestData) {
          // specified destination large enough: transfer data to destination
          memcpy(pvDest, pwkbbxd, (size_t)cbDestData);
        } // if
        // free allocated memory
        BsrFreeMem(pwkbbxd);
        break;
      } // case
      default:
        // codes already checked at begin of function
        ASSERT(0);
        break;
    } // switch
  EXIT_KRNL:
    // free accessed port (if any)
    WkdReleasePort();
  }
RETURN:

  iRet = WkdReturnLength((INT)cbDestData);
  THREAD_API_UNLOCK

  return iRet;
} // WkbQueryEntry2()


/*****************************************************************************
W k d C l o s e
==============================================================================

This function implements the internally WibuBox entry close operation.

Parameters
  hwkbe  contains the handle of an opened WibuBox entry.

*****************************************************************************/

void WkdClose(HWKBENTRY hwkbe)
{
  // determine accessed system, initialize <cstWkd>
  if (!WkdGetEntrySystem(hwkbe, 0)) {
    return;
  } // if
  // destroy complete object
  WkdDestroyHandle(hwkbe);
} // WkdClose()


/*****************************************************************************
W k b C l o s e 2
==============================================================================

This function closes the WibuBox entry which was opened by a previous call
of the function WkbOpen2().

Parameters
  hwkbe  contains the handle of an opened WibuBox entry.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbClose2(HWKBENTRY hwkbe)
{
  INT iRet = 0;


  THREAD_API_LOCK


  WkdClose(hwkbe);
  iRet = WkdReturnBool();

  THREAD_API_UNLOCK

  return iRet;
} // WkbClose2()


// ====================================
// end of clopen.c
// ====================================
