/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clpgm.c                                WIBU-BOX Programming API Implementation
==============================================================================

This module implements the following WibuKey API functions:
  WkbProgram2()

Following, only WIBU-SYSTEMS internally used functions are also implemented
by this module:
  _WkbProgramLow()

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
92Sep13 1.33 mabu: start of development
93Mar18 1.33 mabu: custom source variant released
93Mar21 1.35 mabu: adaption to Windows
93Jun06 1.35 mabu: conversation to large memory model.
93Jun20 1.35 mabu: adaptation to Win32 environment.
93Jul02 1.35 mabu: first adaptations to WibuBox version 2
93Jul22 1.35 mabu: adapation to DOS
93Aug21 2.00 mabu: greater changes in API of new functions
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94Aug10 2.00 mabu: adaptation to new WkNet/WkLAN API
94Oct11 2.00 mabu: implementing the look-ahead-function
94Nov11 2.00 mabu: version 2.00 released
95Feb08 2.02 mabu: bug in CreateHandle() function removed
95Feb09 2.05 mabu: extracted from KRNL.C
95Aug02 2.06 mabu: time out delay time after programming increased to 300 ms
95Nov08 2.09 mabu: extensions for WibuBox 4 implemented
96Jun23 2.13 mabu: renamed from BSPGM.C
97Feb10 2.25 mabu: universal call mechanism for _WkbProgramLow() (only local)
97Jun11 2.31 mabu: litte/big endian adaptation for Macintosh
97Aug06 2.31 mabu: ignoring of hold-errors in WkdReadStatus()
98Feb07 2.41 mabu: integration of new error codes
98Mar16 2.41 mabu: begin of adaptation to new driver class model
00Jan11 2.53 as:   reseting the error code in _WkbProgramLow() (bug fixed)
00Jun06 3.00 as:   updating of Box/U forced to support new box commands
03Jun25 3.33 as:   WkbProgram2()/WKB_PGM_EXTMEM added
06Jan25 5.10 ha:   CodeMerge
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

// flag "programming of WIBU-BOX is currently locked"
static BOOL fPgmLocked;
static REQOPTIONS reqopt;


/*****************************************************************************
_ W k b P r o g r a m L o w
==============================================================================

This function sends a created programming sequence to the WibuBox and returns
the status. This function is called by the API-Function WkbProgram2()

Parameters
  nPort .. contains the port number at which the WibuBox is programmed.
  pwkbsr   contains the serial number of the WibuBox which entry is
           programmed. If this value is NULL or the mask byte of the serial
           number is 0, all sent data bytes are specified in <pabData>.
           This is used for some internal operations of WIBU-SYSTEMS (for
           example to change the serial number of WIBU-BOXes with version 1).
  pabData  points to the programming data sequence. The first byte of this
           sequence contains the entry index (or 0 if configuration
           setting). The further bytes contains the data to be programmed
           and the signature is needed.
  cbData   contains the length of the programming data sequence. The value
           must be 2 or higher.

Return
  The function returns TRUE if the operation was successful and FALSE if not.

Comments
  This function is also used internally by WIBU-SYSTEMS. It is no documented
  API function and does not support memory block handles in pointers or other
  API extensions.

*****************************************************************************/

INT WKAPIENTRY _WkbProgramLow(INT nPort, WKBSERIAL const *pwkbsr,
  const BYTE *pabData, UINT cbData)
{

  // reset the internal error code
  cstWkd.idError = WKB_NO_ERROR;
  {// execute driver operations
    WKDEV_REQUEST *pwkdevreq = NULL;
    DEVOBJECT *pdevobj = NULL;
    BYTE abAnswer[2];
    const BYTE *pbData;
    // access used WibuKey port
    if (!WkdSelectPort(nPort, TRUE)) {
      // port cannot be accessed => return
      goto RET_FALSE;
    } // if
    ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
    // execute via class driver
    // * determine accessed device from serial number
    if (pwkbsr->usMask == 0
      && (cstWkd.flGlobalInternal & DRVC_EVERYSERIAL) == 0) {
      // mask is 0 => set pointer to NULL
      // * special option to modify serial number of WibuBox version 1
      pwkbsr = NULL;
      // set flag that serial number is changed
      // * required for USB class driver
      reqopt.Program.fSetSerialNumber = TRUE;
    } // if
    pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(pwkbsr);
    if (pdevobj == NULL) {
      // device not found => return with error
      goto EXIT_KRNL;
    } // if
    if ( (pabData[1] >= 0xFB) && (pabData[1] <= 0xFF) ) {
      // Possibly reprogramming of serial number: set special flag to update
      // the internal stored device list.
      // * Required for USB class driver.
      // * This commands are secret and must not be published.
      // * They are only used in the manufacturing process.
      reqopt.Program.fSetSerialNumber = TRUE;
    } // if
    // send now programming request command
    // * create request data structure
    pwkdevreq = (WKDEV_REQUEST*)WkdAllocMem(
      FIELDOFFSET(WKDEV_REQUEST, abRequest) + cbData);
    if (pwkdevreq == NULL) {
      // memory too small => return with error
      goto EXIT_KRNL;
    } // if
    // initialize base part of device request structure
    pbData = pabData;
    memset(pwkdevreq, 0, FIELDOFFSET(WKDEV_REQUEST, abRequest));
    pwkdevreq->hwkdev = pdevobj->wkdevobj.hwkdev;
    pwkdevreq->usCmd = BOXCMD_PROGRAM;
    if (pwkbsr != NULL) {
      // no WibuBox verison 1: first byte used as index
      // * first byte is no longer transferred into request data block
      pwkdevreq->usIndex = *pbData++;
      ASSERT(cbData >= 1);
      cbData--;
      pwkdevreq->cbAnswer = sizeof(abAnswer);
    } else {
      // no serial number specified: read only one status byte
      pwkdevreq->cbAnswer = 1;
      // avoid that non-read byte results error
      abAnswer[1] = 0;
    } // if
    pwkdevreq->cbRequest = cbData;
    // set programming sequence into request datagram
    memcpy(pwkdevreq->abRequest, pbData, (size_t)cbData);
    // execute programming command including special options
    if (!cstWkd.pclsdrvSel->pfClDrvRequest(pwkdevreq, abAnswer, &reqopt)) {
      // request not properly executed => return with error
      goto EXIT_KRNL;
    } // if
    // analyse error, if any => set into <cstWkd.idError>
    if (abAnswer[0] != 0) {
      WkdConvertStatusIntoError(abAnswer[0]);
    } else if (abAnswer[1] != 0) {
      WkdConvertStatusIntoError(abAnswer[1]);
    } // if
    // reset flag that serial number programmed (if any)
    reqopt.Program.fSetSerialNumber = FALSE;
    // free allocated memory
EXIT_KRNL:
    if (pwkdevreq != NULL) {
      // memory too small => return with error
      BsrFreeMem(pwkdevreq);
    } // if
    // release access to port
    WkdReleasePort();
  }

  if (cstWkd.idError != WKB_NO_ERROR) {
    // set last error code
    gidWkdLastError = cstWkd.idError;
    // return with error
    return FALSE;
  } // if
  // return without error
  return TRUE;

RET_FALSE:
  // return with error
  return FALSE;
} // _WkbProgramLow()


/*****************************************************************************
W k b P r o g r a m 2
==============================================================================

This function changes an entry of the specified WibuBox or modifies the
configuration of the WibuBox.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To program a WibuBox at the standard local subsystem the
           predefined handle value HWKB_LOCAL can be used.
  flCtrl   contains flags which control the listing of the port. Following
           flags and options are currently supported:
             WKB_PGM_CLEAR        (0x0000) clear complete entry
             WKB_PGM_NEW          (0x0010) set new base entry
             WKB_PGM_ADD          (0x0020) add entry to base entry
             WKB_PGM_MODIFY       (0x0030) modify base entry (user data)
             WKB_PGM_MODIFY_ADDED (0x0040) modify added entry
             WKB_PGM_CONFIG       (0x0080) set WibuBox configuration data
             WKB_PGM_LOCK         (0x0090) lock WibuBox for programming
             WKB_PGM_UNLOCK       (0x00A0) unlock locked WibuBox
             WKB_PGM_EXTMEM       (0x00B0) write ExtMem storage
             WKB_PORTNUMBER       (0x0008) is set to specify a port number
                                  instead of a port name.
  pszPort .. points to a string which contains one single port specification.
             If the flag WKB_PORTNUMBER is set, this parameter is not inter-
             preted as pointer but must contain a valid port specification
             number which is compatible with WibuKey version 1.
  pwkbsr ... contains the serial number of WibuBox which entry is programmed.
  pvpgmData  points to the data structure which contains the data to be
             programmed and the security signature if this is needed. For the
             control flag WKB_PGM_CLEAR, WKB_PGM_ADD, WKB_PGM_MODIFY,
             WKB_PGM_MODIFY_ADDED is this a WKBPROGRAM structure, for the
             control flag WKB_PGM_CONFIG a WKBBOXCONFIG structure.
               The <flCtrl> field of these structures specifies also the
             version of the WibuBox which is programmed. For that, the option
             WKB_SET_BOXVERSION (0x100) must be set and the lower byte
             contains the WibuBox version number, return from WkbListBox2().
               If the WKB_SET_BOXVERSION flag is not set, a WibuBox version 1
             can be specified with WKB_VERSION1 in <flCtrl> and WibuBox
             version 2 or 3 with WKB_VERSION2. This mode is only for
             compatibility with the WibuKey driver version 2.0 up to 2.08 and
             is not not recommended for current and future use.

Return
  The function returns TRUE if the operation was successful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Preliminary Note
  The programming locking feature is currently under discussion.

*****************************************************************************/

INT WKAPIENTRY WkbProgram2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  const char *pszPort, WKBSERIAL const *pwkbsr,
  const void *pvPgmData)
{
  INT idResult = 0;
  UINT uBoxVersion = 0;
  UINT uPgmCmd = 0;

  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pwkbsr)) {
    goto RETURN;
  } // if
  if (!WkdConvertMemHandle(&pvPgmData)) {
    goto RETURN;
  } // if
  uPgmCmd = (UINT)(flCtrl & WKB_PGM_MASK);
  if (uPgmCmd <= WKB_PGM_MODIFY_ADDED) {
    // determine version of WibuBox
    UINT uCtrl = (UINT)((WKBPROGRAM*)pvPgmData)->flCtrl;
    if (uCtrl & WKB_SET_BOXVERSION) {
      uBoxVersion = LOBYTE(uCtrl);
    } else {
      // older specification: support only WibuBox 1 and 3
      uBoxVersion = (uCtrl & WKB_VERSION_MASK) == WKB_VERSION1? 1 : 3;
    } // if
  } // if
  {//// execute driver operations
    BYTE *pabData; // pointer to buffer containing sent programming sequence
    UINT cbData = 0; // number of bytes to send in <pabData>
    ULONG flBoxCtrl = 0;
    UINT uEntry = 0;
    BOOL fUserData = 0;
    UINT uBoxCmd;
    //// validate parameters
    if (uPgmCmd <= WKB_PGM_MODIFY_ADDED) {
      //// read control flags of WKBBOXENTRY structure
      flBoxCtrl = ((WKBPROGRAM*)pvPgmData)->wkbbxe.flCtrl;
      fUserData = (flBoxCtrl & WKB_BE_TYPEMASK) == WKB_BE_DATA;
      uEntry = (UINT)flBoxCtrl & WKB_BE_INDEXMASK;
      cstWkd.idError = WKB_ERROR_INVALID_ENTRY;
      if (uEntry > 9) {
        goto RETURN;
      } // if
    } // if
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    // validate port specification, select & access this port
    if (!WkdAccessPortSpec(pszPort)) {
      goto RETURN;
    } // if
    switch (uPgmCmd) {
      case WKB_PGM_CLEAR:
        //// clear base entry (and added entry if exists)
        // User Data Entry: use variant of WKB_PGM_MODIFY
        uBoxCmd = fUserData? 3 : 0;
        break;
      case WKB_PGM_NEW:
        //// create new base entry (FC/UC, Master or User Data)
        uBoxCmd = 1;
        break;
      case WKB_PGM_ADD:
        //// add Limit Counter or Added Data Entry to base entry
        if (!(flBoxCtrl & (WKB_BE_ADDLIMIT|WKB_BE_ADDDATA))) {
          // no additional entry exists: return with error
          cstWkd.idError = WKB_ERROR_INVALID_ENTRY;
          goto RETURN;
        } // if
        if (uEntry > 4) {
          // bad index of base entry: return with error
          cstWkd.idError = WKB_ERROR_INVALID_ENTRY;
          goto RETURN;
        } // if
        uBoxCmd = 2;
        break;
      case WKB_PGM_MODIFY:
        //// modify base entry: only permitted for user data entries
        if ((flBoxCtrl & WKB_BE_TYPEMASK) != WKB_BE_DATA) {
          // invalid entry for base modification
          goto RETURN;
        } // if
        uBoxCmd = 3;
        break;
      case WKB_PGM_MODIFY_ADDED:
        //// modify added entry: only permitted for limit or data entries
        if (!(flBoxCtrl & (WKB_BE_ADDLIMIT|WKB_BE_ADDDATA))) {
          // no additional entry exists: return with error
          goto RETURN;
        } // if
        if (uEntry > 4) {
          // bad index of base entry: return with error
          cstWkd.idError = WKB_ERROR_INVALID_ENTRY;
          goto RETURN;
        } // if
        uBoxCmd = 4;
        break;
      case WKB_PGM_CONFIG:
        // execute command "configure WibuBox"
        // * no entry used
        uEntry = 0;
        // change configuration
        uBoxCmd = 5;
        break;
      case WKB_PGM_LOCK:
        //// lock specified WibuBox for programming
        if (fPgmLocked) {
          // driver already locked => return with error
          cstWkd.idError = WKB_ERROR_LOCK_VIOLATION;
          goto RETURN;
        } // if
        // set now general flag
        fPgmLocked = TRUE;
        cstWkd.idError = WKB_NO_ERROR;
        goto RETURN;
      case WKB_PGM_UNLOCK:
        // unlock specified WibuBox for programming
        // * reset general flag
        fPgmLocked = FALSE;
        cstWkd.idError = WKB_NO_ERROR;
        goto RETURN;
      case WKB_PGM_EXTMEM: {
        UINT32 ulVersion = WkdGetBoxVersion(pwkbsr);
        if (ulVersion < 6 || ulVersion >= 0x80) {
          cstWkd.idError = WKB_ERROR_OPTION_NOT_SUPPORTED;
          goto RETURN;
        } // if
        // execute command "write ExtMem storage in WibuBox"
        // * no entry used
        uEntry = 0;
        // write ExtMem storage
        uBoxCmd = 6;
        } // case
        break;
      default:
        //// other commands are not supported
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        goto RETURN;
    } // switch

    {// parameters are valid: create parameters for _WkbProgramLow()
      WKBBOXENTRY *pwkbbxe = NULL;
      BYTE *pbSignature = NULL;
      if (uBoxCmd < 5) {
        // no WKB_PGM_CONFIG: calculate scratch pointers
        pwkbbxe = &((WKBPROGRAM*)pvPgmData)->wkbbxe;
        pbSignature = ((WKBPROGRAM*)pvPgmData)->abSignature;
      } // if
      // allocate buffer for data transfer
      // ==>> if uBoxVersion == 1, allocate 1+6+72 bytes
      // ==>> if sub command 0, allocate 2+8 bytes
      // ==>> if sub command 1,2,4, allocate 2+6+8 bytes
      // ==>> if sub command 3, allocate 2+6 bytes
      // ==>> if sub command 4, allocate 2+4 bytes
      // ==>> if sub command 6, allocate 2+2+8+32 bytes
      // ==>> maximum is 79 bytes!!!
      pabData = WkdAllocMem(80);
      if (!pabData) {
        goto RETURN;
      } // if
      // set common part of programming data sequence
      pabData[0] = (BYTE)uEntry;
      // program control code
      pabData[1] = (BYTE)uBoxCmd;
      // the further data format differs now by the value of <uBoxCmd>
      switch (uBoxCmd) {
        case 0:
          //// clear specified entry: send only signature if no user data
          if (uBoxVersion == 1) {
            // version 1: send special clear entry: UC = 0
            pabData[1+0] = LBYTE32(pwkbbxe->ulFirmCode);
            pabData[1+1] = MBYTE32(pwkbbxe->ulFirmCode);
            pabData[1+2] = NBYTE32(pwkbbxe->ulFirmCode);
            memset(pabData+1+3, 0, (size_t)3);
            memcpy(pabData+1+6, pbSignature, (size_t)72);
            cbData = 1 + 6 + 72;
          } else {
            // version 2..4: send 8 bytes signature
            memcpy(pabData+2, pbSignature, (size_t)8);
            cbData = 2 + 8;
          } // if
          break;
        case 1:
        case 3:
          //// create unbundled entry or clear or modify User Data Entry
          //// set new contents & signature if no user data
          if (fUserData) {
            if (uPgmCmd == WKB_PGM_CLEAR) {
              // clear complete User Data Entry
              memset(pabData+2, 0, (size_t)6);
            } else {
              // create or modify User Data Entry
              pabData[2+0] = HBYTE32(pwkbbxe->ulData1);
              pabData[2+1] = LBYTE32(pwkbbxe->ulData2);
              pabData[2+2] = 0xB0;
              pabData[2+3] = LBYTE32(pwkbbxe->ulData1);
              pabData[2+4] = MBYTE32(pwkbbxe->ulData1);
              pabData[2+5] = NBYTE32(pwkbbxe->ulData1);
            } // if
            // no signature required
            cbData = 2 + 6;
          } else if (uBoxVersion == 1) {
            // version 1: set only FC/UC and signature
            pabData[1+0] = LBYTE32(pwkbbxe->ulFirmCode);
            pabData[1+1] = MBYTE32(pwkbbxe->ulFirmCode);
            pabData[1+2] = NBYTE32(pwkbbxe->ulFirmCode);
            pabData[1+3] = LBYTE32(pwkbbxe->ulUserCode);
            pabData[1+4] = MBYTE32(pwkbbxe->ulUserCode);
            pabData[1+5] = NBYTE32(pwkbbxe->ulUserCode);
            memcpy(pabData+1+6, pbSignature, (size_t)72);
            // set length: send long signature
            cbData = 1 + 6 + 72;
          } else {
            // version 2: create Firm/User Code or Master Firm Entry
            pabData[2+0] = LBYTE32(pwkbbxe->ulFirmCode);
            pabData[2+1] = MBYTE32(pwkbbxe->ulFirmCode);
            pabData[2+2] = NBYTE32(pwkbbxe->ulFirmCode);
            pabData[2+3] = LBYTE32(pwkbbxe->ulUserCode);
            pabData[2+4] = MBYTE32(pwkbbxe->ulUserCode);
            pabData[2+5] = NBYTE32(pwkbbxe->ulUserCode);
            memcpy(pabData+2+6, pbSignature, (size_t)8);
            if ((flBoxCtrl & WKB_BE_TYPEMASK) == WKB_BE_MASTER) {
              // set Master Firm Entry option
              pabData[2+2] |= 0xA0;
            } // if
            // set length: signature is required
            cbData = 2 + 6 + 8;
          } // if
          break;
        case 2:
        case 4: {
          // add entry or modify entry
          BYTE *pbEntry;
          BOOL fModify = uBoxCmd == 4;
          ASSERT(uBoxVersion > 1);
          // set length of programming data
          cbData = 2 + 6 + 8;
          //// create or modify added entry:
          ////   only possible for Added Data and Limit Counter entries
          if (uBoxVersion < 4) {
            // set programming signature before entry data
            memcpy(pabData + 2, pbSignature, (size_t)8);
            pbEntry = pabData + 2 + 8;
          } else {
            // set programming signature behind entry data
            memcpy(pabData + 2 + 6, pbSignature, (size_t)8);
            pbEntry = pabData + 2;
          } // if
          if (flBoxCtrl & WKB_BE_ADDDATA) {
            // ==============================
            // add or modify Added Data Entry
            // ==============================
            pbEntry[0] = HBYTE32(pwkbbxe->ulData1);
            pbEntry[1] = LBYTE32(pwkbbxe->ulData2);
            // this entry is always bundle
            pbEntry[2] = 0xB8;
            if (!fModify) {
              // set variation flags
              pbEntry[2] |= (flBoxCtrl & WKB_DATA_VARYMASK) >> 16;
            } // if
            pbEntry[3+0] = LBYTE32(pwkbbxe->ulData1);
            pbEntry[3+1] = MBYTE32(pwkbbxe->ulData1);
            pbEntry[3+2] = NBYTE32(pwkbbxe->ulData1);
          } else {
            // =================================
            // add or modify Limit Counter entry
            // =================================
            ULONG ulCounter;
            ASSERT(flBoxCtrl & WKB_BE_ADDLIMIT);
            // set flag to fix WibuBox bug
            // * see description of REQOPT_PROGRAM structure in <proj.h>
            reqopt.Program.fNoStatusHoldChk = TRUE;
            ulCounter = pwkbbxe->ulData1 & 0xFFFFF;
            // set programming counter only when entry is added
            pbEntry[0] = (BYTE)(fModify? 0 : LBYTE32(pwkbbxe->ulData2));
            pbEntry[1] = MBYTE32(ulCounter);
            pbEntry[2] = (BYTE)(NBYTE32(ulCounter));
            pbEntry[2] |= 0xC0;
            if (fModify && flBoxCtrl & WKB_LIMIT_OFFSET) {
              // set offset addition mode
              pbEntry[2] |= 0x10;
            } // if
            // set lower byte of limit counter into valid entry
            pbEntry[3] = LBYTE32(ulCounter);
            pbEntry[4] = 0;
            pbEntry[5] = 0;
          } // if
          break;
        } // case
        case 5: {
          // ================================
          // change configuration of WibuBox
          // ================================
          UINT uConfig = (UINT)((WKBBOXCONFIG*)pvPgmData)->ulConfig;
          pabData[2+0] = LOBYTE(uConfig);
          pabData[2+1] = HIBYTE(uConfig);
          // set security byte 1
          pabData[2+2] = 0x55;
          // set security byte 2
          pabData[2+3] = 0xAA;
          cbData = 2+4;
          break;
        } // case
        case 6: {
          // ====================
          // write ExtMem storage
          // ====================
          WKBEXTMEM *pwkbxmw = (WKBEXTMEM*)pvPgmData;
          USHORT idPage = pwkbxmw->iPage << 1;
          ASSERT(sizeof(pwkbxmw->abSignature) == 8);
          if (pwkbxmw->fsCtrl & WKB_XM_PROT) {
            // use protected ExtMem access
            idPage |= 1;
            memcpy(pabData + 2 + 2, pwkbxmw->abSignature, sizeof(pwkbxmw->abSignature));
          } else {
            // use unprotected ExtMem access (WKB_XM_USER)
            memset(pabData + 2 + 2, 0, sizeof(pwkbxmw->abSignature));
          } // if
          // set page index
          SET_USHORT_VALUE(pabData + 2, idPage);
          // set content
          memcpy(pabData + 2 + 2 + 8, pwkbxmw->abData, sizeof(pwkbxmw->abData));
          cbData = 2 + 2 + 8 + 32;
          break;
        } // case
        case 7: {
          // read ExtMem storage ==>> see LIST command
          ASSERT(0);
          break;
        } // case
      } // switch
    } // block
    // send complete programming data to WibuBox, return state
    idResult = _WkbProgramLow(cstWkd.nSelPort, pwkbsr, pabData, cbData);
    // reset special options again
    reqopt.Program.fNoStatusHoldChk = FALSE;
    // free memory of allocated buffer
    BsrFreeMem(pabData);
    // return value from _WkbProgramLow()
    return idResult;
  }
RETURN:
  return WkdReturnBool();
} // WkbProgram2()


// ===================================
// end of clpgm.c
// ===================================
