/*****************************************************************************
S r c D r v                               Portable Source Code WibuKey Driver
==============================================================================
clport.c                            WibuKey API Port functions implementation
==============================================================================

This module implements the following WibuKey API functions:
  WkbEnumPorts2, WkbQueryPort2(), WkbSetPort2().

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
95Feb09 2.05 mabu: extracted from KRNL.C
96Jun23 2.13 mabu: renamed from BSPORT.C
97Feb11 2.25 mabu: adapted to Unicode (only WKWIN32.DLL)
97Jun11 2.31 mabu: litte/big endian adaptation for Macintosh
97Jun16 2.31 mabu: WkbEnumPorts2() supports WkNet
98Feb07 2.41 mabu: integration of new error codes
98Mar16 2.41 mabu: begin of adaptation to new driver class model
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"


/*****************************************************************************
W k b E n u m P o r t s 2
==============================================================================

This function determines all available WibuKey ports at the specified
WibuKey subsystem and writes the names of all these ports into the specified
destination string buffer. The names are separated by vertical bars ("|").
A second calling variant returns this list name by name via subsequent
function calls.

Parameters
  hwkbsys  contains the handle of an accessed WibuKey subsystem.
           For using the WIBU-BOXes at the local system, the predefined system
           handle WKB_LOCAL can be used.
  flCtrl   contains flags which controls the execution of the operation.
           Following flags are currently supported:
             WKB_EPT_LIST  (0x0000) The buffer specified by the <pszDest>
                           parameter receives a string containing the names of
                           all available WibuKey ports, separated by the
                           vertical bar and terminated by a 0 character.
             WKB_EPT_NEXT  (0x0001) The buffer specified by the <pszDest>
                           parameter receives a string which corresponds with
                           the next name of an available WibuKey port which
                           follows in the name ordering after the name which
                           is specified in the <pszDest> buffer when the
                           function is called. If the latter name is an empty
                           string, the first available WibuKey port name is
                           returned.

           This parameter contains also standard function control flags.

  pszDest  points to a buffer which receives the string which corresponds with
           the list all port names or with the next port name. In the latter
           case, the buffer must contain the preceeding port name string
           before the function is called or an empty string when the first
           name should be returned. The stored string is terminated by a
           0-character. Never more than <cbDest> bytes are stored into the
           buffer. If <cbDest> is 0, this parameter is not used.
  cchDest  contains the length of the specified buffer in characters.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the
    WkbGetLastError() function returns the corresponding error code.
  * If the length in <cchDest> is too small to receive all returned
    information, the needed buffer length in number of characters is returned.
    A subsequent call of WkbGetLastError() returns the code
    WKB_BUFFER_OVERFLOW.
  * If the length in <cchDest> is large enough, the function returns the
    number of transferred characters without (!) the trailing 0.

*****************************************************************************/

INT WKAPIENTRY WkbEnumPorts2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  char *pszDest, UINT cchDest)
{
  // return value, initialize with error state (0)
  UINT cchLength = 0;
  INT iRet = 0;


  THREAD_API_LOCK

  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pszDest)) {
    goto RETURN;
  } // if
  // check if specified system is valid, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  {//// execute driver operations
    char szName[32];
    UINT cchName = 0;
    switch (flCtrl & WKB_EPT_MASK) {
      case WKB_EPT_LIST: {
        // flag "separator written"
        BOOL fSeparator = FALSE;
        UINT cchWrite = 0;
      #if BSRDEBUG
        // pointer to verify write operation
        const char *pszDestBuf = pszDest;
      #endif
        // select & access first port
        WkdResetPortEnumeration();
        while (WkdSelectNextPort(TRUE)) {
          // determine name of specified port and write into buffer
          // selected port is valid => name is valid
          memset(szName, 0x0, sizeof(szName));
          WkdGetPortName(szName, sizeof(szName), &cchName);
          // determine length including separator
          cchWrite += cchName;
          if (cchWrite < cchDest) {
            // enough memory in buffer: write new string name + separator
            ASSERT(pszDestBuf + cchDest > pszDest);
            memcpy(pszDest, szName, (size_t)cchName);
            pszDest += cchName;
          } // if
          // set separator to next port
          // * will be overwritten at end of list by terminator
          cchWrite++;
          if (cchWrite < cchDest) {
            // write separator character into destination buffer
            ASSERT(pszDestBuf + cchDest > pszDest);
            *pszDest++ = '|';
            // separator set
            fSeparator = TRUE;
          } // if
        } // while
        // release last accessed port
        WkdReleasePort();
        // discard last error
        cstWkd.idError = WKB_NO_ERROR;
        // end of transfer: terminate string if any valid destination
        if (cchDest != 0) {
          if (fSeparator) {
            // any port written: overwrite last set separator by terminator
            pszDest--;
          } // if
          // write terminator
          *pszDest = '\0';
        } // if
        // number of written characters within string
        cchLength = cchWrite;
        if (cchWrite >= cchDest) {
          cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
          // one byte more needed for buffer
          cchLength++;
        } // if
        break;
      } // case
      case WKB_EPT_NEXT: {
        char *pszBuf = pszDest;
        BOOL fValid = FALSE;
        if ( (pszBuf[0] == 0) || ( (pszBuf[0] == '*') && (pszBuf[1] == 0) ) ) {
          // determine first port, select & access it
          WkdResetPortEnumeration();
          fValid = WkdSelectNextPort(TRUE);
        } else {
          // search specified port name
          INT nPort = 0;
          const char *psz = WkdParsePortSpec(pszDest, &nPort);
          cstWkd.idError = WKB_ERROR_INVALID_PORT;
          if ( (psz == NULL) || (*psz != 0) ) {
              ASSERT(cstWkd.idError == WKB_ERROR_INVALID_PORT);
            goto RETURN;
          } // if
          // must be set because of multi tasking ability of WkLAN
          cstWkd.nSelPort = -1;
          cstWkd.iSelPortGroup = -1;
          fValid = WkdSelectPort(nPort, TRUE);
          if ((nPort & 7) == 7) {
            // '*' as part of the specification: reduce to first port of group
            nPort &= ~7;
            // return only valid port name
            if (!WkdSelectPortGroup(nPort)) {
              // specified port does not reflect a port group
              // * not one port valid
              fValid = FALSE;
            } else {
              // determine first valid port within group
              // * pointer to group must be determined
              ASSERT_VALUE(cstWkd.pprtgrpSel);
              ASSERT(nPort >= cstWkd.pprtgrpSel->nFirstPort);
              for (; nPort < cstWkd.pprtgrpSel->nLastPort;
                  nPort++
                ) {
                // try to select and access current port
                fValid = WkdSelectPort(nPort, TRUE);
                if (fValid) {
                  // valid selection => exit loop
                  break;
                } // if
              } // for
            } // if
          } else {
            // determine next port from specified port, select & access it
            fValid = WkdSelectNextPort(TRUE);
          } // if
        } // if
        // default: error state
        cchLength = 0;
        if (fValid) {
          // return name of current selected port
          memset(szName, 0x0, sizeof(szName));
          WkdGetPortName(szName, sizeof(szName), &cchName);
          cchLength = cchName;
          if (cchName < cchDest) {
            // enough memory in buffer: write new string name + separator
            memcpy(pszBuf, szName, (size_t)(cchName + 1));
          } else {
            // memory too small: store number of characters including termin.
            cchName++;
          } // if
          // discard last error
          cstWkd.idError = WKB_NO_ERROR;
        } // if
        // release accessed port (if any)
        WkdReleasePort();
        break;
      } // case
      default:
        cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
        goto RETURN;
    } // switch
  }
RETURN:
  // return transferred/required length in characters
  iRet = WkdReturnLength(cchLength);

  THREAD_API_UNLOCK

  return iRet;
} // WkbEnumPorts2()


/*****************************************************************************
W k b Q u e r y P o r t 2
==============================================================================

This function queries specific control and status information about the
specified WibuKey port. Furthermore this function can be used to convert
a port number in a port name string and vice versa.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  flCtrl   contains flags which control the operation of the function.
           Following flags are currently supported:
             WKB_QP_PHYS (0x0000) returns port specific physical control and
                                  status data.
             WKB_QP_NAME (0x0010) returns the number and the port name of the
                                  specified port.
             WKB_QP_CONFIG_NAMES (0x0020) returns the string of the
                                  configuration bits of this port.
             WKB_PORTNUMBER (0x0008) is set to specify a port number instead
                                  of a port name.
  pszPort  points to a string which contains one single port specification.
           If the flag WKB_PORTNUMBER is set, this parameter is not inter-
           preted as pointer but must contain a valid port specification
           number which is compatible with WibuKey version 1.
  pvDest   points to a destination buffer which receives the port data.
           For this version of the function the structure WKBPORT or
           WKBPORTNAME is used. The length of the WKBPORT structure varies in
           dependence of the type of the port.
  cbDest   contains the length of the destination buffer in bytes. This should
           be the enough to receive a WKBPORTNAME structure and the name
           string of the port or to receive the port dependent WKBPORT
           structure.

Comments
  If the destination buffer is long enough, the WKB_QP_NAME mode stores the
  name of the port behind the WKBPORTNAME structure as a 0 terminated string.
  To address this string from the WKBPORTNAME structure the WKB_USEOFFSETS
  bit is set in the <flCtrl> member and the <pszPort> member contains the
  offset of the string relative to the start of the structure.

  The string which is returned when the WKB_QP_CONFIG_NAMES operation is
  specified, contains for each configuration flag, started with bit 0, a
  partial string. These strings are separated by the vertical bar ('|'). If a
  flag has different names for the inactive (0) and active (1) condition,
  these two names are specified within this substring: at first the active
  name, then the inactive name, separated by a comma (','). If a partial
  string is empty, this configuration bit cannot be changed.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cbDest> is too small to receive all returned
    information, the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cbDest> is large enough, the function returns the number
    of transferred bytes. For the option WKB_QP_CONFIG_NAMES the number of
    characters in the string is returned, without (!) the trailing 0.

*****************************************************************************/

INT WKAPIENTRY WkbQueryPort2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  const char *pszPort, void *pvDest, UINT cbDest)
{
  UINT cbLength = 0;
  BOOL fUseOffset = FALSE;
  INT iRet = 0;


  THREAD_API_LOCK

  // convert handle address to true address if required
  if (cbDest != 0 && !WkdConvertMemHandle((const void**) &pvDest)) {
    // invalid specification specified => return with error
    goto EXIT_FNCT;
  } // if
  // check if specified system is valid, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto EXIT_FNCT;
  } // if
  if ((cstWkd.flCtrl & WKB_QP_MASK) == WKB_QP_NAME) {
    fUseOffset = TRUE;
  } // if
  {// execute kernel operations
    // * validate parameters
    // * validate port specification, select & access this port
    if (!WkdAccessPortSpec(pszPort)) {
      goto EXIT_FNCT;
    } // if
    // execute operation with specified port
    switch (cstWkd.flCtrl & WKB_QP_MASK) {
      case WKB_QP_PHYS: {
        // select specified driver & get specific information
        WKBPORT wkbprt;
        UINT cbCopy = 0;
        // determine port data via class driver
        // * initialize complete structure with 0 bytes
        // * class driver may fill less information than WKBPORT structure
        memset(&wkbprt, 0, sizeof(wkbprt));
        ASSERT_VALUE(cstWkd.pclsdrvSel);
        ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvPortData);
        if (!cstWkd.pclsdrvSel->pfClDrvPortData(&wkbprt, FALSE)) {
          // any error => return
          break;
        } // if
        // transfer read data to destination
        if (cbDest > sizeof(wkbprt)) {
          // destination bigger than defined WKBPORT structure
          // * fill it completely
          memset(pvDest, 0, cbDest);
          // copy only WBKPORT structure
          cbCopy = sizeof(WKBPORT);
        } else {
          // limit copy area to destination length
          cbCopy = cbDest;
        } // if
        // transfer (partial) data to destination
        memcpy(pvDest, &wkbprt, (size_t)cbCopy);
        cbLength = cbDest;
        break;
      } // case
      case WKB_QP_NAME: {
        // determine port number & name and store it
        char szName[20];
        WKBPORTNAME *pwkbprtn = (WKBPORTNAME*) pvDest;
        UINT cch = 0;
        memset(szName, 0x0, sizeof(szName));
        if (!WkdGetPortName(szName, sizeof(szName), &cch)) {
          break;
        } // if
      #if TEST_DEBUG
        printf("Portname: %s (Error = %d/%d)\n", szName, cstWkd.idError, gidWkdLastError);
      #endif
      
        cbLength = (UINT) sizeof(WKBPORTNAME);
        // cbLength = (UINT)(sizeof(WKBPORTNAME) + strlen(szName)) + 1;
        //// determine length of returned data structure
        if (cbDest >= cbLength) {
          // setup fixed part of destination structure
          pwkbprtn->flCtrl = 0;
          pwkbprtn->lPort = cstWkd.nSelPort;
        } else {
          cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
        } // if
        // add port string at end of structure if possible
        // create WKB_ERROR_BUFFER_OVERFLOW error if not enough memory
        WkdSetStructString(pwkbprtn, cbDest, &pwkbprtn->pszPort,
          &cbLength, szName);
        break;
      } // case
      case WKB_QP_CONFIG_NAMES: {
        // select specified port & get specific information
        cbLength = (UINT)strlen(cstWkd.pprtgrpSel->pszConfigNames) + 1;
        if (cbLength > cbDest) {
          // destination buffer not long enough => return with error
          cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
          // one byte more for trailing 0
          cbLength++;
          break;
        } // if
        // copy string into destination buffer if it is long enough
        strcpy(pvDest, cstWkd.pprtgrpSel->pszConfigNames);
        // return in <cbLength> the number of characters (without trail. 0)
        break;
      } // case
      default:
        // other commands are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
    } // switch
    // release accessed port
    WkdReleasePort();
  }
EXIT_FNCT:
  iRet = WkdReturnLength(cbLength);

  THREAD_API_UNLOCK

  return iRet;
} // WkbQueryPort2()


/*****************************************************************************
W k b S e t P o r t 2
==============================================================================

This function sets port specific control data into the specified port.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  flCtrl   contains flags which control the operation of the function.
           Following flags are currently supported:
             WKB_PORTNUMBER (0x0008) is set to specify a port number instead
                                     of a port name.
  pszPort  points to a string which contains one single port specification.
           If the flag WKB_PORTNUMBER is set, this parameter is not inter-
           preted as pointer but must contain a valid port specification
           number which is compatible with WibuKey version 1.
  pvData   points to a structure which contains the port control data.
  cbData   contains the length of the port control data.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbSetPort2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  const char *pszPort, const void *pvData, UINT cbData)
{
  INT iRet = 0;


  THREAD_API_LOCK

  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvData)) {
    goto RETURN;
  } // if
  // check if specified system is valid, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  { // check if FC=0 is allowed ==>> set global state
    WKBPORT *pwkbprt = (WKBPORT*) pvData;
    if (pwkbprt->fsDrvCtrl & WKB_CTRL_NOSRVR) {
      cstWkd.flGlobalInternal |= DRVC_FC0_ALLOWED;
      cstWkd.flGlobalInternal |= DRVC_EVERYSERIAL;
    } else {
      cstWkd.flGlobalInternal &= ~DRVC_FC0_ALLOWED;
      cstWkd.flGlobalInternal &= ~DRVC_EVERYSERIAL;
    } // if
  } // block
  {//// execute kernel operations
    // validate port specification, select & access this port
    if (!WkdAccessPortSpec(pszPort)) {
      goto RETURN;
    } // if
    // execute operation with specified port
    switch (cstWkd.flCtrl & WKB_SP_MASK) {
      case WKB_SP_PHYS: {
        WKBPORT wkbprt;
        if (cbData > sizeof(WKBPORT)) {
          // specified data too long => return with error
          cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
          goto EXIT_KRNL;
        } // if
        // select specified driver & copy specified information
        // * use class driver function
        ASSERT_VALUE(cstWkd.pclsdrvSel);
        ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvPortData);
        if (cbData < sizeof(WKBPORT)) {
          // source structure not fully defined
          // * read port data
          if (!cstWkd.pclsdrvSel->pfClDrvPortData(&wkbprt, FALSE)) {
            // any error => return
            goto EXIT_KRNL;
          } // if
        } // if
        // replace first <cbData> of port structure
        memcpy(&wkbprt, pvData, (size_t)cbData);
        // write back modified data into driver
        if (!cstWkd.pclsdrvSel->pfClDrvPortData(&wkbprt, TRUE)) {
          // any error => return
          goto EXIT_KRNL;
        } // if
      EXIT_KRNL:
        // release accessed port & return
        WkdReleasePort();
        break;
      } // case
      default:
        // other commands are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
    } // switch
  }
RETURN:
  iRet = WkdReturnBool();

  THREAD_API_UNLOCK

  return iRet;
} // WkbSetPort2()


// ====================================
// end of clport.c
// ====================================
