/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clstruct.c                                       WibuKey Structure Filling API
==============================================================================

This module implements the functions to fill the WibuKey API structures for
programming languages which don't support the declaration of structures.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1995 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Markus Balb (marb), Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
95Aug05 2.07 marb: initial version, separated from WibuKey driver
96Feb15 2.10 mabu: integrated into WibuKey driver
96Jun23 2.13 mabu: renamed from BSSTRUCT.C
96Jul31 2.13 mabu: adapted via _ASMDRV constant to WKDOS.LIB library
96Aug27 2.14 mabu: WKB_STC_... constants renamed to WKB_STS_...
96Sep24 2.14 mabu: bug in WkbStructSetData() removed
96Dec20 2.21 mabu: bug in WkbStructGetStruct() removed
97Jun25 2.31 mabu: support of memory handles implemented
97Jul30 2.31 mabu: proper setting of constant and normal pointers in API
97Aug07 2.31 mabu: preprocessor constant _ASMDRV renamed to _DOSASM
97Aug15 2.32 mabu: proper adaptation to assembler driver
99Mar01 2.52 as:   WkbQueryInfo2 included
10Jun24 4.31 ha:   build update

*****************************************************************************************/

#include "proj.h"

// These type codes are used in member definition type entries. The last
// 4 bits contains the length of the type in bytes in the range from 1 to 4.
#define MBTYP_BYTE      1
#define MBTYP_USHORT    2
#define MBTYP_ULONG     4
#define MBTYP_DATA      16 + 4
#define MBTYP_STRING    32 + 4

#if defined(WK32)
  #define MBTYP_HANDLE MBTYP_ULONG
#else
  #define MBTYP_HANDLE MBTYP_USHORT
#endif

/*****************************************************************************
M E M B E R D E F                                  Structure Member Definition
==============================================================================

This structure is used to convert a MemberId into the entry position of the
based structure.

Members
-------
nId
    contains the MemberId which characteristic should be determined.
idType
    contains the type of the member. This is one of the MBTYP_... constants
    if this is a base type or one of the WKB_STS_... constants.
offs
    contains the byte offset position within the based structure.

*****************************************************************************/

typedef struct {
  BYTE   nId;
  BYTE   idType;
  USHORT offs;
} MEMBERDEF; // prefix "mbd"


/*****************************************************************************
S T R U C T D E F                                         Structure Definition
==============================================================================

This structure describes the members of WibuKey-API structure. It contains
the number of members, the length of the structure and the pointer to the
member definition array.

Properties
----------
cMember
    contains the number of members in the structure.
cbStruct
    contains the number of bytes in the WibuKey-API structure.
pambd
    points to the member definition array.
fuCtrl
    control flags which define several structure characteristics. Following
    flags are defined:
      STDEF_STDCTRL  the structure contains standard control flags at the
                     begin of the structure


*****************************************************************************/

typedef struct {
  INT cMember;
  UINT cbStruct;
  const MEMBERDEF *pambd;
  UINT fuCtrl;
} STRUCTDEF; // prefix "stdf"

// flags for <fuCtrl>
#define STDEF_STDCTRL 0x0001


/*****************************************************************************
S T R U C T B L K                                        StructBlock Structure
==============================================================================

This structure describes the internal StructBlock structure.

Members
-------
nStructId
    contains the StructId value which corresponds with the StructBlock.
cbAddedArea
    contains the length of the added structure area.
cbStringArea
    contains the length of the string area.
cbStringData
    contains the length of the filled string space area
abData
    contains the variable area of a StructBlock memory area. It is declared
    as a 1 byte dummy which represents a variable length buffer.

Comments
--------
The <abData> member of the structure has following entries:
    +-----------------+
    |fixed struct area| length determined via <nStructId>
    +-----------------+
    |added area       | length specified by <cbAddedArea>
    +-----------------+
    |string space     | length specified by <cbStringArea>
    +-----------------+

*****************************************************************************/

typedef struct {
  INT  nStructId;
  UINT cbAddedArea;
  UINT cbStringArea;
  UINT cbStringData;
  BYTE abData[1];
} STRUCTBLK; // prefix "stblk"


/*****************************************************************************
S T R U C T P O S                                           Structure Position
==============================================================================

This structure describes the ULONG structure which manages a structure
position

Members
-------
usOffset
    contains the offset of the position within the structure, relative to the
    beginning of the structure.
bStructId
    contains the StructId of the StructBlock for which the position is
    determined.
idMemberType
    contains the type of the member which is positioned by the StructPos.

*****************************************************************************/

typedef struct {
  USHORT usOffset;
  BYTE bStructId;
  BYTE idMemberType;
} STRUCTPOS; // prefix "stps"


// ============================
// Member setting lookup tables
// ============================

// type WKBSERIAL
MEMBERDEF ambdSerial[] = {
  {WKB_STM_ULSERIAL, MBTYP_ULONG,  FIELDOFFSET(WKBSERIAL, ulSerial)},
  {WKB_STM_USMASK,   MBTYP_USHORT, FIELDOFFSET(WKBSERIAL, usMask)},
  {WKB_STM_USRE,     MBTYP_USHORT, FIELDOFFSET(WKBSERIAL, usReserved)}
};


// type WKBBOXLIST
MEMBERDEF ambdBoxList[] = {
  {WKB_STM_FLCTRL, MBTYP_ULONG,  FIELDOFFSET(WKBBOXLIST, flCtrl)},
  {WKB_STM_CBOX,   MBTYP_USHORT, FIELDOFFSET(WKBBOXLIST,cBox)},
  {WKB_STM_USRE,   MBTYP_USHORT, FIELDOFFSET(WKBBOXLIST, usReserved)},
  {WKB_STM_AWKBSR, WKB_STS_SERIAL, FIELDOFFSET(WKBBOXLIST, awkbsr)}
};


// type WKBBOXENTRY
MEMBERDEF ambdBoxEntry [] = {
  {WKB_STM_FLCTRL,     MBTYP_ULONG, FIELDOFFSET(WKBBOXENTRY, flCtrl)},
  {WKB_STM_ULFIRMCODE, MBTYP_ULONG, FIELDOFFSET(WKBBOXENTRY, ulFirmCode)},
  {WKB_STM_ULUSERCODE, MBTYP_ULONG, FIELDOFFSET(WKBBOXENTRY, ulUserCode)},
  {WKB_STM_ULD1,       MBTYP_ULONG, FIELDOFFSET(WKBBOXENTRY, ulData1)},
  {WKB_STM_ULD2,       MBTYP_ULONG, FIELDOFFSET(WKBBOXENTRY, ulData2)}
};


// type WKBBOXDATA
MEMBERDEF ambdBoxData[] = {
  {WKB_STM_FLCTRL,     MBTYP_ULONG, FIELDOFFSET(WKBBOXDATA, flCtrl)},
  {WKB_STM_USVERSION,  MBTYP_USHORT, FIELDOFFSET(WKBBOXDATA, usVersion)},
  {WKB_STM_CENTRY,     MBTYP_USHORT, FIELDOFFSET(WKBBOXDATA, cEntry)},
  {WKB_STM_AWKBBXE,    WKB_STS_BOXENTRY, FIELDOFFSET(WKBBOXDATA, awkbbxe)}
};


// type WKBBOXCONFIG
MEMBERDEF ambdBoxConfig[] = {
  {WKB_STM_FLCTRL,         MBTYP_ULONG, FIELDOFFSET(WKBBOXCONFIG, flCtrl)},
  {WKB_STM_USVERSION,      MBTYP_USHORT, FIELDOFFSET(WKBBOXCONFIG, usVersion)},
  {WKB_STM_USRE,           MBTYP_USHORT, FIELDOFFSET(WKBBOXCONFIG, usReserved)},
  {WKB_STM_ULPROGRAMCOUNT, MBTYP_ULONG, FIELDOFFSET(WKBBOXCONFIG, ulProgramCount)},
  {WKB_STM_ULCONFIG,       MBTYP_ULONG, FIELDOFFSET(WKBBOXCONFIG, ulConfig)}
};


// type WKBOPENENTRY
MEMBERDEF ambdOpenEntry [] = {
  {WKB_STM_FLCTRL,     MBTYP_ULONG, FIELDOFFSET(WKBOPENENTRY, flCtrl)},
  {WKB_STM_HWKBSYS,    MBTYP_HANDLE, FIELDOFFSET(WKBOPENENTRY, hwkbsys)},
#if !defined(WK32)
  {WKB_STM_USFILL,     MBTYP_USHORT, FIELDOFFSET(WKBOPENENTRY, usFill)},
#endif
  {WKB_STM_LPORT,      MBTYP_ULONG, FIELDOFFSET(WKBOPENENTRY, lPort)},
  {WKB_STM_PSZPORT,    MBTYP_STRING, FIELDOFFSET(WKBOPENENTRY, pszPort)},
  {WKB_STM_WKBSERIAL,  WKB_STS_SERIAL, FIELDOFFSET(WKBOPENENTRY, wkbsr)},
  {WKB_STM_USVERSION,  MBTYP_USHORT, FIELDOFFSET(WKBOPENENTRY, usVersion)},
  {WKB_STM_USENTRY,    MBTYP_USHORT, FIELDOFFSET(WKBOPENENTRY, usEntry)},
  {WKB_STM_ULFIRMCODE, MBTYP_ULONG, FIELDOFFSET(WKBOPENENTRY, ulFirmCode)},
  {WKB_STM_ULUSERCODE, MBTYP_ULONG, FIELDOFFSET(WKBOPENENTRY, ulUserCode)}
};


// type WKBSELECT
MEMBERDEF ambdSelect[] = {
  {WKB_STM_FLCTRL,    MBTYP_ULONG, FIELDOFFSET(WKBSELECT, flCtrl)},
  {WKB_STM_ULCACHEID, MBTYP_ULONG, FIELDOFFSET(WKBSELECT, ulCacheId)},
  {WKB_STM_ULCRC,     MBTYP_ULONG, FIELDOFFSET(WKBSELECT, ulCrc)},
  {WKB_STM_CBINIT,    MBTYP_USHORT, FIELDOFFSET(WKBSELECT, cbInit)},
  {WKB_STM_ABINIT,    MBTYP_BYTE, FIELDOFFSET(WKBSELECT, abInit)}
};


// type WKBAREA
MEMBERDEF ambdArea[] = {
  {WKB_STM_FLCTRL,       MBTYP_ULONG, FIELDOFFSET(WKBAREA, flCtrl)},
  {WKB_STM_FLSELCTRL,    MBTYP_ULONG, FIELDOFFSET(WKBAREA, flSelCtrl)},
  {WKB_STM_ULSELECTCODE, MBTYP_ULONG, FIELDOFFSET(WKBAREA, ulSelectCode)},
  {WKB_STM_ULCRC,        MBTYP_ULONG, FIELDOFFSET(WKBAREA, ulCrc)},
  {WKB_STM_ULRE,         MBTYP_ULONG, FIELDOFFSET(WKBAREA, ulReserved)},
  {WKB_STM_PBGAPMAP,     MBTYP_DATA,  FIELDOFFSET(WKBAREA, pbGapMap)},
  {WKB_STM_PVCTRL,       MBTYP_DATA,  FIELDOFFSET(WKBAREA, pvCtrl)}
};


// type WKBPORT
MEMBERDEF ambdPort[] = {
  {WKB_STM_ULADDRESS, MBTYP_ULONG, FIELDOFFSET(WKBPORT, ulAddress)},
  {WKB_STM_FSCTRL,    MBTYP_USHORT, FIELDOFFSET(WKBPORT, fsPortCtrl)}
};


// type WKBPORTNAME
MEMBERDEF ambdPortName[] = {
  {WKB_STM_FLCTRL,  MBTYP_ULONG, FIELDOFFSET(WKBPORTNAME, flCtrl)},
  {WKB_STM_LPORT,   MBTYP_ULONG, FIELDOFFSET(WKBPORTNAME, lPort)},
  {WKB_STM_PSZPORT, MBTYP_STRING, FIELDOFFSET(WKBPORTNAME, pszPort)}
};


// type WKBPROGRAM
MEMBERDEF ambdProgram[] = {
  {WKB_STM_FLCTRL,      MBTYP_ULONG, FIELDOFFSET(WKBPROGRAM, flCtrl)},
  {WKB_STM_WKBBXE,      WKB_STS_BOXENTRY, FIELDOFFSET(WKBPROGRAM, wkbbxe)},
  {WKB_STM_ABSIGNATURE, MBTYP_BYTE, FIELDOFFSET(WKBPROGRAM, abSignature)}
};


// type WKBSTATUS
MEMBERDEF ambdStatus[] = {
  {WKB_STM_USFLAGS,  MBTYP_USHORT, FIELDOFFSET(WKBSTATUS, usFlags)},
  {WKB_STM_USDRVINT, MBTYP_USHORT, FIELDOFFSET(WKBSTATUS, usDrvInt)}
};


// type WKBACCESS
MEMBERDEF ambdAccess[] = {
  {WKB_STM_FLCTRL,        MBTYP_ULONG, FIELDOFFSET(WKBACCESS, flCtrl)},
  {WKB_STM_LFIRMCODE,     MBTYP_ULONG, FIELDOFFSET(WKBACCESS, lFirmCode)},
  {WKB_STM_LUSERCODE,     MBTYP_ULONG, FIELDOFFSET(WKBACCESS, lUserCode)},
  {WKB_STM_FLSTDCTRL,     MBTYP_ULONG, FIELDOFFSET(WKBACCESS, flStdCtrl)},
  {WKB_STM_BVERLOW,       MBTYP_BYTE, FIELDOFFSET(WKBACCESS, bVersionLow)},
  {WKB_STM_BVERHIGH,      MBTYP_BYTE, FIELDOFFSET(WKBACCESS, bVersionHigh)},
  {WKB_STM_IDNETUSER,     MBTYP_USHORT, FIELDOFFSET(WKBACCESS, idNetUser)},
  {WKB_STM_PSZAPPNAME,    MBTYP_STRING, FIELDOFFSET(WKBACCESS, pszAppName)},
  {WKB_STM_PSZSERVERNAME, MBTYP_STRING, FIELDOFFSET(WKBACCESS, pszServerName)},
  {WKB_STM_PSZWKNETFILE,  MBTYP_STRING, FIELDOFFSET(WKBACCESS, pszWkNetFile)},
  {WKB_STM_USREFRESHTIME, MBTYP_USHORT, FIELDOFFSET(WKBACCESS, usRefreshTime)},
  {WKB_STM_USMAXSTARTDELTA, MBTYP_USHORT,
                          FIELDOFFSET(WKBACCESS, usMaxStartDelta)},
  {WKB_STM_PBENCRYPTSEQ,  MBTYP_DATA, FIELDOFFSET(WKBACCESS, pbEncryptSeq)},
  {WKB_STM_ULRE2,         MBTYP_ULONG, FIELDOFFSET(WKBACCESS, pwkbai)},
  {WKB_STM_ULRE3,         MBTYP_ULONG, FIELDOFFSET(WKBACCESS, ulReserved3)}
};

// type WKBSYSTEM
MEMBERDEF ambdSystem[] = {
  {WKB_STM_FLCTRL,        MBTYP_ULONG, FIELDOFFSET(WKBSYSTEM, flCtrl)},
  {WKB_STM_PSZSERVERNAME, MBTYP_STRING, FIELDOFFSET(WKBSYSTEM, pszServerName)},
  {WKB_STM_IDNETUSER,     MBTYP_USHORT, FIELDOFFSET(WKBSYSTEM, idNetUser)},
  {WKB_STM_CMAXUSER,      MBTYP_USHORT, FIELDOFFSET(WKBSYSTEM, cMaxUser)},
  {WKB_STM_CFREEUSER,     MBTYP_USHORT, FIELDOFFSET(WKBSYSTEM, cFreeUser)},
  {WKB_STM_MAXDATA,       MBTYP_USHORT, FIELDOFFSET(WKBSYSTEM, cbMaxData)},
  {WKB_STM_CREFRESHERROR, MBTYP_USHORT, FIELDOFFSET(WKBSYSTEM, cRefreshError)},
  {WKB_STM_USRE1,         MBTYP_USHORT, FIELDOFFSET(WKBSYSTEM, usReserved1)}
};


// type WKNET_BASE
MEMBERDEF ambdNet_Base[] = {
  {WKB_STM_ABSIGNATURE,   MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, abSignature)},
  {WKB_STM_BHIGHVERSION,  MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, bHighVersion)},
  {WKB_STM_BLOWVERSION,   MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, bLowVersion)},
  {WKB_STM_CBBASE,        MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, cbBase)},
  {WKB_STM_USMAXUSER,     MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, usMaxUser)},
  {WKB_STM_CBSLOT,        MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, cbSlot)},
  {WKB_STM_MAXBLOCK,      MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, cbMaxBlock)},
  {WKB_STM_CBCUSTOMAREA,  MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, cbCustomArea)},
  {WKB_STM_ULCUSTOMADDR,  MBTYP_ULONG, FIELDOFFSET(WKNET_BASE, ulCustomAddr)},
  {WKB_STM_ULBOXLISTADDR, MBTYP_ULONG, FIELDOFFSET(WKNET_BASE, ulBoxListAddr)},
  {WKB_STM_USDUPD,        MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, usDateUpd)},
  {WKB_STM_USTUPD,        MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, usTimeUpd)},
  {WKB_STM_USCMAXUSER,    MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, usCMaxUser)},
  {WKB_STM_USDPROT,       MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, usDateProt)},
  {WKB_STM_USTPROT,       MBTYP_USHORT, FIELDOFFSET(WKNET_BASE, usTimeProt)},
  {WKB_STM_BUPDTICK,      MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, bUpdTick)},
  {WKB_STM_ABRE2,         MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, abReserved2)},
  {WKB_STM_ABCUSTOM1,     MBTYP_BYTE, FIELDOFFSET(WKNET_BASE, abCustom1)},
  {WKB_STM_ULUPDSECURITY, MBTYP_ULONG, FIELDOFFSET(WKNET_BASE, ulUpdSecurity)}
};

// type WKNET_SLOT
MEMBERDEF ambdNet_Slot[] = {
  {WKB_STM_SZSYSTEMNAME,  MBTYP_BYTE, FIELDOFFSET(WKNET_SLOT, szSystemName)},
  {WKB_STM_USDLOGIN,      MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, usDateLogIn)},
  {WKB_STM_USTLOGIN,      MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, usTimeLogIn)},
  {WKB_STM_SDACCESS,      MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, usDateAccess)},
  {WKB_STM_STACCESS,      MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, usTimeAccess)},
  {WKB_STM_USLOGINSCRIPT, MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, usLogInScript)},
  {WKB_STM_ABRE1,         MBTYP_BYTE, FIELDOFFSET(WKNET_SLOT, abReserved1)},
  {WKB_STM_ULFC,          MBTYP_ULONG, FIELDOFFSET(WKNET_SLOT, ulFC)},
  {WKB_STM_ULUC,          MBTYP_ULONG, FIELDOFFSET(WKNET_SLOT, ulUC)},
  {WKB_STM_ULSC,          MBTYP_ULONG, FIELDOFFSET(WKNET_SLOT, ulSC)},
  {WKB_STM_CBBLOCK,       MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, cbBlock)},
  {WKB_STM_USCONTROL,     MBTYP_USHORT, FIELDOFFSET(WKNET_SLOT, usControl)},
  {WKB_STM_ULACCSECURITY, MBTYP_ULONG, FIELDOFFSET(WKNET_SLOT, ulAccSecurity)},
  {WKB_STM_BSTATUS,       MBTYP_BYTE, FIELDOFFSET(WKNET_SLOT, bStatus)},
  {WKB_STM_ABRE2,         MBTYP_BYTE, FIELDOFFSET(WKNET_SLOT, abReserved2)}
};

// type WKBOPENFIND
MEMBERDEF ambdOpenFind[] = {
  {WKB_STM_FLCTRL,      MBTYP_ULONG, FIELDOFFSET(WKBOPENFIND, flCtrl)},
  {WKB_STM_ULFIRMCODE,  MBTYP_ULONG, FIELDOFFSET(WKBOPENFIND, ulFirmCode)},
  {WKB_STM_ULUSERCODE,  MBTYP_ULONG, FIELDOFFSET(WKBOPENFIND, ulUserCode)},
  {WKB_STM_ULENTRYSKIP, MBTYP_ULONG, FIELDOFFSET(WKBOPENFIND, ulEntrySkip)}
};

// type WKBVERSION
MEMBERDEF ambdVersion[] = {
  {WKB_STM_USVERSION,     MBTYP_USHORT, FIELDOFFSET(WKBVERSION, usVersion)},
  {WKB_STM_USSUBVERSION,  MBTYP_USHORT, FIELDOFFSET(WKBVERSION, usSubVersion)},
  {WKB_STM_USBUILD,       MBTYP_USHORT, FIELDOFFSET(WKBVERSION, usBuild)},
  {WKB_STM_USCOUNT,       MBTYP_USHORT, FIELDOFFSET(WKBVERSION, usCount)},
  {WKB_STM_USYEAR,        MBTYP_USHORT, FIELDOFFSET(WKBVERSION, usYear)},
  {WKB_STM_BMONTH,        MBTYP_BYTE, FIELDOFFSET(WKBVERSION, bMonth)},
  {WKB_STM_BDAY,          MBTYP_BYTE, FIELDOFFSET(WKBVERSION, bDay)},
  {WKB_STM_ULRE,          MBTYP_ULONG, FIELDOFFSET(WKBVERSION, ulReserved1)}
};



// generate structure table
static STRUCTDEF astdTable[] = {
  {// WKB_STS_ACCESS
    DIMOF(ambdAccess),    sizeof(WKBACCESS),    ambdAccess, STDEF_STDCTRL
  },
  {// WKB_STS_AREA
    DIMOF(ambdArea),      sizeof(WKBAREA),      ambdArea, STDEF_STDCTRL
  },
  {// WKB_STS_BOXCONFIG
    DIMOF(ambdBoxConfig), sizeof(WKBBOXCONFIG), ambdBoxConfig, STDEF_STDCTRL
  },
  {// WKB_STS_BOXDATA
    DIMOF(ambdBoxData),   sizeof(WKBBOXDATA),   ambdBoxData, STDEF_STDCTRL
  },
  {// WKB_STS_BOXENTRY
    DIMOF(ambdBoxEntry),  sizeof(WKBBOXENTRY),  ambdBoxEntry, 0
  },
  {// WKB_STS_BOXLIST
    DIMOF(ambdBoxList),   sizeof(WKBBOXLIST),   ambdBoxList, STDEF_STDCTRL
  },
  {// WKB_STS_OPENENTRY
    DIMOF(ambdOpenEntry), sizeof(WKBOPENENTRY), ambdOpenEntry, STDEF_STDCTRL
  },
  {// WKB_STS_PORT
    DIMOF(ambdPort),      sizeof(WKBPORT),      ambdPort, 0
  },
  {// WKB_STS_PORTNAME
    DIMOF(ambdPortName),  sizeof(WKBPORTNAME),  ambdPortName, STDEF_STDCTRL
  },
  {// WKB_STS_PROGRAM
    DIMOF(ambdProgram),   sizeof(WKBPROGRAM),   ambdProgram, STDEF_STDCTRL
  },
  {// WKB_STS_SELECT
    DIMOF(ambdSelect),    sizeof(WKBSELECT),    ambdSelect, STDEF_STDCTRL
  },
  {// WKB_STS_SERIAL
    DIMOF(ambdSerial),    sizeof(WKBSERIAL),    ambdSerial, 0
  },
  {// WKB_STS_STATUS
    DIMOF(ambdStatus),    sizeof(WKBSTATUS),    ambdStatus, 0
  },
  {// WKB_STS_SYSTEM
    DIMOF(ambdSystem),    sizeof(WKBSYSTEM),    ambdSystem, STDEF_STDCTRL
  },
  {// WKB_STS_NET_BASE
    DIMOF(ambdNet_Base),  sizeof(WKNET_BASE),   ambdNet_Base, 0
  },
  {// WKB_STS_NET_SLOT
    DIMOF(ambdNet_Slot),  sizeof(WKNET_SLOT),   ambdNet_Slot, 0
  },
  {// WKB_STS_OPENFIND
    DIMOF(ambdOpenFind),  sizeof(WKBOPENFIND),  ambdOpenFind, STDEF_STDCTRL
  },
  {// WKB_STS_VERSION
    DIMOF(ambdVersion),   sizeof(WKBVERSION),   ambdVersion, 0
  }
}; // astdTable[]


/*****************************************************************************
C h e c k S t r u c t I d
==============================================================================

This function checks if the specified StructId is available or not.

Parameter
  nStructId  contains the structure identification which is checked.

Return
  The function returns TRUE if the StructId is available and FALSE if not.

*****************************************************************************/

static BOOL CheckStructId(INT nStructId)
{
  if (nStructId >= WKB_STS_BASE && nStructId <= WKB_STS_LAST) {
    // return without error
    return TRUE;
  } // if
  // otherwise: set an error code, return FALSE
  WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
  return FALSE;
} // CheckStructId()


/*****************************************************************************
G e t M e m b e r I n f o
==============================================================================

This function checks if the specified MemberId is compatible with the
specified StructId. If so, a pointer to the member definition is returned.

Parameter
  nStructId   contains the StructId of the structure which contains
              the member.
  nMemberId   contains the MemberId which is checked.

Return
  The function returns a pointer to the member definition structure and
  NULL if the MemberId is not valid.

Attributes
  Local

*****************************************************************************/

static const MEMBERDEF *GetMemberInfo(INT nStructId, INT nMemberId)
{
  const MEMBERDEF *pmbd = NULL; // pointer to the member definition
  INT cMember = 0;
  INT i = 0;
  if (!CheckStructId(nStructId)) {
    // bad structure identification => return with error
    return NULL;
  } // if
  nStructId -= WKB_STS_BASE;
  pmbd = astdTable[nStructId].pambd;
  // load address of member table
  cMember = astdTable[nStructId].cMember;
  // load count of members used in the structure
  for (i = 0; i < cMember; i++, pmbd++) {
    if (pmbd->nId == nMemberId) {
      // MemberId found => return TRUE
      return pmbd;
    } // if
  }  // for
  // MemberId in structure definition not found => return with error
  WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
  return NULL;
} // GetMemberInfo()


/*****************************************************************************
G e t P o s i t i o n
==============================================================================

This function calculates the offset in the StructBlock structure from the
specified StructPos and returns it. It checks also the validity of the
specified StructBlock and sets the StructId which is

Parameter
  pstblk ..... points to the StructBlock structure memory.
  ulStructPos  contains the StructPos value of the caller function. It is 0,
               if no StructPos is specified.
  pnStructId   points to an INT variable which receives the StructId which is
               used to determine the offset of a specified MemberId.

Return
  The function returns (UINT)-1 if the StructPos is invalid. Otherwise the
  returned value is the member offset. If no StructPos is specifed, 0 is
  returned.

Attributes
  Local

*****************************************************************************/

static UINT GetPosition(const STRUCTBLK *pstblk, ULONG ulStructPos,
  INT *pnStructId)
{
  UINT cbMaxPos;
  INT nStructId;
  UINT offs = 0;
  nStructId = pstblk->nStructId;
  if (!CheckStructId(nStructId)) {
    // invalid structure => return with error
    return (UINT)-1;
  } // if
  if (ulStructPos != 0) {
    // determine position within specified position
    STRUCTPOS stps;
    stps = *(STRUCTPOS*)&ulStructPos;
    cbMaxPos = astdTable[nStructId - WKB_STS_BASE].cbStruct +
      pstblk->cbAddedArea -
      astdTable[stps.idMemberType - WKB_STS_BASE].cbStruct;
    if (stps.bStructId != (BYTE)nStructId || stps.usOffset > cbMaxPos) {
      // StructPos doesn't match StructBlock: return with error
      WkbSetLastError(WKB_ERROR_INVALID_FUNCTION);
      return (UINT)-1;
    } // if
    if (stps.idMemberType >= WKB_STS_BASE) {
      // search member now in local structure
      nStructId = stps.idMemberType;
    } // if
    offs = stps.usOffset;
  } // if
  // store used StructId, return offset
  *pnStructId = nStructId;
  return offs;
} // GetPosition()


/*****************************************************************************
G e t M e m b e r S i z e
==============================================================================

This function determines the length of the specified member. For array it
determines the size of each entry and the complete length of the array.
If the member is at the end of the structure, the added data space is also
included into the size calculation.

Parameter
  pstblk .. points to the StructBlock structure memory.
  pbmd .... points to the member description.
  pcbEntry  points to an UINT variable which receives the length of a single
            entry.

Return
  The function returns the length of the complete member including the
  added data area when the member is at the end of the structure.

Attributes
  Local

*****************************************************************************/

static UINT GetMemberSize(const STRUCTBLK *pstblk,
  const MEMBERDEF *pmbd, UINT *pcbEntry)
{
  INT nStructId = 0;
  UINT cbMember = 0;
  INT iMember = 0;
  INT idType = 0;
  STRUCTDEF *pstdf = NULL;

  nStructId = pstblk->nStructId;
  pstdf = astdTable + (nStructId - WKB_STS_BASE);
  iMember = (INT)(pmbd - pstdf->pambd);
  if (iMember + 1 < pstdf->cMember) {
    // member size determined by position of next member in structure
    cbMember = pstdf->pambd[iMember + 1].offs - pmbd->offs;
  } else {
    // at end of structure: length determined by structure length and ...
    // ... optional added data area
    cbMember = pstdf->cbStruct - pmbd->offs + pstblk->cbAddedArea;
  } // if
  ASSERT((INT)cbMember >= 0);
  // now determine size of one member
  idType = pmbd->idType;
  if (idType < WKB_STS_BASE) {
    // base type: length is determined by last 4 bits of constant
    *pcbEntry = idType & 0x0F;
  } else {
    // structured type: determine length via structure definition table
    *pcbEntry = astdTable[idType - WKB_STS_BASE].cbStruct;
  } // if
  return cbMember;
} // GetMemberSize()


/*****************************************************************************

---- Exported WibuKey API functions -----------------------------------------

*****************************************************************************/

/*****************************************************************************
W k b S t r u c t S i z e O f
==============================================================================

This function determines the size of the structure which corresponds with the
specified StructId.

Parameters
  nStructId  contains the StructId of the desired WibuKey API structure.

Return
  The length of the structure in bytes is returned. The length is the true
  WibuKey-API structure size, not the size which is required for a
  StructBlock of this structure. If an error occurred, 0 is returned.
  In this case, a subsequent call of the WkbGetLastError() function returns
  the corresponding error code.

*****************************************************************************/

UINT WKAPIENTRY WkbStructSizeOf(INT nStructId)
{
  if (!CheckStructId(nStructId)) {
    // invalid structure => return with length 0
    return 0;
  } // if
  // determine size and return this value
  return astdTable[nStructId - WKB_STS_BASE].cbStruct;
} // WkbStructSizeOf()


/*****************************************************************************
W k b S t r u c t C a l c S i z e
==============================================================================

This function calculates the length of a StructBlock in bytes with the
specified characteristic.

Parameters
  nStructId ... contains the StructId of the desired structure block.
  cbAddedArea   contains the size of the variable structure area in bytes.
  cbStringArea  contains the size of the variable string area in bytes.

Return
  The complete length of the StructBlock in bytes is returned. If an error
  occurred, 0 is returned. In this case, a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.

Comments
  The 16 bit version of the driver doesn't support added areas or string areas
  with more than 12 KBytes per area.

*****************************************************************************/

UINT WKAPIENTRY WkbStructCalcSize(INT nStructId, UINT cbAddedArea,
  UINT cbStringArea)
{
  if (!CheckStructId(nStructId)) {
    // invalid structure => return with length 0
    return 0;
  } // if
#if !defined(WK32)
  if (cbAddedArea > 12288 || cbStringArea > 12288) {
    // invalid length => return with error
    return 0;
  } // if
#endif
  // determine size and return this value
  return astdTable[nStructId - WKB_STS_BASE].cbStruct +
    FIELDOFFSET(STRUCTBLK, abData) + cbAddedArea + cbStringArea;
} // WkbStructCalcSize()


/*****************************************************************************
W k b S t r u c t I n i t
==============================================================================

This function initializes the specified StructBlock. It writes the specified
values into the header of the StructBlock and fills the data area of the
struct block with 0 bytes. The string area is set to empty (no strings
stored). This function may be called again to an existing StructBlock to
cleanup its contents.

Parameters
  pvStructBlk   points to the StructBlock which is initialized.
  nStructId ... contains the StructId of the desired structure block. If this
                value is 0, the StructBlock is reinitialized again and the
                parameters <cbAddedArea> and <cbStringArea> are ignored.
  cbAddedArea   contains the size of the variable structure area in bytes.
  cbStringArea  contains the size of the variable string area in bytes.

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In this case, a subsequent call of the WkbGetLastError() function returns
  the corresponding error code.

Comments
* If a structure is initialized at first and nStructId is not 0, then all
  parameters except the first must contain the same values as the
  parameters of the WkbStructCalcSize() function, which was called to
  determine the length of the StructBlock which is specified in the
  first parameter.
* The 16 bit version of the driver doesn't support added areas or string
  areas with more than 12 KBytes per area.

*****************************************************************************/

BOOL WKAPIENTRY WkbStructInit(void *pvStructBlk, INT nStructId,
  UINT cbAddedArea, UINT cbStringArea)
{
  STRUCTBLK *pstblk;
  STRUCTDEF *pstdf;
  if (!CheckStructId(nStructId)) {
    // invalid structure => return with length 0
    return FALSE;
  } // if
#if !DOSASM
  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pvStructBlk)) {
    // any error => return with error
    return FALSE;
  } // if
#endif
#if !defined(WK32)
  if (cbAddedArea > 12288 || cbStringArea > 12288) {
    // invalid length => return with error
    return FALSE;
  } // if
#endif
  // initialize header
  pstblk = pvStructBlk;
  pstblk->nStructId = nStructId;
  pstblk->cbAddedArea = cbAddedArea;
  pstblk->cbStringArea = cbStringArea;
  pstblk->cbStringData = 0;
  // filling variable area of the StructBlk with 0 bytes
  pstdf = astdTable + (nStructId - WKB_STS_BASE);
  memset(pstblk->abData, 0, pstdf->cbStruct + cbAddedArea + cbStringArea);
  // return without error
  return TRUE;
} // WkbStructInit()


/*****************************************************************************
W k b S t r u c t G e t A d d r
==============================================================================

This function returns the address of the internal WibuKey-API structure
within the specified StructBlock after the StructBlock header. This address
is used in several WibuKey-API functions.

Parameters
  pvStructBlk  points to the StructBlock from which the address is
               determined.

Return
  The address is returned. If an error occurred, NULL is returned. In this
  case, a subsequent call of the WkbGetLastError() function returns the
  corresponding error code.

Comments
  The <pvStructBlk> may also be a memory block handle. The returned value
  is always an address and never a memory block handle.

*****************************************************************************/

void *WKAPIENTRY WkbStructGetAddr(const void *pvStructBlk)
{
  const STRUCTBLK *pstblk;
#if !DOSASM
  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvStructBlk)) {
    // any error => return with error
    return NULL;
  } // if
#endif
  pstblk = pvStructBlk;
  // determine address and return this value
  return (void*)pstblk->abData;
} // WkbStructGetAddr()


/*****************************************************************************
W k b S t r u c t G e t S i z e
==============================================================================

This function returns the size of the internal WibuKey-API structure in bytes
within the specified StructBlock without the StructBlock header. This size is
used in several WibuKey-API functions. This size contains also the added area
and the string area of the StructBlock.

Parameters
  pvStructBlk  points to the StructBlock which structure size is returned.

Return
  The size is returned. If an error occurred, 0 is returned. In this case, a
  subsequent call of the WkbGetLastError() function returns the corresponding
  error code.

*****************************************************************************/

UINT WKAPIENTRY WkbStructGetSize(const void *pvStructBlk)
{
  const STRUCTBLK *pstblk;
#if !DOSASM
  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvStructBlk)) {
    // any error => return with error
    return 0;
  } // if
#endif
  pstblk = pvStructBlk;
  // determine address and return this value.
  return astdTable[pstblk->nStructId - WKB_STS_BASE].cbStruct +
    pstblk->cbAddedArea + pstblk->cbStringArea;
} // WkbStructGetSize()


/*****************************************************************************
W k b S t r u c t G e t P o s i t i o n
==============================================================================

This function returns the position of a substructure in the specified
StructBlock and returns a StructPos which can be used to address members
in this substructures by functions like WkbStructGetValue() etc.
For the addressing of complex members, this function can also receive another
StructPos value as start and/or an index value.

Parameter
  pvStructBlk  points to the StructBlock in which the position is determined.
  ulStructPos  contains the StructPos to specify a substructure, which is used
               as start point for the new StructPos calculation. To address a
               member on the highest level in the structure block, 0 must be
               specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as an array, 0 may be specified here.
  nIndex ..... contains an index to address an entry in an array member.
               The first entry has the index 0, the next 1 etc. This parameter
               must be set to 0 for members which are not an array.

Return
  The function returns the calculated StructPos value or 0 if an error
  occurred. In this case, a subsequent call of the WkbGetLastError() function
  returns the corresponding error code.

*****************************************************************************/

LONG WKAPIENTRY WkbStructGetPosition(const void *pvStructBlk,
  LONG ulStructPos, INT nMemberId, INT nIndex)
{
  const STRUCTBLK *pstblk = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  STRUCTPOS stpsDest;
#if !DOSASM
  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvStructBlk)) {
    // any error => return with error
    return 0;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  nStructId = pstblk->nStructId;
  if (!CheckStructId(nStructId) || nIndex < 0) {
     // invalid structure or bad index => return with error
    return 0;
  } // if
  // get member characteristics
  pmbd = GetMemberInfo(nStructId, nMemberId);
  if (pmbd == NULL) {
    // member invalid => return
    return 0;
  } // if
  // initialize returned StructPos
  stpsDest.usOffset = pmbd->offs;
  stpsDest.bStructId = (BYTE)nStructId;
  stpsDest.idMemberType = pmbd->idType;
  if (nIndex != 0) {
    // check if index value is permitted
    UINT cbMember = 0;
    UINT cbEntry = 0;
    // determine size of member & entry
    cbMember = GetMemberSize(pstblk, pmbd, &cbEntry);
    ASSERT(cbMember != 0);
    ASSERT(cbEntry != 0);
    ASSERT(nIndex > 0);
    if ((ULONG)(nIndex - 1) * cbEntry > (ULONG)cbMember) {
      // index out of addressable member area: return with error
      WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
      return 0;
    } // if
    // add index position to member offset
    stpsDest.usOffset += (USHORT)(nIndex * cbEntry);
  } // if
  // retun created StructPos
  return *(ULONG*)&stpsDest;
} // WkbStructGetPosition()


/*****************************************************************************
W k b S t r u c t G e t V a l u e
==============================================================================

This function gets the contents of the specified (single) value member in the
specified structure block and returns this contents as integer value.

Parameter
  pvStructBlk  points to the StructBlock from which the value member
               is read.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here.

Return
  The function returns the integer value as LONG if the operation was
  successful. It returns -1 if an error occurred. In this case, a subsequent
  call of the WkbGetLastError() function returns the corresponding error code.
  When -1 is returned as a valid member value, a subsequent call of
  WkbGetLastError() returns WKB_NO_ERROR (0), especially in this case.

*****************************************************************************/

LONG WKAPIENTRY WkbStructGetValue(const void *pvStructBlk,
  ULONG ulStructPos, INT nMemberId)
{
  const STRUCTBLK *pstblk = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  const void *pvSrc;
  UINT offs;
#if !DOSASM
  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvStructBlk)) {
    // any error => return with error
    return 0;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  offs = GetPosition(pstblk, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return -1;
  } // if
  if (nMemberId != 0) {
    // get member characteristics
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return -1;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  // calculate source data position
  pvSrc = pstblk->abData + offs;
  // return data in dependence of data type
  switch (pmbd->idType) {
    case MBTYP_BYTE:
      // return byte value
      return *(BYTE*)pvSrc;
    case MBTYP_USHORT:
      // return 16 bits value
      return *(USHORT*)pvSrc;
    case MBTYP_ULONG:
    case MBTYP_DATA:
    case MBTYP_STRING: {
      // return 32 bits value, also valid for pointers
      ULONG ul = *(ULONG*)pvSrc;
      if (ul == (ULONG_MAX)-1) {
        // -1 returned but no error: reset error flag
        WkbSetLastError(WKB_NO_ERROR);
      } // if
      return ul;
    } // case
    default:
      // type which is not supported: set error code
      WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
      return -1;
      break;
  } // switch
} // WkbStructGetValue()


/*****************************************************************************
W k b S t r u c t S e t V a l u e
==============================================================================

This function sets the specified value into the specified (single) value
member in the specified structure.

Parameter
  pvStructBlk  points to the StructBlock in which the value member
               is written.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here.
  lValue ..... contains the source value which is stored.

Return
  The function returns TRUE if the operation was successful and
  FALSE if an error occurred. In this case a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.

*****************************************************************************/

BOOL WKAPIENTRY WkbStructSetValue(void *pvStructBlk,
  ULONG ulStructPos, INT nMemberId, LONG lValue)
{
  STRUCTBLK *pstblk = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  const void *pvDest;
  UINT offs;
#if !DOSASM
  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pvStructBlk)) {
    // any error => return with error
    return FALSE;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  offs = GetPosition(pstblk, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return FALSE;
  } // if
  if (nMemberId != 0) {
    // get member characteristics
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return FALSE;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  // calculate destination data position
  pvDest = pstblk->abData + offs;
  // transfer data in dependence of data type, return
  switch (pmbd->idType) {
    case MBTYP_BYTE:
      // store byte value
      *(BYTE*)pvDest = (BYTE)lValue;
      return TRUE;
    case MBTYP_USHORT:
      // store 16 bits value
      *(USHORT*)pvDest = (USHORT)lValue;
      return TRUE;
    case MBTYP_ULONG:
    case MBTYP_DATA:
    case MBTYP_STRING:
      // store 32 bits value, also valid for pointers
      *(ULONG*)pvDest = (ULONG)lValue;
      return TRUE;
    default:
      // type which is not supported: set error code & return
      WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
      return FALSE;
  } // switch
} // WkbStructSetValue()


/*****************************************************************************
W k b S t r u c t G e t S t r i n g
==============================================================================

This function gets the contents of the specified string value member in the
specified structure block and writes the contents into the specified
destination buffer.

Parameter
  pvStructBlk  points to the StructBlock from which the value member
               is read.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here.
  pszDest .... points to a buffer which receives the determined string.
               The stored string is terminated by a 0-character. Never more
               than <cchDest> characters are stored into the buffer. If
               <cchDest> is 0, this parameter is not used and may be set to
               NULL.
  cchDest .... contains the length of the specified buffer in characters.

Return
  The function returns one of the following values:
  * 0 if any error occurred. In this case a subsequent call of the
    WkbGetLastError() function returns the corresponding error code.
  * If the length in <cchDest> is too small to receive all returned
    information, the needed buffer length in number of characters is returned.
  * A subsequent call of WkbGetLastError() returns the WKB_BUFFER_OVERFLOW
    code. If the length in <cchDest> is large enough, the function returns
    the number of transferred characters without (!) the trailing 0.

Comments
  Unicode is currently not supported. Later a single bit in the <nMemberId>
  parameter will be used to store Unicode characters.

*****************************************************************************/

INT WKAPIENTRY WkbStructGetString(const void *pvStructBlk,
  ULONG ulStructPos, INT nMemberId, char *pszDest, UINT cchDest)
{
  const STRUCTBLK *pstblk = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  UINT cbString = 0;
  const char *pszSrc = NULL;
  UINT offs = 0;
#if !DOSASM
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle(&pvStructBlk)) {
    // any error => return with error
    return FALSE;
  } // if
  if (!WkdConvertMemHandle((const void**) &pszDest)) {
    // any error => return with error
    return FALSE;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  offs = GetPosition(pstblk, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return 0;
  } // if
  if (nMemberId != 0) {
    // get member characteristics, return if member invalid
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return 0;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  if (pmbd->idType != MBTYP_STRING) {
    // type which is not supported: set error code & return
    WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
    return 0;
  } // if
  // determine string address
  if (*(ULONG*)pstblk->abData & WKB_USEOFFSETS) {
    // string is specified as offset: calculate address
    pszSrc = (const char*)(pstblk->abData + *(ULONG*)(pstblk->abData + offs));
  } else {
    // string stored by pointer (not typical for StructBlocks)
    pszSrc = (const char*)(pstblk->abData + offs);
  } // if
  cbString = strlen(pszSrc) + 1;
  // transfer string into destination
  if (cchDest < cbString) {
    // destination too small => return with error
    WkbSetLastError(WKB_ERROR_BUFFER_OVERFLOW);
    return cbString;
  } // if
  // transfer string into string space, increase size of string space
  memcpy(pszDest, pszSrc, cbString);
  // return number of transferred characters without trailing 0
  return cbString - 1;
} // WkbStructGetString()


/*****************************************************************************
W k b S t r u c t S e t S t r i n g
==============================================================================

This function sets the specified string value into the specified value member
in the specified structure. This function uses automatically the next free
writing position in the string area of the StructBlock. The function manages
also the setting of the <flCtrl> entry of the stored WibuKey API structure
that strings are stored by their structure offsets.

Parameter
  pvStructBlk  points to the StructBlock in which the value member
               is written.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here.
  pszSrc ..... points to the 0-terminated string which is stored.

Return
  The function returns TRUE if the operation was successful and
  FALSE if an error occurred. In this case a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.

Comments
  Unicode is currently not supported. Later a single bit in the <nMemberId>
  parameter will be used to store Unicode characters.

Internal Comments
  The current version doesn't support the dynamic storing of strings. A future
  implementation could do following steps:
  * Blocks which are written via WkbStructSetData() are stored at the
    beginning of the string area as "fixed blocks".
  * If strings or blocks are changed, they are stored at the old location.
    If the block length has changed, all following strings are moved and
    their offsets in the WibuKey-API-structure are updated.
  * A problem occurs if a port string is replaced: This could be also a
    port number (flCtrl has set the WKB_PORTNUMBER flags) and cannot be
    differed from offset values. That's why the complete structure contents
    must be checked to solve such conflicts.

*****************************************************************************/

BOOL WKAPIENTRY WkbStructSetString(void *pvStructBlk, ULONG ulStructPos,
  INT nMemberId, const char *pszSrc)
{
  STRUCTBLK *pstblk = NULL;
  STRUCTDEF *pstdf = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  ULONG *pulOffset = NULL;
  UINT cbString = 0;
  BYTE *pbString = NULL;
  UINT offs = 0;

#if !DOSASM
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pvStructBlk)) {
    // any error => return with error
    return FALSE;
  } // if
  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pszSrc)) {
    // any error => return with error
    return FALSE;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  offs = GetPosition(pstblk, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return 0;
  } // if
  if (nMemberId != 0) {
    // get member characteristics
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (!pmbd) {
      // member invalid => return
      return 0;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  pstdf = astdTable + (nStructId - WKB_STS_BASE);
  if (pmbd->idType != MBTYP_STRING || !(pstdf->fuCtrl & STDEF_STDCTRL)) {
    // type which is not supported: set error code & return
    WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
    return FALSE;
  } // if
  // calculate destination data position (offset or string address)
  pulOffset = (ULONG*)(pstblk->abData + offs);
  // determine if enough memory in string space, return if not
  cbString = strlen(pszSrc) + 1;
  if ((ULONG)pstblk->cbStringData + cbString + 1 >
      (ULONG)pstblk->cbStringArea
    ) {
    WkbSetLastError(WKB_ERROR_MEMORY_TOO_SMALL);
    return FALSE;
  } // if
  // calculate string destination address
  pbString = pstblk->abData + pstdf->cbStruct +
    pstblk->cbAddedArea + pstblk->cbStringData;
  // transfer string into string space, increase size of string space
  memcpy(pbString, pszSrc, cbString);
  pstblk->cbStringData += cbString;
  // set offset mode for string addresses
  *(ULONG*)pstblk->abData |= WKB_USEOFFSETS;
  // store offset of string
  *pulOffset = pbString - pstblk->abData;
  // return without error
  return TRUE;
} // WkbStructSetString()


/*****************************************************************************
W k b S t r u c t G e t D a t a
==============================================================================

This function copies the structure data with the specified MemberId into the
specified destination buffer. If the MemberId represents a structure member,
the contents of the member is transferred into the buffer. Array members are
copied as one data block. If the MemberId represents a pointer to a block in
the string space, the data at this pointer is copied until the destination
buffer is full or the end of the string space is reached.

Parameter
  pvStructBlk  points to the StructBlock from which the value member is read.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here.
  pvDest ..... points to the buffer which receives the structure data.
  cbDest ..... contains the length of the destination buffer in bytes.

Return
  The function returns one of the following values:
    0 if any error occurred. In this case a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.
    If the length in <cbDest> is too small to receive all returned
  information, the needed buffer length in number of bytes is returned.
  A subsequent call of WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
    If the returned value is not 0 and not larger than <cbDest>, the
  operation was successful and the number of received bytes is returned.

*****************************************************************************/

UINT WKAPIENTRY WkbStructGetData(void *pvStructBlk,
  ULONG ulStructPos, INT nMemberId, void *pvDest, UINT cbDest)
{
  const STRUCTBLK *pstblk = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  UINT cbData = 0;
  UINT cbEntry = 0;
  const BYTE *pbSrc = NULL;
  UINT offs = 0;

#if !DOSASM
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pvStructBlk)) {
    // any error => return with error
    return 0;
  } // if
  if (!WkdConvertMemHandle((const void**) &pvDest)) {
    // any error => return with error
    return 0;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  offs = GetPosition(pstblk, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return 0;
  } // if
  if (nMemberId != 0) {
    // get member characteristics
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return 0;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  // calculate source data position
  pbSrc = pstblk->abData + offs;
  if (pmbd->idType == MBTYP_DATA) {
    // data pointer reference
    UINT offs = *(UINT*)pvDest;
    if (offs == 0) {
      // no data available => return without error
      WkbSetLastError(WKB_NO_ERROR);
      return 0;
    } // if
    cbData = astdTable[nStructId - WKB_STS_BASE].cbStruct + 
      pstblk->cbAddedArea + pstblk->cbStringData;
    if (cbData < offs) {
      // no data available (invalid offset?)
      WkbSetLastError(WKB_NO_ERROR);
      return 0;
    } // if
    // determine data length
    cbData = offs - cbData;
    // set pointer to begin of data
    pbSrc = pstblk->abData + offs;
  } else {
    // determine size of member (arrays: complete member)
    cbData = GetMemberSize(pstblk, pmbd, &cbEntry);
  } // if
  if (cbData < cbDest) {
    // reduce copied area
    cbDest = cbData;
  } // if
  // transfer data
  memcpy(pvDest, pbSrc, cbDest);
  // return number of available data (independence of copied length)
  return cbDest;
} // WkbStructGetData()


/*****************************************************************************
W k b S t r u c t S e t D a t a
==============================================================================

This function copies the specified source data with the specified length into
the structure member with the specified MemberId, If the MemberId represents
a structure member, the complete structure is overwritten with the specified
data. Array members are seen as one data block. If the MemberId represents a
data pointer, the data are stored with the number of specified bytes at the
next free position in the string space.

Parameter
  pvStructBlk  points to the StructBlock from which the value member is read.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here.
  pvSrc ...... points to a data area which is stored.
  cbSrc ...... contains the length in bytes of the data area which is stored.

Return
  The function returns the number of transferred bytes. If the function
  failed, 0 is returned. In this case a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.

*****************************************************************************/

UINT WKAPIENTRY WkbStructSetData(void *pvStructBlk, ULONG ulStructPos,
  INT nMemberId, const void *pvSrc, UINT cbSrc)
{
  STRUCTBLK *pstblk = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  UINT cbData = 0;
  UINT cbEntry = 0;
  BYTE *pbData = NULL;
  UINT offs = 0;

#if !DOSASM
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pvStructBlk)) {
    // any error => return with error
    return 0;
  } // if
  if (!WkdConvertMemHandle(&pvSrc)) {
    // any error => return with error
    return 0;
  } // if
#endif
  // get structure block contents
  pstblk = pvStructBlk;
  offs = GetPosition(pstblk, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return 0;
  } // if
  if (nMemberId != 0) {
    // get member characteristics, return if member invalid
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return 0;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  // calculate source data position
  pbData = pstblk->abData + offs;
  if (pmbd->idType == MBTYP_DATA) {
    // data pointer reference: add data at end of string area
    if ((ULONG)pstblk->cbStringData + cbSrc >= (ULONG)pstblk->cbStringArea) {
      WkbSetLastError(WKB_ERROR_MEMORY_TOO_SMALL);
      return 0;
    } // if
    // calculate destination offset
    offs = astdTable[nStructId - WKB_STS_BASE].cbStruct +
      pstblk->cbAddedArea + pstblk->cbStringData;
    // save offset value to data
    *(ULONG*)pbData = offs;
    // set offset mode for addressing
    *(ULONG*)pstblk->abData |= WKB_USEOFFSETS;
    // increase size of string space
    pstblk->cbStringData += cbSrc;
    // set pointer to begin of data
    pbData = pstblk->abData + offs;
    // set size of data to size of specified data
    cbData = cbSrc;
  } else {
    // determine size of member (arrays: complete member)
    cbData = GetMemberSize(pstblk, pmbd, &cbEntry);
  } // if
  if (cbSrc < cbData) {
    // not enough data: fill remainder with 0 bytes
    memset(pbData + cbSrc, 0, cbData - cbSrc);
    cbData = cbSrc;
  } // if
  // transfer data
  memcpy(pbData, pvSrc, (size_t)cbData);
  // return number of available data (independence of copied length)
  return cbData;
} // WkbStructSetData()


/*****************************************************************************
W k b S t r u c t G e t S t r u c t
==============================================================================

This function gets the contents of the specified structure member in the
specified structure block and writes the contents into the specified
destination buffer.

Parameter
  pvStructBlk  points to the StructBlock from which the value member
               is read.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here. If both <ulStructPos> and <nMemberId>
               parameters are 0, the complete structure including added data
               and strings is copied to the destination StructBlock.
  pvStructBlkDest  points to an initialized StructBlk which receives the data.
               Its type must match the addressed structure member and the size
               must be large enough to hold the specified data. Any previous
               contents of the StructBlk is cleared.

Return
  The function returns TRUE if the operation was successful and
  FALSE if an error occurred. In this case a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.

*****************************************************************************/

BOOL WKAPIENTRY WkbStructGetStruct(const void *pvStructBlk,
  ULONG ulStructPos, INT nMemberId, void *pvStructBlkDest)
{
  const STRUCTBLK *pstblkSrc = NULL;
  STRUCTBLK *pstblkDest = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  UINT cbData = 0;
  UINT cbEntry = 0;
  UINT offs = 0;

#if !DOSASM
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle(&pvStructBlk)) {
    // any error => return with error
    return FALSE;
  } // if
  if (!WkdConvertMemHandle((const void**) &pvStructBlkDest)) {
    // any error => return with error
    return FALSE;
  } // if
#endif
  // get structure block contents
  pstblkSrc = pvStructBlk;
  offs = GetPosition(pstblkSrc, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return FALSE;
  } // if
  if (nMemberId != 0) {
    // get member characteristics
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return FALSE;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  // get destination information
  pstblkDest = pvStructBlkDest;
  if (pstblkDest->nStructId != pmbd->idType) {
    // incompatible structure types => return with error
    WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
    return FALSE;
  } // if
  // determine size of source (arrays: complete member)
  cbData = GetMemberSize(pstblkSrc, pmbd, &cbEntry);
  if (astdTable[nStructId - WKB_STS_BASE].cbStruct +
      pstblkDest->cbAddedArea < cbData
    ) {
    // not enough memory in destination => return with error
    WkbSetLastError(WKB_ERROR_MEMORY_TOO_SMALL);
    return FALSE;
  } // if
  // transfer data into destination StructBlock
  memcpy(pstblkDest->abData, pstblkSrc->abData + offs, cbData);
  // return without error
  return TRUE;
} // WkbStructGetStruct()


/*****************************************************************************
W k b S t r u c t S e t S t r u c t
==============================================================================

This function sets the complete contents of the specified StructBlock into
the specified value member in the specified structure.

Parameter
  pvStructBlk  points to the StructBlock in which the value member
               is written.
  ulStructPos  contains the StructPos which specifies a substructure. To
               address a member on the highest level in the structure block,
               0 must be specified.
  nMemberId .. contains the MemberId of the desired structure member. If
               a member is specified in <ulStructPos> which can be addressed
               directly as a value (for example an entry of an array), 0
               may be specified here. If both <ulStructPos> and <nMemberId>
               parameters are 0, the complete structure including added data
               and strings is copied from the source StructBlock.
  pvStructBlkSrc  points to an initialized StructBlk which contains the data
               to transfer into the StructBlk destination. Its type must match
               the addressed structure member and the size should be not
               larger than the specified in the destination.

Return
  The function returns TRUE if the operation was successful and
  FALSE if an error occurred. In this case a subsequent call of the
  WkbGetLastError() function returns the corresponding error code.

*****************************************************************************/

BOOL WKAPIENTRY WkbStructSetStruct(void *pvStructBlk, ULONG ulStructPos,
  INT nMemberId, const void *pvStructBlkSrc)
{
  STRUCTBLK *pstblkDest = NULL;
  const STRUCTBLK *pstblkSrc = NULL;
  INT nStructId = 0;
  const MEMBERDEF *pmbd = NULL;
  UINT cbData = 0;
  UINT cbEntry = 0;
  UINT offs = 0;

#if !DOSASM
  // convert handle addresses to true addresses if required
  if (!WkdConvertMemHandle((const void**) &pvStructBlk)) {
    // any error => return with error
    return FALSE;
  } // if
  if (!WkdConvertMemHandle(&pvStructBlkSrc)) {
    // any error => return with error
    return FALSE;
  } // if
#endif
  // get structure block contents
  pstblkDest = pvStructBlk;
  offs = GetPosition(pstblkDest, ulStructPos, &nStructId);
  if (offs == (UINT)-1) {
    // invalid StructPos specified => return with error
    return FALSE;
  } // if
  if (nMemberId != 0) {
    // get member characteristics
    pmbd = GetMemberInfo(nStructId, nMemberId);
    if (pmbd == NULL) {
      // member invalid => return
      return FALSE;
    } // if
    // add member offset to StructPos or begin of StructBlock
    offs += pmbd->offs;
  } // if
  // get source information
  pstblkSrc = pvStructBlkSrc;
  if (pstblkSrc->nStructId != nStructId) {
    // incompatible structure types => return with error
    WkbSetLastError(WKB_ERROR_INVALID_PARAMETER);
    return FALSE;
  } // if
  // determine size of source (arrays: complete member)
  cbData = astdTable[nStructId - WKB_STS_BASE].cbStruct +
    pstblkSrc->cbAddedArea;
  if (GetMemberSize(pstblkSrc, pmbd, &cbEntry) < cbData) {
    // not enough memory in destination => return with error
    WkbSetLastError(WKB_ERROR_MEMORY_TOO_SMALL);
    return FALSE;
  } // if
  // transfer data into destination StructBlock
  memcpy(pstblkDest + offs, pstblkSrc->abData, cbData);
  // return without error
  return TRUE;
} // WkbStructSetStruct()


/*****************************************************************************
W k b B i t R e s e t S e t
==============================================================================

This function changes the bits in an 32 bit integer value by logical
operations. At first, the bits in the specified source value are "and" masked
with the negation of the reset mask value. Then, the result of this operation
is "or" masked by the setting mask value. The final result is returned.

Parameter
  ulSrc ...... contains the 32 bit value which is modified.
  ulResetMask  contains the reset mask: All bits which are set in this value
               resets the corresponding bits in the source value.
  ulSetMask .. contains the setting mask: All bits which are set in this value
               sets also the corresponding bits in the source value.

Return
  The determined value is returned.

*****************************************************************************/

ULONG WKAPIENTRY WkbBitResetSet(ULONG ulSrc, ULONG ulResetMask,
  ULONG ulSetMask)
{
  ulSrc &= ~ulResetMask;
  ulSrc |= ulSetMask;
  return ulSrc;
} // WkbBitResetSet()


// ======================================
// end of clstruct.c
// ======================================
