/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
clsubs.c                                          Subsystem API Implementation
==============================================================================

This module implements the following WibuKey API functions:
  WkbGetLastError(), WkbSetLastError(), WkbAccess2(), WkbRelease2(),
  WkbQuerySystem2(), WkbSetCompletionEvent(), WkbGetCompletionEvent(),
  WkbSetCompletionFunction(), WkbGetCompletionFunction(), WkbIsComplete(),
  WkbEnumServers2(), WkbControlSystem2(), WkbReset2(), WkbGetVersion2(),
  WkbQueryStatus2().

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52 D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All POSIX OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
92Sep13 1.33 mabu: start of development
93Mar18 1.33 mabu: custom source variant released
93Mar21 1.35 mabu: adaption to Windows
93Apr13 1.35 mabu: WkbUniversalCall() included (Module UNICALL.C)
93Jun06 1.35 mabu: conversation to large memory model.
93Jun20 1.35 mabu: adaptation to Win32 environment.
93Jul02 1.35 mabu: first adaptations to WibuBox version 2
93Jul22 1.35 mabu: adapation to DOS
93Aug21 2.00 mabu: greater changes in API of new functions
94Mar08 2.00 mabu: low level driver variants added
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94Aug10 2.00 mabu: adaptation to new WkNet/WkLAN API
94Oct11 2.00 mabu: implementing the look-ahead-function
94Nov11 2.00 mabu: version 2.00 released
95Feb08 2.02 mabu: bug in CreateHandle() function removed
95Feb09 2.02 mabu: extracted from KRNL.C
95Jun02 2.06 mabu: implementation of WkbSetLastError()
96Jun23 2.13 mabu: renamed from BSSYS.C
96Aug28 2.14 mabu: valid server name for local systems in WkbQuerySystem2()
96Oct19 2.15 mabu: redirection of WkbAccess2()/WkbRelease2() for Win16->Win32
96Dec04 2.21 mabu: WKWIN.DLL: proper setting of WkbQueryStatus2() kernel mode
96Dec06 2.21 mabu: saving of subsystem Firm/User Code in base system structure
97Feb11 2.25 mabu: Unicode adaptation in WkbEnumServers2() function
97Feb25 2.30 mabu: adding for scanning dialog for WkbEnumServers2() function
97Jun11 2.31 mabu: big/little endian adaptation for Macintosh
97Oct19 2.40 mabu: implementation of WKB_CTS_CANCEL_USERS
98Jan28 2.41 mabu: implementation of WKB_QSY_IPADDR
98Feb06 2.41 mabu: integration of new error codes
99Feb19 2.52 as:   implementation of WkbQueryInfo2()
99Mar10 2.52 as:   WKB_QSY_DIALOG added to WkbQuerySystem2()
99Oct27 2.52 as:   WKB_QSY_OPENENTRY added to WkbQuerySystem2()
99Oct27 2.52 as:   implementation of WkbApplicationInfo2()
00May25 3.00 as:   implementation of WkbGetVersion3()
00Aug28 3.00 as:   implementation of WKB_ACC_STATIONSHARE
01mar19 3.01 as:   bugs in WkbApplicationInfo2 fixed
10Jun24 4.31 ha:   build update

*****************************************************************************************/

#include "proj.h"

#if _UNIX
  #include <netdb.h>
#endif




/*****************************************************************************

---- Info Dialog Management --------------------------------------------------

*****************************************************************************/


/*****************************************************************************

---- Internal Functions ------------------------------------------------------

*****************************************************************************/

/*****************************************************************************
WkdDrvGetVersion
==============================================================================

This function is identically with the WkbGetVersion2() call except that the
error message printing is not suppressed in a Windows calling driver.

Parameters
  see WkbGetVersion2()

Return
  see WkbGetVersion2()

*****************************************************************************/

INT WkdDrvGetVersion (HWKBSYSTEM hwkbsys)
{
  INT idResult;

  if (hwkbsys == NULL) {
    // return version of driver containing API
    WkdResetStatus();
    idResult = 256 * DRV_SUBVERSION + DRV_VERSION;
    goto TRANSFORM_VERSION;
  } // if
  // determine accessed system, initialize <cstWkd>
  if (!WkdGetSystem(hwkbsys, 0)) {
    // any error => return 0
    idResult = 0;
    goto RETURN_VERSION;
  } // if
  {//// execute driver operations
    ASSERT(cstWkd.fuSysType & STYPE_LOCAL);
    // store version number
    idResult = 256 * DRV_SUBVERSION + DRV_VERSION;
  }

TRANSFORM_VERSION:
  if (idResult != 0) {
    // calculate the the vice versa version value
    INT nSubVer = HIBYTE(idResult);
    INT nVer = LOBYTE(idResult);

    // Check if version higher than 3.0. If the version
    // is smaler than 3.0 then is nothing to do.
    if ((256 * nVer + nSubVer) >= (256 * 3 + 0)) {
      // calculate the corret version (2.60 ...)
      nSubVer += (nVer - 2) * 60;
      idResult = (2 | (nSubVer << 8));
    } // if
  } // if

RETURN_VERSION:
  return WkdReturnChecked(idResult);
} // WkdDrvGetVersion()


/*****************************************************************************
SetSubSystemEntry
==============================================================================

This function stores the specification of the specified entry into an
subsystem entry object.

Parameters
  hwkbe ... contains the handle of the opened entry.
  psysDest  contains the pointer to the subsystem object which receives the
            entry description.

*****************************************************************************/

static void SetSubSystemEntry(HWKBENTRY hwkbe, SYSENTRY *psysDest)
{
  const OPENENTRY *pope;
  ASSERT(WkdGetHandleType(hwkbe) == HTYP_ENTRY);
  // determine address of WibuBox entry object
  pope = WkdGetHandlePtr(hwkbe);
  ASSERT(pope != NULL);
  ASSERT(pope->lPort >= 0 && pope->lPort < 65535);
  psysDest->lFirmCode = pope->ulFirmCode;
  psysDest->lUserCode = pope->ulUserCode;
  psysDest->usPort = (USHORT)pope->lPort;
  psysDest->usEntry = pope->iEntry;
  psysDest->wkbsr = pope->wkbsr;
  psysDest->fMaster = (pope->flCtrl & WKB_OE_MASTER) != 0;
} // SetSubSystemEntry()


/*****************************************************************************
CreateSubSystem
==============================================================================

This function creates a simple subsystem with the specified parameters.
It is called by WkbAccess2() when following parameters are set:
* WKB_ACC_CREATE with WKB_ACC_LOCAL
* WKB_ACC_SHELL
* WKB_ACC_GLOBAL

Parameters
  flCtrl contains the control flags of the WkbAccess2() call. They are
         used for the following constants:
         * WKB_ACC_NONETLIMIT
         * WKB_ACC_EXCLUSIVE
         * WKB_ACC_STATIONSHARE
         * WKB_ACC_SHELL
  pvCtrl points to the subsystem parameters. In dependence of the control
         code in <flCtrl>, this may be a WKBACCESS or a WKBOPENENTRY
         structure.

Return
  The handle of the created subsystem is returned. It is NULL if the subsystem
  could not be created.

*****************************************************************************/

static HWKBSYSTEM CreateSubSystem(ULONG flCtrl, const void *pvCtrl)
{
  const WKBACCESS *pwkbacc;
  HWKBSYSTEM hwkbsys;
  SYSENTRY *psys = NULL;
  UINT uType = 0;
  BOOL fGlobal = FALSE;

  uType = (UINT)flCtrl & WKB_ACC_MASK;
  fGlobal = (uType == WKB_ACC_GLOBAL);
  // <pakbacc> only valid if fGlobal == FALSE
  pwkbacc = pvCtrl;
  ASSERT(cstWkd.idError == 0);
  // create & initialize local subsystem
  psys = WkdCreateHandle(HTYP_SYS, sizeof(SYSENTRY), (HANDLE*) &hwkbsys);
  if (hwkbsys == NULL) {
    // no handle available => return with error
    return NULL;
  } // if
  // initialize created system block: copy type from HWKB_LOCAL system
  psys->fuType = gsysWkdLocal.fuType;
  if (!fGlobal) {
    if (flCtrl & WKB_ACC_NONETLIMIT) {
      // set flag that opened without net limitation
      psys->fuFlags |= SFLAG_NONETLIMIT;
    } // if
    if (flCtrl & WKB_ACC_EXCLUSIVE) {
      // set flag that opened in exclusive mode
      psys->fuFlags |= SFLAG_EXCLUSIVE;
    } // if
    if (flCtrl & WKB_ACC_STATIONSHARE) {
      // set flag that opened in station share mode
      psys->fuFlags |= SFLAG_STATIONSHARE;
    } // if
  } // if
  {//// execute driver operations
    UINT32 flOpenCtrl = 0;
  #if ASYNCMODE
    if (cstWkd.fAsyncCall) {
      flOpenCtrl |= WKB_ASYNCHRONOUS;
    } // if
  #endif
    if (cstWkd.fUnicodeCall) {
      flOpenCtrl |= WKB_UNICODE;
    } // if
    // set also current level for call
    flOpenCtrl |= (cstWkd.bLevel << 8);
    if (fGlobal) {
      // ==============
      // WKB_ACC_GLOBAL
      // ==============
      HWKBENTRY hwkbe;
      WKBOPENENTRY wkboe;
      ASSERT(uType == WKB_ACC_GLOBAL);
      wkboe = *(WKBOPENENTRY*)pvCtrl;
      // reset all flags (open with WKB_VERSION1 ...) but set WKB_OPEN_GLOBAL
      wkboe.flCtrl = WKB_OE_GLOBAL;
      // open WibuBox with specified entry (UserCode 0 <=> unspecified)
      hwkbe = WkbOpen2(hwkbsys, flOpenCtrl | WKB_OPEN_ENTRY,
        NULL, 0, 0, &wkboe);
      if (hwkbe == NULL) {
        // specified WibuBox not found: set valid error code again
        cstWkd.idError = gidWkdLastError;
        goto RETURN;
      } // if
      // store properties of the entry into subsystem object
      SetSubSystemEntry(hwkbe, psys);
    } else {
      // ==============================
      // WKB_ACC_CREATE / WKB_ACC_SHELL
      // ==============================
      ASSERT(uType == WKB_ACC_CREATE || uType == WKB_ACC_SHELL);
      flOpenCtrl |= WKB_VERSION_RED;
      if (uType != WKB_ACC_SHELL) {
        // analyse driver version & search WibuBox
        HWKBENTRY hwkbe;
        if ( (pwkbacc->bVersionHigh > DRV_VERSION) ||
             ((pwkbacc->bVersionHigh == DRV_VERSION) &&
             (pwkbacc->bVersionLow > DRV_SUBVERSION))
          ) {
          // bad version of local system => return with error
          cstWkd.idError = WKB_ERROR_INVALID_VERSION;
          goto RETURN;
        } // if
        // search WibuBox at local subsystem
        // * force share/user mode flags to WkbOpen2() flags
        if (flCtrl & WKB_ACC_NONETLIMIT) {
          flOpenCtrl |= WKB_OPEN_NONETLIMIT;
        } // if
        if (flCtrl & WKB_ACC_EXCLUSIVE) {
          flOpenCtrl |= WKB_OPEN_EXCLUSIVE;
        } // if
        if (flCtrl & WKB_ACC_STATIONSHARE) {
          flOpenCtrl |= WKB_OPEN_STATIONSHARE;
        } // if
        // open entry for this subsystem
        hwkbe = WkbOpen2(hwkbsys, flOpenCtrl | WKB_OPEN_FIND,
          NULL, pwkbacc->lFirmCode, pwkbacc->lUserCode, NULL);
        // Firm/Code Code in local subsystem not found: return error
        if (hwkbe == NULL) {
          // set valid error code again
          cstWkd.idError = gidWkdLastError;
          goto RETURN;
        } // if
        // store properties of the entry into subsystem object
        SetSubSystemEntry(hwkbe, psys);
      } // if
      // initialize created system block: copy type from HWKB_LOCAL system
      psys->fuType = gsysWkdLocal.fuType;
      psys->fuFlags = 0;
      if (flCtrl & WKB_ACC_NONETLIMIT) {
        // set flag that opened without net limitation
        psys->fuFlags |= SFLAG_NONETLIMIT;
      } // if
      if (flCtrl & WKB_ACC_EXCLUSIVE) {
        // set flag that opened in exclusive mode
        psys->fuFlags |= SFLAG_EXCLUSIVE;
      } // if
      if (flCtrl & WKB_ACC_STATIONSHARE) {
        // set flag that opened in station share mode
        psys->fuFlags |= SFLAG_STATIONSHARE;
      } // if
    } // if
  }
RETURN:
  // set pointer to created subsystem into status structure
  // * this may not be set previously because WkbOpen2() may overwrite it!
  cstWkd.psys = psys;
  // analyse error condition
  if (cstWkd.idError && hwkbsys != NULL) {
    // error condition: free created handle
    WkdDestroyHandle(hwkbsys);
    // return with error
    return NULL;
  } // if
  // return without error
  return hwkbsys;
} // CreateSubSystem()


/*****************************************************************************
DestroySubSystem
==============================================================================

This function destroys the subsystem system which corresponds with the
specified handle.

Parameters
  hwkbsys  contains the handle of the system which is destroyed.

*****************************************************************************/

static void DestroySubSystem(HWKBSYSTEM hwkbsys)
{
  SYSENTRY *psys = NULL;
  if (hwkbsys == HWKB_LOCAL) {
    // local system: never destroyed, ignore call
    return;
  } // if
  ASSERT(WkdGetHandleType(hwkbsys) == HTYP_SYS);
  // determine corresponding handle address
  psys = WkdGetHandlePtr(hwkbsys);
  ASSERT(psys != NULL);
  if (psys->fuType & (STYPE_KERNEL|STYPE_LOCAL)) {
    // local system: nothing to do
  } else {
    ASSERT(0);
  } // block/if
  // close all opened WibuBox entries which a connect with this subsystem
  WkdCloseAllOfSubsystem(hwkbsys);
  //!P:mabu: destroy all memory blocks which are connected with this subsystem
#if ASYNCMODE
  // destroy all actions which are connected with this subsystem
  WkdRemoveAllActionsOfSubSystem(psys);
#endif
  // destroy now handle itself
  WkdDestroyHandle(hwkbsys);
} // DestroySubSystem()



// flag "subsystem specific profile read", set to FALSE by WKB_CTS_RESET
BOOL fWkdSubSysProfileRead = FALSE;

// ordering of subsystem scanning
static BYTE abSubSysOrder[4];

/*****************************************************************************
ReadSubSysProfile
==============================================================================

This function reads the contents of the driver setting variable SubSystem
into the abSubSysOrder[] array. This function is called only once when the
system is started or after the profile should be read again after calling
the WkbControlSystem2() function with WKB_CTS_RESET.

Comments
  The Macintosh drivers uses always the standard subsystem order
  Local/WkLAN/WkNet.

*****************************************************************************/

static void ReadSubSysProfile(void)
{
  char *pszBuf = NULL;
  char *pszParse = NULL;
  UINT cchLength = 0;
  BYTE *pbSubSysOrder = NULL;

  if (fWkdSubSysProfileRead) {
    // already read => nothing to do
    return;
  } // if
  // set flag that read
  fWkdSubSysProfileRead = TRUE;
  // set default values: kernel, WkLAN; WkNet
  abSubSysOrder[0] = WKB_ACC_LOCAL;
  abSubSysOrder[1] = WKB_ACC_WKLAN;
  abSubSysOrder[2] = 0;
  abSubSysOrder[3] = 0;

  pszBuf = BsrAllocMem(256 * sizeof(char));
  if (pszBuf == NULL) {
    // memory too small => use default
    return;
  } // if
  // search specified environment variable (system specific)
  cchLength = WkdGetProfileValue(WKPRFTPC_DRIVER_CLIENT, "SubSystem", pszBuf, 256);
  if (cchLength == 0 || cchLength > 256) {
    // variable not found in environment or argument too long => use default
    goto RETURN;
  } // if
  // set destination position
  pbSubSysOrder = abSubSysOrder;
  // set search position
  pszParse = pszBuf;
  for (;;) {
    char *pszNext = NULL;
    INT idSubSys = 0;
    // search specified type name in environment string
    // * skip leading spaces
    WkdSkipSpace(&pszParse);
    // check next string for specific name
    pszNext = WkdCompString(pszParse, "KERNEL");
    if (pszNext != NULL) {
      // string matched: store next specification
      idSubSys = WKB_ACC_LOCAL;
    } else {
      pszNext = WkdCompString(pszParse, "WKLAN");
      if (pszNext != NULL) {
        // string matched: store next specification
        idSubSys = WKB_ACC_WKLAN;
      } else {
        pszNext = WkdCompString(pszParse, "WKNET");
        if (pszNext != NULL) {
          // string matched: store next specification
          idSubSys = WKB_ACC_WKNET;
        } // if
      } // if
    } // if
    if (pszNext != NULL) {
      // string matched: check if valid end of name
      // * skip trailing spaces
      WkdSkipSpace(&pszNext);
      if (*pszNext == ',' || *pszNext == 0) {
        // end of name (comma follows or end of string) => name matched
        *pbSubSysOrder++ = (BYTE)idSubSys;
      } // if
    } // if
    // search next comma to check next entry
    while (*pszParse != 0 && *pszParse != ',') {
      pszParse++;
    } // if
    if (*pszParse == 0 || pbSubSysOrder == abSubSysOrder + 3) {
      // end of string or destination buffer full: exit loop
      break;
    } // if
    // to character after comma
    pszParse++;
  } // for cycle, exit by break;
  // set end of suborder list
  *pbSubSysOrder = 0;

RETURN:
  ASSERT(pszBuf != NULL);
  BsrFreeMem(pszBuf);
  return;
} // ReadSubSysProfile()



/*****************************************************************************

---- WibuKey API Functions --------------------------------------------------

*****************************************************************************/

/*****************************************************************************
WkbGetLastError
==============================================================================

This function returns the error code which was set at last by the WibuKey
driver for the current task. This code was produced by the last called
function which returned an error state. Functions which occurred successfully
don't change this value.

Return
  The function returns the last set WibuKey error code.

Preliminary Note
  The storing of task different values is not implemented.

*****************************************************************************/

INT WKAPIENTRY WkbGetLastError(void)
{
  INT nRet = 0;
  THREAD_API_LOCK
  nRet = gidWkdLastError;

	THREAD_API_UNLOCK

  return nRet;
} // WkbGetLastError()


/*****************************************************************************
WkbSetLastError
==============================================================================

This function sets the specified error code as the driver internal last
error code which is normally set by driver functions when an error occurs.

Parameters
 idErrorCode  contains the error code which is set.

Preliminary Note
 The storing of task different values is not implemented.

*****************************************************************************/

void WKAPIENTRY WkbSetLastError(INT idErrorCode)
{
  THREAD_API_LOCK

  gidWkdLastError = idErrorCode;

  THREAD_API_UNLOCK

} // WkbSetLastError()


/*****************************************************************************
WkbAccess2
==============================================================================

This function accesses a WibuKey subsystem locally or within the network.
A subsystem is determined locally by the specified Firm and User code or
by a WkLAN or WkNet server and one or more WIBU-BOXes with doubled entries:
The first Firm Code/User Code must be the same like the specified Firm and
User Code and is the access entry, the entry with the same Firm Code and the
next higher User Code limits the concurrent access to this subsystem.

Parameters
  flCtrl  contains a combination of flags which controls the execution of the
          function. Following command values are possible:
            WKB_ACC_CREATE   (0x00) creates a new subsystem.
            WKB_ACC_RESTORE  (0x10) restores a subsystem which status was
                             saved via the function WkbQuerySystem2 and the
                             WKB_QSY_STATUS control command.
            WKB_ACC_SHELL .. (0x20) creates a local subsystem which is not
                             attached to a specific box. It is only used to
                             free all opened WibuBox entries via this
                             subsystem.
          For the WKB_ACC_CREATE command, one or more of the following option
          flags must be used in combination:
            WKB_ACC_LOCAL (0x01) tries to open a local WibuBox with the
                    specified Firm and User Code. If such a WibuBox is
                    found, a subsystem is used which is compatible to the
                    standard subsystem, addressed by the HWKB_LOCAL constant,
                    except that the default setting of the standard bits in
                    the control flag parameters and members may be specified.
            WKB_ACC_WKLAN (0x02) tries to access a WkLAN network subsystem
                    which corresponds with the specified control parameters.
            WKB_ACC_WKNET (0x04) tries to access a WkNet network subsystem
                    which corresponds with the specified control parameters.

          For the other commands, none of these flags may be used. The
          restore command sets the same subsystem which was created before
          the WkbQuerySystem2 function was called which returns the last
          state.

          If two or more of these flags are opened, the function tries at
          first the local WibuBox, then the WkLAN subsystem and at least
          the WkNet subsystem. This ranking guarantees the selection of the
          fastest possible subsystem.

          Following flag may be additionally used:
          * WKB_ACC_NONETLIMIT ... (0x200000) addresses the specified network
                            subsystem directly without allocating any user.
          * WKB_ACC_EXCLUSIVE .... (0x400000) addresses the specified network
                            subsystem in exclusive mode.
          * WKB_ACC_STATIONSHARE . (0x800000) addresses the specified network
                            subsystem in station shre mode.

  pvCtrl  points to a buffer of different type in dependence of the specified
          command option:
          * If WKB_ACC_CREATE is set, <pvCtrl> points to a WKBACCESS
            structure, which contains a subsystem specification.
          * If WKB_ACC_RESTORE is set, <pvCtrl> points to an unspecified
            buffer, which was returned by the WkbQuerySystem2() function
            with the WKB_QSY_STATUS flag.

Return
  The function returns the handle of the opened subsystem or NULL if an error
  occurs. In this case a subsequent call of the function WkbGetLastError()
  returns the corresponding error code. The function doesnot return never the
  HWKB_LOCAL constant.

Preliminary Note
  The access to a WkLAN subsystem is currently not realized.

*****************************************************************************/

HWKBSYSTEM WKAPIENTRY WkbAccess2(ULONG flCtrl, const void *pvCtrl)
{
  HWKBSYSTEM hwkbsys = NULL;
  UINT idSubSysError = WKB_NO_ERROR;
  INT cSubSys = 0; // number of accessed subsystems

  // set standard in control flags
  WkdResetStatus();

  // check & store specified control flags, error => return
  if (!WkdGetCtrlFlags(flCtrl)) {
    goto RETURN;
  } // if
  
  // common setting of the following flags are not permitted
  if (flCtrl & WKB_ACC_STATIONSHARE) {
    if (flCtrl & (WKB_ACC_EXCLUSIVE | WKB_ACC_NONETLIMIT)) {
      // station share must be only set if the other
      // possible options are not set
      cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
      goto RETURN;
    } // if
  } // if
  // read contents of SubSystem driver setting variable if not already done
  ReadSubSysProfile();
  // convert handle address to true address if required
  if (!WkdConvertMemHandle(&pvCtrl)) {
    goto RETURN;
  } // if
  switch ((UINT)flCtrl & WKB_ACC_MASK) {
    case WKB_ACC_CREATE: {
      // create new WibuKey subsystem
      const WKBACCESS *pwkbacc;
      BYTE *pbNextSubSys = NULL;
      pwkbacc = pvCtrl;
    
      // access to subsystem variants in specified search order
      for (pbNextSubSys = abSubSysOrder;
          *pbNextSubSys != 0;
          pbNextSubSys++
        ) {
        // start without error
        cstWkd.idError = WKB_NO_ERROR;
        // try one or more of the specified subsystems
        if (flCtrl & *pbNextSubSys & WKB_ACC_LOCAL) {
          // try to access local subsystem
          cSubSys++;
          hwkbsys = CreateSubSystem(flCtrl, pwkbacc);
          if (hwkbsys) {
            goto SETSUBSYS;
          } // if
          // FB6912
          if (!WkdGetCtrlFlags(flCtrl)) {
            goto RETURN;
          } // if
        } // if
        if (cstWkd.idError != WKB_NO_ERROR) {
          // save error of accessed subsystem
          // * don't overwrite critical subsystem error code
          //   by harmless error codes
          if (idSubSysError == WKB_NO_ERROR ||
              idSubSysError == WKB_ERROR_ENTRY_NOT_FOUND ||
              idSubSysError == WKB_ERROR_SERVER_NOT_FOUND
            ) {
            // store if no error or general error in the past
            idSubSysError = cstWkd.idError;
          } // if
        } // if
      } // for
      // no subsystem accessed sucessfully: determine returned error code
      if (cSubSys == 0) {
        // no subsystem accessed => use special error code
        cstWkd.idError = WKB_ERROR_INVALID_SUBSYSTEM;
      } else if (cSubSys > 1 && idSubSysError == WKB_ERROR_ENTRY_NOT_FOUND) {
        // several subsystems accessed, not success but no critical error
        cstWkd.idError = WKB_ERROR_SERVER_NOT_FOUND;
      } else {
        // exactly one subsystem called or critical subsystem error
        // * return also some subsystem specific errors instead general code
        ASSERT(idSubSysError != WKB_NO_ERROR);
        cstWkd.idError = idSubSysError;
      } // if
      goto RETURN;
    SETSUBSYS:
      // final setting of subsystem: save also Firm/User Code
      cstWkd.psys->flStdCtrl = pwkbacc->flStdCtrl & STDFNFLAGS;
      cstWkd.psys->lFirmCode = pwkbacc->lFirmCode;
      cstWkd.psys->lUserCode = pwkbacc->lUserCode;
      goto RETURN;
    } // case
    case WKB_ACC_SHELL: {
      // create a local subsystem without a real WibuBox access
      const WKBACCESS *pwkbacc;
      pwkbacc = pvCtrl;
      hwkbsys = CreateSubSystem(flCtrl, pwkbacc);
      break;
    } // case
    case WKB_ACC_GLOBAL: {
      // =========================================
      // Global access to specified WibuBox entry
      // =========================================
      const WKBOPENENTRY *pwkboe;
      pwkboe = pvCtrl;
      hwkbsys = CreateSubSystem(flCtrl, pwkboe);
      break;
    } // case
    default:
      // other commands are not defined
      cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
  } // switch
RETURN:
  return (HWKBSYSTEM)(long)WkdReturnChecked((INT)(long)hwkbsys);
} // WkbAccess2()


/*****************************************************************************
WkbRelease2
==============================================================================

This function releases the access to a WibuKey subsystem.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem. The
           function doesn't nothing if the predefined HWKB_LOCAL handle is
           specified.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbRelease2(HWKBSYSTEM hwkbsys)
{
  INT iRet = 0;


  THREAD_API_LOCK

  if (!WkdGetSystem(hwkbsys, 0)) {
    goto RETURN;
  } // if
  {//// execute driver operations: nothing to do
  }
#if ASYNCMODE
  if (cstWkd.fAsyncCall && cstWkd.idError == WKB_NO_ERROR) {
    // if asynchronous mode: destroy this until WkbIsComplete() is called
    ASSERT(cstWkd.pasacThis != NULL);
    cstWkd.pasacThis->hwkbsysDestroy = hwkbsys;
  } else
#endif
  {// error or no asynchronous call: destroy now subsystem
    DestroySubSystem(hwkbsys);
  #if ASYNCMODE
    // avoid further destroying of asynchronous action
    cstWkd.fAsyncCall = FALSE;
  #endif // #if ASYNCMODE
  } // if
RETURN:

  iRet = WkdReturnBool();
  THREAD_API_UNLOCK

  return iRet;
} // WkbRelease2()





















/*****************************************************************************

---- WibuKey API Functions --------------------------------------------------

*****************************************************************************/



/*****************************************************************************
SetVersionToString
==============================================================================

This function sets a given version (given in WKBVERSION) to a string.
Before this string you can set another string which you can also give
as parameter. 

Parameters
  idString .... contains the ID of the string whichg is set before the
      version text.
  pwkbver ..... contains the version which should be set as a text.
  pszVersion .. contains the result of the text setting.
  cbVersion ... contains the length of the buffer.
  fDate ....... says if the date should be printed behind the version.

Return
  the length of the set text.

*****************************************************************************/

UINT SetVersionToString (UINT idString, WKBVERSION *pwkbver,
    char *pszVersion, UINT cbVersion, BOOL8 fDate)
{
  UINT cbLength = 0;
  char *pszBuf = NULL;
  char szFullText[200];
  char szVersion[100], szDate[100], szModule[100];

  switch (idString) {
    case DRVC_IDS_VERINFO_TEXT1:
      strcpy(szModule, "Local Driver:\t");
      break;
    case DRVC_IDS_VERINFO_TEXT2:
      strcpy(szModule, "Server:\t\t");
      break;
    case DRVC_IDS_VERINFO_TEXT3:
      strcpy(szModule, "Server Driver:\t");
      break;
    case DRVC_IDS_VERINFO_TEXT4:
      strcpy(szModule, "Calling Driver:\t");
      break;
    case DRVC_IDS_VERINFO_TEXT5:
      strcpy(szModule, "Kernel Driver:\t");
      break;
    default:
      ASSERT(idString == 0);
      strcpy(szModule, "");
  } // switch
  sprintf(szVersion, "%s%u.%02u.%u.%03u", szModule, pwkbver->usVersion, pwkbver->usSubVersion,
      pwkbver->usBuild, pwkbver->usCount);
  pszBuf = szVersion;
  if (fDate) {
    sprintf(szDate, "%s of %02u/%02u/%04u", szVersion, pwkbver->bMonth, pwkbver->bDay, pwkbver->usYear);
    pszBuf = szDate;
  } // if
  sprintf (szFullText, "%s", pszBuf);

  cbLength = (UINT)strlen(szFullText);
  szFullText[cbLength] = 0;

  // set the return values
  if (cbVersion >= cbLength) {
    memcpy (pszVersion, szFullText, (size_t)(cbLength + 1));
  } else {
    cbLength = 0;
  } // if

  return cbLength;
} // SetVersionToString()



/*****************************************************************************
WkbQuerySystem2
==============================================================================

This function returns information status information about the specified
subsystem. This is normally a WkNet or a WkLAN system.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
  flCtrl   contains flags which controls the execution of the function.
           Currently only the following flag is defined:
             WKB_QSY_SYSTEM (0x00) returns the current system status in a
               WKBSYSTEM structure.
             WKB_QSY_STATUS (0x10) returns the full status of the system
               status in an unspecified byte buffer which must be long enough
               to hold the complete information. This buffer can be used later
               to restore the subsystem via the function WkbAccess2 and the
               WKB_ACC_RESTORE option.
             WKB_QSY_IPADDR (0x20) returns a string containing the TCP/IP
             address of a WkLAN subsystem. For WkNet or local subsystems, the
             function returns an empty string.
  pvDest   points to a structure of the WKBSYSTEM type or an unspecified byte
           buffer which receives the current state of the accessed subsystem.
  cbDest   contains the length of the destination buffer.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cbDest> is too small to receive all returned
    information, the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cbDest> is large enough, the function returns the number
    of transferred bytes.

*****************************************************************************/

INT WKAPIENTRY WkbQuerySystem2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  void *pvDest, UINT cbDest)
{
  UINT cbLength = 0;
  INT iRet = 0;

  THREAD_API_LOCK


  // check system & specified control flags, return if error
  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  // convert handle address to true address if required
  if (cbDest != 0 && !WkdConvertMemHandle((const void**) &pvDest)) {
    // invalid specification => return with error
    goto RETURN;
  } // if
  {//// determine system data
    BOOL fWkLan = (cstWkd.fuSysType & STYPE_WKLAN) != 0;
    switch ((UINT)flCtrl & WKB_QSY_MASK) {
      case WKB_QSY_SYSTEM: {
        //// get information about subsystem
        BOOL fUpdate = (flCtrl & WKB_QSY_UPDATE) ? 1 : 0;
        WKBSYSTEM *pwkbsys = (WKBSYSTEM*)pvDest;
        cbLength = sizeof(WKBSYSTEM);
        // determine length of returned data structure
        if (cbDest >= sizeof(WKBSYSTEM)) {
          // setup fixed part of destination structure
          memset(pwkbsys, 0, sizeof(*pwkbsys));
          pwkbsys->flCtrl = cstWkd.psys->fuType &
            (WKB_ACC_LOCAL|WKB_ACC_WKLAN|WKB_ACC_WKNET);
          if (cstWkd.psys->fuFlags & SFLAG_NONETLIMIT) {
            pwkbsys->flCtrl |= WKB_ACC_NONETLIMIT;
          } // if
          if (cstWkd.psys->fuFlags & SFLAG_EXCLUSIVE) {
            pwkbsys->flCtrl |= WKB_ACC_EXCLUSIVE;
          } // if
          if (cstWkd.psys->fuFlags & SFLAG_STATIONSHARE) {
            pwkbsys->flCtrl |= WKB_ACC_STATIONSHARE;
          } // if
          {// set local subsystem parameters
            ASSERT(cstWkd.fuSysType & STYPE_LOCAL);
            pwkbsys->cMaxUser = 1;
            pwkbsys->cbMaxData = 65520;
          } // if
        } // if
        } // case
        break;
      case WKB_QSY_STATUS: {
        //// return complete subsystem status
        switch (cstWkd.psys->fuType) {
          case STYPE_LOCAL:
            // nothing to store
            cbLength = 0;
            break;
          default:
            ASSERT(0);
        } // if
        // add prolog and epilog values to destination data
        cbLength += sizeof(USHORT) + sizeof(USHORT);
        if (cbLength > cbDest) {
          cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
        } else {
          // generate data, protected by simple 16 bit checksum
          BYTE *pbData = NULL;
          UINT uSum = 0;
          UINT i = 0;
          // save netto length
          *(USHORT*)pvDest = (USHORT)(cbLength - 2 * sizeof(USHORT));
          // save data in dependence of subsystem
          switch (cstWkd.psys->fuType) {
            case STYPE_LOCAL:
              // nothing to store
              break;
            default:
              ASSERT(0);
          } // if
          // calculate simple check sum
          pbData = (BYTE*)pvDest + sizeof(USHORT);
          // validate received subsystem data
          for (i = 0; i < cbLength - 2 * sizeof(USHORT); i++) {
            uSum += *pbData++;
          } // for
          *(USHORT*)((BYTE*)pvDest + cbLength - sizeof(USHORT)) =
            (USHORT)uSum;
        } // if
        } // case
        break;
      case WKB_QSY_OPENENTRY: {
#if 0 //!P:as This implementation is not final !!!
        WKBOPENENTRY *pwkboe = (WKBOPENENTRY*) pvDest;
        cbLength = sizeof(WKBOPENENTRY);
        // validate the parameter flCtrl
        if (flCtrl & (WKB_QSY_DIALOG | WKB_QSY_UPDATE)) {
          // WKB_QSY_DIALOG & WKB_QSY_UPDATE are not allowed
          cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
          goto RETURN;
        } // if

        // determine length of returned data structure
        memset (pvDest, 0, cbDest);
        if (cbDest >= sizeof(WKBOPENENTRY)) {
          {//// execute driver operations
            OPENENTRY *pope = NULL;

            ASSERT(cstWkd.fuSysType & STYPE_LOCAL);
            // search a local entry handle which matches the given
            // system handle
            pope = WkdSearchHandle (hwkbsys, HTYP_ENTRY);
            if (pope == NULL) {
              cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
              goto RETURN;
            } // if

            // set the values from the internal structure
            // to the output structure
            pwkboe->flCtrl = pope->flCtrl;
            pwkboe->hwkbsys = pope->hwkbsys;
          #if !defined(WK32)
            pwkboe->usFill = pope->usFill;
          #endif
            pwkboe->lPort = pope->lPort;
            pwkboe->pszPort = pope->pszPort;
            pwkboe->wkbsr = pope->wkbsr;
            pwkboe->usVersion = pope->usBoxVersion;
            pwkboe->usEntry = pope->iEntry;
            pwkboe->ulFirmCode = pope->ulFirmCode;
            pwkboe->ulUserCode = pope->ulUserCode;
          } //if
        } else {
          cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
        } // if
#else
        // other command codes are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_FUNCTION;
#endif //!P:as end of not final implementation
        break;
      } // case
      default: {
        // other command codes are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
      } // case
    } // switch
  } // block

RETURN:

  iRet = WkdReturnLength(cbLength);
  THREAD_API_UNLOCK

  return iRet;
} // WkbQuerySystem2()


/*****************************************************************************
WkbSetCompletionEvent
==============================================================================

This function sets the completion event value for the specified subsystem. In
dependence of the used operation system, this function sets or clears the
state of this event if an function terminates which was called asynchronously
previously.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  hEvent   contains the handle of a valid event object. If the value is NULL,
           then no event is set.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occurred. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Comments
  The WibuKey driver supports for a single task/process only a single event.
  If the function is called multiply with different event handles, only the
  last set event is set if an asynchronous function terminates.

*****************************************************************************/

INT WKAPIENTRY WkbSetCompletionEvent(HWKBSYSTEM hwkbsys, HANDLE hEvent)
{
  gidWkdLastError = WKB_ERROR_INVALID_FUNCTION;
  return FALSE;
} // WkbSetCompletionEvent()


/*****************************************************************************
WkbGetCompletionEvent
==============================================================================

This function returns the handle of the last set completion event value
for the specified subsystem.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle
           value HWKB_LOCAL can be used.

Return
  The function returns the handle of the set event or NULL if no such event
  exists. In this case a subsequent call of the function WkbGetLastError()
  returns the corresonding error code.

Preliminary Note
  This function is currently not implemented.

*****************************************************************************/

HANDLE WKAPIENTRY WkbGetCompletionEvent(HWKBSYSTEM hwkbsys)
{
  // not implemented
  gidWkdLastError = WKB_ERROR_INVALID_FUNCTION;
  return NULL;
} // WkbGetCompletionEvent()


/*****************************************************************************
WkbSetCompletionFunction
==============================================================================

This function sets the completion callback function for the specified
subsystem.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  pfn .... contains the address of a valid completion callback function. If
           this value is NULL, then no function is set.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Comments
  The WibuKey driver supports for a single task/process only a single 
  completion function. If the function is called multiply with different
  completion functions, only the last set function is called if an
  asnychronous driver function terminates.

Preliminary Note
  This function is currently not implemented.

*****************************************************************************/

INT WKAPIENTRY WkbSetCompletionFunction(HWKBSYSTEM hwkbsys, WKBCOMPLFUNC *pfn)
{
  gidWkdLastError = WKB_ERROR_INVALID_FUNCTION;
  return FALSE;
} // WkbSetCompletionFunction()


/*****************************************************************************
WkbGetCompletionFunction
==============================================================================

This function returns the address of the last set completion callback function
for the specified subsystem.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  ppfn ... points to a variable which receives the address of the last set
           completion callback function. If no such event exists,
           NULL is stored.

Return
  The function returns the address of the set completion function or NULL
  if no such event exists. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

Preliminary Note
  This function is currently not implemented.

*****************************************************************************/

WKBCOMPLFUNC *WKAPIENTRY WkbGetCompletionFunction(HWKBSYSTEM hwkbsys)
{
  gidWkdLastError = WKB_ERROR_INVALID_FUNCTION;
  return NULL;
} // WkbGetCompletionFunction()


/*****************************************************************************
WkbIsComplete
==============================================================================

This function checks if the function with the specified action identification
has terminated.

Parameters
  nActionID  returns the action identification which was returned of a function
             which was called asynchronously.
  pidReturn  points to an integer or handle variable which receives the return
             value of the function which was executed and terminated
             asynchronously.

Return
  If the function which corresponds with the specified action identification
  has terminated, the WkbIsComplete function returns TRUE and the variable
  which is pointered by the second parameter receives the return value of the
  terminated function. This can be a status code, a handle or any other
  function dependent value. If this value corresponds with an error condition,
  the error code can be returned via the WkbGetErrorCode function.
  If the specified function has not terminated the WkbIsComplete function
  returns FALSE and a subsequent call of WkbGetErrorCode returns
  WKB_ERROR_IN_ACTION. If the specified action is undefined, the error code
  WKB_ERROR_UNDEFINED_ACTION is returned otherwise.

Comments
  For each asynchronous called function this function returns only once a
  completion value. A subsequent call results in the error value
  WKB_ERROR_UNDEFINED_ACTION.

*****************************************************************************/

BOOL WKAPIENTRY WkbIsComplete(INT nActionID, INT *pidReturn)
{

#if ASYNCMODE
  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pidReturn)) {
    goto RETURN;
  } // if
  // init default call status
  WkdResetStatus();
  // search specified action
  if (WkdFindCallerAction(nActionID)) {
    // action found => analyse if complete
    if (cstWkd.pasacThis->idState != ACT_COMPLETE) {
      // action is still running: set special error
      cstWkd.idError = WKB_ERROR_IN_ACTION;
    } else {
      // action has completed: get state & delete action
      if (cstWkd.pasacThis->idError != WKB_NO_ERROR) {
        // set error code of action for later WkbGetLastError() call
        gidWkdLastError = cstWkd.pasacThis->idError;
      } // if
      *pidReturn = cstWkd.pasacThis->idReturn;
      cstWkd.idError = WKB_NO_ERROR;
      if (cstWkd.pasacThis->hwkbsysDestroy != NULL) {
        // execute last step of delayed WkbRelease2: destroy subsystem
        // this call destroys also the current action object
        DestroySubSystem(cstWkd.pasacThis->hwkbsysDestroy);
      } else {
        // no subsystem removed: remove current action object
        WkdRemoveAction(cstWkd.pasacThis);
      } // if
      ASSERT(cstWkd.idError == WKB_NO_ERROR);
    } // if
  } // if
RETURN:
#else
  cstWkd.idError = WKB_ERROR_INVALID_FUNCTION;
#endif
  return WkdReturnBool();
} // WkbIsComplete()


/*****************************************************************************
WkbEnumServers2
==============================================================================

This function determines all available WibuKey WkLAN servers in the network
and writes the names of all these systems into the specified destination
string buffer. The names are separated by vertical bars ("|").

Parameters
  flCtrl   contains flags which control the operation of the function. Such
           flags is currently not used and this parameter must be set to 0.
  pszDest  points to a buffer which receives the string with all WkLAN
           server names. The stored string is terminated by a 0-character.
           Never more than <cchDest> characters are stored into the buffer.
           If <cchDest> is 0, this parameter is not used.
  cchDest  contains the length of the specified buffer.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cchDest> is too small to receive all returned
    information, the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cchDest> is large enough, the function returns the
    number of transferred characters without (!) the trailing 0.

*****************************************************************************/

INT WKAPIENTRY WkbEnumServers2(ULONG flCtrl, char *pszDest, UINT cchDest)
{
  UINT cchLength = 0;
  INT iRet = 0;


  THREAD_API_LOCK


  cstWkd.idError = WKB_ERROR_INVALID_FUNCTION;

  iRet = WkdReturnLength(cchLength);
  THREAD_API_UNLOCK

  return iRet;
} // WkbEnumServers2()


/*****************************************************************************
WkbControlSystem2
==============================================================================

This function controls the specified subsystem.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem. To
           address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.

  flCtrl   contains flags which controls the execution of the function.
           Currently following command flags are defined:
             WKB_CTS_RESET ....... (0x00) reset some low-level specific parts
                                   of the specified system. The exact reset
                                   operation is specified in a bit map in the
                                   <cbCtrl> parameter which is constructed by
                                   following values:

               WKB_RESET_SYSTEM .. (0x01) is reserved for future use and must
                                   be set to 0.
               WKB_RESET_ADDRESS   (0x02) rereads the addressing
                                   specifications in the driver setting
                                   (profile).
               WKB_RESET_TIMING .. (0x04) recalculates the complete WibuBox
                                   timing.
               WKB_RESET_BOXSCAN   (0x08) restarts the elementar scanning of
                                   all WIBU-BOXes when the next access
                                   function is started. This option is only
                                   needed for the Macintosh-ADB-variant of
                                   the driver.
               WKB_RESET_FULL .... (0xFE) is a mask which forces all current
                                   and future reset options.

             WKB_CTS_CANCEL_USERS  (0x10) resets the specified users in the
                                   WkNet or WkLAN subsystem.

             WKB_CTS_FREE_ACCESS   (0x20) releases temporarily the access of
                                   the specified subsystem to reaccess later
                                   by the WkbAccess2() function via the option
                                   WKB_ACC_RESTORE.

             WKB_CTS_KILL_ACTION   (0x30) kills the actions with the
                                   identification in <cbCtrl> of all actions
                                   of the specified subsystem.

  pvCtrl   points to the data structure which controls the execution of the
           function. For the WKB_CTS_CANCEL_USERS option, this is a array
           containing USHORT (16 bit) values. Two values in this array
           specify a pair of minimum and maximum user identifications in which
           range (including the boundaries) all users are freed in the WkNet
           file or in the WkLAN server structure. If a single user entry is
           specified at the end of the array, only this one user is cancelled.
           The length of the array in bytes is specified by the <cbCtrl>
           parameter. For the options WKB_CTS_RESET, WKB_CTS_FREE_ACCESS or
           WKB_CTS_KILL_ACTION this parameter is currently not used and must
           be set to NULL.
  cbCtrl   contains the length of the data structure which is specified via
           the <pvCtrl> pointer if this pointer is not NULL. If this pointer
           is NULL and the WKB_CTS_RESET option is specified, the <cbCtrl>
           parameter contains the specified WKB_RESET_... flags; if the
           WKB_CTS_KILL_ACTION is specified this parameter contains the
           identification of the action which should be killed or the special
           value WKB_ALL_ACTIONS if all actions of the subsystem should be
           cancelled.

Return
  The function returns TRUE if the operation was sucessful and FALSE if an
  error occured. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbControlSystem2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  void *pvCtrl, UINT cbCtrl)
{
  UINT uFuncCmd = 0;
  INT iRet = 0;


  THREAD_API_LOCK

  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto EXITFN_RET;
  } // if
  // convert handle address to true address if required
  if (!WkdConvertMemHandle((const void**) &pvCtrl)) {
    goto EXITFN_RET;
  } // if
  uFuncCmd = (UINT)flCtrl & WKB_CTS_MASK;
  // check parameters
  switch (uFuncCmd) {
    case WKB_CTS_RESET:
    case WKB_CTS_FREE_ACCESS:
    case WKB_CTS_KILL_ACTION:
      if (pvCtrl != NULL) {
        cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
        goto EXITFN_RET;
      } // if
      break;
    case WKB_CTS_CANCEL_USERS:
      if ((cbCtrl & 1) || cbCtrl > 4096) {
        // invalid length of specified array => return with error
        cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
        goto EXITFN_RET;
      } // if
      break;
    default:
      // other commands not permitted
      cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
      goto EXITFN_RET;
  } // switch
#if ASYNCMODE
  // execute subsystem-type independent subcommands
  switch (uFuncCmd) {
    case WKB_CTS_KILL_ACTION: {
      // kill (all) action(s)
      UINT nActionID = cbCtrl;
      if (cstWkd.fAsyncCall && nActionID == WKB_ALL_ACTIONS) {
        // asynchronous call not permitted here
        cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
        goto EXITFN_RET;
      } // if
      if (nActionID == WKB_ALL_ACTIONS) {
        // kill all actions of current subsystem
        WkdRemoveAllActionsOfSubSystem(cstWkd.psys);
      } else {
        ASYNCACTION *pasac = WkdFindAction(nActionID);
        if (pasac == NULL) {
          // action not found => return with error
          goto EXITFN_RET;
        } // if
        // remove found action
        WkdRemoveAction(pasac);
      } // if
      // all done => return
      goto EXITFN_RET;
    } // case
    default:
      // all other commands: not executed here
      break;
  } // switch
#endif // if ASYNCMODE
  {//// execute driver operations
    switch (uFuncCmd) {
      case WKB_CTS_RESET: {
        // reset specified subsystem options
        if (pvCtrl != NULL) {
          // no control data may be specified
          cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
          goto EXITFN_RET;
        } // if
        WkdResetDriver(cbCtrl);
        // end of operation: return without error
        break;
      } // case
      case WKB_CTS_CANCEL_USERS:
        //!P
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
      case WKB_CTS_FREE_ACCESS:
        //!P
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
      case WKB_CTS_KILL_ACTION:
        //!P
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
      default:
        // other functions are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
        break;
    } // switch
  }
EXITFN_RET:
  iRet = WkdReturnBool();
  THREAD_API_UNLOCK

  return iRet;
} // WkbControlSystem2()


/*****************************************************************************
WkbGetVersion3
==============================================================================

This function determines the version numbers of the WibuBox driver of the
specified subsystem and stores them into the specified variable.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To determine the version of the standard local subsystem the
           predefined handle value HWKB_LOCAL can be used. To determine
           the version of the driver part which contains the code of this
           function, NULL must be specified. This value differs from the
           HWBK_LOCAL specification only if the kernel of the driver is
           dynamically linked (under Windows NT or DOS with WKDOS.EXE).

Return
  The function returns the number of the version in the minor byte (bit 0..7)
  and the lower number in the next byte (bit 8..15). Under the WK32-API, the
  bits 16..31 of the value are set all to 0. If the function returns 0, an
  error occurred. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbGetVersion3 (HWKBSYSTEM hwkbsys)
{
  INT idResult = 0;

  THREAD_API_LOCK

  // call internal GetVersion() without printing error message
  idResult = WkdDrvGetVersion (hwkbsys);
  {// calculate the right format ==>> 3.x instead of 2.(60+x) etc.
    INT nVer = LOBYTE(idResult) + (HIBYTE(idResult) / 60);
    INT nSubVer = HIBYTE(idResult) % 60;
    idResult = (nVer << 8) | nSubVer;
  } // block

  THREAD_API_UNLOCK

  return idResult;
} // WkbGetVersion3()


/*****************************************************************************
WkbGetVersion2
==============================================================================

This function determines the version numbers of the WibuBox driver of the
specified subsystem and stores them into the specified variable.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To determine the version of the standard local subsystem the
           predefined handle value HWKB_LOCAL can be used. To determine
           the version of the driver part which contains the code of this
           function, NULL must be specified. This value differs from the
           HWBK_LOCAL specification only if the kernel of the driver is
           dynamically linked (under Windows NT or DOS with WKDOS.EXE).

Return
  The function returns the number of the version in the lower byte (bit 0..7)
  and the minor number in the next byte (bit 8..15). Under the WK32-API, the
  bits 16..31 of the value are set all to 0. If the function returns 0, an
  error occurred. In this case a subsequent call of the function
  WkbGetLastError() returns the corresponding error code.

*****************************************************************************/

INT WKAPIENTRY WkbGetVersion2 (HWKBSYSTEM hwkbsys)
{
  INT idResult = 0;
  THREAD_API_LOCK

  // call internal GetVersion() without printing error message
  idResult = WkdDrvGetVersion (hwkbsys);

  THREAD_API_UNLOCK

  return idResult;
} // WkbGetVersion2()


/*****************************************************************************
WkbQueryStatus2
==============================================================================

This function returns the status of the specified subsystem.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  flctrl   contains flags which controls the execution of the function.
           This parameter is currently not used and should be set to 0.
  pwkbst   points to a destination buffer which receives the status data.
           For this version of the function the structure WKBSTATUS is used.
  cbDest   contains the length of the destination buffer in bytes. This should
           be the length of the WKBSTATUS structure.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cbDest> is too small to receive all returned information,
    the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cbDest> is large enough, the function returns the number
    of transferred bytes.

*****************************************************************************/

INT WKAPIENTRY WkbQueryStatus2(HWKBSYSTEM hwkbsys, ULONG flCtrl,
  WKBSTATUS *pwkbst, UINT cbDest)
{
  UINT cbLength = 0;
  UINT uFlags = 0;
  INT iRet = 0;


  THREAD_API_LOCK

  if (!WkdGetSystem(hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  // add constants of supported features
#if ASYNCMODE
  uFlags |= WKB_ST_ASYNCHRONOUS;
#endif
  uFlags |= WKB_ST_AREACRYPT;
  uFlags |= WKB_ST_EXPDATE;
  {// determine flags of supported subsystems
   // * read contents of SubSystem driver setting variable if not already done
    INT i = 0;
    ReadSubSysProfile();
    for (i = 0; abSubSysOrder[i] != 0; i++) {
      switch (abSubSysOrder[i]) {
        case WKB_ACC_LOCAL:
          // local is always supported
          break;
        case WKB_ACC_WKLAN:
          // acccept not if no WkLAN supported
          goto EXITLB;
          break;
        case WKB_ACC_WKNET:
          // acccept not if no WkNet supported
          goto EXITLB;
          break;
        default:
          ASSERT(0);
          break;
      } // switch
      // move specified bits to correct position
      uFlags |= ((USHORT)abSubSysOrder[i] * (WKB_ST_WKNET/WKB_ACC_WKNET));
    EXITLB:;
    } // for
  } // for
  {//// execute driver operations
    switch ((UINT)flCtrl & WKB_QST_MASK) {
      case WKB_QST_DRIVER: {
        //// get information about subsystem
        cbLength = sizeof(WKBSTATUS);
        // determine length of returned data structure
        if (cbDest >= sizeof(WKBSTATUS)) {
          // setup fixed part of destination structure
          memset(pwkbst, 0, sizeof(*pwkbst));
        #if WIN16DLL
          if (gfKrnlCall) {
            // calling of kernel driver => set flag
            uFlags |= WKB_ST_KRNL;
          } // if
        #elif WINKRNL
          // VXD and NT kernel driver: set always
          uFlags |= WKB_ST_KRNL;
        #endif
        } else {
          cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
        } // if
        // store determined flags into structure
        pwkbst->usFlags = (USHORT)uFlags;
        } // case
        break;
      default:
        // other commands are not implemented
        cstWkd.idError = WKB_ERROR_INVALID_COMMAND;
    } // switch
  }
RETURN:

  iRet = WkdReturnLength((INT)cbLength);
  THREAD_API_UNLOCK

  return iRet;
} // WkbQueryStatus2()


/*****************************************************************************
ConvertVersionToString
==============================================================================

This fuction convert teh version of a diver or a server to a string.

Parameters
  flCtrl .... contains flags which controls the execution of the function.
              This parameter is currently not used and should be set to 0.
  pwkbver ... contain one or more version in a array of the struct WKBVERSION.
              If two versions should be converted, the first version in the 
              array is always a driver version and the second is a server 
              version. More than two version conversion are not supplied.
  pszVersion  gives the converted string back to the caller.
  cbVersion   number of reserved bytes in pszVersion.
Return
  The function returns the number of characters containing pszVersion.

*****************************************************************************/

static UINT ConvertVersionToString (HWKBSYSTEM hwkbsys, ULONG flCtrl,
    WKBVERSION *pwkbver, char *pszVersion, UINT cbVersion)
{
  char szFullText[200];
  UINT cbLength = 0;
  UINT i = 0;

  ASSERT((flCtrl & WKB_QI_VERTEXT)
      || ((flCtrl & WKB_QI_DIALOG) && (flCtrl & WKB_QI_VERVALUE)));
  ASSERT(pwkbver != NULL);
  ASSERT(pszVersion != NULL);

  //==========================================
  // create the text for the driver version
  //==========================================
  if (((flCtrl & WKB_QI_DRIVER) == WKB_QI_DRIVER)
      || ((flCtrl & (WKB_QI_DRIVER | WKB_QI_SERVER)) == 0)) {
    // The version of the server is in the 1st Element of
    // the WKBVERSION array.
    cbLength = SetVersionToString(0, &pwkbver[0], szFullText, 200, TRUE);
    i++;
  } // if

  //==========================================
  // create the text for the server version
  //==========================================
  if (((flCtrl & WKB_QI_SERVER) == WKB_QI_SERVER)
    && (hwkbsys != HWKB_LOCAL)
    && !(pwkbver[1].usVersion == 0 && pwkbver[1].usSubVersion == 0)) {
    char *pszBuffer = NULL;                                                      

    // check if a driver version string is set before to insert a seperator
    if ((flCtrl & WKB_QI_DRIVER)) { 
      // insert a seperator, if the driver version already inserted
      if (i == 1) {
        szFullText[cbLength++] = '|';
        szFullText[cbLength] = 0;
        pszBuffer = szFullText + cbLength;
      } // if
    } else {                 
      pszBuffer = szFullText;
    } // if

    // The version of the server is in the 2nd Element of
    // the WKBVERSION array.
    cbLength += SetVersionToString (0, &pwkbver[1],
        pszBuffer, (200 - cbLength), TRUE);
  } // if
  szFullText[cbLength] = 0;   // set end of line

  // copy the version text back into the parameter
  if (cbVersion >= cbLength) {
    // everthing is all right
    memcpy (pszVersion, szFullText, (size_t)(cbLength + 1));
  } // if

  return cbLength;
} // ConvertVersionToString()


// local resource storage for all error strings in english
 // ==>> for none Windows Operating Systems!!!
#include "errorcode.h"

/*****************************************************************************
WkbQueryInfo2
==============================================================================

This function returns some information of the current driver like
the current error text or the current version.

Parameters
  hwkbsys  contains the handle of a valid accessed WibuKey subsystem.
           To address the standard local subsystem the predefined handle value
           HWKB_LOCAL can be used.
  flCtrl   contains flags which controls the execution of the function.
           This parameter is currently not used and should be set to 0.
  pvDest   points to a destination buffer which receives the information data.
           If WKB_QI_xxxTEXT is set, this value points to a charcter string.
           If WKB_OI_VERVALUE is set, this value points to a 
           WKBVERSION-structure.
  cbDest   contains the length of the destination buffer in bytes. This should
           be the length of the WKBSTATUS structure.

Return
  The function returns one of the following values:
  * 0 if any error occured. In this case a subsequent call of the function
    WkbGetLastError() returns the corresponding error code.
  * If the length in <cbDest> is too small to receive all returned information,
    the needed buffer length is returned. A subsequent call of
    WkbGetLastError() returns the code WKB_BUFFER_OVERFLOW.
  * If the length in <cbDest> is large enough, the function returns the number
    of transferred bytes.

*****************************************************************************/

INT WKAPIENTRY WkbQueryInfo2 (HWKBSYSTEM hwkbsys, ULONG flCtrl,
  void *pvDest, UINT cbDest)
{
  UINT cbData = 0;
  BOOL8 fIsServer = TRUE;
  void *pvBuffer = NULL;
  WKBVERSION wkbver[2];
  char szBuffer[200];
  ULONG flRaw = flCtrl & WKB_QI_MASK;
  INT iRet = 0;
  
  THREAD_API_LOCK

  // reset internal error code
  cstWkd.idError = WKB_NO_ERROR;
#if _UNIX
  // initialize the internal variables
  memset (szBuffer, 0, sizeof(szBuffer) );
  memset (wkbver, 0, sizeof(wkbver));
  // handle the local calling driver version query
  if (hwkbsys == NULL) {
    // give version of Calling Driver
    if (!(flCtrl & (WKB_QI_VERVALUE | WKB_QI_VERTEXT)) || (flCtrl & WKB_QI_SERVER)) {
      cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
      goto RETURN;
    } // if

    // copy the current version in the struct WKBVERSION (first element)
    wkbver->usVersion = DRV_VERSION;
    wkbver->usSubVersion = DRV_SUBVERSION;
    wkbver->usBuild = DRV_VER_BUILD;
    wkbver->usCount = DRV_VER_COUNT;
    wkbver->usYear = DRV_VER_YEAR4;
    wkbver->bMonth = DRV_VER_MONTH;
    wkbver->bDay = DRV_VER_DAY;
    wkbver->ulReserved1 = 0;
    if (flCtrl & WKB_QI_VERTEXT) {
      // select the current language of the operation system
      // give version as text
      cbData = ConvertVersionToString (HWKB_LOCAL, flCtrl, wkbver, szBuffer, 200);
      szBuffer[cbData] = 0; // temination byte
      pvBuffer = szBuffer;
    } else {                 
      // give version in WKBVERSION struct
      cbData = sizeof(WKBVERSION);
      pvBuffer = wkbver;
    } // if

    goto RETURN_SETVAL;
  } // if
#endif // #if _UNIX

  // check if specified system is valid, initialize <cstWkd>
  if (!WkdGetSystem (hwkbsys, flCtrl)) {
    goto RETURN;
  } // if
  // convert handle address to true address if required
  if (cbDest != 0 && !WkdConvertMemHandle((const void**) &pvDest)) {
    // invalid specification => return with error
    goto RETURN;
  } // if
  // analyse & store Unicode flag
  cstWkd.fUnicodeCall = ((flCtrl & WKB_UNICODE) != 0);

  {//// execute driver operations
    ASSERT(cstWkd.fuSysType & STYPE_LOCAL);
    if (cstWkd.fuSysType & STYPE_LOCAL) {
      // execute local driver operations
      if ((flCtrl & (WKB_QI_SERVER | WKB_QI_DRIVER)) == WKB_QI_SERVER) {
        // no local version request wanted
        cbData = 0;
        goto RETURN;
      } // if

      if (flRaw & (WKB_QI_VERVALUE | WKB_QI_VERTEXT)) {
        // >>> version wanted <<<
        pvBuffer = wkbver;
        cbData = sizeof(WKBVERSION);

        // copy the current version in the struct WKBVERSION
        wkbver[0].usVersion = DRV_VERSION;
        wkbver[0].usSubVersion = DRV_SUBVERSION;
        wkbver[0].usBuild = DRV_VER_BUILD;
        wkbver[0].usCount = DRV_VER_COUNT;
        wkbver[0].usYear = DRV_VER_YEAR4;
        wkbver[0].bMonth = DRV_VER_MONTH;
        wkbver[0].bDay = DRV_VER_DAY;
        wkbver[0].ulReserved1 = 0;
      } // if
    } // if
  } //if

  // select the current language of the operation system

  if (flRaw == WKB_QI_ERRTEXT) {
    char *pszBuf = NULL;
    // char *szLang = NULL;
    // char pszBuf2[256];
    INT nLastError = 0;

      nLastError = gidWkdLastError;

    // definition - all error-strings are UTF-8 encoded
    if (nLastError < WKB_NO_ERROR || nLastError > WKB_ERROR_KRNL_COMMUNICATION) {
      // no such a error code exists
      cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
      goto RETURN;
    } // if

    pszBuf = gszErrorTextStrings_us[nLastError];
    cbData = strlen(pszBuf);

    if (cbData == 0) {
      // no such a error message exists
      cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
      goto RETURN;
    } // if
    memcpy (szBuffer, pszBuf, cbData);
    szBuffer[cbData] = 0; // temination byte
    pvBuffer = szBuffer;
  } else if ((flRaw & WKB_QI_VERTEXT)
    || ((flCtrl & WKB_QI_DIALOG) && (flRaw & WKB_QI_VERVALUE))) {
    if (fIsServer) {
      // convert the version to a text
      cbData = ConvertVersionToString (hwkbsys, flCtrl, wkbver, szBuffer, 200);
    } // if
    if (cbData == 0) {
      // no such a error message exists
      cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
      goto RETURN;
    } // if
    pvBuffer = szBuffer;
  } // if

#if _UNIX
RETURN_SETVAL:
#endif

  // general post work
  // copy the error text in the output buffer
  if ((pvDest != NULL) && (cbDest >= (cbData)) 
    && ((flRaw == WKB_QI_ERRTEXT) || (flRaw & WKB_QI_VERTEXT))) {
    // copy string to destination, convert to Unicode if required
    WkdSetApiString (pvDest, pvBuffer, cbData, cstWkd.fUnicodeCall);
  } else if (pvDest != NULL && cbDest >= cbData) {
    // copy the data back to the caller
    ASSERT(pvBuffer != NULL);
    cstWkd.idError = WKB_NO_ERROR;
    memcpy(pvDest, pvBuffer, (size_t)cbData);
  } else {
    if (cbDest == 0) {
      cstWkd.idError = WKB_NO_ERROR;
      if (flCtrl & (WKB_QI_ERRTEXT | WKB_QI_VERTEXT)) {
        // destination too short: return required length including 0-term.
        cbData++;
      } // if
    } else {
      cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
    } // if
  } // if

RETURN:
  iRet = WkdReturnLength((UINT)cbData);

  THREAD_API_UNLOCK

  return iRet;
} // WkbQueryInfo2()


// ====================================
// end of clsubs.c
// ====================================
