/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
getfcuc.c                                       Firm Code/User Code Management
==============================================================================

This C source file implements the WibuKey API functions WkbGetFC() and
WkbGetUC() and contains the signature to set the Firm Code and the User Code
by the WKCRYPT application.

This module is no part of the dynamic drivers WKWIN.DLL, WKWIN32.DLL,
WIBUKEY.SYS or WIBUKEY.VXD but is part of the API linking libraries
WKWIN.LIB and WKWIN32.LIB.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/c
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
93Jan31 1.33 mabu: separated from KERNEL.C
93Aug28 2.00 mabu: adaptation to Large memory model
94Oct11 2.00 mabu: bug in signature data removed
95Aug01 2.06 mabu: bug in reading of Firm And User Code removed
95Aug01 2.06 mabu: BSR_NODEFLIB specified
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

// ======================================================================
// Signature for setting Firm and User Code, initialized by (WK)CRYPT.EXE
// ======================================================================

static struct {
  BYTE azCRight[30];
  BYTE abSign[5];
  BYTE fPlain;
  BYTE abFC[3];
  BYTE abUC[3];
  BYTE bVersion;
} FcUcSign =
{ "(C) 1989-2010 by WIBU-SYSTEMS", // abCRight
  {11, 72, 29, 15, 237}, // abSign
  145,                   // fPlain
  {';', 'F', 'C'},       // abFC
  {':', 'U', 'C'},       // abUC
  'A'                    // bVersion
};


/*****************************************************************************
WkbGetFC
==============================================================================

This function returns the Firm Code of the current program. This code is
stored into the program specific WibuKey signature via the (WK)CRYPT.EXE
program by using the /CX option.

Parameters
  plFirmCode  points to a LONG variable which receives the Firm Code.
  ulReserved  is currently not used and should be set to 0.

Return
  The function returns a WibuKey error code which can be
  one of the following values:
     0  WKB_NO_ERROR
    10  WKB_ERROR_NOT_CRYPTIC

*****************************************************************************/

INT WKAPIENTRY WkbGetFC(LONG *plFirmCode, ULONG ulReserved)
{
  if (FcUcSign.fPlain != 0) {
     // no Firm Code set
     return WKB_ERROR_NOT_CRYPTIC;
  } // if
  // determine & save set Firm Code
  *plFirmCode = (FcUcSign.abFC[2] * 256L + FcUcSign.abFC[1]) * 256L
                + FcUcSign.abFC[0];
  // return without error
  return WKB_NO_ERROR;
} // WkbGetFC()


/*****************************************************************************
WkbGetUC
==============================================================================

This function returns the User Code of the current program. This code is
stored into the program specific WibuKey signature via the (WK)CRYPT.EXE
program by using the /CX option.

Parameters
  plUserCode  points to a LONG variable which receives the Firm Code.
  ulReserved  is currently not used and should be set to 0.

Return
  The function returns a WibuKey error code which can be
  one of the following values:
     0  WKB_NO_ERROR
    10  WKB_ERROR_NOT_CRYPTIC

*****************************************************************************/

INT WKAPIENTRY WkbGetUC(LONG *plUserCode, ULONG ulReserved)
{
  if (FcUcSign.fPlain != 0) {
     // no User Code set
     return WKB_ERROR_NOT_CRYPTIC;
  } // if
  // determine & save set User Code
  *plUserCode = (FcUcSign.abUC[2] * 256L + FcUcSign.abUC[1]) * 256L
                + FcUcSign.abUC[0];
  // return without error
  return WKB_NO_ERROR;
} // WkbGetUC()


// =====================================
// end of getfcuc.c
// =====================================
