/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
kecommon.c:                                   Hardware Access Common Functions
==============================================================================

This C source file implements driver internal functions which are called from
the specific low level drivers for the different supported interface and
functions to management the handling in the low level part of the driver.

==============================================================================

The portable WibuKey driver source is used under DOS and Windows to build
the internally used driver WKDOS2.LIB and the shipped drivers WKWIN.DLL,
WKWIN32.DLL, WIBUKEY.SYS and WIBUKEY.VXD. It is also used to build the
Macintosh and Netware NLM drivers.

Furthermore, it is also the origin of the SRCDRV project. All sources are
written in ANSI C, except the using of C++ style single line comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-0-721-93172-0, Fax: +49-0-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
96Jan29 2.10 mabu: initial version, built from CTRL.C, NTINIT.C and NTDISP.C
96Apr09 2.11 mabu: toupper() -> WkdToUpper to avoid including MS-C ctype array
96May07 2.11 mabu: optional disabling of critical serial number checking
96Jun24 2.13 mabu: renamed from LDCOMMON.C
97Feb15 2.25 mabu: unpacking of expiration dates
97Feb26 2.30 mabu: collecting of partial WIBU-BOX contents
98Jan12 2.41 mabu: support for PCCARD by WIBUKEY.SYS/NT (Award CardWare)
98Mar16 2.41 mabu: begin of adaptation to new driver class model
98May04 2.41 mabu: separation of sequential device code to KESEQ.C
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

/*****************************************************************************

---- Access to Platform Specific Driver Part ---------------------------------

*****************************************************************************/

// ======================================
// Table of supported WibuKey port types
// ======================================

// standard driver supports LPT and COM
PORTGROUP * apprtgrpWkd[] = {
  &prtgrpWkdSrl
#if USE_USB
  ,
  &prtgrpWkdUsb
#endif
};

// Store size of port table into global variable.
INT gcWkdPortGroup = DIMOF(apprtgrpWkd);


/*****************************************************************************
W k d S e l e c t P o r t G r o u p
==============================================================================

This function accesses and sets the class driver and the pointer of the group
of the specified port but does not access the port itself.

Parameters
  nPort  contains the number of the port which is accessed.

Return
  The function returns TRUE if the port group of the port is available and
  FALSE if not.

Comments
* The function does not checks if the specified port is really available.
* The function does not access the specified port.

*****************************************************************************/

BOOL WkdSelectPortGroup(INT nPort)
{
  if (cstWkd.iSelPortGroup != -1) {
    // already a port group selected: check if this can be used for new port
    ASSERT_VALUE(cstWkd.pprtgrpSel);
    if (nPort >= cstWkd.pprtgrpSel->nFirstPort &&
        nPort <= cstWkd.pprtgrpSel->nLastPort
      ) {
      // port match last select port group: nothing to do
      return TRUE;
    } // if
  } // if
  // no longer any port selected
  // * required because <nSelPort> and <iSelPortGroup> depend vice versa
  cstWkd.nSelPort = -1;
  {// specified port needs other port group: search sequential in list
    INT iPortGroup;
  #if MULTIMACH
    // check machine configuration, switch driver addresses if required
    WkdCheckMachine();
  #endif
    for (iPortGroup = 0; iPortGroup < gcWkdPortGroup; iPortGroup++) {
      PORTGROUP *pprtgrp = apprtgrpWkd[iPortGroup];
    #if MULTIMACH
      if (pprtgrp == NULL) {
        // port class driver inactivated: check next
        continue;
      } // if
    #endif
      if (nPort >= pprtgrp->nFirstPort && nPort <= pprtgrp->nLastPort) {
        // port group found: select it
        cstWkd.iSelPortGroup = iPortGroup;
        cstWkd.pprtgrpSel = pprtgrp;
        cstWkd.pclsdrvSel = pprtgrp->pclsdrv;
        return TRUE;
      } // if
      // check next port
    } // for
    // port group not found => port is invalid, return with error
    // * no longer any port group selected
    cstWkd.iSelPortGroup = -1;
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
  } // block
  return FALSE;
} // WkdSelectPortGroup()


/*****************************************************************************
W k d S e l e c t P o r t
==============================================================================

This is the main entry function to select or access a specific port. It sets
all port specific information in <cstWkd> and calls for an access also the
port class specific function but does not access the port physically.

Parameters
  nPort ..... contains the number of the port which is selected or accessed.
  fForAccess  contains the flag "prepare the selected driver for access".
              Usually this flag is set except the port is selected for a
              reset operation. This permits also to select ports which are
              dynamically invalidated, for example by the system profile.

Globals
  cstWkd*

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

Comments
* If the port number is identical with the current selected port,
  this function does not a re-selection but terminates really fast.
* If the last selected port is accessed it is released before another
  port is selected.

*****************************************************************************/

BOOL WkdSelectPort(INT nPort, BOOL fForAccess)
{
  if (cstWkd.fAccessed) {
    // any port is accessed => release it before other port is selected
    WkdReleasePort();
  } // if
  if (nPort != cstWkd.nSelPort) {
    // selected port invalid or must be changed
    // * change also selected port group if required
    if (!WkdSelectPortGroup(nPort)) {
      // port group does not exist => return with error
      cstWkd.idError =  WKB_ERROR_INVALID_SELECTION;
      return FALSE;
    } // if
    // select specified port now within port class driver
    // * this port is now selected
    cstWkd.nSelPort = nPort;
    // * this function checks also if <nPort> is a valid port
    ASSERT(cstWkd.iSelPortGroup >= 0);
    ASSERT_VALUE(cstWkd.pclsdrvSel);
    ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvSelect);
    if (fForAccess && !cstWkd.pclsdrvSel->pfClDrvSelect(nPort)) {
      // specific port cannot be selected => return with error
      // * <nSelPort> and <iSelPortGroup> no longer valid
      cstWkd.nSelPort = -1;
      cstWkd.iSelPortGroup = -1;
      return FALSE;
    } // if
  } // if
  // check selected port
  ASSERT(cstWkd.nSelPort >= cstWkd.pprtgrpSel->nFirstPort && 
    cstWkd.nSelPort <= cstWkd.pprtgrpSel->nLastPort);
  ASSERT_VALUE(cstWkd.pclsdrvSel);
  if (fForAccess) {
    // set flag that port is accessed
    cstWkd.fAccessed = TRUE;
  } // if
  // return without error
  return TRUE;
} // WkdSelectPort()


/*****************************************************************************
W k d R e l e a s e P o r t
==============================================================================

This function terminates the access of a port for the WibuBox addressing.

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code except it is
  already filled with another error from an earlier operation.

*****************************************************************************/

BOOL WkdReleasePort(void)
{
  INT idError = 0;
  if (!cstWkd.fAccessed) {
    // not accessed => nothing to do
    return TRUE;
  } // if
  // release selected device & save error code of release
  idError = cstWkd.pclsdrvSel->pfClDrvRelease();
  // port no longer accessed
  cstWkd.fAccessed = FALSE;
  // analyse error & return state
  if (idError == WKB_NO_ERROR) {
    // proper termination of device release => return without error
    return TRUE;
  } // if
  if (cstWkd.idError == WKB_NO_ERROR) {
    // any error and no error set => save new error
    cstWkd.idError = idError;
  } // if
  // return with error
  return FALSE;
} // WkdReleasePort()


/*****************************************************************************
W k d G e t P o r t N a m e
==============================================================================

This function determines the name of the specified port number and writes
it into the specified buffer.

Parameters
  pszPort   points to a buffer which gets the specified name. This buffer
            should be long enough to get 20 or more characters.
  pcch .... points to an UINT variable which gets the number of characters in
            the name (without the trailing 0 terminator).

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

BOOL WkdGetPortName(char *pszPort, UINT cbLen, UINT *pcch)
{
  UINT cch = 0;
  // no port group selected => name cannot be determined
  if (cstWkd.iSelPortGroup < 0) {
    return FALSE;
  } // if
  ASSERT_VALUE(cstWkd.pprtgrpSel);
  // create & copy string into destination
  cch = (UINT)strlen(cstWkd.pprtgrpSel->pszName);
  if ( cch > cbLen ) {
    cch = cbLen;
    ASSERT(0);
  }
  // get base part of port name
  memcpy(pszPort, cstWkd.pprtgrpSel->pszName, (size_t)cch);
  if (cstWkd.pprtgrpSel->nFirstPort != cstWkd.pprtgrpSel->nLastPort) {
    // add index value of port
    pszPort[cch++] = (char)(cstWkd.nSelPort -
      cstWkd.pprtgrpSel->nFirstPort + '1');
  } // if
  // set end of string
  pszPort[cch] = 0;
  // save number of stored characters
  *pcch = cch;
  return TRUE;
} // WkdGetPortName()


/*****************************************************************************
W k d C o n v e r t S t a t u s I n t o E r r o r
==============================================================================

This function converts the specified WibuBox status byte into an error code
if it reflects an error condition and stores it into <cstWkd.idError>.

Parameters
  uData  contains the WibuBox status code in the range 0..255. If the value
         is 128 or greater, an error is detected.

*****************************************************************************/

void WkdConvertStatusIntoError(UINT uData)
{
  if (uData >= 128) {
    // error status read from WibuBox: convert error code
    switch (uData) {
      case 255:
        // 255 looks like communication error
        cstWkd.idError = WKB_ERROR_NOT_READY;
        break;
      case 0x87:
        // invalid function: return other error
        cstWkd.idError = WKB_ERROR_OPTION_NOT_SUPPORTED;
        break;
      default:
        // convert error code into range 16..25
        cstWkd.idError = uData + WKB_ERROR_NOT_SELECTED - 128;
        break;
    } // switch
  } // if
} // WkdConvertStatusIntoError()


/*****************************************************************************
WkdResetDriver
==============================================================================

This function resets parts of the driver in dependence of the specified flags.

Parameters
  fuReset  contains the flags which control the reset flags. Following flags
           are defined:
             WKB_RESET_SYSTEM .. (0x01) reserved for future use.
             WKB_RESET_ADDRESS   (0x02) resets the driver's profiling caching
             WKB_RESET_TIMING .. (0x04) resets the driver's timing caching
             WKB_RESET_BOXSCAN   (0x08) resets the driver's WibuBox scanning
             WKB_RESET_BOXLOCKS  (0x10) frees all select-locked WIBU-BOXes

Return
  The function returns TRUE if the serial number is valid and FALSE if not.

*****************************************************************************/

void WkdResetDriver(UINT fuReset)
{
  // call reset routine of all available low level drivers
  if (fuReset & WKB_RESET_TIMING) {
    // reset determined system time constants
  #if !ABSTIME
    guWkdSysT = 0;
    //!P:as  gnWkdMinT = -1;
  #endif
  } // if
#if MULTIMACH
  if (fuReset & WKB_RESET_ADDRESS) {
    // determine machine type again
    gfWkdMachineChecked = FALSE;
  } // if
#endif // #if MULTIMACH

  if (fuReset & WKB_RESET_BOXLOCKS) {
    // close all opened WibuBox entries
    WkdFreeLockedBoxes();
  } // if
  // call the reset routine in the low level driver for all ports
  WkdResetPortEnumeration();
  // select next port (no preparation for access)
  while (WkdSelectNextPort(FALSE)) {
    // next port exist: reset it
    ASSERT_VALUE(cstWkd.pprtgrpSel);
    ASSERT_VALUE(cstWkd.pprtgrpSel->pclsdrv);
    ASSERT_VALUE(cstWkd.pprtgrpSel->pclsdrv->pfClDrvReset);
    // call reset routine of class driver
    (cstWkd.pprtgrpSel->pclsdrv->pfClDrvReset)(cstWkd.nSelPort, fuReset);
  } // while
} // WkdResetDriver()


/*****************************************************************************
WkdGetBoxVersion
==============================================================================

This function gets the BoxVersion from the box.

Parameters
  pwkbsr ... contains the serial number of the box.

Return
  The function returns the version of the box and 0 if an error occurred.

*****************************************************************************/

UINT32 WkdGetBoxVersion (const WKBSERIAL *pwkbsr)
{
  WKDEV_REQUEST wkdevreq;
  DEVOBJECT *pdevobj = NULL;
  BYTE bVersion = 0;


  // execute via class driver
  // * determine accessed device from serial number
  ASSERT_VALUE(cstWkd.pclsdrvSel);
#if !BSR16
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
#endif
  pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(pwkbsr);
  if (pdevobj == NULL) {
    // device not found => return with error
    return 0;
  } // if
  // read now determined number of bytes from device
#if !BSR16
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
#endif
  // setting request parameter
  wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
  wkdevreq.usCmd = BOXCMD_LIST;
  // control code for getting the version of the WIBU-BOX
  wkdevreq.usIndex = 0x02;
  wkdevreq.cbRequest = 0;
  // get only one 1 byte ==>> version byte
  wkdevreq.cbAnswer = 1;
  if (!cstWkd.pclsdrvSel->pfClDrvRequest(&wkdevreq, &bVersion, NULL)) {
    // request not properly executed => return with error
    return 0;
  } // if
  // return the WIBUI-BOX version
  return bVersion;
} // WkdGetBoxVersion()



/*****************************************************************************
WkdGetChipInfo
==============================================================================

This function gets the Chip Info from the box (Size of ExtMem & ASIC number).

Parameters
  pwkbsr .......... contains the serial number of the box.
  pusAsic ......... receives the ASIC info.
  pusExtendedMem .. recieves the size of the extended memory.

*****************************************************************************/

void WkdGetChipInfo (const WKBSERIAL *pwkbsr, 
    USHORT *pusAsic, USHORT *pusExtendedMem)
{
  // fixme-chang ha:original:  BYTE abAnswer[5];
  BYTE abAnswer[5];
  WKDEV_REQUEST wkdevreq;
  DEVOBJECT *pdevobj = NULL;
  memset(abAnswer, 0x0, sizeof(abAnswer) );

  // execute via class driver
  // * determine accessed device from serial number
  ASSERT_VALUE(cstWkd.pclsdrvSel);
#if !BSR16
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
#endif
  pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(pwkbsr);
  if (pdevobj == NULL) {
    // device not found => return with error
    // if the box does not support chip info ==>> no error will returned
    cstWkd.idError = WKB_NO_ERROR;
    return;
  } // if
  // WORKAROUND of a WORKAROUND
  // * Box/U:       6-5129000  .. 6-5169999
  // * Box/RU:      7-5053000  .. 7-5065999
  // * Box/U:       6-4960000  .. 6-4960500
  // * SecuriyKey:  25-2008900 .. 25-2009200
  if (((pwkbsr->usMask == 6) && ((pwkbsr->ulSerial >= 5129000) && (pwkbsr->ulSerial <= 5169999)))
    || ((pwkbsr->usMask == 7) && ((pwkbsr->ulSerial >= 5053000) && (pwkbsr->ulSerial <= 5065999)))
    || ((pwkbsr->usMask == 6) && ((pwkbsr->ulSerial >= 4960000) && (pwkbsr->ulSerial <= 4960500)))
    || ((pwkbsr->usMask == 25) && ((pwkbsr->ulSerial >= 2008900) && (pwkbsr->ulSerial <= 2009200)))
    ){
    *pusAsic = 0;
    *pusExtendedMem = 0;
    cstWkd.idError = WKB_NO_ERROR;
    return;
  } // if
  // read now determined number of bytes from device
#if !BSR16
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
#endif
  // setting request parameter
  wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
  wkdevreq.usCmd = BOXCMD_LIST;
  // control code for getting the version of the WibuBox
  wkdevreq.usIndex = 0x03;  // Cmd=06  Request=03 (GetChipInfo only Version>6)
  wkdevreq.cbRequest = 0;
  // get only one 5 byte ==>> chip info
  wkdevreq.cbAnswer = sizeof(abAnswer);
  if (cstWkd.pclsdrvSel->pfClDrvRequest(&wkdevreq, abAnswer, NULL)) {
    *pusAsic = GET_USHORT_VALUE(abAnswer + 1);
    *pusExtendedMem = GET_USHORT_VALUE(abAnswer + 1 + 2);
    // the box returns the size in a multiple of 8 KB
    // ==>> convert it to a multiple of 32 bytes
    // ==>> return the number of pages
    *pusExtendedMem <<= 8;
  } else {
    // error ==>> set parameter to 0
    *pusAsic = 0;
    *pusExtendedMem = 0;
  } // if
  // if the box does not support chip info ==>> no error will returned
  cstWkd.idError = WKB_NO_ERROR;
} // WkdGetChipInfo()


/*****************************************************************************
WkdCheckBoxVersion
==============================================================================

This function checks if the box has the wanted version of the driver.

The following configurations are supported at the moment:
  BoxVersion 1 supports Alg 1
  BoxVersion 2 supports Alg 2
  BoxVersion 3 supports Alg 1,2
  BoxVersion 4 supports Alg 1,2
  BoxVersion 5 supports Alg 3
  BoxVersion 6 supports Alg 1,2,4
  BoxVersion 7 supports Alg 3,5
The following assumption are done for future Box Versions (> 7):
  even numbers supports at least Alg 1,2,4
  odd numbers supports at least Alg 3,5

Parameters
  pflCtrl ....... contains the wanted version of the driver.
  usBoxVersion .. contains the box version.

Return
  The function returns TRUE if the version is valid and FALSE if not.

*****************************************************************************/

BOOL8 WkdCheckBoxVersion (ULONG *pflCtrl, USHORT usBoxVersion, WKBSERIAL *pwkbser)
{

  switch (*pflCtrl & WKB_VERSION_MASK) {
    case WKB_VERSION1:
      if (usBoxVersion == 1 || usBoxVersion == 3 || usBoxVersion == 4
          || (usBoxVersion >= 6 && ((usBoxVersion & 1) == 0))
        ) {
        return TRUE;
      } // if
      break;
    case WKB_VERSION2:
      if ((usBoxVersion >= 2 && usBoxVersion <= 4)
          || (usBoxVersion >= 6 && ((usBoxVersion & 1) == 0))
        ) {
        return TRUE;
      } // if
      break;
    case WKB_VERSION3:
      if (usBoxVersion == 5 || (usBoxVersion >= 7 && ((usBoxVersion & 1) == 1))) {

        return TRUE;
      } // if
      break;
    case WKB_VERSION4:
      if (usBoxVersion >= 6 && ((usBoxVersion & 1) == 0)) {

        return TRUE;
      } else if (usBoxVersion == 4) {
        // Box/4c (chip info exists (> 0)) also supports WKB_VERSION4
        USHORT usAsic = 0;
        USHORT usExtMem = 0;
        WkdGetChipInfo(pwkbser, &usAsic, &usExtMem);
        if (usAsic > 0) {
          return TRUE;
        } // if
      } // if
      break;
    case WKB_VERSION5:
      if (usBoxVersion >= 7 && ((usBoxVersion & 1) == 1)) {

        return TRUE;
      } else if (usBoxVersion == 5) {
        // Box/5a (chip info exists (> 0)) also supports WKB_VERSION5
        USHORT usAsic = 0;
        USHORT usExtMem = 0;
        WkdGetChipInfo(pwkbser, &usAsic, &usExtMem);
        if (usAsic > 0) {
          return TRUE;
        } // if
      } // if
      break;
    case WKB_VERSION_RED:
      // returns the best algorithm for the current box
      *pflCtrl &= ~WKB_VERSION_MASK;
      if (usBoxVersion >= 1 && usBoxVersion <= 7) {
        switch (usBoxVersion) {
          case 1:
            *pflCtrl |= WKB_VERSION1;
            break;
          case 2:
          case 3:
            *pflCtrl |= WKB_VERSION2;
            break;
          case 4: {
            USHORT usAsic = 0;
            USHORT usExtMem = 0;
            // Box/4c (chip info exists (> 0)) also supports WKB_VERSION4
            WkdGetChipInfo(pwkbser, &usAsic, &usExtMem);
            *pflCtrl |= (usAsic == 0) ? WKB_VERSION2 : WKB_VERSION4;
            } // case
            break;
          case 5: {
            USHORT usAsic = 0;
            USHORT usExtMem = 0;
            // Box/5a (chip info exists (> 0)) also supports WKB_VERSION5
            WkdGetChipInfo(pwkbser, &usAsic, &usExtMem);
            *pflCtrl |= (usAsic == 0) ? WKB_VERSION3 : WKB_VERSION5;
            } // case
            break;
          case 6:
           *pflCtrl |= WKB_VERSION4;
            break;
          case 7:
            *pflCtrl |= WKB_VERSION5;
            break;
        } // switch
      } else {
        // NOTE: This must be modified for future box versions!!! 
        if (usBoxVersion & 1) {
          // odd box Version ==>> use max. reduced algorithm
          *pflCtrl |= WKB_VERSION5;
        } else {
          // even box Version ==>> use max. standard algorithm
          *pflCtrl |= WKB_VERSION4;
        } // if
      } // if
      // all WIBU-BOX versions allowed
      return TRUE;
    case WKB_VERSION_STD:
      // returns the standard algorithm for the current box
      *pflCtrl &= ~WKB_VERSION_MASK;
      if (usBoxVersion >= 1 && usBoxVersion <= 7) {
        switch (usBoxVersion) {
          case 1:
			*pflCtrl |= WKB_VERSION1;
            break;
          case 2:
          case 3:
          case 4:
          case 6:
		    *pflCtrl |= WKB_VERSION2;
            break;
          case 5:
          case 7:
            *pflCtrl |= WKB_VERSION3;
            break;
          default:
            ASSERT(0);
            *pflCtrl |= WKB_VERSION2;
        } // switch
      } else {
        // NOTE: This must be modified for future box versions!!! 
        if (usBoxVersion & 1) {
          // odd box Version ==>> use default reduced algorithm
          *pflCtrl |= WKB_VERSION3;
        } else {
          // even box Version ==>> use default standard algorithm
          *pflCtrl |= WKB_VERSION2;
        } // if
      } // if
      // all WIBU-BOX versions allowed
      return TRUE;
  } // switch
  // not the right version ==>> return with error
  return FALSE;
} // WkdCheckBoxVersion()


/*****************************************************************************
C m p M e m N o C a s e
==============================================================================

This function compares the specified memory area without checking the
letter case.

Parameters
  psz1 ... points to the "left" character sequence.
  psz2 ... points to the "right" character sequence.
  cchArea  contains the number of characters which are compared.

Return
  The function returns TRUE if both memory area contains the same characters
  except a difference in upper/lower case and FALSE if both areas contains
  different characters.

*****************************************************************************/

static BOOL CmpMemNoCase(const char *psz1,
  const char *psz2, INT cchArea)
{
  while (cchArea-- > 0) {
    INT ch1 = WkdToUpper(*psz1);
    INT ch2 = WkdToUpper(*psz2);
    if (ch1 != ch2) {
      return FALSE;
    } // if
    psz1++;
    psz2++;
  } // while
  return TRUE;
} // CmpMemNoCase()


/*****************************************************************************
W k d P a r s e P o r t S p e c
==============================================================================

This function parses a port string, searches its port number and sets returns
this value. This function detects also wildcard specifications like
"LPT*" which are set to a port number with the least 3 significant bits set
all to 1.

Parameters
  pszPort ... points to a port specification string.
  pnPortDest  receives the port index of the parsed port name.

Return
  The function returns the string address of the next valid character after
  the specified string or NULL if the specified port is invalid.

Comments
  Leading and trailing space characters or tabulate characters are skipped.

*****************************************************************************/

const char * WkdParsePortSpec(const char *pszPort, INT *pnPortDest)
{
  INT iPortGroup = 0;
  while (*pszPort == ' ' || *pszPort == '\t') {
    pszPort++;
  } // if
  // search name body in table with low level drivers
  for (iPortGroup = 0; iPortGroup < gcWkdPortGroup; iPortGroup++) {
    PORTGROUP *pprtgrp = apprtgrpWkd[iPortGroup];
    INT cch;
  #if MULTIMACH
    if (pprtgrp == NULL) {
      // port class driver inactivated: check next
      continue;
    } // if
  #endif
    cch = (INT)strlen(pprtgrp->pszName);
    if (CmpMemNoCase(pprtgrp->pszName, pszPort, cch)) {
      INT chIndex = pszPort[cch];
      if (chIndex == '*') {
        // wildcard: set special offset 7
        *pnPortDest = pprtgrp->nFirstPort + 7;
      } else if (chIndex >= '1' && chIndex <= '7') {
        // digit, check range, set offset 0..6
        INT n = chIndex - '1';
        if (n > pprtgrp->nLastPort - pprtgrp->nFirstPort) {
          continue;
        } // if
        *pnPortDest = pprtgrp->nFirstPort + n;
      } else if (chIndex == 0 || chIndex == ' ' || chIndex == '\t' ||
          chIndex == '|'
        ) {
        // no index specified but valid termination: set to first port
        *pnPortDest = pprtgrp->nFirstPort;
        // one character less
        cch--;
      } else {
        // other follows: not permitted, continue search
        continue;
      } // if
      // driver found: port is valid, set selected driver & class
      cstWkd.pprtgrpSel = pprtgrp;
      cstWkd.pclsdrvSel = pprtgrp->pclsdrv;
      // skip trailing spaces
      pszPort += cch + 1;
      while (*pszPort == ' ' || *pszPort == '\t') {
        pszPort++;
      } // while
      // return position after parsed specification
      return pszPort;
    } // if
  } // for
  // end of search operation: return error condition
  return NULL;
} // WkdParsePortSpec()


/*****************************************************************************
W k d R e s e t P o r t E n u m e r a t i o n
==============================================================================

This function resets the port enumeration process. After calling this function
the first port may be returned by calling the WkdSelectNextPort() function.

*****************************************************************************/

void WkdResetPortEnumeration(void)
{
  ASSERT(gcWkdPortGroup != 0);
#if MULTIMACH
  // check machine configuration, switch driver addresses if required
  WkdCheckMachine();
#else
  ASSERT(gcWkdPortGroup != 0);
#endif
  // no port & no port group selected
  // * set index values to "pre-increment" values
  cstWkd.nSelPort = -1;
  cstWkd.iSelPortGroup = -1;
} // WkdResetPortEnumeration()


/*****************************************************************************
W k d S e l e c t N e x t P o r t
==============================================================================

This function selects and possibly accesses the next port which is available
in the list of available ports.

Parameters
  fForAccess  contains the flag "prepare the selected driver for access".
              Usually this flag is set except the port is selected for a
              reset operation. This permits also to select ports which are
              dynamically invalidated, for example by the system profile.

Return
  The function returns TRUE if a port is selected and FALSE if not. In the
  latter case, not further port is available.

Globals
  cstWkd.nSelPort

Comments
  This function assumes that the <iSelPortGroup> and <nSelPort> members of
  <cstWkd> address the same port and that this port is valid. An exception
  is when both members are -1. This is the initial enumeration state after
  calling WkdResetPortEnumeration().

*****************************************************************************/

BOOL WkdSelectNextPort(BOOL fForAccess)
{
  INT nSelPort = 0;
  INT iSelPortGroup = 0;
  ASSERT( (cstWkd.nSelPort == -1) ||
          (cstWkd.iSelPortGroup != -1) && 
          (cstWkd.pprtgrpSel != NULL) );
  // continue with last selected port or start with -1
  nSelPort = cstWkd.nSelPort;
  iSelPortGroup = cstWkd.iSelPortGroup;
  for (;;) {
    // check if next port is part of the current driver
    // * when called WkdResetPortEnumeration() before: set <nSelPort> to 0
    nSelPort++;
    ASSERT(nSelPort >= 0);
    if (iSelPortGroup == -1 || nSelPort > cstWkd.pprtgrpSel->nLastPort) {
      // no port group selected or new port number out of range of port group
      // * select next valid port group
      for (;;) {
        // select next port group
        // * when called WkdResetPortEnumeration() before: set iPortGroup to 0
        iSelPortGroup++;
        if (iSelPortGroup >= gcWkdPortGroup) {
          // no next driver exists => no next port exists
          // * end of port enumeration reached
          return FALSE;
        } // if
      #if MULTIMACH
        if (apprtgrpWkd[iSelPortGroup] != NULL) {
          // this port group is not dynamically invalidated => exit loop
          break;
        } // if
      #else
        // next port group determined: exit loop
        break;
      #endif
        // check next port group
      } // for
      // next port group is established: set status
      cstWkd.iSelPortGroup = iSelPortGroup;
      cstWkd.pprtgrpSel = apprtgrpWkd[iSelPortGroup];
      cstWkd.pclsdrvSel = cstWkd.pprtgrpSel->pclsdrv;
      // get first port of determined port group
      nSelPort = cstWkd.pprtgrpSel->nFirstPort;
    } // if
    if (WkdSelectPort(nSelPort, fForAccess)) {
      // selection of port is valid => return this port
      return TRUE;
    } // if
    // <nSelPort> not valid: check next
    // * possibly no port selected: <cstWkd> members may be -1
    // * the last search position is stored in <nSelPort> and <iSelPortGroup>
  } // for cycle, exit by return
  // this loop is never exit by break
  ASSERT(0);
  return FALSE;
} // WkdSelectNextPort()


/*****************************************************************************
W k d A c c e s s P o r t S p e c
==============================================================================

This function checks the port specification and sets the corresponding
internal port number into the <cstWkd.nSelPort> variable. This port is also
selected and accessed. This function can
be only used to specify single ports.

Parameters
  pszPort  points to a port specification string or contains in the lower byte
           a port specification number (0..126). If a string is specified,
           it's letter case is not significant.

Globals
  cstWkd*

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

BOOL WkdAccessPortSpec(const char *pszPort)
{
  INT nPort = 0;
  if (cstWkd.flCtrl & WKB_PORTNUMBER) {
    // old port specification (value 0..127 with least 3 bits not all 1)
    if (HIUSHORT( (unsigned long)pszPort) != 0) {
      cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
      return FALSE;
    } // if
    nPort = LOUSHORT((unsigned long)pszPort);
  } else {
    // port string specified: convert into memory block if required
    if (!WkdConvertMemHandle((const void**) &pszPort)) {
      return FALSE;
    } // if
    pszPort = WkdParsePortSpec(pszPort, &nPort);
    // any error => return
    if (pszPort == NULL || *pszPort != '\0') {
      cstWkd.idError = WKB_ERROR_INVALID_PORT;
      return FALSE;
    } // if
  } // if
  // check if specified/determined port number is valid & no wildcard
  if (nPort < 0 || nPort > 126 || (nPort & 7) == 7) {
    // bad port value specified
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if
  // select & access determined port & return status
  return WkdSelectPort(nPort, TRUE);
} // WkdAccessPortSpec()


/*****************************************************************************
E x p a n d E x p D a t e
==============================================================================

This function decrypts an expiration date in the specified entry. If the
entry is an added data entry with describe such an entry the contents is
decrypted and stored as an absolute and a day-relative date.

Parameters
  pwkbbxe  points an initialized WibuBox entry structure.

Globals
  cstWkd.pope  points to an initialized WibuBox open entry description.

*****************************************************************************/

static void ExpandExpDate(WKBBOXENTRY *pwkbbxe)
{
  UINT uDate = 0;
  UINT uCurrentDate = 0;
  SYSTEMTIME st;
  if (cstWkd.bLevel < 1) {
    // level 0 call => no support of expiration dates
    return;
  } // if
  if (!(pwkbbxe->flCtrl & WKB_BE_ADDDATA)
    || (LOBYTE(pwkbbxe->ulData2) != 0xFF)
    || (HIUSHORT(pwkbbxe->ulData1) != 0xFFFF)) {
    // no encrypted expiration date => nothing to do
    return;
  } // if
  // get packed date value
  uDate = (UINT) LOUSHORT(pwkbbxe->ulData1);
  // determine date count of today
  uCurrentDate = BsrGetCurrentDateCount();
  // convert now entry's date count to date
  if (!BsrConvertDateCount(uDate, &st)) {
    // no valid year ==>> no Expdate ==>> let Added Data Entry
    return;
  } // if
  // valid year ==>> convert to an ExpDate
  pwkbbxe->ulData1 = MAKEULONG(MAKEUSHORT(st.wDay, st.wMonth),
      (st.wYear & 0xFFF) | (st.wDayOfWeek << 12) );
  // store difference between specified and current date (negative => 0xFFFF)
  pwkbbxe->ulData2 = uCurrentDate > uDate ?
      USHRT_MAX : (USHORT)(uDate - uCurrentDate);
  // high 16 bits of <ulData2> remains 0
  ASSERT(HIUSHORT(pwkbbxe->ulData2) == 0);
  // reset added data, set expiration date
  pwkbbxe->flCtrl &= ~WKB_BE_ADDDATA;
  pwkbbxe->flCtrl |= WKB_BE_ADDEXPDATE;
} // ExpandExpDate()


/*****************************************************************************
W k d S e t B o x E n t r y
==============================================================================

This function converts 6 or 12 bytes physically stored entry information in
the WibuBox into a corresponding API specific data structure of type
WKBBOXENTRY. This function is used by the API function WkbListBox2 and by
the WKB_QE_ENTRY option of the function WkbQueryEntry2.

Parameters
  pbBox ....... points to the 6 bytes base entry which is read directly from
                the WibuBox.
  pbBoxAdd .... points to the 6 bytes added entry whichis read directly from
                the WibuBox. This entry is only used if the base entry is
                a FC/UC or a master entry and it index less than 5. If this
                entry is consumed, it is overwritten with 6 bytes to avoid
                that it is reread again.
  nVersion .... contains the version of the WibuBox which is currently a
                value in range 1..5.
  nIndex ...... contains the index of the entry in the range 0..9.
  pwkbboxeDest  points to the WibuKey API specific WibuBox entry which is
                filled by this function.
  fIsFC0Allowed specifies if the access to FC=0 is allowed.

Return
  The function returns TRUE if the entry is read and FALSE if the entry is
  empty.

*****************************************************************************/

BOOL WkdSetBoxEntry(const BYTE *pbBox, BYTE *pbBoxAdd,
  INT nVersion, INT nIndex, WKBBOXENTRY *pwkbboxeDest, BOOL8 fIsFC0Allowed)
{
  WKBBOXENTRY wkbbxe; // local entry for creations
  ULONG ulFC = MAKEULONG(MAKEUSHORT(pbBox[0], pbBox[1]), pbBox[2]);
  ULONG ulUC = MAKEULONG(MAKEUSHORT(pbBox[3], pbBox[4]), pbBox[5]);
  UINT uType = pbBox[2] >> 4;
  ASSERT(nIndex >= 0 && nIndex <= 9);
  ASSERT(nVersion >= 1 && nVersion <= 7);
  if (ulFC == 0 && ulUC == 0 && !fIsFC0Allowed) {
    // empty entry => don't store
    return FALSE;
  } // if
  // init complete entry
  memset(&wkbbxe, 0, sizeof(wkbbxe));
  // set index value into flags
  wkbbxe.flCtrl = nIndex;
  // analyse type of entry
  if ( (nVersion == 1) || (uType == 0) || (uType == 0xA) ) {
    // standard FC/UC or Master Entry
    if ( (nVersion == 1) || (uType == 0) ) {
      // store three bytes as Firm Code Entry
      wkbbxe.flCtrl |= WKB_BE_FCUC;
      wkbbxe.ulFirmCode = ulFC;
    } else {
      // store Master Firm Entry
      wkbbxe.flCtrl |= WKB_BE_MASTER;
      wkbbxe.ulFirmCode = ulFC & 0x0FFFFF;
    } // if
    wkbbxe.ulUserCode = ulUC;
    if (nIndex < 5 && nVersion > 1) {
      // analyse if bundled entry exists
      UINT uBundleType = pbBoxAdd[2] >> 4;
      if (uBundleType == 0xB && (pbBoxAdd[2] & 8) != 0) {
        // bundled data entry
        wkbbxe.flCtrl |= WKB_BE_ADDDATA;
        wkbbxe.ulData1 = MAKEULONG(
          MAKEUSHORT(pbBoxAdd[3], pbBoxAdd[4]),
          MAKEUSHORT(pbBoxAdd[5], pbBoxAdd[0]));
        wkbbxe.ulData2 = pbBoxAdd[1];
        // transfer special added data flags (WKB_DATA_xxxVARY)
        wkbbxe.flCtrl |= ((ULONG)(pbBoxAdd[2] & 7)) << 16;
        // added entry is consumed: set to empty
        memset(pbBoxAdd, 0, (size_t)6);
      } else if (uBundleType >= 0xC && uBundleType <= 0xE) {
        // bundled limit count entry
        wkbbxe.flCtrl |= WKB_BE_ADDLIMIT;
        wkbbxe.ulData1 = MAKEULONG(
          MAKEUSHORT(pbBoxAdd[3+uBundleType-0xC], pbBoxAdd[1]),
          pbBoxAdd[2] & 0xF);
         // programming counter
        wkbbxe.ulData2 = pbBoxAdd[0];
        // added entry is consumed: set to empty
        memset(pbBoxAdd, 0, (size_t)6);
      } // if
    } // if
    if (cstWkd.bLevel >= 1) {
      // expand a specified expiration date entry
      ExpandExpDate(&wkbbxe);
    } // if
  } else if (uType == 0xB && (pbBox[2] & 8) == 0) {
    // User Data entry (unbundled)
    wkbbxe.flCtrl |= WKB_BE_DATA;
    wkbbxe.ulData1 = MAKEULONG(
      MAKEUSHORT(pbBox[3], pbBox[4]),
      MAKEUSHORT(pbBox[5], pbBox[0]));
    wkbbxe.ulData2 = pbBox[1];
    // note: contents of bit 0..2 of pbBox[2] is ignored!
  } else {
    // defect or invalid entry
    wkbbxe.ulFirmCode = ulFC;
    wkbbxe.ulUserCode = ulUC;
    if (uType == 0xF) {
      wkbbxe.flCtrl |= WKB_BE_DEFECT;
    } else {
      wkbbxe.flCtrl |= WKB_BE_INVALID;
    } // if
  } // if
  // entry valid: update destination buffer, return TRUE
  *pwkbboxeDest = wkbbxe;
  return TRUE;
} // WkdSetBoxEntry()


/*****************************************************************************

---- Reading of Port Address Driver Setting Variables ------------------------

*****************************************************************************/

/*****************************************************************************
W k d R e a d P o r t A d d r
==============================================================================

This function reads a valid port specification. Under all operating system
except Windows NT this is a single 16 bit hexadecimal number. Under Windows
NT, optional a bus type and a bus index can be specified with the following
syntax:

  BusType[BusIndex]:Address

with a <BusType> in the following range:
  Internal (0), Isa (1), Eisa (2), MicroChannel (3), TurboChannel (4),
  PCIBus (5), VMEBus (6), NuBus (7), PCMCIABus (8), CBus (9), MPIBus (10),
  MPSABus (11), ProcessorInternal (12), InternalPowerBus (13)
and <BusIndex> as an index, started with 0.

Parameters
  ppszParse  points to the current parsing location.
  pptsDest   points to the port specification which receives the port data.

Return
  The function returns TRUE if a valid address was read and FALSE if no valid
  specification was found or the address is 0.

*****************************************************************************/

BOOL WkdReadPortAddr(char **ppszParse, PORTSPEC *pptsDest)
{
  UINT uAddr = 0;
  char *pszParse = *ppszParse;
  // skip leading spaces
  WkdSkipSpace(&pszParse);
  uAddr = (UINT)WkdReadHexNumber(&pszParse);
  {// create valid port specification
    // transfer only port/address number
    pptsDest->pta = (USHORT)uAddr;
  } // block
  // save new parsing position
  *ppszParse = pszParse;
  // return TRUE if address is not 0
  return uAddr != 0;
} // WkdReadPortAddr()


/*****************************************************************************
WkdQueryEnvWkCom
==============================================================================

This function searches the Variable WKCOM in the environment, interprets
the argument and is setting an Arry of BOOL whether the port is usable or
not.

Parameters
  pszVarName ... points to the port type string. The string must be specified
                 in uppercase.

Return
  The function returns:
  0    if there's no Port selectable by WKCOM (e.g WKCOM='/D,/D,/D,/D')
  >0   the number of available Ports at success
  <0   Error case, something's wrong
       -1   no memory
       -2   WKCOM not found
       -3   wrong entry in WKCOM

*****************************************************************************/

INT WkdQueryEnvWkCom(char *pszVarName)
{
  char *pszBuf = NULL;
  char *pszParse = NULL;
  UINT   cchLength = 0;
  BOOL  *pgafPorts;
  INT    cPorts = 0;
  INT    i = 0;

  pgafPorts = gafPorts;
  // if Ports checked before, count and return them
  if(gfPortsChecked){
    //array walk
    for (i = 0; i < C_SERIAL_PORT; i++){
      if (*(pgafPorts + i))
        cPorts++;
    }//for
    return (cPorts);
  } //if

  // clean the global boolean array
  memset(gafPorts, 0, sizeof(gafPorts));
  pszBuf = WkdAllocMem(256 * sizeof(char));
  if (pszBuf == NULL) {
    cstWkd.idError = WKB_ERROR_MEMORY_TOO_SMALL;
    // set global marker
    gfPortsChecked = FALSE;
    return (-1);
  } // if

  // search specified environment variable (system specific)
  cchLength = WkdGetProfileValue(WKPRFTPC_DRIVER_KERNEL, pszVarName, pszBuf, 256);
  if (!cchLength) {
    cPorts = -2;
    // set global marker
    gfPortsChecked = FALSE;
    goto EXIT_ENVWKCOM;
  } // if
  // set search position
  pszParse = pszBuf;
  while (*pszParse != '\0') {
    // Port is selected
    if (*pszParse == '*') {
      *pgafPorts = TRUE;
      pgafPorts++;
      cPorts++;
    } // else 
    if (*pszParse == '/') {
      // Port is deselected
      if (WkdToUpper(*(++pszParse)) == 'D') {
        *pgafPorts = FALSE;
        pgafPorts++;
      } else {
        cPorts = -3;
        // set global marker
        gfPortsChecked = FALSE;
        goto EXIT_ENVWKCOM;
      } // if
    } // if
    pszParse++;
  } // while
  // set global marker
  gfPortsChecked = TRUE;

EXIT_ENVWKCOM:
  if (pszBuf != NULL ) { free(pszBuf); pszBuf = NULL; }
  return (cPorts);
} // WkdQueryEnvWkCom()

/*****************************************************************************
W k d Q u e r y E n v P o r t s
==============================================================================

This function searches the specified port type in the environment, interprets
the argument as addresses and stores the addresses into the specified buffer.
Never more addresses are stored than the number of entries in the buffer
specifies. Before the buffer is filled, it is cleared (all addresses are set
to -1), independent if an enviroment specification exists or not.

Parameters
  pszVarName ... points to the port type string. The string must be specified
                 in uppercase.
  pszOptionName  points to a 0-separated string with uppercase characters
                 which represents options after the address specification,
                 separated by the sharp ('/').
  paptsDest .... points to the buffer which is cleared and which gets the
                 specified port address values.
  pauOptions ... points to a array of integer values which gets the specified
                 option bits. Bit 0 corresponds with the first character in
                 <pszOptionName>, bit 1 with the second character and so far.
                 This array is initially filled with 0 bytes.
  cDestBuf ..... contains the number of entries in the buffer. This value must
                 not be 0.

Comments
  Adresses which are set to -1 after return must be changed by the caller with
  default values from the BIOS or other sources or set to invalid (0) in
  dependency of the port type.

*****************************************************************************/

void WkdQueryEnvPorts(char *pszVarName, char *pszOptionName,
  PORTSPEC *paptsDest, UINT *pauOptions, INT cDestBuf)
{
  char *pszBuf = NULL;
  char *pszParse = NULL;
  UINT cchLength = 0;
  INT i = 0;
  // set complete destination address buffer to -1 values
  memset(paptsDest, -1, cDestBuf * sizeof(PORTSPEC));
  // set complete destination option buffer to 0 values
  memset(pauOptions, 0, cDestBuf * sizeof(UINT));
  // search specified environment variable (system specific)
  pszBuf = BsrAllocMem(256 * sizeof(char));
  if (pszBuf == NULL) {
    // memory too small => return
    return;
  } // if
  // search specified environment variable (system specific)
  cchLength = WkdGetProfileValue(WKPRFTPC_DRIVER_KERNEL, pszVarName, pszBuf, 256);
  // variable not found in environment => return default setting
  if (cchLength == 0 || cchLength > 256) {
    goto RETURN;
  } // if
  // set search position
  pszParse = pszBuf;
  for (i = 0; i < cDestBuf; i++) {
    // parse next entry and store into destination buffer
    WkdSkipSpace(&pszParse);
    if (*pszParse == '*') {
      // '*': use already set default value -1
      pszParse++;
    } else {
      // get address specification, 0 if invalid
      WkdReadPortAddr(&pszParse, paptsDest + i);
    } // if
    // skip spaces after read number/'*'
    WkdSkipSpace(&pszParse);
    // check for options (started with '/')
    while (*pszParse == '/') {
      char ch;
      char *psz;
      // parse now options
      pszParse++;
      // get single character
      ch = (char)WkdToUpper(*pszParse);
      // search in character list (string)
      psz = strchr(pszOptionName, ch);
      if (psz != NULL) {
        // option found: set bit in destination
        SET_BIT(pauOptions[i], psz - pszOptionName);
      } // if
      // this character is always consumed
      pszParse++;
      // ignore further characters which follows directly except terminators
      while (*pszParse != 0 && !strchr("\t ,/", *pszParse)) {
        pszParse++;
      } // if
      // space characters are permitted
      WkdSkipSpace(&pszParse);
    } // while
    if (*pszParse == ',') {
      // skip parsed comma
      pszParse++;
    } // if
  } // for
RETURN:
  BsrFreeMem(pszBuf);
} // WkdQueryEnvPorts()

/*****************************************************************************

---- Handling of WibuBox contents data and opened entries -------------------

*****************************************************************************/

/*****************************************************************************
W k d G e t B o x C o n t e n t s
==============================================================================

This function reads the complete contents of the specified WibuBox into
a created WKBBOXDATA data structure which is returned.

Parameters
  pwkbsr .... points to the serial number of the WibuBox.
  uEntryMask  contains a mask of all entries which contents should be read.
              Bit 0 represents entry 0, bit 1 entry 1 etc. The function
              WkbQueryEntry2() uses this mask typically with the multifind
              bit structure of WkbOpen2()/WKB_OPEN_MULTIFIND.
  pcbData ... points to a UINT variable which receives the valid bytes in the
              created WKBBOXDATA list.

Return
  The created contents structure is returned as a pointer to a block on the
  heap. If NULL is returned, an error occured which is stored in
  cstWkd.idError.

*****************************************************************************/

WKBBOXDATA *WkdGetBoxContents(const WKBSERIAL *pwkbsr,
  UINT uEntryMask, UINT *pcbData)
{
  DEVOBJECT *pdevobj = NULL;
  WKDEV_REQUEST wkdevreq;
  // created destination buffer
  WKBBOXDATA *pwkbbxdDest = NULL;
  const UINT cbBoxData = 60 + 1;
  BYTE *pabBoxData = NULL;
  // allocate read data buffer
  pabBoxData = (BYTE*) WkdAllocMem(cbBoxData);
  if (pabBoxData == NULL) {
    // memory too small => return with error
    goto RET_ERROR;
  } // if
  // allocate destination buffer
  pwkbbxdDest = (WKBBOXDATA*) WkdAllocMem(sizeof(WKBBOXDATA));
  if (pwkbbxdDest == NULL) {
    // memory too small => return with error
    goto RET_ERROR;
  } // if
  // clear complete structure
  memset(pwkbbxdDest, 0, sizeof(WKBBOXDATA));
  // execute via class driver
  // * determine accessed device from serial number
  ASSERT_VALUE(cstWkd.pclsdrvSel);
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
  pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(pwkbsr);
  if (pdevobj == NULL) {
    // device not found => return with error
    goto RET_ERROR;
  } // if
  wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
  wkdevreq.usCmd = BOXCMD_LIST;
  // get complete contents of WibuBox
  wkdevreq.usIndex = 0;
  wkdevreq.cbRequest = 0;
  wkdevreq.cbAnswer = 60 + 1;
  // read now determined number of bytes from device
  ASSERT(cstWkd.fAccessed);
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
  if (!cstWkd.pclsdrvSel->pfClDrvRequest(&wkdevreq, pabBoxData, NULL)) {
    // request not properly executed => return with error
    goto RET_ERROR;
  } // if
  // interpret first byte of read data as WibuBox version value
  pwkbbxdDest->usVersion = pabBoxData[0];
  {// expand packed WibuBox data into WKBBOXDATA structure
    INT i = 0;
    BYTE *pbSrc = NULL;
    INT cEntry = 0;
    // initialize destination store position
    WKBBOXENTRY *pwkbbxeDest = pwkbbxdDest->awkbbxe;
    // init complete entry
    memset(pwkbbxeDest, 0, sizeof(pwkbbxdDest->awkbbxe));
    for (i = 0, pbSrc = pabBoxData + 1;
        i < 10;
        i++, pbSrc += 3*2
      ) {
      if (IS_BIT(uEntryMask, i)) {
        // expand WibuBox contents (convert also expiration dates)
        if (WkdSetBoxEntry(pbSrc, pbSrc + 5*6, pabBoxData[0], i,
          pwkbbxeDest, FALSE)) {
          // created entry is stored: increment number of entries
          pwkbbxeDest++;
          cEntry++;
        } // if
      } // if
    } // for
    // store number of valid entries
    pwkbbxdDest->cEntry = (USHORT)cEntry;
    // store number of written bytes
    *pcbData = (UINT)((BYTE*)pwkbbxeDest - (BYTE*)pwkbbxdDest);
    // return without error
    return pwkbbxdDest;
  } // block
RET_ERROR:
  // free allocated memory (if any)
  BsrFreeMem(pwkbbxdDest);
  BsrFreeMem(pabBoxData);
  // return with error
  return NULL;
} // WkdGetBoxContents()


/*****************************************************************************
W k d S e t M u l t i F i n d O p t i o n s
==============================================================================

This function determines the used Expiration Date and the used Limit Counter
from the read data. Only entries are interpreted which match with the entries
in the multiopen list.

Parameters
  pwkbxd  points to a list of read WibuBox data entries.
  pope .. points to an opened WibuBox entry structure for which the
          best Expiration Date and the used Limit Counter entry is determined.

Comments
  If this function is called by WkbOpen2(), all WibuBox entries are loaded.
  In this case, <pope->fuEntryOpen> is used to determine if a loaded entry
  matches the specified Firm and User Code. If this function is later called
  by WkbSelect2() to determine another Limit Counter or Expiration Date, it
  is assumed that only valid entries are loaded.

Important point
  Please note, that if all entries are loaded by WkbOpen2(), <pope->iEntry>
  must point to a valid entry (range 0..9). An invalid entry (-1) is only
  permitted in <pope->iEntry>, if the first entry matches the opened entry
  which requires the calling outside of the WkbOpen2() function.
  <pope->iEntry> is always determined in WkbOpen2() by the search algorithm
  in the WibuBox itself.

*****************************************************************************/

void WkdSetMultiFindOptions(const WKBBOXDATA *pwkbbxd,
  OPENENTRY *pope)
{
  UINT iEntry = 0;
  const WKBBOXENTRY *pwkbbxe = NULL;
  const WKBBOXENTRY *pwkbbxeSel = NULL;
  BOOL fSelLimitCounter = FALSE;
  BOOL fSelMaster = FALSE;
  ASSERT(pope->fMultiFind);
  ASSERT(pwkbbxd->cEntry > 0);
  // ============================================
  // determine pointer to selected entry (if any)
  // ============================================
  if (pope->iEntry >= 0) {
    // determine pointer to selected entry
    for (iEntry = 0, pwkbbxe = pwkbbxd->awkbbxe;
        iEntry < pwkbbxd->cEntry;
        iEntry++, pwkbbxe++
      ) {
      if (((UINT)pwkbbxe->flCtrl & WKB_BE_INDEXMASK) == (UINT)pope->iEntry) {
        // entry found: select it by pointer
        pwkbbxeSel = pwkbbxe;
        break;
      } // if
    } // for
    if (iEntry == pwkbbxd->cEntry) {
      // mysterious error: entry not found, set to invalid
      ASSERT(0);
      pope->iEntry = -1;
    } // if
  } // if
  if (pope->iEntry == -1) {
    // entry not determined: use first entry (all are compatible)
    pwkbbxeSel = pwkbbxd->awkbbxe;
    // set preliminary entry index
    pope->iEntry = (UINT)pwkbbxeSel->flCtrl & WKB_BE_INDEXMASK;
    // this entry must match (don't come to this point from WkbOpen2() call!)
    ASSERT( (pwkbbxeSel->ulFirmCode == pope->ulFirmCode) &&
            ( ((pwkbbxeSel->flCtrl & WKB_BE_TYPEMASK) == WKB_BE_FCUC) &&
            (pwkbbxeSel->ulUserCode == pope->ulUserCode) ||
            ((pwkbbxeSel->flCtrl & WKB_BE_TYPEMASK) == WKB_BE_MASTER) &&
            ( (pope->ulUserCode & pwkbbxeSel->ulUserCode) == pope->ulUserCode)) );
  } // if
  // define flag that selected entry has Limit Counter or is Master Entry
  fSelLimitCounter = (pwkbbxeSel->flCtrl & WKB_BE_ADDLIMIT) != 0;
  fSelMaster = (pwkbbxeSel->flCtrl & WKB_BE_TYPEMASK) == WKB_BE_MASTER;
  // ==============================================
  // search now alternative entry to selected entry
  // ==============================================
  for (iEntry = 0, pwkbbxe = pwkbbxd->awkbbxe;
      iEntry < pwkbbxd->cEntry;
      iEntry++, pwkbbxe++
    ) {
    BOOL fLimitCounter = FALSE;
    BOOL fIsMaster = FALSE;
    UINT uEntry = (UINT)pwkbbxe->flCtrl & WKB_BE_INDEXMASK;
    if (!IS_BIT(pope->fuEntryOpen, uEntry)) {
      // entry does not match entry => continue
      continue;
    } // if
    ASSERT( (pwkbbxe->ulFirmCode == pope->ulFirmCode) &&
				    (((pwkbbxe->flCtrl & WKB_BE_TYPEMASK) == WKB_BE_FCUC) &&
					  (pwkbbxe->ulUserCode == pope->ulUserCode) ||
            ((pwkbbxe->flCtrl & WKB_BE_TYPEMASK) == WKB_BE_MASTER) &&
            ((pope->ulUserCode & pwkbbxe->ulUserCode) == pope->ulUserCode)) );
    // Master and Limit Counter flags for checked entry
    fIsMaster = (pwkbbxe->flCtrl & WKB_BE_TYPEMASK) == WKB_BE_MASTER;
    fLimitCounter = (pwkbbxe->flCtrl & WKB_BE_ADDLIMIT) != 0;
    if (fLimitCounter) {
      // current entry has Limit Counter: check for replace Selection
      if (fSelLimitCounter) {
        // selected entry has Limit Counter: check for replace
        if (pwkbbxe->ulData1 == 0) {
          // counter of this entry 0 => no replace
          continue;
        } // if
        if (pwkbbxeSel->ulData1 > 0 && (fIsMaster || !fSelMaster)) {
          // selected entry's counter not 0, current is master or
          // selected is no master: no replace
          continue;
        } // if
      } // if
      // selected entry no Limit Counter: replace always
    } else {
      // current entry has no Limit Counter: check for replace Selection
      if (fSelLimitCounter) {
        // selected entry has Limit Counter: no replace
        continue;
      } // if
      if (fIsMaster || !fSelMaster) {
        // this entry is Master or Selected is no master: no replace
        continue;
      } // if
    } // if
    // found entry is candidate for replacing: do it
    pope->iEntry = uEntry;
    fSelLimitCounter = fLimitCounter;
    fSelMaster = fIsMaster;
    pwkbbxeSel = pwkbbxe;
  } // for
  // set further options in open entry
  pope->flCtrl &= ~WKB_OE_MASTER;
  if (fSelMaster) {
    pope->flCtrl |= WKB_OE_MASTER;
  } // if
  // no expiration date determined
  pope->uExpDate = NO_EXPDATE;
  { // ==========================================
    // get highest Expiration Date of all entries
    // ==========================================
    UINT uBestExpDate = 0; // initial: 70Jan01
    for (iEntry = 0, pwkbbxe = pwkbbxd->awkbbxe;
        iEntry < pwkbbxd->cEntry;
        iEntry++, pwkbbxe++
      ) {
      UINT uEntry = (UINT)pwkbbxe->flCtrl & WKB_BE_INDEXMASK;
      if (!IS_BIT(pope->fuEntryOpen, uEntry)) {
        // entry does not match entry => analyze next entry
        continue;
      } // if
      if (pwkbbxe->flCtrl & WKB_BE_ADDEXPDATE) {
        // check if specified Expiration Date is better
        UINT uExpDate = (UINT) LOUSHORT(pwkbbxe->ulData2);
        // calculate absolute date
        if (uExpDate == USHRT_MAX) {
          // out of date: set to 0
          uExpDate = 0;
        } else {
          // otherwise: add relative value to current date
          uExpDate += BsrGetCurrentDateCount();
        } // if
        if (uExpDate >= uBestExpDate) {
          // higher than specified expiration => replace this
          uBestExpDate = uExpDate;
          // set copy of used expiration date into open entry structure
          pope->uExpDate = uBestExpDate;
        } // if
      } // if
    } // for
  } // block
} // WkdSetMultiFindOptions()


/*****************************************************************************
W k d G e t O p e n B o x E n t r y
==============================================================================

This function determines the WibuBox contents of the specified open entry
and returns it in a WKBBOXENTRY structure. This function is called by the
WkbQueryEntry2() function with the WKB_QE_BOX option, by the WkbOpen2()
function with the WKB_OPEN_ENTRY option, by WkbOpen2()/WKB_OPEN_FIND/
WKB_OPEN_MULTI_FIND when a real subsystem is used and by WkbSelect2()
to determine if an Expiration Date is specified.

Parameters
  pwkbsr ....... points to the serial number of the requested WibuBox.
  iEntry  ...... contains the entry of the determined entry in the range 0..9.
  pwkbbxeDest .. points to a buffer which receives the WibuBox entry data
  pnVersion .... points to a integer variable which receives the version
                 of the WibuBox.
  fIsFC0Allowed  specifies if the access to FC=0 is allowed.

Globals
  cstWkd.nPort

Return
  The function returns TRUE if the WibuBox entry is determined and
  FALSE if not.

*****************************************************************************/

BOOL WkdGetOpenBoxEntry(const WKBSERIAL *pwkbsr, INT iEntry,
  WKBBOXENTRY *pwkbbxeDest, INT *pnVersion, BOOL8 fIsFC0Allowed)
{
  DEVOBJECT *pdevobj = NULL;
  WKDEV_REQUEST wkdevreq;
  BYTE abBuf[1 + 2 * 6];
  BOOL fReturn = FALSE;
  UINT uVersion = 0;
  REQOPTIONS reqopt;
  if (iEntry < 0 || iEntry > 9) {
    // invalid entry specified
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    return FALSE;
  } // if
  // execute via class driver
  // * determine accessed device from serial number
  ASSERT_VALUE(cstWkd.pclsdrvSel);
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial);
  ASSERT(cstWkd.fAccessed);
  pdevobj = cstWkd.pclsdrvSel->pfClDrvGetDeviceFromSerial(pwkbsr);
  if (pdevobj == NULL) {
    // device not found => return with error
    goto EXIT_FN;
  } // if
  wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
  wkdevreq.usCmd = BOXCMD_LIST,
  wkdevreq.usIndex = 0;
  wkdevreq.cbRequest = 0;
  wkdevreq.cbAnswer = 13;
  // set request option to required entry index
  reqopt.List.iEntry = iEntry;
  // read now determined number of bytes from device
  ASSERT_VALUE(cstWkd.pclsdrvSel->pfClDrvRequest);
  if (!cstWkd.pclsdrvSel->pfClDrvRequest(&wkdevreq, abBuf, &reqopt)) {
    // request not properly executed => return with error
    goto EXIT_FN;
  } // if
  // convert WibuBox raw data into structure data
  // * abBuf[0] contains WibuBox version
  uVersion = abBuf[0];
  if (!WkdSetBoxEntry(abBuf + 1, abBuf + 6 + 1,
      uVersion, iEntry, pwkbbxeDest, fIsFC0Allowed)
    ) {
    // entry not found => return WKB_ERROR_ENTRY_NOT_FOUND
    cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
    goto EXIT_FN;
  } // if
  *pnVersion = uVersion;
  // data now read
  fReturn = TRUE;
EXIT_FN:
  // return determined state
  return fReturn;
} // WkdGetOpenBoxEntry()


// ======================================
// end of kecommon.c
// ======================================
