/*****************************************************************************
S r c D r v                               Portable Source Code WibuKey Driver
==============================================================================
kemach.c                                    Kernel Specific Machine Adaptation
==============================================================================

This module implements machine (hardware platform) specific functions which
separate dynamically between the IBM PC and other platforms, for example the
NEC PC-98xx.

This module is only implemented in
WKWIN.DLL, WKWIN32.DLL, WIBUKEY.SYS and WIBUKEY.VXD.

==============================================================================

The portable WibuKey driver source is used under DOS and Windows to build
the internally used driver WKDOS2.LIB and the shipped drivers WKWIN.DLL,
WKWIN32.DLL, WIBUKEY.SYS and WIBUKEY.VXD. It is also used to build the
Macintosh and Netware NLM drivers.

Furthermore, it is also the origin of the SRCDRV project. All sources are
written in ANSI C, except the using of C++ style single line comments.

==============================================================================

Copyright  1997 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
97Mar09 2.31 mabu: initial version, separated from KESYS.C
97Mar09 2.31 mabu: adding of NEC PC-98xx support (WIBU-BOX/SX)
97Apr05 2.31 mabu: automatic detection of NEC PC-98xx on Windows NT
97May08 2.31 mabu: automatic switching of IBM/NEC routine on Windows 95
97May09 2.31 mabu: automatic switching of IBM/NEC routine on Windows 3.1
97Oct05 2.32 mabu: automatic switching of IBM/NEC routine modified
98Aug04 2.51 mabu: bug in stall execution function setting removed
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

#if MACH_NEC98

static BOOL gfNecPc = FALSE;

// flag "machine variant is checked"
BOOL gfWkdMachineChecked = FALSE;

/*****************************************************************************
I s I b m P c B i o s
==============================================================================

This function checks if the BIOS of the system is typical for IBM-PC. For this
the BIOS date at the end of the system ROM is checked. This date is only
available on an IBM PC at address F000:FFF5 or FFFF:5 in following format:

MM/DD/YY

where MM means the decimal month, DD the decimal day and YY the decimal year.

Return
  The function returns TRUE if an IBM PC is detected and FALSE if not.

Comments
  On some Toshiba systems, the second slash ("/") is replaced by a "j".
  Therefore, the digits and not the slashs are checked.

*****************************************************************************/

static BOOL IsIbmPcBios(void)
{
  // pointer to address FFFF:0
  BYTE *pbRom;
  // determine type of PC by reading specific BIOS ROM parts
  // method: check digits of year, month and day
  pbRom = (BYTE *)MAKEULONG(0, 0xFFFF);
  if (pbRom[5] < '0' || pbRom[5] > '9') {
    // high value/year not 0 => no IBM PC
    return FALSE;
  } // if
  if (pbRom[6] < '0' || pbRom[6] > '9') {
    // high value/year not 0 => no IBM PC
    return FALSE;
  } // if
  if (pbRom[8] < '0' || pbRom[8] > '9') {
    // high value/year not 0 => no IBM PC
    return FALSE;
  } // if
  if (pbRom[9] < '0' || pbRom[9] > '9') {
    // high value/year not 0 => no IBM PC
    return FALSE;
  } // if
  if (pbRom[11] < '0' || pbRom[11] > '9') {
    // high value/year not 0 => no IBM PC
    return FALSE;
  } // if
  if (pbRom[12] < '0' || pbRom[12] > '9') {
    // high value/year not 0 => no IBM PC
    return FALSE;
  } // if
  // this is an IBM PC
  return TRUE;
} // IsIbmPcBios()


/*****************************************************************************
W k d C h e c k M a c h i n e
==============================================================================

This function checks the type of the machine. If uses the WKMACHINE
environment constant (supports IBMPC or NECPC) or try to determine the machine
by analyzing hardware specific data.

Return
  The function returns TRUE if the current computer is a NEC and FALSE
  if it is a IBM.

*****************************************************************************/

BOOL8 WkdCheckMachine (void)
{
  if (gfWkdMachineChecked) {
    // machine type already determined: nothing to do
    return gfNecPc;
  } // if

  // determine system by environment variable (if any)
  gfNecPc = FALSE;
  if (WkdCheckEnvVar("WKMACHINE", "NECPC") == +1) {
    // NEC PC specified explicitly: address this machine type
    gfNecPc = TRUE;
  } else if (WkdCheckEnvVar("WKMACHINE", "IBMPC") != +1) {
    // determine type of PC by reading specific BIOS ROM parts
    if (!IsIbmPcBios()) {
      // IBM PC not determined properly => assume NEC PC
      gfNecPc = TRUE;
    } // if
  } // if

  // modify setting of low level kernel drivers
  if (gfNecPc) {
    //// set NEC PC drivers
    // activate NEC LPT ports
    WkdDefineLowDriver(0, &lowdrvWkdLptNec);
    // activate NEC COM ports
    WkdDefineLowDriver(8, &lowdrvWkdComNec);
  } else {
    // set (standard) IBM PC drivers
    // this drivers are normally set but by a registry switch from
    //   NEC -> IBM, these drivers must be set again
    // activate IBM LPT ports
    WkdDefineLowDriver(0, &lowdrvWkdLpt);
    // activate IBM COM ports
    WkdDefineLowDriver(8, &lowdrvWkdCom);
  } // if
  // machine now checked
  gfWkdMachineChecked = TRUE;
} // WkdCheckMachine()

#endif // #if MACH_NEC98

#if !ABSTIME

// variable for setting determined system time
UINT guWkdSysT = 0; // 0 => is not determined


/*****************************************************************************
W k d G e t S y s T i m e
==============================================================================

This function determines the delay-loop count which corresponds with an
absolute delay of 50 microseconds. This value is used for the initialize
time out delay if no WibuBox is connected at a WibuKey port. For this,
the returned value may not be determined exactely but it must should be
rounded up to a "safe" value.

Parameters
  ptaSync  contains the port address which is used to delay the port access.

Globals
  guWkdSysT  is set with the determined value.

Return
  The function returns TRUE if the specified port number is valid and FALSE
  if not.

*****************************************************************************/

UINT WkdGetSysTime(PORTADDR ptaSync)
{
  if (guWkdSysT == 0) {
    UINT cCycle;
  #if MACH_NEC98
    if (!gfNecPc)
  #endif
  #if !defined(_BORLANDC) &&  !UNIX
    {// determine 50 ms delay count value
    #if defined(_PHARLAP)
      BYTE bTimerCount;
    #else
      volatile BYTE *pbTimerCount = WkdGetBiosPtr(BIOS_PTR(0x6C));
    #endif
      volatile UINT cCycle = 0;
    #if defined(_PHARLAP)
      BYTE bLastVal;
      bTimerCount = PeekRealByte((REALPTR)(16*0x40+0x6C));
      bLastVal = bTimerCount;
    #else
      BYTE bLastVal = *pbTimerCount;
    #endif
      // interrupt must not be disabled
      ENABLE_INTERRUPT();
      // wait for start of new timer interval
    #if defined(_PHARLAP)
      while (bTimerCount == bLastVal) {
        bTimerCount = PeekRealByte((REALPTR)(16*0x40+0x6C));
      } // while
      bLastVal = bTimerCount;
    #else
      while (*pbTimerCount == bLastVal) {
      } // while
      bLastVal = *pbTimerCount;
    #endif
      // count delay cycles until end of started timer interval
      do {
        cCycle++;
        // dummy access
        READ_PORT_BYTE(ptaSync);
    #if defined(_PHARLAP)
        bTimerCount = PeekRealByte((REALPTR)(16*0x40+0x6C));
      } while (bTimerCount == bLastVal);
    #else
      } while (*pbTimerCount == bLastVal);
    #endif
      // typical values for cCycle:
      // 386SX,  16 MHz, MS C/C++ 7.0 with /Oaxz:   10000
      // 486DX2, 66 MHz, MS C/C++ 7.0 with /Oaxz:   30000
      // 386SX,  16 MHz, Watcom C 9.0 with /oaxt /4r: 10000
      // 486DX2, 66 MHz, Watcom C 9.0 with /oaxt /4r: 30000
    }
  #endif // #if !defined(_BORLANDC)
  #if MACH_NEC98
    else
  #endif
  #if MACH_NEC98 || defined(_BORLANDC)
    {// set hypothetical value
      cCycle = 30000;
    }
  #endif // #if defined(_BORLANDC)
    // determine 50 microsecs from 50 ms
    guWkdSysT = cCycle / 1000;
  #if defined(TEST)
    printf("Determined 50 microsecs system count: %u\n", cWkdSysT);
  #endif
  } // if
  // return determined value
  return guWkdSysT;
} // WkdGetSysTime()

#endif // #if !ABSTIME


// ====================================
// end of kemach.c
// ====================================
