/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
keseq.c:                                        Sequential Device Class Driver
==============================================================================

This C source file implements the sequential class driver which calls the
low level driver at LPT, COM and PCCARD.

==============================================================================

The portable WibuKey driver source is used under DOS and Windows to build
the internally used driver WKDOS2.LIB and the shipped drivers WKWIN.DLL,
WKWIN32.DLL, WIBUKEY.SYS and WIBUKEY.VXD. It is also used to build the
Macintosh and Netware NLM drivers.

Furthermore, it is also the origin of the SRCDRV project. All sources are
written in ANSI C, except the using of C++ style single line comments.

==============================================================================

Copyright  1998-2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-0-721-93172-0, Fax: +49-0-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
98May04 2.41 mabu: separated from KECOMMON.C
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

#if ABSTIME

// contains max. delay for port acquiring
// UINT guWkdAcquireWait = (UINT)-1;  // -1 => is not determined

#else

// variable for setting minimum T value
INT gnWkdMinT = -1;   // -1 => is not determined

#endif // #if !ABSTIME

// selected device object
static DEVOBJECT devobj;

// last selected variant, -1 if port is not accessed
static INT iSelVariant = - 1;

// last selected low level driver function group
static LOWDRIVER *plowdrvSel;

// last selected low level status
static LOWSTATUS *plwstSel;

#if UNIX

/*****************************************************************************
W k d R e a d P o r t P r o p e r l y
==============================================================================

This function reads a data byte from the specified port address and reread
the data again until the read data is stable. By the doubled reading,
indefinite polls on the signal lines are not read as true data.

Parameters
  ptadr  contains the port address from which data are read.
  bMask  contains the mask which specify the bits to be read.

Return
  The function returns the read bits in the data byte.

Comments
  This function reads the port at least for two times. A wait loop which
  timeout delay bases on the number of port reads should reduce the timeout
  counter by 2.

*****************************************************************************/

BYTE WkdReadPortProperly(PORTADDR ptadr, BYTE bMask)
{
  for (;;) {
    // continuous loop until stable result
    BYTE bData = (BYTE)(READ_PORT_BYTE(ptadr) & bMask);
    if (((BYTE)READ_PORT_BYTE(ptadr) & bMask) == bData) {
      // stable signal: return read data
      return bData;
    } // if
    // otherwise: continue in loop
  } // for
  ASSERT(0);
  return 0;
} // WkdReadPortProperly()

#endif // if UNIX

#if MULTIMACH

/*****************************************************************************
W k d D e f i n e L o w D r i v e r
==============================================================================

This function sets the specified low level driver for the specified port
in the low driver list of the kernel driver.

Parameters
  nPort .. contains the first port number of the driver class, for example 0
           for LPT, 8 for COM and 32 for PcCard.
  plowdrv  contains the address of the new low level driver. If this pointer
           is NULL, the port is deactivated.

*****************************************************************************/

void WkdDefineLowDriver(INT nPort, LOWDRIVER *plowdrv)
{
  INT iPortGroup;
  for (iPortGroup = 0; iPortGroup < gcWkdPortGroup; iPortGroup++) {
    if (nPort == apprtgrpWkd[iPortGroup]->nFirstPort) {
      // driver found: replace it & return
      apprtgrpWkd[iPortGroup]->plowdrv = plowdrv;
      return;
    } // if
  } // for
  // no driver found
  ASSERT(0);
} // WkdDefineLowDriver()

#endif // #if MULTIMACH


#if !ABSTIME

/*****************************************************************************
S e t M i n T
==============================================================================

This function sets the lower limit of the specified T value to the MinT value
if this value is specified in the environment. In the diagnosis variant of the
driver, a diagnosis entry is generated for the first when the T value is
increased to MinT for a specific port.

*****************************************************************************/

static void SetMinT(void)
{
  if (gnWkdMinT == -1) {
    // check if WKMINWAIT is specified in profiling: set <gnWkdMinT> if so
    gnWkdMinT = WkdGetProfileInt(WKPRFTPC_DRIVER_KERNEL, "WkMinWait");
    ASSERT(gnWkdMinT >= 0);
  } // if
  // minimum T correction => replace <uT> if minimum is smaller
  if (plwstSel->uT < (UINT)gnWkdMinT) {
    plwstSel->uT = gnWkdMinT;
  } // if
} // SetMinT()


/*****************************************************************************
WakeLow
==============================================================================

This function "wakes" up at lower level the WibuBox at the selected port
by calling the suitable low-level function. This function returns the original
determined T value or 0 if an error occured. The T value is compared with the
optionally determined MinT value and is set to this minimum value if needed.

Return
  The function returns TRUE if the wakeup was correct and FALSE if not.
  The determined (and corrected T value) is stored in <plwstSel->uT>.

*****************************************************************************/

static BOOL WakeLow(void)
{
  cstWkd.idError = (plowdrvSel->pfLowWakeUp)(cstWkd.nSelPort);
  // reset bit that first wake up
  plwstSel->fuStatus &= ~LWST_FIRSTWAKEUP;
  // any error => return FALSE
  if (cstWkd.idError != WKB_NO_ERROR) {
    return FALSE;
  } // if
  // increase T to MinT if desired and required
  SetMinT();
  return TRUE;
} // WakeLow()

#endif // #if !ABSTIME


/*****************************************************************************
WakeBox
==============================================================================

This function "wakes" up the WibuBox at the selected port by calling the
suitable low-level function. This function also checks also the stability of
the determined T factor by comparing it length between the determined value
of a previous wake up and the value of the new wake up.

Parameters
  iVariant  contains the variant which is waken up.

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

static BOOL WakeBox(INT iVariant)
{
#if !ABSTIME
  UINT uRefT; // (local) reference T value
  INT cWake;  // wake up counter
#endif
  ASSERT_VALUE(plwstSel);
#if ABSTIME
  // simple call of low level wake up function, retry if any error
  cstWkd.idError = (plowdrvSel->pfLowWakeUp)(cstWkd.nSelPort);
  if (cstWkd.idError != WKB_NO_ERROR) {
    // any error => retry operation
    cstWkd.idError = (plowdrvSel->pfLowWakeUp)(cstWkd.nSelPort);
  } // if
#else
  uRefT = plwstSel->uRefT;
  // set flag that first wake up is executed
  plwstSel->fuStatus |= LWST_FIRSTWAKEUP;
  for (cWake = 0; cWake < 5; cWake++) {
    //// try complete and validated WibuBox wakeup
    INT nDelta;
    UINT uNewT1;
    if (uRefT == 0) {
      //// no reference T value determined
      // get default T value from system timer (50 microseconds value)
      PORTADDR ptadr;
      (plowdrvSel->pfLowGetPortData)(cstWkd.nSelPort, &ptadr,
        GLPD_ADDRESS);
      plwstSel->uT = WkdGetSysTime(ptadr);
      // lower limit is <MinT> value
      SetMinT();
    } else {
      // use determined reference value as T value
      plwstSel->uT = uRefT;
    } // if
    // execute wake-up cycle of WibuBox via low-level routine
    if (!WakeLow()) {
      //// wrong wakeup: determine if retry make sense
      if (cstWkd.idError != WKB_ERROR_NOT_READY) {
        // no timeout error: exit loop & abort WakeUp
        break;
      } // if
      if (cWake == 0) {
        // first wakeup at this box:
        // try again with after longer wake up cycle and WibuBox is ready
        continue;
      } // if
      if (!IS_BIT(plwstSel->fuStatus, iVariant)) {
        // this variant was not invalid in the past: check again
        continue;
      } // if
      // WibuBox not ready: exit loop & abort WakeUp
      break;
    } // if
    // new wakeup was successful: analyse determined T value
    ASSERT(cstWkd.idError == WKB_NO_ERROR);
    // get determined T value
    uNewT1 = plwstSel->uT;
    ASSERT(uNewT1 != 0);
    if (uRefT == 0) {
      // no previous T reference determined: set determined T as new reference
      // make second wakeup for validation
      uRefT = uNewT1;
      // error if end of retries
      cstWkd.idError = WKB_ERROR_NOT_READY;
      continue;
    } // if
    // determine <uTRef|uNewT1> delta
    ASSERT(uRefT != 0);
    nDelta = uRefT - uNewT1;
    nDelta = abs(nDelta);
    // <uRefT|uNewT1> delta less 12.5% difference => exit loop (uNewT valid)
    if ((nDelta * 8) / uRefT == 0 || nDelta <= 1) {
      break;
    } // if
    {// <uRefT|uNewT1> delta is bad: get uNewT2 by further wakeup
      UINT uNewT2;
      // set <uRefT> again as start T value
      plwstSel->uT = uRefT;
      if (!WakeLow()) {
         // error in validation wakeup => further retry
        continue;
      } // if
      ASSERT(cstWkd.idError == WKB_NO_ERROR);
      uNewT2 = plwstSel->uT;
      // determine <uRefT|uNewT2> delta
      nDelta = uRefT - uNewT2;
      nDelta = abs(nDelta);
      if ((nDelta * 8) / uRefT != 0 && nDelta > 1) {
        // <uRefT|uNewT2> delta OK: uNewT was wrong, no changes
        break;
      } // if
      // determine <uNewT1|uNewT2> delta
      nDelta = uNewT2 - uNewT1;
      nDelta = abs(nDelta);
      if ((nDelta * 8) / uNewT1 != 0 || nDelta > 1) {
        // * <uTNew|uTNew2> delta also bad: very unstable, further retry
        // error if end of retries
        cstWkd.idError = WKB_ERROR_NOT_READY;
        continue;
      } // if
      // <uNewT1> and <uNewT2> looks better than <uRefT>: set new reference
      uRefT = (uNewT1 + uNewT2 + 1) / 2;
      // this value is now valid
      break;
    } // block
  } // for
#endif // #else ABSTIME
  if (cstWkd.idError == WKB_NO_ERROR) {
    //// no error: this variant is now valid
    CLEAR_BIT(plwstSel->fuStatus, iVariant);
    // this wakeup was not valid
    plwstSel->fuStatus |= LWST_GOODWAKEUP;
  #if !ABSTIME
    if (plwstSel->uRefT != uRefT) {
      // store <uRefT> as new determined reference T
      plwstSel->uRefT = uRefT;
      plwstSel->uT = uRefT;
    } // if
  #endif
  } else {
    //// no WibuBox is available: in future no additional retries
    // set invalidation flag for this variant
    SET_BIT(plwstSel->fuStatus, iVariant);
    // convert error "TimeOut" to error "No WibuBox connected"
    if (cstWkd.idError == WKB_ERROR_NOT_READY) {
      cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
    } // if
  } // if
  // return determined error state
  return cstWkd.idError == WKB_NO_ERROR;
} // WakeBox()


UINT ReleaseSeqPort(void);



/*****************************************************************************
R e l e a s e S e q P o r t
==============================================================================

This function terminates the selection of a port for the WibuBox
addressing. The termination handshake is executed and the port is released.

Return
  The function returns an error code or WKB_NO_ERROR if the operation was
  successful.

Implementation
  The termination handshake is executed and the port is released.

*****************************************************************************/

static UINT ReleaseSeqPort(void)
{
  INT idError;
  if (iSelVariant < 0) {
    // not selected => nothing to do, return without error
    return WKB_NO_ERROR;
  } // if
  ASSERT_VALUE(plowdrvSel);
  ASSERT_VALUE(plwstSel);
  // low level driver must be selected
  ASSERT(plwstSel->fuStatus & LWST_SELECTED);
  // final wait for ready (ignore any error, timeout 100 milliseconds)
  if (cstWkd.idError != WKB_ERROR_NOT_READY &&
      cstWkd.idError != WKB_ERROR_NOT_CONNECTED
    ) {
    (plowdrvSel->pfLowWaitReady)(cstWkd.nSelPort, 1000);
  } // if 
  // release WibuKey-use of specified port, return error code
  idError = (plowdrvSel->pfLowReleasePort)(cstWkd.nSelPort);
  // this port is no longer selected
  plwstSel->fuStatus &= ~LWST_SELECTED;
  // no longer any accessed port
  iSelVariant = -1;
  // return determined error code
  return idError;
} // ReleaseSeqPort()


/*****************************************************************************
W k d A c c e s s S e q P o r t
==============================================================================

This function access the last selected port if this is possible. If so,
the function saves the current status of this port, modifies it
for the WibuKey adressing and sends a wake up cycle to all WIBU-BOXes at
the port.

Parameters
  iVariant  contains the variant for the port access.

Globals
  cstWkd*

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

BOOL WkdAccessSeqPort(INT iVariant)
{
  ASSERT(iVariant >= 0);
  if ( (iSelVariant >= 0) && (iVariant != iSelVariant) ) {
    // changing of selected variant: release port before selection
    ReleaseSeqPort();
    // no variant selected
    iSelVariant = -1;
  } // if
  if (iSelVariant < 0) {
    // currently no selection: select specified port/variant
    ASSERT_VALUE(plowdrvSel);
    ASSERT(iVariant < plowdrvSel->cVariant);
    ASSERT_VALUE(plwstSel);
    // access specified port via selected low level driver
    cstWkd.idError = (plowdrvSel->pfLowAccessPort)(cstWkd.nSelPort,
      iVariant);
    if (cstWkd.idError != WKB_NO_ERROR) {
      return FALSE;
    } // if
    // this port is now selected: set flag & store variant
    plwstSel->fuStatus |= LWST_SELECTED;
    iSelVariant = iVariant;
  } // if
  // wake up selected WibuBox (do multiple retries)
  return WakeBox(iVariant);
} // WkdAccessSeqPort()


/*****************************************************************************
WkdSendSeqByte
==============================================================================

This function sends one byte to the selected WibuKey port in <cstWkd>.

Parameters
  uData  contains the byte which is sent in the range 0..255.

Globals
  cstWkd*

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

BOOL WkdSendSeqByte(BYTE bData)
{
  ASSERT_VALUE(plowdrvSel);
  // send byte to port via selected low level driver
  cstWkd.idError = (plowdrvSel->pfLowSendByte)(cstWkd.nSelPort, bData);
  return cstWkd.idError == WKB_NO_ERROR;
} // WkdSendSeqByte()


/*****************************************************************************
WkdReadSeqByte
==============================================================================

This function reads one byte from the specified WibuKey port.

Parameters
  pbDest  points to a BYTE variable which receives the read value in the
          range 0..255 if no error was read.

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

Comments
  This function waits until a byte value is reached or the timeout interval
  has passed. In the last case an error WKB_ERROR_NOT_READY is returned.

*****************************************************************************/

BOOL WkdReadSeqByte(BYTE *pbDest)
{
  UINT uDest;
  ASSERT_VALUE(plowdrvSel);
  // read byte from port via selected low level driver
  cstWkd.idError = (plowdrvSel->pfLowReadByte)(cstWkd.nSelPort, &uDest);

  if (cstWkd.idError == WKB_NO_ERROR) {
    // no error: store read byte into destination
    *pbDest = (BYTE)uDest;
    return TRUE;
  } // if
  // return with error
  return FALSE;
} // WkdReadSeqByte()


/*****************************************************************************
L o o k A h e a d
==============================================================================

This function calls the low level driver look ahead function of the specified
WibuKey port.

Parameters
  cbRead  contains the number of read in near future before the next send or
          release operation is done.
  fuCtrl  contains the look-ahead flags (LOW_RELEASE or LOW_CLUSTER).

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

static BOOL LookAhead(UINT cbRead, UINT fuCtrl)
{
  ASSERT_VALUE(plowdrvSel);
  // call look-ahead operation
  cstWkd.idError = (plowdrvSel->pfLowLookAhead)(cstWkd.nSelPort,
    cbRead, fuCtrl);
  return (cstWkd.idError == WKB_NO_ERROR);
} // LookAhead()


/*****************************************************************************
S e n d S e q D a t a
==============================================================================

This function sends the data sequence in the specified buffer to the WibuBox
via the specified port. The sequence must be sent after a previously sent
command byte. Before the byte is sent it is waiting for the busy cycle of
the WibuBox.

Parameters
  pbSrcData  points to a buffer which contains the bytes to be sent.
  cbSrcData  contains the number of bytes to be sent.
  pcbSent .. points to a variable which receives the transferred byte. If this
             pointer is NULL, no value is stored.

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

*****************************************************************************/

static BOOL SendSeqData(const BYTE *pbSrcData, UINT cbSrcData,
  UINT *pcbSent)
{
  UINT iByte = 0;

  for (iByte = 0; iByte < cbSrcData; iByte++, pbSrcData++) {
    if (!WkdSendSeqByte(*pbSrcData)) {
      // send next byte, any error => exit loop
      break;
    } // if
  } // for
  if (pcbSent != NULL) {
    // store number of transferred bytes
    *pcbSent = iByte;
  } // if
  // return FALSE if not all data sent
  return iByte == cbSrcData;
} // SendSeqData()


/*****************************************************************************
B U F S P E C                           Buffer specification for data transfer
==============================================================================

This structure is used to transfer data via a buffer with a specific length

Members
-------
cbFree
    contains the number of bytes in the specified buffer.
pbData
    points to the specified buffer.

*****************************************************************************/

typedef struct {
  UINT cbFree;
  BYTE *pbData;
} BUFSPEC; // prefix "bf"


/*****************************************************************************
R e a d S e q D a t a
==============================================================================

This function reads a data sequence with the specified maximum length in the
specified buffer from the WibuBox via the selected port.

Parameters
  pbfspec  points to a buffer specification with a destination buffer read
           address and the buffer length in bytes.
  cbData   contains the number of bytes to be read.

Return
* The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.
* The destination buffer read address and the buffer length in <pbfspec> is
  updated.

*****************************************************************************/

static BOOL ReadSeqData(BUFSPEC *pbfspec, INT cbData)
{
  if ( NULL == pbfspec ) {
    return FALSE;
  } // if 

  for (; cbData > 0; cbData--, pbfspec->cbFree--, pbfspec->pbData++) {
    BYTE bData;
    // return error if buffer is full
    if (pbfspec->cbFree == 0) {
      cstWkd.idError = WKB_ERROR_BUFFER_OVERFLOW;
      return FALSE;
    } // if
    if (!WkdReadSeqByte(&bData)) {
      return FALSE;
    } // if
    *pbfspec->pbData = bData;
  } // for
  // all done: return without error
  return TRUE;
} // ReadSeqData()


/*****************************************************************************
S k i p S e q D a t a
==============================================================================

This function skips the specified number of bytes in the data stream which
is read from the WibuBox via the selected port.

Parameters
  puData  points to an integer variable which receives the read data.
  cbData  contains the number of bytes to be read.

Return
  The function returns the number of not skipped bytes. When the operation
  was sucessful and bytes are skipped, 0 is returned.

*****************************************************************************/

static UINT SkipSeqData(INT cbData)
{
  for (; cbData > 0; cbData--) {
    BYTE bData;
    if (!WkdReadSeqByte(&bData)) {
      return cbData;
    } // if
  } // for
  // all done: return 0
  return 0;
} // SkipSeqData()


/*****************************************************************************
R e a d S e q S t a t u s
==============================================================================

This function reads a status byte from a WibuBox and interprets it as a
valid value or an error value.

Parameters
  pbDest  points to a BYTE variable which receives the read value in the
          range 0..255. This byte is also stored if an error occurred.

Return
  The function returns TRUE if the operation was successful and FALSE if not.
  In the latter case, <cstWkd.idError> contains an error code.

Comments
  The received status byte is also stored if an error occurs. This is required
  because of a bug in the WibuKey ASIC sometimes an hold-error must be
  ignored but the status byte must be analyzed.

*****************************************************************************/

static BOOL ReadSeqStatus(BYTE *pbDest)
{
  BYTE bData = 0;
  BOOL fValid = FALSE;


  ASSERT_VALUE(pbDest);
  if ( NULL == pbDest ) {
    return FALSE;
  } // if

  // read next byte from low-level driver
  fValid = WkdReadSeqByte(&bData);
  // store read value into buffer
  // * store also if read error
  *pbDest = bData;
  if (!fValid) {
    // any error => set into diagnosis and return with error
    return FALSE;
  } else {
    // convert status byte to error code in <cstWkd.idError> if any
    WkdConvertStatusIntoError(bData);
  } // if
  // return error condition
  return (cstWkd.idError == WKB_NO_ERROR);
} // ReadSeqStatus()


/*****************************************************************************
I s V a l i d S e r i a l
==============================================================================

This function checks if the received serial number of a WibuBox is valid.
A valid serial number contains a serial mask byte in the range between 1 and
31 and a serial number which is not 0 or -1.

Parameters
  pab  points to the serial number block to be checked as received from the
       WibuBox. pab[4] points to the mask byte, pab[0..3] to the serial
       number.

Return
  The function returns TRUE if the serial number is valid and FALSE if not.

*****************************************************************************/

static BOOL IsValidSerial(const BYTE *pab)
{
  ULONG ul = 0;
  UINT uDrvCtrl = 0;


  ASSERT_VALUE(plowdrvSel);
  if (NULL == pab) {
    return FALSE;
  } // if 

  // check if verifying of serial number is disabled
  (plowdrvSel->pfLowGetPortData)(cstWkd.nSelPort,
    &uDrvCtrl, GLPD_CTRLFLAGS);
  if (cstWkd.flGlobalInternal & DRVC_EVERYSERIAL) {
    // no checking of serial number (for WKFILL.EXE etc): return TRUE
    return TRUE;
  } // if
  // check mask byte of serial number
  if ((pab[4] < 1) || (pab[4] > 31)) {
    // bad mask value => invalid serial number
    return FALSE;
  } // if
  // check serial number
  // trick: works also with big endian because only 0/-1 is checked
  ul = *(ULONG*)pab;
  // serial number 0 or 0xFFFFFFFF: return with error
  if ((ul == 0) || (ul == UINT_MAX)) {
    return FALSE;
  } // if
  // serial number block is valid
  return TRUE;
} // IsValidSerial()


/*****************************************************************************
C l e a n U p D e v i c e L i s t
==============================================================================

This function cleans up the device table of the specified low level status
structure.

Parameters
  plwst  points to the the low level status structure which device list is
         destroyed.

*****************************************************************************/

static void CleanUpDeviceList(LOWSTATUS *plwst)
{
  DEVOBJECT *pdevobj;
  DEVOBJECT *pdevobjNext;

  if (NULL == plwst ) {
    return;
  } // if 

  for (pdevobj = plwst->pdevobjList; pdevobj != NULL; pdevobj = pdevobjNext) {
    ASSERT_VALUE(pdevobj);
    // save pointer to next element
    pdevobjNext = pdevobj->pNext;
    // destroy current element
    BsrFreeMem(pdevobj);
  } // for
  // list now empty
  plwst->pdevobjList = NULL;
} // CleanUpDeviceList()


/*****************************************************************************
S e q R e s e t
==============================================================================

This function resets the specified port and other class driver specific data.

Parameters
  nPort   contains the number of the port in the range 0..127.
  fuCtrl  contains the reset control flags.

Attributes
  Class driver export function

Implementation
  The function calls the low level driver and determines the minimum T value
  from the profile.

Comments
  It may be that a port may be reset which may be not selected for example
  if it invalidated by the environment. That's why this function manages
  their own port number.

*****************************************************************************/

static void SeqReset(INT nPort, UINT fuCtrl)
{
  LOWDRIVER *plowdrv;
  ASSERT_VALUE(cstWkd.pprtgrpSel);
  ASSERT(nPort >= cstWkd.pprtgrpSel->nFirstPort &&
    nPort <= cstWkd.pprtgrpSel->nLastPort);
  plowdrv = cstWkd.pprtgrpSel->plowdrv;
#if !ABSTIME
  if (fuCtrl & WKB_RESET_TIMING) {
    UINT idError;
    LOWSTATUS *plwst;
    // reset all data in the public low level driver status data
    idError = (plowdrv->pfLowGetPortData)(nPort, &plwst, GLPD_STATUS);
    if (idError == WKB_NO_ERROR) {
      ASSERT_VALUE(plwst);
      // set all LOWSTATUS data to 0
      memset(plwst, 0, sizeof(LOWSTATUS));
    } // if
  } // if
#endif // #if !ABSTIME
  // call low level driver function.
  (plowdrv->pfLowReset)(nPort, fuCtrl);
} // SeqReset()


/*****************************************************************************
S e q S e l e c t
==============================================================================

This function selects the selected port in <cstWkd.nPort> within the
class driver if this is possible.

Parameters
  nPort  contains the number of the port in the range 0..127.

Return
  The function returns TRUE if the port is available and can be selected.
  and returns FALSE if not.
  In the latter case, the last error code is set to WKB_ERROR_INVALID_PORT;

Attributes
  Class driver export function

Comments
  This function does not address the physcial hardware and is really fast.

*****************************************************************************/

static BOOL SeqSelect(INT nPort)
{
  ASSERT_VALUE(cstWkd.pprtgrpSel);
  plowdrvSel = cstWkd.pprtgrpSel->plowdrv;
  ASSERT_VALUE(plowdrvSel);
  ASSERT(nPort >= cstWkd.pprtgrpSel->nFirstPort &&
    nPort <= cstWkd.pprtgrpSel->nLastPort);
  // check if port can be selected and return address to low level status
  cstWkd.idError = (plowdrvSel->pfLowGetPortData)(nPort,
    &plwstSel, GLPD_STATUS);
  // set flag "this port is not accessed"
  iSelVariant = -1;
  if (cstWkd.idError != WKB_NO_ERROR) {
    // any error => return FALSE
    return FALSE;
  } // if
  ASSERT_VALUE(plwstSel);
  // return without error
  return TRUE;
} // SeqSelect()


/*****************************************************************************
S e q R e l e a s e
==============================================================================

This function terminates the last accessed port of the class. It may be called
for non-accessed ports too.

Return
  The function returns the error code of the release operation. It is set to
  WKB_NO_ERROR (0) if no error occurred.

Attributes
  Class driver export function

Implementation
  The termination handshake is executed and the port is released.

*****************************************************************************/

static UINT SeqRelease(void)
{
  if (iSelVariant >= 0) {
    // port is properly selected => clean up built device table (if any)
    // * to save memory for drivers with not automatic device-state detection
    ASSERT_VALUE(plwstSel);
    CleanUpDeviceList(plwstSel);
  } // if
  // release selected port if any & return error code
  return ReleaseSeqPort();
} // SeqRelease()


/*****************************************************************************
S e q G e t D e v i c e L i s t
==============================================================================

This function returns a pointer to a list of all available serial numbers
at the last selected port.

Parameters
  ppdevobj  points to a variable which receives the first element of a list
            of found WibuBox devices.

Return
  The function returns the number of found devices or 0 if no device was
  found. In the latter case, the last error code is set to
  WKB_ERROR_NOT_CONNECTED or to another serious error code.

Attributes
  Class driver export function

*****************************************************************************/

static UINT SeqGetDeviceList(DEVOBJECT **ppdevobj)
{
  UINT cDevice = 0;

  // check if specified port is selected
  ASSERT_VALUE(cstWkd.pprtgrpSel);
  ASSERT_VALUE(plowdrvSel);
  ASSERT_VALUE(plwstSel);
  // free last read device table (if any)
  CleanUpDeviceList(plwstSel);
  // now no devices exist
  cDevice = 0;
  { // read all WibuBox devices from selected port
    DEVOBJECT **ppdevobjAppend;
    BYTE bStatus;  // returned status of box communication
    UINT idCmd;    // current list subcommand
    INT  cVariant; // number of available port addressing variants
    INT  iVariant; // current port addressing variant
    BOOL fValid;   // any read operation was true
    // set append position to begin of list
    ppdevobjAppend = &plwstSel->pdevobjList;
    // first command is FSEARCH
    idCmd = BOXCMD_FSEARCH;
    // select low level driver to determine number of variants
    // * read number of variants and start search operation
    cVariant = plowdrvSel->cVariant;
    fValid = FALSE;
    iVariant = 0;
    for (;;) {
      // execute next partial search command
      BYTE abSerial[5];
      BUFSPEC bfSerial;
      bfSerial.cbFree = sizeof(abSerial);
      bfSerial.pbData = abSerial;
      // new port access is required to guarantee stable signal state
      if (!WkdAccessSeqPort(iVariant)) {
        // bad selection: check next variant
        // * no error if no WibuBox found at this variant
        if (cstWkd.idError == WKB_ERROR_NOT_CONNECTED) {
          fValid = TRUE;
        } // if
        goto NEXT_VARIANT;
      } // if
      // send command byte to WibuBox (FSEARCH or NSEARCH)
      if (!WkdSendSeqByte((BYTE)idCmd)) {
        // any ignore => ignore this variant
        goto NEXT_VARIANT;
      } // if
      // set look-ahead state for reading & release
      if (!LookAhead(6, 0)) {
        goto NEXT_VARIANT;
      } // if
        // read data (if any)
        if (!ReadSeqData(&bfSerial, sizeof(abSerial))) {
          goto NEXT_VARIANT;
        } // if
        // read final status of executed command
        if (!ReadSeqStatus(&bStatus)) {
          goto NEXT_VARIANT;
        } // if
        // accept only proper status value 0 or 1
        // * higher values are coming from other dongles!
        if (bStatus < 2 && IsValidSerial(abSerial)) {
          // new WibuBox detected: copy serial number into buffer
          DEVOBJECT *pdevobjNew;
          WKBSERIAL wkbsr;
          // * one WibuBox read more
          cDevice++;
          // fill serial number structure
          wkbsr.ulSerial = GET_UINT32_LE(abSerial);
          // set mask value
          wkbsr.usMask = abSerial[4];
          wkbsr.usReserved = 0;
          // create new device entry
          pdevobjNew = WkdAllocMem(sizeof(DEVOBJECT));
          if (pdevobjNew == NULL) {
            // memory too small => return 0, exit loop
            // * cleanup built device table
            CleanUpDeviceList(plwstSel);
            cDevice = 0;
            goto EXIT_BUILD_LIST;
          } // if
          pdevobjNew->wkbsr = wkbsr;
          // handle to device here not used
          pdevobjNew->wkdevobj.hwkdev = NULL;
          // set preliminary end of list
          // * required when partially list is destroyed (memory too small)
          pdevobjNew->pNext = NULL;
          // append list at end of list
          *ppdevobjAppend = pdevobjNew;
          ppdevobjAppend = &pdevobjNew->pNext;
          // any port variant read without error => no sum error
          fValid = TRUE;
          if (bStatus == 0) {
            // further WibuBox follows: next command is NSEARCH
            idCmd = BOXCMD_NSEARCH;
            continue;
          } // if
        } // if
      NEXT_VARIANT:
      // no further WibuBox in this variant
      // * continue with next addressing variant (if available)
      // * start at next variant with FSEARCH command again
      iVariant++;
      if (iVariant >= cVariant) {
        ASSERT(iVariant == cVariant);
        // all variants scanned: exit loop
        break;
      } // if
      // next variant: start WibuBox search operation again
      idCmd = BOXCMD_FSEARCH;
    } // for
    // all variants scanned: return with from function with <cbData>
    ReleaseSeqPort();
  EXIT_BUILD_LIST:
    // end of of list must be set
    ASSERT(*ppdevobjAppend == NULL);
    if (fValid) {
      // at least one device variant was read properly
      if (cDevice > 0) {
        // any devices found => no error
        cstWkd.idError = WKB_NO_ERROR;
      } else {
        // no devices found but no wrong device state: set special error
        cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
      } // if
    } // if
  } // block
  // any error code must be set if no devices found
  ASSERT(cDevice != 0 || cstWkd.idError != WKB_NO_ERROR);
  // set pointer of first element of created list
  *ppdevobj = plwstSel->pdevobjList;

  // return number of found devices
  return cDevice;
} // SeqGetDeviceList()


/*****************************************************************************
S e q G e t D e v i c e F r o m S e r i a l
==============================================================================

This function searches the specified serial number in the list of read serial
numbers at the last selected port and returns a pointer to the corresponding
entry.

Parameters
  pwkbsr  points to the serial number of the device. It is assumed that the
          serial number is valid. An exception is setting this pointer to
          NULL, which is used for modifying a serial number of WIBU-BOXes
          with version 1.

Return
  The function returns a reference to the WibuKey device object or NULL if
  the device is not found.

Attributes
  Class driver export function

*****************************************************************************/

static DEVOBJECT * SeqGetDeviceFromSerial(const WKBSERIAL *pwkbsr)
{
  if (pwkbsr == NULL) {
    // no serial number set
    // * used to modify a serial number of WibuBox version 1
    // * set us mask value to 0
    devobj.wkbsr.usMask = 0;
  } else if ((pwkbsr->usMask == 0 || pwkbsr->usMask > 255)
    && (cstWkd.flGlobalInternal & DRVC_EVERYSERIAL) == 0) {
    // only byte storable value possible: return error
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    return NULL;
  } else {
    // set serial number into device object
    devobj.wkbsr = *pwkbsr;
  } // if
  
  // return address of this device object
  return &devobj;
} // SeqGetDeviceFromSerial()


/*****************************************************************************
S e q F i n d D e v i c e
==============================================================================

This function searches a WibuBox at the last selected port with an entry
which matches the specified Firm and User Code and WibuBox version.

Parameters
  pope  points to the open entry structure. This is partially initialized:
        * <flCtrl> defines all flags except WKB_OE_MASTER which is reset.
        * <hwkbsys> contains the handle of the subsystem.
        * <lPort> contains the port number.
        * <pszPort> is undefined.
        * <ulFirmCode> contains the required Firm Code
        * <ulUserCode> contains the required Firm Code
        These entry must not be changed by this function except
        the WKB_OE_MASTER flag.

Return
  The function returns TRUE if the WibuBox was found and FALSE if not. If
  the WibuBox entry was found, its serial number and entry index is stored
  into the structure.

Attributes
  Class driver export function

*****************************************************************************/

static BOOL SeqFindDevice(OPENENTRY *pope)
{
  INT iVariant = 0;   // current search variant
  INT cVariant = 0;   // number of variants for selected port
  BYTE abEntry[6];    // received serial number + entry index from WibuBox
  BYTE bVersion = 0;  // determined version of the WibuBox
  BUFSPEC bfEntry;    // WibuBox read buffer structure
  WKBSERIAL wkbsr;    // serial number of found WibuBox
  // read number of variants and start search operation
  ASSERT_VALUE(plowdrvSel);
  cVariant = plowdrvSel->cVariant;
  // port is determined
  pope->lPort = cstWkd.nSelPort;
  for (iVariant = 0; iVariant < cVariant; iVariant++) {
    BYTE abBuf[6];
    BYTE bStatus;
    // access port with current variant
    if (!WkdAccessSeqPort(iVariant)) {
      // any error => try next variant
      goto CONT_VARIANT;
    } // if
    // send command byte (find FC/UC, code 1)
    if (!WkdSendSeqByte(BOXCMD_FIND)) {
      goto CONT_ERROR;
    } // if
    // send Firm and User code to WibuBox(es) at <nPort>
    abBuf[0] = LBYTE32(pope->ulFirmCode);
    abBuf[1] = MBYTE32(pope->ulFirmCode);
    abBuf[2] = NBYTE32(pope->ulFirmCode);
    abBuf[3] = LBYTE32(pope->ulUserCode);
    abBuf[4] = MBYTE32(pope->ulUserCode);
    abBuf[5] = NBYTE32(pope->ulUserCode);
    if (!SendSeqData(abBuf, 6, NULL)) {
      goto CONT_ERROR;
    } // if
    // set look-ahead state for reading & release
    if (!LookAhead(8, LOW_RELEASE)) {
      goto CONT_ERROR;
    } // if
    // read status from WibuBox
    if (!ReadSeqStatus(&bStatus)) {
      goto CONT_ERROR;
    } // if
    // WibuBox entry found: read serial number & index
    bfEntry.cbFree = sizeof(abEntry);
    bfEntry.pbData = abEntry;
    if (!ReadSeqData(&bfEntry, sizeof(abEntry))) {
      goto CONT_ERROR;
    } // if
    // check if serial number is valid, error => ignore WibuBox
    if (!IsValidSerial(abEntry)) {
      goto CONT_VARIANT;
    } // if
    // set default "version 1"
    bVersion = 1;
    if (abEntry[4] > 1) {
      // check if further byte follows (only WibuBox with version > 1)
      if (!WkdReadSeqByte(&bVersion)) {
        goto CONT_ERROR;
      } // if
      // check if true (!) ready state - an imaginary byte could be read!
      // the delay is 20 ms
      if ((plowdrvSel->pfLowWaitReady)(cstWkd.nSelPort, 200) != WKB_NO_ERROR) {
        goto CONT_ERROR;
      } // if
      // version byte read: validate & store it
      // * invalid byte value: assume that Box not found
      if (bVersion < 0xE0) {
        goto CONT_ERROR;
      } // if
      if (!(bVersion & 0x10)) {
        // opened via master entry: set flag
        pope->flCtrl |= WKB_OE_MASTER;
      } // if
      // calculate version value (0xXF = 2, 0xXE = 3)
      bVersion = 0x11 - (bVersion & 0x0F);
    } // if
    // WibuBox found => return without error
    cstWkd.idError = WKB_NO_ERROR;
    // exit loop
    goto EXIT_VARIANT;
  CONT_ERROR:
    // exception handling
    ASSERT(cstWkd.idError != WKB_NO_ERROR);
    // this port variant not suitable => test next
  CONT_VARIANT:;
  } // for
EXIT_VARIANT:
  // release selected WibuBox port
  ReleaseSeqPort();

  // any error => return
  if (cstWkd.idError != WKB_NO_ERROR) {
    return FALSE;
  } // if

  // get serial number & WibuBox entry
  wkbsr.ulSerial = GET_UINT32_LE(abEntry);
  wkbsr.usMask = abEntry[4];
  // clear reserved data (required to transfer serial number to diagnosis)
  wkbsr.usReserved = 0;

  // check if version of WibuBox is OK
  if (!WkdCheckBoxVersion((ULONG*)&(pope->flCtrl), (USHORT)bVersion, &wkbsr, TRUE)) {
    // invalid version of the found WibuBox => return with error
    cstWkd.idError = WKB_ERROR_INVALID_VERSION;
    return FALSE;
  } // if

  // open operation was successful
  // * set read data into open entry structure
  pope->wkbsr = wkbsr;
  pope->iEntry = abEntry[5];
  pope->usBoxVersion = bVersion;

  // return without error
  return TRUE;
} // SeqFindDevice()


/*****************************************************************************
S e q R e q u e s t
==============================================================================

This function returns a pointer to a list of all available serial numbers
at the specified port.

Parameters
  pwkdevreq  points to the request block which contains the request header
             and an optional request byte structure.
  pabDest .. points to a memory block which receives the read data.
  preqopt .. points to the request option structure or is NULL if no such
             structure is required. This structure contains or receives
             additional members which are request type specific. Please note
             that the BOXCMD_PROGRAM and BOXCMD_CRYPT commands always assume
             a non-NULL valid pointer.

Return
  The function returns TRUE if it was successful and FALSE if not.

Attributes
  Class driver export function

*****************************************************************************/

static BOOL SeqRequest(WKDEV_REQUEST *pwkdevreq, BYTE *pabDest,
  REQOPTIONS *preqopt)
{
  UINT uCmd = pwkdevreq->usCmd;
  UINT cbRequest = pwkdevreq->cbRequest;
  UINT cbAnswer = pwkdevreq->cbAnswer;
  UINT uEntry = pwkdevreq->usIndex;
  INT iUsedVariant = 0;
  const BYTE *pbRequest = pwkdevreq->abRequest;
  // pointer to current destination position
  BYTE *pbDest = pabDest;
  // verify value of entry
  if (uEntry > 9) {
    // invalid entry: return with error
    cstWkd.idError = WKB_ERROR_INVALID_ENTRY;
    return FALSE;
  } // if
  if ((devobj.wkbsr.usMask != 0)
    || (cstWkd.flGlobalInternal & DRVC_EVERYSERIAL)) {
    // select WibuBox by serial number (check all variants if required)
    BYTE abData[7];
    INT iVariant;
    INT cVariant;
    UINT uSent; // number of sent bytes
    ASSERT_VALUE(plowdrvSel);
    // determine variant search area
    if (iSelVariant >= 0) {
      // variant determined (from previous call): use only this once
      iVariant = iSelVariant;
      cVariant = iSelVariant + 1;
    } else {
      // no variant determined: search all
      iVariant = 0;
      cVariant = plowdrvSel->cVariant;
    } // if
    // create 6 bytes of command code & serial number
    abData[0] = (BYTE)uCmd;
    SET_UINT32_LE(abData + 1, devobj.wkbsr.ulSerial);
    abData[5] = (BYTE)devobj.wkbsr.usMask;
    abData[6] = (BYTE)uEntry;
    for (; iVariant < cVariant; iVariant++) {
      uSent = 0;
      iUsedVariant = iVariant;
      // access specified WibuBox port, return if error
      if (!WkdAccessSeqPort(iVariant)) {
        goto NEXT_VARIANT;
      } // if
      // send command code, serial number, mask value, entry or control code
      if (!SendSeqData(abData, 7, &uSent)) {
        goto NEXT_VARIANT;
      } // if
      // set look-ahead state for reading
      if (!LookAhead(1, 0)) {
        goto NEXT_VARIANT;
      } // if
      // read next byte as first destination byte, return if no error
      if (ReadSeqStatus(pbDest)) {
        goto EXIT_SELECT;
      } // if
    NEXT_VARIANT:;
    } // for
  EXIT_SELECT:
    // all port variants scanned
    // * return with error state
    if (cstWkd.idError != WKB_NO_ERROR) {
    // create diagnostic event
      return FALSE;
    } // if
    // one byte less to read
    ASSERT(cbAnswer > 0);
    cbAnswer--;
    pbDest++;
  } else {
    // no serial number specified
    // * used to program new serial number of WibuBox version 1
    // * access port only without sending serial number
    // * use highest variant (= unidirectional mode for WibuBox 1)
    if (!WkdAccessSeqPort(plowdrvSel->cVariant - 1)) {
      // port selection not valid => return with error
      return FALSE;
    } // if
  } // if
  // execute rest of request transfer in dependence of command
  switch (uCmd) {
    case BOXCMD_CRYPT: {
      UINT cbSend;
      UINT cbSent = 0;
      const BYTE *pbSrc;
      BYTE *pbDest;
      UINT cbCluster = 0; // length of one cluster
      UINT cbInCluster;   // length of last cluster
      // calculate data length from request data length
      // * length at pbRequest[0] already overwritten by first status byte
      // * (pabDest and pbRequest may point to same memory block)
      ASSERT(cbRequest >= 1);
      cbSend = cbRequest - 1;
      ASSERT(cbSend <= 256);
      // 0 bytes to encrypt: nothing to do (but WibuBox entry was checked)
      if (cbSend != 0) {
        // one or more byte to encrypt: send length byte to WibuBox
        // * 256 is represented as 0 byte
        cbCluster = plowdrvSel->cbCluster;
        // send length byte to WibuBox: 256 <-> 0
        if (!WkdSendSeqByte(LOBYTE(cbSend))) {
          // any error => abort
          goto EXIT_CRYPT;
        } // if
      } // if
      // transfer source bytes to WibuBox, store encrypted bytes
      pbSrc = pbRequest + 1;
      pbDest = pabDest + 1;
      for (cbSent = 0; cbSent < cbSend; cbSent += cbInCluster) {
        UINT i;
        UINT uLookAhead = LOW_CLUSTER;
        UINT cbRest = cbSend - cbSent;
        cbInCluster = cbCluster;
        if (cbInCluster >= cbRest) {
          // end of current command: reduce size, set release flag
          cbInCluster = cbRest;
          uLookAhead |= LOW_RELEASE;
        } // if
        if (!LookAhead(cbInCluster, uLookAhead)) {
          goto EXIT_CRYPT;
        } // if
        // send source bytes to WibuBox within cluster
        for (i = 0; i < cbInCluster; i++, pbSrc++) {
          if (!WkdSendSeqByte(*pbSrc)) {
            goto EXIT_CRYPT;
          } // if
        } // for
        // read encrypted bytes from WibuBox within cluster
        for (i = 0; i < cbInCluster; i++, pbDest++) {
          if (!WkdReadSeqByte(pbDest)) {
            goto EXIT_CRYPT;
          } // if
        } // for
      } // for
    EXIT_CRYPT:
      // store number of properly encrypted bytes into request option block
      ASSERT_VALUE(preqopt);
      preqopt->Crypt.cbEncrypted = cbSent;
      // release access to WibuBox (required for proper next wakeup)
      // * next SeqReset/BOXCMD_CRYPT call when data size > 256 bytes
      ReleaseSeqPort();
      if (cstWkd.idError != WKB_NO_ERROR) {
        // error during encryption: return with error
        // return with error
        return FALSE;
      } // if
      // return without error
      return TRUE;
    } // case
    case BOXCMD_LIST: {
      // ============================================
      // read WibuBox contents or configuration data
      // ============================================
      BOOL fConfig = uEntry == 1;
      INT nVersion = pabDest[0];
      BUFSPEC bfData;
      // set pointer to destination buffer
      bfData.pbData = pbDest;
      if (preqopt != NULL) {
        // =====================================
        // read only contents of specified entry
        // =====================================
        INT iEntry;
        INT cbSkipToBase;
        UINT cbRmd;
        UINT cbRead;
        BOOL fValid;
        INT iAddedRead;
        // determine entry to read, reset read byte counter
        ASSERT_VALUE(preqopt);
        iEntry = preqopt->List.iEntry;
        cbRead = 0;
        // determine number of bytes to skip
        cbSkipToBase = iEntry * 6;
        if (iEntry < 5 && nVersion >= 2) {
          // calculate position of added entry (if available)
          iAddedRead = (iEntry + 5) * 6;
        } else {
          // read only one entry
          iAddedRead = cbSkipToBase;
        } // if
        // set look-ahead state for reading & release
        if (!LookAhead(iAddedRead + 6, LOW_RELEASE)) {
          goto EXIT_READ_ENTRY;
        } // if
        // skip data until base data
        cbRmd = SkipSeqData(cbSkipToBase);
        cbRead += cbSkipToBase - cbRmd;
        if (cbRmd != 0) {
          // not all data skipped: error
          goto EXIT_READ_ENTRY;
        } // if
        // get contents of base entry
        bfData.cbFree = 6;
        fValid = ReadSeqData(&bfData, 6);
        cbRead += 6 - bfData.cbFree;
        if (!fValid) {
          goto EXIT_READ_ENTRY;
        } // if
        if (cbSkipToBase != iAddedRead) {
          // get contents of added entry
          BOOL fValid;
          bfData.cbFree = 6;
          cbRmd = SkipSeqData(5 * 6 - 6);
          cbRead += 5 * 6 - 6 - cbRmd;
          if (cbRmd != 0) {
            // not all data skipped: error
            goto EXIT_READ_ENTRY;
          } // if
          fValid = ReadSeqData(&bfData, 6);
          cbRead += 6 - bfData.cbFree;
          if (!fValid) {
            goto EXIT_READ_ENTRY;
          } // if
        } // if
        if (cbRead < 6 * 10) {
          // not all box data were read => terminate command
          // * this is required especially for bidirectional mode
          // * simulate by new wakeup command
          WakeBox(iSelVariant);
        } // if
      EXIT_READ_ENTRY:
        if (cstWkd.idError != WKB_NO_ERROR) {
          return FALSE;
        } // if
      } else {
        // ================================================
        // read complete WibuBox contents of configuration
        // ================================================
        if (fConfig && nVersion <= 1) {
          // config command not supported for WibuBox with version 1
          if (nVersion <= 1) {
            // proper terminating of WibuBox requires new WakeUp by access
            WkdAccessSeqPort(iUsedVariant);
            // return with error
            // * WKLIB expects WKB_ERROR_OPTION_NOT_SUPPORTED: don't change
            cstWkd.idError = WKB_ERROR_OPTION_NOT_SUPPORTED;
            return FALSE;
          } // if
        } // if
        // read packed WibuBox data/configuration into destination buffer
        bfData.cbFree = cbAnswer;
        if (!ReadSeqData(&bfData, cbAnswer)) {
          // error in reading data: return with error
          return FALSE;
        } // if
      }
      return TRUE;
    } // case
    case BOXCMD_PROGRAM: {
      // ====================================================================
      // send WibuBox entry, configuration or serial number data to WibuBox
      // ====================================================================
      UINT cWait = 1;
      switch ((BYTE) *pbRequest) {
        case 0xFD:
          // wait max. 30 sec
          cWait = 300000;
          break;
        case 0xFE:
          // wait max. 1,4 sec
          cWait = 14000;
          break;
      } // switch
    #if TEST_DEBUG
      printf("Enter Program (cWait = %d/Control Code = %d)\n", cWait, *pbRequest);
    #endif
      while (cbRequest > 0) {
        if (!WkdSendSeqByte(*pbRequest++)) {
          // any error => abort sending
          goto EXIT_SEND;
        } // if
        // add extra delay of 300 ms for ready wait
        (plowdrvSel->pfLowWaitReady)(cstWkd.nSelPort, 3000);
        cbRequest--;
      } // while
      // set look-ahead state for reading & release
      if (!LookAhead(1, LOW_RELEASE)) {
        goto EXIT_SEND;
      } // if
      // read next byte as final status data into next destination byte
      //!I:as for special command CtrlCode = 0xfe & 0xfd wait longer (see above)
      (plowdrvSel->pfLowWaitReady)(cstWkd.nSelPort, cWait);
      if (!ReadSeqStatus(pbDest)) {
        // any error: check if must be ignored
        // * for this the read byte is checked independent of timeout error
      #if TEST_DEBUG
        printf("ReadSeqStatus failed (Error = %d)\n", cstWkd.idError);
      #endif
        if (preqopt != NULL) {
          ASSERT_VALUE(preqopt);
          if (preqopt->Program.fNoStatusHoldChk && *pbDest == 0 &&
              cstWkd.idError == WKB_ERROR_NOT_READY
            ) {
            // special condition (see description of <fNoStatusHoldChk>)
            cstWkd.idError = WKB_NO_ERROR;
          } // if
        } // if
      } // if

      // secial case for writing ExtMem storage ==>> read 32 bytes
      cbAnswer--;
      if (cbAnswer > 0) {
        BUFSPEC bfData;
        pbDest++;
        // set pointer to destination buffer
        bfData.pbData = pbDest;
        bfData.cbFree = cbAnswer;
        if (!ReadSeqData(&bfData, cbAnswer)) {
          goto EXIT_SEND;
        } // if
      } // if
      cbAnswer--;


      // return from function with <idResult>
    EXIT_SEND:
      // reselect all variants after changing the configuration flags
      if (pwkdevreq->abRequest[0] == 5) {
        // configuration change: is updated not before next wakeup
        INT iVariant;
        INT cVariant;
        INT idError;
        // get number of variants
        cVariant = plowdrvSel->cVariant;
        // save current error for later restore
        idError = cstWkd.idError;
        // * required for:
        //   WibuBox/P: bidirectional/unidirectional access
        //   WibuBox/M: attribute memory protection
        // * force change to hardware by WakeUp in all variants now
        // execute release operation
        ReleaseSeqPort();
        for (iVariant = 0; iVariant < cVariant; iVariant++) {
          // access specified WibuBox port, ignore any error
          WkdAccessSeqPort(iVariant);
        } // for
        // final release operation done via WkdReleasePort()
        // * restore error code (WkdAccessSeqPort() sets any errors)
        cstWkd.idError = idError;
      } // if
      // return without error
      return TRUE;
    } // case
    case BOXCMD_SELECT1: {
      // ===============================================================
      // select WibuBox entry at currently selected WibuBox, version 1
      // ===============================================================
      // * send two bytes of Selection Code to WibuBox
      if (!SendSeqData(pbRequest, 2, NULL)) {
        // error in sending Selection Code => return with error
        break;
      } // if
      // lock ahead for BOXCMD_SELECT2 status byte
      if (!LookAhead(0, LOW_RELEASE)) {
        break;
      } // if
      // return without error
      return TRUE;
    } // case
    case BOXCMD_SELECT2: {
      // ===================================================================
      // select WibuBox entry at currently selected WibuBox, version 2/3/4
      // ===================================================================
      // * further argument of WibuBox: generator 1-3 selection & User Code
      // save the control flag
      BOOL fSlctVersion2 = pbRequest[0];
      if (!SendSeqData(pbRequest, 4, NULL)) {
        // error in sending control flags and User Code => return with error
        break;
      } // if
      // set look-ahead state for reading
      if (!LookAhead(1, 0)) {
        break;
      } // if
      // read next status byte into destination
      if (!ReadSeqStatus(pbDest)) {
        // error in reading second status byte => return with error
        break;
      } // if
      // one byte read more
      ASSERT(cbAnswer > 0);
      cbAnswer--;
      pbDest++;
      // send four/two bytes of Selection Code to WibuBox
      ASSERT(cbRequest == 8 || cbRequest == 6);
      if (!SendSeqData(pbRequest + 4, cbRequest - 4, NULL)) {
        // error in sending Selection Code => return with error
        break;
      } // if
      if (fSlctVersion2) {
        // read status byte 3 from WibuBox (algo. 2/3 have 3 status codes!!!)
        // * set look-ahead state for reading & release
        if (!LookAhead(1, LOW_RELEASE)) {
          break;
        } // if
        // add extra delay of 1 seconds for ready wait
        (plowdrvSel->pfLowWaitReady)(cstWkd.nSelPort, 10000);
        ASSERT(cbAnswer > 0);
        if (!ReadSeqStatus(pbDest)) {
          // any error => return with error
          break;
        } // if
      } else {
        // look ahead for BOXCMD_SELECT2 status byte
        if (!LookAhead(0, LOW_RELEASE)) {
          break;
        } // if
      } // if
      // return without error
      return TRUE;
    } // case
    case BOXCMD_RESET:
    case BOXCMD_FIND:
    default:
      ASSERT(0);
      break;
  } // switch
  // error return
  ASSERT(cstWkd.idError != WKB_NO_ERROR);
  return FALSE;
} // SeqRequest()


/*****************************************************************************
S e q P o r t D a t a
==============================================================================

This function reads or updates the port control data structure.

Parameters
  pwkbprt  points to a WKBPORT structure which contains the source data
           (fSet = FALSE) receives the destinatino data (fSet = TRUE).
  fSet ... contains the flag "set data instead of get data".

Return
  The function returns TRUE if it was successful and FALSE if not. In the
  latter case, <cstWkd.idError> contains an error code.

Attributes
  Class driver export function

Implementaiton
  The sequential class driver calls the low level driver for reading and
  updating the port information.

*****************************************************************************/

static BOOL SeqPortData(WKBPORT *pwkbprt, BOOL fSet)
{
  ASSERT(fSet == TRUE || fSet == FALSE);
  ASSERT_VALUE(pwkbprt);
  ASSERT_VALUE(plowdrvSel);
  if (fSet) {
    // update: support is optional
    // * check if function is supported
    if (plowdrvSel->pfLowSetup == NULL) {
      // not supported => return with error
      cstWkd.idError = WKB_ERROR_INVALID_FUNCTION;
    } else {
      // execute low level function for updating port data
      cstWkd.idError = (plowdrvSel->pfLowSetup)(cstWkd.nSelPort, pwkbprt);
    } // if
  } else {
    // execute low level function for reading port data
    // * function must be defined
    ASSERT_VALUE(plowdrvSel->pfLowGetPortData);
    cstWkd.idError = (plowdrvSel->pfLowGetPortData)(cstWkd.nSelPort,
        pwkbprt, GLPD_WKBPORT);
  } // if
  // return error status
  return (cstWkd.idError == WKB_NO_ERROR);
} // SeqPortData()


/*****************************************************************************
S e q C l e a n U p
==============================================================================

This function cleans up the class driver before the complete driver is
terminated.

Parameters
  nPort  contains the port which is cleaned up.

Attributes
  Class driver export function

Implementation
  The function does nothing.

*****************************************************************************/

static void SeqCleanUp(INT nPort)
{
  // currently nothing to do
  // * the built device list is already destroyed in SeqRelease()
} // SeqCleanUp()


// ====================================================
// sequential devices class driver definition structure
// ====================================================

// the NULL addresses of the functions is only preliminary

CLASSDRIVER clsdrvWkdSeq = {
  SeqReset,
  SeqSelect,
  SeqRelease,
  SeqGetDeviceList,
  SeqGetDeviceFromSerial,
  SeqFindDevice,
  SeqRequest,
  SeqPortData,
  SeqCleanUp,
  256
}; // clsdrvWkdSeq


// ===================================
// end of keseq.c
// ===================================
