/*****************************************************************************
D r v C                                  Portable WibuKey Driver, Written in C
==============================================================================
kesrl.c                                   Middle-Level Driver for WIBU-BOX /ST
==============================================================================

This module implements the specific middle level driver parts for the Serial
interface of the WIBU-BOX/ST.

This module is only implemented in WKDOS.LIB, WIBUKEY.VXD, WIBUKEY.SYS

==============================================================================

The portable WibuKey driver source is used under DOS and Windows to build
the internally used driver WKDOS2.LIB and the shipped drivers WKWIN.DLL,
WKWIN32.DLL, WIBUKEY.SYS and WIBUKEY.VXD. It is also used to build the
Macintosh and Netware NLM drivers.

Furthermore, it is also the origin of the SRCDRV project. All sources are
written in ANSI C, except the using of C++ style single line comments.

==============================================================================

Copyright  1998-2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/c
Author ... Marcellus Buchheit (mabu), Alexander Schmitt (as),
           Markus Neumann (mn), Heiko Abraham (ha)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
93Mar13 1.33 mabu: Version 1.33 released
93Mar14 1.34 mabu: Start of implementing the serial commmuncation
93Jun24 1.35 mabu: Adaptation of LptGetPortData() to WkbQueryPort()
93Aug02 1.35 mabu: Structure of PORTDATA changed
93Aug28 2.00 mabu: Adaptation to WibuKey version 2
93Sep10 2.00 mabu: Port name added
94Feb16 2.00 mabu: Adding of configuration name string
94Sep16 2.00 mabu: Adaptation to buffered communication like WibuBox/A
95Dec22 2.10 mabu: Renamed from DRVC.C
96Jun24 2.13 mabu: Renamed from LDSRL.C
98Aug20 2.51 as:   Completely new design (Lowlevel to Middellevel)
00May12 2.52 mn:   Impl. quasi-'timeout' with usleep() to RecvDirect()
00Aug04 3.00 mn:   WkdQueryEnvCom(), AccesCom(), changed RecvDirect
04Dec14 3.01 ha:   fix fopen deadlock on AIX
05Apr15 3.02 ha:   some fixes for HP-UX, thanks to Jan Kellmer, synaps-ing
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>

#if HPUX_PA
  #include <sys/termiox.h>
  #include <sys/modem.h>
#endif
#if LINUX || defined(__APPLE__)
  #include <termios.h>
#endif

#if defined(AIX)
  #ifndef CRTSCTS
    #define CRTSCTS 0 /* for AIX 5 */
  #endif
#endif 

/* IO-Adaptation */
#ifndef IUCLC
  // #define IUCLC    0001000
  #define IUCLC    0000000
#endif

#ifndef BSDLY
  // #define BSDLY    0020000
  #define BSDLY 000000
#endif

#ifndef CRDLY
  // #define CRDLY    0003000
  #define CRDLY    0000000
#endif

#ifndef FFDLY
  // #define FFDLY    0100000
  #define FFDLY    0000000
#endif

#ifndef NLDLY
  // #define NLDLY    0000400
  #define NLDLY    0000000
#endif

#ifndef OCRNL
  // #define OCRNL    0000010
  #define OCRNL    0000000
#endif

#ifndef OFDEL
  // #define OFDEL    0000200
  #define OFDEL    0000000
#endif

#ifndef OFILL
  // #define OFILL    0000100
  #define OFILL    0000000
#endif

#ifndef OLCUC
  // #define OLCUC    0000002
  #define OLCUC    0000000
#endif

#ifndef ONLRET
  // #define ONLRET   0000040
  #define ONLRET   0000000
#endif

#ifndef TABDLY
  // #define TABDLY   0014000
  #define TABDLY   0000000
#endif

#ifndef VTDLY
  // #define VTDLY    0040000
  #define VTDLY    0000000
#endif

#ifndef XCASE
  // #define XCASE    0000004
  #define XCASE    0000000
#endif

#ifndef CBAUD
  // #define CBAUD    0010017
  #define CBAUD    0000000
#endif

#ifndef ONOCR
  #define ONOCR    0000000
  // #define ONOCR    0000020
#endif

// first & last serial port number
#define N_MINPORT       16
#define N_MAXPORT       (N_MINPORT + C_SERIAL_PORT - 1)

// in- & output range [48, 111]
#define LOWER_RANGE     48    // ASCII = '0'
#define UPPER_RANGE     111   // ASCII = 'o'

// command for the function AccessCOM
#define FREE_COM       -1    // Command for checking the port

// special commands of the WibuBox/ST (send direct without uu-encoding)
#define CMDST_ACKNOWLEDGE   0x21     // BOX/ST-Command '!'
#define CMDST_WAKEUP        0x22     // BOX/ST-Command '"'
#define CMDST_TERMINATE     0x23     // BOX/ST-Command '#'
#define CMDST_DATASTOP      0x25     // BOX/ST-Command '%'
#define CMDST_STARTRECEIVE  0x27     // BOX/ST-Command '\''
#define CMD_SEQ             "\"\"'40%" // WakeUp + FSEARCH (=0x04) + STOP

// macros for the baudrate (the divisor latch value)
#define BAUD_DIVISOR_4800    24
#define BAUD_DIVISOR_9600    12
#define BAUD_DIVISOR_19200    6
#define BAUD_DIVISOR_38400    3
#define BAUD_DIVISOR_57600    2
#define BAUD_DIVISOR_115200   1

// real baudrate constances
#define BR_4800    4800
#define BR_9600    9600
#define BR_19200   19200
#define BR_38400   38400
#define BR_57600   57600
#define BR_115200  115200

// timeout constants
#if _UNIX
  #define TIMEOUT_US_SEND   50000L // =  50 ms (waiting by sending)
  #define TIMEOUT_US_RECV   50000L // = 130 ms (waiting by receiving)
#endif

#define TIMEOUT_SEND           50 // =  50 ms (waiting by sending)
#define TIMEOUT_RECV          130 // = 130 ms (waiting by receiving)
#define TIMEOUT_ACC             2 // =   2 ms (waiting by accessing)


/*****************************************************************************
   The following macros are for UU en-/decoding.
   Mention a 24 bit word:

   23              16              8              0
   ------------------------------------------------
   |     byte 3     |   byte 2     |   byte 1     |
   ------------------------------------------------
                        /\
                        ||
          transformed into (and vice versa)
                        ||
                        \/
   23          18          12          6          0
   ------------------------------------------------
   |    byte 3  |  byte 2   |  byte 2  |  byte 1  |
   ------------------------------------------------
*****************************************************************************/

// 3 input bytes ==>> 4 output bytes
#define MGET_BYTE1_8TO6(a) (*(a) & 0x3f)
#define MGET_BYTE2_8TO6(a) (((*(a+1) & 0x0f) << 2) | ((*(a) & 0xc0) >> 6))
#define MGET_BYTE3_8TO6(a) (((*(a+2) & 0x03) << 4) | ((*(a+1) & 0xf0) >> 4))
#define MGET_BYTE4_8TO6(a) ((*(a+2) & 0xfc) >> 2)
// 4 input bytes ==>> 3 output bytes
#define MGET_BYTE1_6TO8(a) (((*(a+1) & 0x03) << 6) | (*(a) & 0x3f))
#define MGET_BYTE2_6TO8(a) (((*(a+2) & 0x0f) << 4) | ((*(a+1) & 0x3c) >> 2))
#define MGET_BYTE3_6TO8(a) (((*(a+3) & 0x3f) << 2) | ((*(a+2) & 0x30) >> 4))

#define OFFSET    0x30    // = 48


// ==========================================
// Handle of the selected port (OS dependent)
// ==========================================
#if 1
// ===================
// for UNIX, others OS
// ===================

#include <stdio.h>
#include <fcntl.h>
#include <memory.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>

#if _UNIX
#include <sys/time.h>
#include <sys/types.h>

#if DEBUG

  #define SRLBSRMSGM(...) fprintf (stderr, __VA_ARGS__)

#else
  //undef
  #define SRLBSRMSGM(...)

#endif //DEBUG

#define EXEC_BOX_CON 1


/* Change only the #defines with 'WKVAR_...' and 'DEFAULT_...' in proj.h.
 * Entry '{-1, NULL, NULL, NULL}' is an End-of-Array-Marker
 *   and must reside always at the end as the last Array Element!
 */
//global environment array
ENVS gaevWK[] = {
    { ENV_PORTNAME,   WKVAR_PORTNAME,   NULL, DEFAULT_PORTNAME   },
    { ENV_COM,        WKVAR_COM,        NULL, DEFAULT_COM        },
    { ENV_SYSTEM,     WKVAR_SYSTEM ,    NULL, DEFAULT_SYSTEM     },
    { ENV_HOSTNAME,   WKVAR_HOSTNAME,   NULL, DEFAULT_HOSTNAME   },
    { ENV_SERVERNAME, WKVAR_SERVERNAME, NULL, DEFAULT_SERVERNAME },
    { ENV_IPPORT,     WKVAR_IPPORT,     NULL, DEFAULT_IPPORT     },
    { ENV_BAUDRATE,   WKVAR_BAUDRATE,   NULL, DEFAULT_BAUDRATE   },
    { ENV_PORTENUM,   WKVAR_PORTENUM,   NULL, DEFAULT_PORTENUM   },
    { ENV_PORT,       WKVAR_PORT,       NULL, DEFAULT_PORT       },
    { ENV_LPT,        WKVAR_LPT,        NULL, DEFAULT_LPT        },
    { ENV_LANG,       WKVAR_LANG,       NULL, DEFAULT_LANG       },
    { ENV_TIMEOUT,    WKVAR_TIMEOUT,    NULL, DEFAULT_TIMEOUT    }
};

#endif // #if LINUX

BOOL gfPortsChecked = FALSE;
BOOL gafPorts[C_SERIAL_PORT];  //[4]; Port is available if True
char gszDevName[30] = "";      // empty preset global Unix/Linux-DeviceName


#define INVALID_HANDLE_COM    -1
static INT pDev = INVALID_HANDLE_COM;
struct termios oldtio;

// ================
// Global variables
// ================

// in- & output-buffer for uuen-/decoding
static BYTE aBufCOM[4];
// current number of bytes stored in aBufCOM
static INT cBuf = 0;

// store the current set baudrate
static ULONG gnBaudrate = 0;
// number of the selected port 0..127
static INT nPortSel = FREE_COM;
// list of all detected WibuBox/ST
static DEVOBJECT *pdevobjlist;
// number of scaned boxes
static INT cdevicelist = 0;
// selected device object
static DEVOBJECT devobj;
// structure as dummy for saving & returning data
static WKBPORT wkbprtSrl;

// This two values are set from ComGetSelPort in kecom.c
// or from ComNGetSelPort in kecomN.c.
// contains the port address of COM1 until COM4
USHORT pusComAddr[C_SERIAL_PORT] = {0x3F8, 0x2F8, 0, 0};


// *************
// **** IBM ****
// *************
#define B_RTS_IBM   BIT1  // BOXWR signal
#define B_DTR_IBM   BIT0  // BOXCLK signal

// COM relative addresses (IBM)
#define PRT_DATA       0 // use BITx values
#define PRT_INTENABLE  1
#define PRT_INTIDENT   2
#define PRT_LINECTRL   3
#define PRT_MODCTRL    4
#define PRT_LINESTATUS 5
#define PRT_MODSTATUS  6


// get port data for the serial port (IBM-PC/defined in kecom.c)
typedef struct {
  PORTSPEC pts;
  USHORT fsCtrl;
  USHORT fsDrvCtrl;
  LOWSTATUS lwst;
  BYTE bOrgIntEna;  // original value of interrupt enable register
  BYTE bOrgModCtrl; // original value of modem control register
  BYTE bLineCtrl;   // current line control register value
  BYTE bModCtrl;    // current modem control register value
} COM_PORTDATA; // prefix "prtd"

COM_PORTDATA * ComGetSelPort (INT nPort);

// *************
// **** NEC ****
// *************
#define B_RTS_NEC   BIT5  // BOXWR signal
#define B_DTR_NEC   BIT1  // BOXCLK signal

// COM relative addresses (NEC)
#define PRT_CONTROL    2 // use BITx values
#define PRT_STATUS     3

// get port data for the serial port (NEC-PC/defined in kecomN.c)
typedef struct {
  PORTSPEC pts;
  USHORT fsCtrl;
  USHORT fsDrvCtrl;
  LOWSTATUS lwst;
  BYTE bOrgIntEna;  // original value of interrupt enable register
  BYTE bOrgControl; // original value of modem control register
  BYTE bControl;    // current modem control register value
} COMN_PORTDATA; // prefix "prtd"

COMN_PORTDATA * ComNGetSelPort (INT nPort);

#if _UNIX


BOOL IsReadableTimeOut(int fd, long sec, long usec)
{
  //P:mn See select(2): synchronous I/O-multiplexing
  fd_set              rfds; // set of read  fd
  struct timeval        tv; // timeout for select()
  int               fValid; // returncheck
  #if DEBUG
  int pid = getpid();
  #endif

  // reset all fdset-pointer-bits
  FD_ZERO(&rfds);
  // add fd to a set of readfds
  FD_SET(fd, &rfds);
  // timeout
  tv.tv_sec = sec;//long
  tv.tv_usec = usec;//long
  // SVR4 and BSD not POSIX.1
  // non-blocking check if read() is possible
  fValid = select(fd + 1, &rfds, NULL, NULL, &tv);
  // value of tv is UNDEFINED now!
  if (fValid > 0){
    // success
    // SRLBSRMSGM("(%d) IsReadableTimeOut/Data arrived.\n", pid);
    // only read if pDev is ready for reading
    if ( FD_ISSET(fd, &rfds) ){
      // READ HERE
      return TRUE;
    }// if
  } else if (fValid == 0){
    // time is out
    SRLBSRMSGM("(%d) IsReadableTimeOut/Timeout reached after %f sec.\n",
              pid, ((usec/1e6) + sec));
  } else {
    // Error
    SRLBSRMSGM("(%d) IsReadableTimeOut/select error.\n", pid);
  } // if - elseif - else
  return FALSE;
// ------------------- threadsave down -----------------------
} // IsReadableTimeOut()


BOOL IsWriteableTimeOut(int fd, long sec, long usec)
{
  // See select(2): synchronous I/O-multiplexing
  fd_set                         wfds; // set of write fd
  struct timeval                   tv; // timeout for select()
  int                          fValid; // returncheck
  
  // reset all fdset-pointer-bits
  FD_ZERO(&wfds);
  // add fd to a set of readfds
  FD_SET(fd, &wfds);//FD_SET(0, &wfds);
  // timeout
  tv.tv_sec = sec;//long
  tv.tv_usec = usec;//long
  // SVR4 and BSD not POSIX.1
  // non-blocking check if write() is possible
  fValid = select(fd + 1, NULL, &wfds, NULL, &tv);
  // value of tv is UNDEFINED now!
  if (fValid > 0){
    // success, write() seems possible
    if ( FD_ISSET(fd, &wfds) ){
      // only write if pDev is ready for writing
      // BSRMSGM("(%d) IsWriteableTimeOut/Data sendable.\n", pid);
      return TRUE;

    }// if
  } else if (fValid == 0){
    // time is out
  } else {
    // Error
  }
  return FALSE;
} // IsWriteableTimeOut()



/*****************************************************************************
S i z e o f E n v V a r A r r a y
==============================================================================

This function returns the Size of gaevWK - Array

Return
  The size of gaevWK-Array as INT

*****************************************************************************/

INT SizeofEnvVarArray (void)
{
  return ( (INT) ( sizeof(gaevWK) / sizeof(gaevWK[0]) ) );
}//SizeofEnvVarArray()


/*****************************************************************************
S e t E n v V a r A r r a y
==============================================================================

This function sets the Value of the global Array of struct ENVS.
  These are the Name-Value-Pairs of the WIBUKEY-Environment-Variables
  under Linux. The Default Values are also stored.

Parameters
  id
      Is the ID of the Environment-Value in struct ENVS:

        typedef struct {
          INT   id;           // This might be obsolete, but :-)
          char *pszName;     // Environment Variables' Names
          char *pszValue;    // The apropriate Values
          char *pszDefValue; // The Default Value
        } ENVS;
        extern ENVS gaevWK[];

  pszString
      Points to the '\0'-terminated String-Constant which is to set.

      If it is set to "" or NULL the default Value (as stored in the structure)
      is set. Otherwise the given pszString ist set to ENVS[].*pszValue.

Globals
  cstWkd.flCtrl, cstWkd.idError, gaevWK[].

Return
  -1  for Error, cstWkd.idError is set to WKB_ERROR_INVALID_PARAMETER or
                                         WKB_ERROR_MEMORY_TOO_SMALL in Error case.
 >=0  otherwise (the copied chars)

Comments

*****************************************************************************/

INT SetEnvVarArrayEntry(INT id, const char *pszString)
{
  INT iLen = 0;
  const INT ERR = -1;
  // Number of ArrayElements
  INT iEnt = SizeofEnvVarArray();

  //overwrites WKB_ERROR_INVLID_PORT
  cstWkd.idError = WKB_NO_ERROR;
  ASSERT(gaevWK != NULL);

  // check id's range is ok
  if( (id >= iEnt) || (id < 0) ){
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    return (ERR);
  }//if

  //decide which string to copy:
  if((pszString == NULL) || pszString[0] == '\0'){
    // No given String: set default (Pointercopy)
    iLen = strlen(gaevWK[id].pszDefValue);
    gaevWK[id].pszValue = (char*) strdup( gaevWK[id].pszDefValue ); // fixme - memleak
    return (iLen);
  } else {
    iLen = strlen(pszString);
    // String is given: Check if it equates with stored one.
    // * Copy String into gaevWK[] if it differs else do nothing.
    if ((gaevWK[id].pszValue != NULL) &&
      !(strncmp(gaevWK[id].pszValue, pszString, iLen))){
      // OK (nothing to do)
      return (iLen);
    }//if
    // ENV found: set the given String
    gaevWK[id].pszValue = (char*) strdup(pszString);    // fixme - memleak
    return (iLen);
  }//if
  return (ERR);
}// SetEnvVarArrayEntry()



/*****************************************************************************
C h e c k P o r t 
==============================================================================

This function checks, if a serial port can be opend.

Return
  The function returns TRUE, if the serial port is opened
  and if no device is connected to the serial port and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL CheckPort (char *szDevice)
{
  int fd = 0;
  if ( (NULL == szDevice ) || ( szDevice[0] == '\0' ) ) {
    return FALSE;
  } // if 
  fd = open( szDevice, O_RDWR | O_NONBLOCK );
  if ( 0 == fd ) {
    return FALSE;
  } // if 
  close(fd);
  return TRUE;
} // CheckPort()


/*****************************************************************************
C h e c k C O M
==============================================================================

This function checks, if a serial port is opened.

Return
  The function returns TRUE, if the serial port is opened
  and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL CheckCOM ()
{
  if (pDev != INVALID_HANDLE_COM && nPortSel != FREE_COM) {
    return TRUE;
  }// port validation

  return FALSE;
} // CheckCOM()
#endif //#if LINUX


/*****************************************************************************
SetComAddr
==============================================================================

This function sets the physical address of the given serial port. 
The functions ComGetSelPort & ComNGetSelPort are implemented in
the file kecom.c & kecomN.c for the serial communication with 
te WibuBox/SP on IBM-PCs & NEXC-PCs.

*****************************************************************************/

static BOOL8 SetComAddr (INT nPort, USHORT *fsCtrl)
{
  BOOL8 fExist = FALSE;
#if !_UNIX
  BOOL8 fSet = FALSE;

  // get the right port address
  if (!WkdCheckMachine ()) {
    // ********************************
    // *** IBM *** serial port register
    // ********************************
    COM_PORTDATA *pprtdIbm;
    pprtdIbm = ComGetSelPort (nPort - 8);

    // Set current port address to WKPORT structure
    wkbprtSrl.ulAddress = pusComAddr[nPort - N_MINPORT];
    *fsCtrl = 0;
    if (pprtdIbm != NULL) {
      // Check if the option /B is set in the environment variable
      // WKCOM ==>> disable this SRL port.
      fExist = !((pprtdIbm->fsCtrl & (WKB_CTRL_BIN | WKB_CTRL_CHAR)) == WKB_CTRL_BIN);
      *fsCtrl = pprtdIbm->fsCtrl;
    } else {
      return FALSE;
    } // if

    pprtdIbm->bModCtrl &= ~(B_DTR_IBM | B_RTS_IBM);
    if (pprtdIbm->fsCtrl & WKB_CTRL_RTSA) {
      // /R flag set: activate RTS permanently in future
      pprtdIbm->bModCtrl |= B_RTS_IBM;
      fSet = TRUE;
    } // if
    if (pprtdIbm->fsCtrl & WKB_CTRL_DTRA) {
      // /T flag set: activate DTR permanently in future
      pprtdIbm->bModCtrl |= B_DTR_IBM;
      fSet = TRUE;
    } // if
    if (fSet) {
      WRITE_PORT_BYTE(pprtdIbm->pts.pta + PRT_MODCTRL, pprtdIbm->bModCtrl);
    } // if
  } else {
    // ********************************
    // *** NEC *** serial port register
    // ********************************
    COMN_PORTDATA *pprtdNec;
    pprtdNec = ComNGetSelPort (nPort - 8);

    // Set current port address to WKPORT structure
    wkbprtSrl.ulAddress = pusComAddr[nPort - N_MINPORT];
    *fsCtrl = 0;
    if (pprtdNec != NULL) {
      // Check if the option /B is set in the environment variable
      // WKCOM ==>> disable this SRL port.
      fExist = !((pprtdNec->fsCtrl & (WKB_CTRL_BIN | WKB_CTRL_CHAR)) == WKB_CTRL_BIN);
      *fsCtrl = pprtdNec->fsCtrl;
    } else {
      return FALSE;
    } // if

    pprtdNec->bControl &= ~(B_DTR_NEC | B_RTS_NEC);
    if (pprtdNec->fsCtrl & WKB_CTRL_RTSA) {
      // /R flag set: activate RTS permanently in future
      pprtdNec->bControl |= B_RTS_NEC;
      fSet = TRUE;
    } // if
    if (pprtdNec->fsCtrl & WKB_CTRL_DTRA) {
      // /T flag set: activate DTR permanently in future
      pprtdNec->bControl |= B_DTR_NEC;
      fSet = TRUE;
    } // if
    if (fSet) {
      WRITE_PORT_BYTE(pprtdNec->pts.pta + PRT_CONTROL, pprtdNec->bControl);
    } // if
  } // if

#else  // #if !LINUX
  fExist = TRUE;
#endif  // #if !LINUX

  return fExist;
} // SetComAddr()


/*****************************************************************************
SendDirect
==============================================================================

This function sends cIn Bytes direct to the serial port.

Parameters
  pbIn ... contains the bytes which will send to the serial port.
  cIn .... contains the number of bytes, which will be send to the port.

Return
  The function returns TRUE, if the byte sent correctly to the serial port
  and FALSE if not.
  
Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL8 SendDirect (BYTE *pbIn, INT cIn)
{
#if   defined(_DOS) || WIN16DLL || WINVXD
// ====================
// for DOS
// ====================

  if (!WkdCheckMachine ()) {
    UINT nPort, nRet, i;
    ULONG nStartTime, nDiffTime;

    // set the right port index
    nPort = nPortSel - N_MINPORT;
    // read the data register to flush the DR bit
    READ_PORT_BYTE(pusComAddr[nPort] + PRT_DATA);
    for (i = 0; i < (UINT) cIn; i++) {
      // write one byte in the send register
      nRet = WRITE_PORT_BYTE(pusComAddr[nPort], pbIn[i]);
      if (nRet != pbIn[i]) {
        return FALSE;
      } // if

      // get the reference time for the timeout
      nStartTime = GetTickCount();
      // wait until the byte is sent
      do {
        // Bit 5 (Transmitter Holding Register) and Bit 6 (Transmitter
        // Shift Register Empty) of the Line status register must be set.
        nRet = READ_PORT_BYTE(pusComAddr[nPort] + PRT_LINESTATUS);
        nDiffTime = GetTickCount() - nStartTime;
      } while (((nRet & 0x60) != 0x60) && nDiffTime < TIMEOUT_SEND);
    } // for
  } else {
    // No support of the WibuBox/ST driver for DOS and 16 Bit Windows.
    return FALSE;
  } // if

#else
// ====================
// for UNIX, others OS
// ====================

  INT i, nSent;

#if !_UNIX
  // This call send all bytes single by running the loop.
  nSent = 0;
  // OLD WRITE, delete lateron
  for (i = 0; i < cIn; i++) {
    nSent = (INT) write (pDev, &(pbIn[i]), 1);
    if (nSent == -1) {
      return FALSE;
    } // if
  } // for
#else 
    nSent = 0;
    if( IsWriteableTimeOut(pDev, 0L, TIMEOUT_US_SEND) ){
      for (i = 0; i < cIn; i++) {
        nSent = (INT) write (pDev, &(pbIn[i]), 1);
        if (nSent == -1) {
          return FALSE;
        } // if
      } // for
    } else {
      // Error
    } //if-elsif-else

#endif 

#endif // #if NTKRNL

  return TRUE;
} // SendDirect()


/*****************************************************************************
RecvDirect
==============================================================================

This function receives 1 Byte direct from the serial port.

Parameters
  bOut ... contains the byte received from the serial port.

Return
  The function returns TRUE, if the byte received correctly to the serial port
  and FALSE if not.
  
Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL8 RecvDirect (BYTE *bOut)
{
#if _UNIX
  //To avoid little/big-endian problems
  BYTE  bBuffer = 0;
#else
  char bBuffer = 0;
#endif

#if 1
// ====================
// for UNIX, others OS
// ====================

  // NEW POLLING: nonblocking read, return immediately; pause; do again.
  INT nRecv = 0;
  // To avoid Little/Big-endian problems just read one byte
  BYTE nRead = 0;
  // fixme: check, it timeout enougth if ( IsReadableTimeOut(pDev, 0L, TIMEOUT_US_RECV * 3L) ) {
  if ( IsReadableTimeOut(pDev, 0L, TIMEOUT_US_RECV ) ) {

    // receive only one byte
    nRecv = read (pDev, &nRead, 1);

  } else {
    // Error
  }//if - else

  if (nRecv <= 0) {
    // Maximum of reads is reached
    //!F:mn ask errno too! It's not granted whether -1 comes first!
    goto RET_NOT_OK;
  }
  #if _UNIX
    bBuffer = nRead;
  #else
    bBuffer = (char) nRead;
  #endif

#endif // #if NTKRNL

  *bOut = (BYTE) bBuffer;
  return TRUE;

RET_NOT_OK:
    return FALSE; 
} // RecvDirect()


/*****************************************************************************
SendBytes
==============================================================================

This function sends a byte to COM. But first she waits until 3 bytes should
be send and stores the given byte in a buffer. If the buffer has 3 bytes, 
they will be UU-encoded and then the result of 4 bytes will send to COM.

Parameters
  bIn ... contains the byte which will be sent to the serial port.

Return
  The function returns TRUE, if the byte sent correctly to the serial port
  and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL8 SendBytes (BYTE bIn)
{
  // store the byte in the buffer
  aBufCOM[cBuf] = bIn;
  cBuf++;
  if (cBuf == 3) {        // if cBuf == 3, send the whole global buffer to COM
    BYTE pabBuf[4];
    BOOL8 fValid = TRUE;

    pabBuf[0] = MGET_BYTE1_8TO6(aBufCOM) + OFFSET;   // UU-encode
    pabBuf[1] = MGET_BYTE2_8TO6(aBufCOM) + OFFSET;   // UU-encode
    pabBuf[2] = MGET_BYTE3_8TO6(aBufCOM) + OFFSET;   // UU-encode
    pabBuf[3] = MGET_BYTE4_8TO6(aBufCOM) + OFFSET;   // UU-encode
    cBuf = 0;
    fValid = SendDirect (pabBuf, 4);
    if (!fValid) {
      return FALSE;
    } // if
  } // if
  
  return TRUE;
} // SendBytes()


/*****************************************************************************
RecvBytes
==============================================================================

This function reads bytes from the serial port. If 1 byte received, it will
be send back to the port, because the Box/ST want to know, if the byte is
sent correctly. But the problem is, that the bytes, which are received from 
the Box /ST are uu-encoded. This function received the bytes from the Box,
confirmed this byte to the Box, uu-decoded the byte(s) and give them back to
the calling function. The number of wanted bytes from the box should not
have to be more than 3 bytes.

Parameters
  pOut ... contains the bytes received from the serial port. The bytes are
           already uu-decode. This variable have to allocate cOut bytes
           memory.
  cOut ... contains the number of bytes, which should be received from the
           serial port.

Return
  The function returns TRUE, if the byte received correctly from the serial
  port and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL8 RecvBytes (BYTE *pOut, UINT *cOut)
{
  BOOL8 fValid = FALSE;
  BYTE pabBuf[4], bOut;

  ASSERT(*cOut >= 0 && *cOut <= 3);
  memset (pabBuf, 0, 4);

  // get the bytes only, if 1 until 3 bytes wanted
  if (*cOut >= 1) {
    //==========================
    // receive 1. byte
    //==========================
    bOut = 0;
    fValid = RecvDirect (&bOut);      // receive one byte
    if (fValid && bOut != 0) {
      fValid = SendDirect (&bOut, 1); // send received byte as acknowledge
      pabBuf[0] = bOut - OFFSET;      // UU-decode-offset
      if (bOut == CMDST_TERMINATE) {
        // termination byte received ==>> exit
        *cOut = 0;
        return TRUE;
      } // if
    } else {       
      // nothing received ==>> exit with error
      return FALSE;
    } // if

    //==========================
    // receive 2. byte
    //==========================
    bOut = 0;
    fValid = RecvDirect (&bOut);      // receive one byte
    if (fValid && bOut != 0) {
      fValid = SendDirect (&bOut, 1); // send received byte as acknowledge
      pabBuf[1] = bOut - OFFSET;      // UU-decode-offset
      if (bOut == CMDST_TERMINATE) {
        // termination byte received ==>> exit
        // ATTENTION: at this position a termination byte should never received
        ASSERT(!(fValid && bOut == CMDST_TERMINATE));
        *cOut = 0;
        return TRUE;
      } // if
    } else {       
      // nothing received ==>> exit with error
      return FALSE;
    } // if
    // check the first complete received byte
    pOut[0] = MGET_BYTE1_6TO8(pabBuf); // now UU-decode the 1.bytes

    //==========================
    // receive 3. byte
    //==========================
    if (*cOut >= 2) {
      bOut = 0;
      fValid = RecvDirect (&bOut);      // receive one byte
      if (fValid && bOut != 0) {
        fValid = SendDirect (&bOut, 1); // send received byte as acknowledge
        pabBuf[2] = bOut - OFFSET;      // UU-decode-offset
        if (bOut == CMDST_TERMINATE) {
          // termination byte received ==>> exit
          *cOut = 1;
          return TRUE;
        } // if
      } else {       
        // nothing received ==>> exit with error
        return FALSE;
      } // if
      // check the first complete received byte
      pOut[1] = MGET_BYTE2_6TO8(pabBuf); // now UU-decode the 1.bytes

      //==========================
      // receive 4. byte
      //==========================
      if (*cOut == 3) {
        bOut = 0;
        fValid = RecvDirect (&bOut);      // receive one byte
        if (fValid && bOut != 0) {
          fValid = SendDirect (&bOut, 1); // send received byte as acknowledge
          pabBuf[3] = bOut - OFFSET;      // UU-decode-offset
          if (bOut == CMDST_TERMINATE) {
            // termination byte received ==>> exit
            *cOut = 2;
            return TRUE;
          } // if
        } else {       
          // nothing received ==>> exit with error
          return FALSE;
        } // if
        // check the first complete received byte
        pOut[2] = MGET_BYTE3_6TO8(pabBuf); // now UU-decode the 1.bytes
      } // if
    } // if
  } // if

  return TRUE;
} // RecvBytes()


/*****************************************************************************
FlushSend
==============================================================================

This function sends the rest of the buffer uu-encoded to COM.

Parameters

Return
  The function returns TRUE, if the bytes sent correctly to the serial port
  and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL8 FlushSend (void)
{
  BYTE pbBuf[3];
  BOOL8 fValid = TRUE;

  // send the rest of the buffer to COM
  switch (cBuf) {
    case 0:       // nothing to do
      break;
    case 1:
      // one byte received ==>> two bytes returned
      pbBuf[0] = MGET_BYTE1_8TO6(aBufCOM) + OFFSET;
      pbBuf[1] = (MGET_BYTE2_8TO6(aBufCOM) & 0x03) + OFFSET;
      fValid = SendDirect (pbBuf, 2);
      if (!fValid) {
        cBuf = 0;
        return FALSE;
      } // if
      break;
    case 2:
      // two bytes received ==>> three bytes returned
      pbBuf[0] = MGET_BYTE1_8TO6(aBufCOM) + OFFSET;
      pbBuf[1] = MGET_BYTE2_8TO6(aBufCOM) + OFFSET;
      pbBuf[2] = (MGET_BYTE3_8TO6(aBufCOM) & 0x0f) + OFFSET;
      fValid = SendDirect (pbBuf, 3);
      if (!fValid) {
        cBuf = 0;
        return FALSE;
      } // if
      break;
    default:
      // This case sould never happens ==>> Shit happens :-).
      ASSERT(0);  // ERROR
      fValid = FALSE;
  } // switch

  cBuf = 0;
  return fValid;
} // FlushSend()


/*****************************************************************************
ClearCOMBuffers
==============================================================================

This function clears the receive buffer of the opened serial port.

Parameters
  fRead .... purge the recieve buffer
  fWrite ... purge the send buffer

Return
  The function returns TRUE, if the serial port clears correctly
  and FALSE if not.

*****************************************************************************/

static BOOL8 ClearCOMBuffers (BOOL8 fRead, BOOL8 fWrite)
{

  return TRUE;
} // ClearCOMBuffers()


/*****************************************************************************
ReleaseCOM
==============================================================================

This function closes the serial port to get them free for other devices.

Parameters
  nPort ... contains the port number in the range of 0..127.

Return
  The function returns TRUE, if the serial port closes correctly
  and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

BOOL8 ReleaseCOM (INT nPort)
{
  BOOL8 bReturn = TRUE;

  // if the port is already closed
  if (nPortSel == FREE_COM) {
    return TRUE;
  } else if (nPort < N_MINPORT || nPort > N_MAXPORT) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if

  // close device (serial port) if he is opened
  { // block (begin)
#if 1
// ===================
// for UNIX, others OS
// ===================

  INT nClose = 0;

  // set all resources back to the origin values and give the port free.
#if _UNIX
  //!E:mn LED-box blinks 5times
  // * Reset the port settings of the chosen serial port NOT immediately,
  // * wait until all chars reached - then change mode!
  tcsetattr (pDev, TCSADRAIN, &oldtio);
#else
  tcsetattr (pDev, TCSANOW, &oldtio);
#endif
  //!F:mn Check success with tcgetattr
  nClose = close (pDev);
  pDev = INVALID_HANDLE_COM;
  bReturn = (nClose == 0) ? TRUE : FALSE;

#endif // #if NTKRNL
  } // block (end)

  nPortSel = FREE_COM;  // global variable
  return bReturn;       // all okay
} // ReleaseCOM()

#if _UNIX
#if (EXEC_BOX_CON)
/*****************************************************************************
CheckBoxConnected
==============================================================================

Call a AccessCOM() first !!!
This function checks (opens) the port and sends the RESET command to the
specified port. All received bytes are echoed.

Parameters
  pcPortString  0 terminated Portname

Return
  The function returns TRUE if a Box is connected, otherwise FALSE.

*****************************************************************************/

BOOL CheckBoxConnected(char *pcPortString)
{
  return TRUE;
} // CheckBoxConnected()
#endif //#if EXEC_BOX_CON


/*****************************************************************************
TranslateLinuxBaudRate
==============================================================================
Used by SrlSelect() in this file:
 This function translates given human readable Baudrates
 into Linux Constants from <termios.h>.
   e.g.: for a given 19200 constant, B19200 (= 0000016) is returned.
   See <bits/termios.h> (#define B19200 0000016)

Parameters
  ulBaudrate  Baudrate: 4800, 9600, 19200, 38400 or 57600

Return
  The function returns <bits/termios.h> constants for "'B'"ulBaudrate"".
  B19200 (WIBUKEY default) is returned in any errorcase.

*****************************************************************************/
speed_t TranslateLinuxBaudRate(ULONG ulBaudrate)
{
#define MN_BAUD_FAKE
    if(ulBaudrate==0){
        gnBaudrate = WkdGetProfileInt (WKPRFTPC_DRIVER_KERNEL, "Baudrate");
        if (gnBaudrate == 0) {
            // set default baudrate = 19200
            gnBaudrate = BR_19200;
        } // if
    }//if
/* there are DIFFENRENCES for 'B' constants in
   /usr/include/asm/termbits.h:103:#define  B9600  0000015
   /usr/include/bits/termios.h:132:#define  B9600  0000015      <- THIS SCHULD IT BE
#define  B4800  0000014
#define  B9600  0000015
#define  B19200 0000016
#define  B38400 0000017
#define  B57600 0010001
   AND
   /usr/include/rpcsvc/rex.h:66:   #define B9600 13             <- BUT THIS IS TAKEN ??
#define B4800 12
#define B9600 13
#define B19200 14
#define B38400 15
   /usr/include/sys/ttydefaults.h:50:#define TTYDEF_SPEED  (B9600)
*/
    switch((int)ulBaudrate){
#if defined(MN_BAUD_FAKE)
    case BR_4800:
        return ((speed_t) 0000014); /* 0000014 */
        break;
    case BR_9600:
        return ((speed_t) 0000015); /* 0000015 */
        break;
    case BR_19200:
        return ((speed_t) 0000016);/* 0000016 */
        break;
    case BR_38400:
        return ((speed_t) 0000017);/* 0000017 */
        break;
    case BR_57600:
        return ((speed_t) 0010001);/* 0010001 */
        break;
    default:
        return ((speed_t) 0000016);/* 0000016 */
        break;
#else
    case BR_4800:
        return ((speed_t)B4800); /* 0000014 */
        break;
    case BR_9600:
        return ((speed_t)B9600); /* 0000015 */
        break;
    case BR_19200:
        return ((speed_t)B19200);/* 0000016 */
        break;
    case BR_38400:
        return ((speed_t)B38400);/* 0000017 */
        break;
    case BR_57600:
        return ((speed_t)B57600);/* 0010001 */
        break;
    default:
        return ((speed_t)B19200);/* 0000016 */
        break;
#endif
    }//switch
} // TranslateLinuxBaudRate()

static UINT SrlRelease (void);
#endif // #if LINUX

/*****************************************************************************
AccessCOM
==============================================================================

This function opens the serial port to access them. After accessing the
port some properties are set.

Parameters
  nPort ... contains the port number in the range of 0..127.

Return
  The function returns TRUE, if the serial port opens correctly
  and FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

BOOL8 AccessCOM (INT nPort, INT nDivisor)
{  
  BOOL8 fValid;
  USHORT fsCtrl;

  // set baudrate if nDivisor contains a valid value
  if (nDivisor != 0) {
    gnBaudrate = nDivisor;
  } // if

  // if the port is aready opened
  if ((nPort == nPortSel) && (nPortSel != FREE_COM)) {
    return TRUE;    // nothing to do
  } else if (nPort < N_MINPORT || nPort > N_MAXPORT) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if

  ASSERT(gnBaudrate != 0);
  // Set the physical Address of the serial port.
  // No more settings if the port is not available.
  fValid = SetComAddr (nPort, &fsCtrl);
  if (!fValid) {
    return FALSE; // This port is not available.
  } // if

  // open serial interface with specific name
  { // block (begin)
  // ===================
  // for UNIX, other OS
  // ===================
  // port name definition if no variable is used default value is
  // /dev/ttyS
  char szDevName[32+1], szDefDevName[] = DEFAULT_PORTNAME;
  char szBuf[32+1], *pszPortName, *pszPortEnum;//15
  speed_t           s_tBaudrate;  //local (Linux like) Baudrate
  UINT                 ucLength;
  INT        cPorts, fError = 0;
  struct termios         newtio;
  BOOL fInvalidPort = TRUE;

  // nPort range validation internal WibuKey range
  if (nPort < N_MINPORT || nPort > N_MAXPORT) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if
  // Decide whether or not the environment variable is to check
  if (gfPortsChecked && ((int) strlen(gaevWK[ENV_PORTNAME].pszValue))) {
    // -> yes, Ports were checked before, get the name which was last time read
    strcpy(szDevName, gaevWK[ENV_PORTNAME].pszValue);
    // if the global list is TRUE, port exists
    if (gafPorts[nPort-N_MINPORT]) {
      // if the port is already opened
      if (pDev != INVALID_HANDLE_COM && nPort == nPortSel && CheckCOM()) {
        return TRUE; // nothing to do
      }
    } else { // if (gafPorts[nPort-N_MINPORT])
      cstWkd.idError = WKB_ERROR_INVALID_PORT;
      return FALSE;
    } // else if (gafPorts[nPort-N_MINPORT])
  } else { // if (gfPortsChecked): -> no, portname was never checked before
    // ALL possible Ports now will be checked once with each call of AccessCOM.
    // first check the correct name. Look for the enviromentvariable WKPORTNAME
    #if _UNIX
      ucLength = WkdGetProfileValue (WKPRFTPC_GENERAL,
        gaevWK[ENV_PORTNAME].pszName, szBuf, sizeof(szBuf));
    #else
      ucLength = WkdGetProfileValue (WKPRFTPC_GENERAL, WKVAR_PORTNAME, szBuf,
        sizeof(szBuf));
    #endif //#if LINUX
    pszPortName = szBuf;
    // No WKPORTNAME set
    if (ucLength == 0) {
      // if no environment variable is set use the default name.
      strcpy(szDevName, szDefDevName); //setting the global default name
      #if _UNIX
        // Set global ENVS-array as default also
        SetEnvVarArrayEntry(ENV_PORTNAME, NULL);
      #endif //#if LINUX
      //!F:mn Error Handling
    } else {
      //!F:mn ToDo: Here is to check if WKPORTNAME is set correct !
      strcpy(szDevName, pszPortName);
      #if _UNIX
        // Set global ENVS-array with found value also
        SetEnvVarArrayEntry(ENV_PORTNAME, pszPortName);
      #endif //#if LINUX
    } // else
    // The global var remembers the correct name if it once was read from
    // * the environment.
    #if _UNIX
      // WKPORTNAME to global portname
      strcpy (gszDevName, gaevWK[ENV_PORTNAME].pszValue);
    #else //!N:mn
      strcpy (gszDevName, szDevName);  // WKPORTNAME to global portname
    #endif //#if LINUX

    // Now fill the global list 'gafPorts' with TRUEs for ports which
    // could be enabled. Read WKCOM with a format like: '*,*,/D,/D'.
    // fixme: check this 
    fError=0; 
  
    fInvalidPort = WkdCheckEnvVar("WKPORT", "COM") < 0;
    if ( !fInvalidPort ) {
      cPorts = WkdQueryEnvWkCom(WKVAR_COM); // count of Ports
    } else {
      cPorts = 0;
    }
    if (cPorts <= 0) { // error handling (0 is NO '*' in WKCOM!)
      switch (cPorts) {
      case  0:
        // No Ports Accessible by WKCOM=/D,/D,/D,/D
        // BugFix: even with WKCOM=/D,/D,/D,/D there IS an access (should not)!
        fError=1;
        break;
      case -1:
        //No memory left!
        cstWkd.idError = WKB_ERROR_MEMORY_TOO_SMALL;
        fError=1;
        break;
      case -2:
        //WKCOM not found: Checking ports anyway!
        fError=0;
        break;
      case -3:
        //WKCOM has wrong entry: Checking ports anyway!
        break;
      default:
        fError=0;
        break;
      } //switch
      // anyway all possible ports checkable
      if (fError == 1){
        // global array = all false
        memset(gafPorts, 0, sizeof(gafPorts));
        return FALSE;
      }//if
      // global array = all true (all possible)
      memset(gafPorts, 1, sizeof(gafPorts));
    }//if (cPorts < 0)
    //!N:mn }
  } //if gfPortsChecked

  ucLength = WkdGetProfileValue (WKPRFTPC_GENERAL, WKVAR_PORTENUM, szBuf,
                                 sizeof(szBuf));

  // (default: DIGIT)
  // build complete Portname adding the portnumber
  // Format: /dev/cua[0123].
  sprintf (szDevName, "%s%d", szDevName, nPort - N_MINPORT);

  // perhaps overwrite the default, and build ALPHA-names:
  if (ucLength != 0){
      pszPortEnum = szBuf;
      if(!strncmp(pszPortEnum, "ALPHA", 5)){// if ALPHA
          // sun has /dev/ttya, /dev/ttyb and so on.
          // So add 'a' (97=0x61) offset of ASCII to get a  /dev/tty[abcd].
          sprintf (szDevName, "%s%c", szDevName, (char)(nPort - N_MINPORT + 0x61));
      }//if
  }//if

  // Now, check if the chosen port is available. This checking has to be done
  // because the function "open" does return a valid handle if the port is
  // available. It also return a valid handle if the port is already used
  // by another device, so that the settings of the port will be changed in
  // this function for this device and this device cannot work any more.
  // If this device is a mouse you cannot work with the system any more :-(.

  // If the port is not available or if there is another device
  // connected to the port ==>> No access available.

  // Only check TRUE ports in global array, which is set by WKCOM variable.
  // If no WKCOM is set in environment the whole array is TRUE!

  if (gafPorts[nPort - N_MINPORT]) { // check only TRUE-set WKCOM Ports!
    if (!CheckPort (szDevName)) {
      //Port not available
      gafPorts[nPort - N_MINPORT] = FALSE;
      cstWkd.idError = WKB_ERROR_INVALID_PORT;
      return FALSE;
    }else{
      gafPorts[nPort - N_MINPORT] = TRUE;
    } // if (!CheckPort (szDevName))
  }// if (gafPorts[nPort-N_MINPORT])

  // **** open the chosen serial port

  // to return immediately from receving: 
  if ( TRUE == gafPorts[nPort - N_MINPORT] ) {
    pDev = open (szDevName, O_RDWR | O_NOCTTY | O_NDELAY | O_NONBLOCK);
  } else {
    pDev = INVALID_HANDLE_COM;
  }

  if (pDev == INVALID_HANDLE_COM) { //!Ch:mn (pDev == -1)
    //Port not openable
    gafPorts[nPort - N_MINPORT] = FALSE;
    cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
    return FALSE; // ERROR
  }else{
    gafPorts[nPort - N_MINPORT] = TRUE;
  } // if

  // configure the opened serial port
  errno = 0;
  if(tcgetattr (pDev, &oldtio) < 0){
      // ask errno
  }
#if defined(sun) // Thanks to Bruce Fitzsimons
  tcgetattr (pDev, &newtio);
#else
  memset (&newtio, 0, sizeof(newtio));
#endif 
  // Set raw in- & output flow control, no parity, 8 databits, 2 stopbits.
  newtio.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | IXON);
  newtio.c_oflag &= ~OPOST;
  newtio.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
  newtio.c_cflag &= ~(CSIZE | PARENB);  // **** no parity checking
  newtio.c_cflag |= CS8;  // **** 8 databits

  // These settings are some complements for the WibuBox/ST.
  newtio.c_cflag &= ~CRTSCTS;
  newtio.c_cflag |= CLOCAL;    //!C:mn should always be enabled
  newtio.c_cflag |= CREAD;     //!C:mn should always be enabled
  newtio.c_cflag |= CSTOPB;    // **** 2 stopbits
  // line discipline

  /* Baudrate is now got by WKBAUDRATE (19200 is default).
   * Now we have to translate baudrate (only digits) into 'B'-Baudrate (Linux)
   * See SrlSelect() in kesrl.c.
   */
  s_tBaudrate = TranslateLinuxBaudRate(gnBaudrate);
  cfsetospeed (&newtio, s_tBaudrate); //!N:mn Output-Speed
  cfsetispeed (&newtio, s_tBaudrate); //!N:mn Input-Speed

  // set now all settings to the port immediately
  errno = 0;
  if(tcsetattr (pDev, TCSANOW, &newtio) < 0){
      //!F:mn ask errno
  }
  //!F:mn check if the Options are set correctly
  /* --> HERE there is to implement and check with tcgetattr (pDev,  &newtio);
   * and a following comparison of the values that changed
   */
  /*** I/O-ADAPTATION } ***/

  cBuf = 0; // global variables
  nPortSel = nPort;

#if (EXEC_BOX_CON)
  if(!(CheckBoxConnected(szDevName))){
      // release the Port (fd) immediately to hinder
      // * wasting of filedesciptors (fd).
      SrlRelease();//!N:20030326:mn FD massing
      nPortSel = FREE_COM;//!N:mn
      return FALSE;
  }
#endif
  return TRUE; // all okay

  } // block (end)

  cBuf = 0;         // reset the send buffer
  nPortSel = nPort; // global variable
  return TRUE;      // all okay
} // AccessCOM()


/*****************************************************************************
ControlCOM
==============================================================================

This function sends a stream of bytes to the serial port (COM) and waits
then for the answer and gives them back. This function realize the 
character oriented communication protokoll with the Box /ST

Parameters
  pbInStream    points to the the input byte stream.
  cbIn ........ contains the length of the input byte stream.
  pbOutStream   points to the the output byte stream.
  cbOutLen .... contains the length of the output byte stream.

Return
  the function returns TRUE, if sending and receiving worked correctly and 
  FALSE if not.

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

BOOL8 ControlCOM (BYTE *pbInStream, INT cbIn,
                         BYTE *pbOutStream, INT cbOut)
{
  BOOL8 fValid = TRUE;
  UINT i, j, k;
  UINT cLoop, cRest, nOut;
  BYTE bOut = 0, pabBuf[3];

  // if any port is opened
  if (nPortSel == FREE_COM) {
    goto EXIT_CONTROL;
  } // if

  memset (pbOutStream, 0, (size_t)(cbOut));

  //==================
  // SEND-sequence
  //==================
  // send the 3 wake up commands (>> ""' <<)
  // 3 tries allowed
  j = 0;
  do {
    pabBuf[0] = CMDST_WAKEUP;         // '"'
    pabBuf[1] = CMDST_WAKEUP;         // '"'
    pabBuf[2] = CMDST_STARTRECEIVE;   // '\''
    fValid = SendDirect (pabBuf, 3);
    // exit when no true sending could be done
    if (!fValid) {
      goto TOEND_LOOP;
    } // if
    // send the input (command) stream to the box
    fValid = TRUE;
    for (i = 0; i < (UINT) cbIn; i++) {
      fValid = SendBytes (pbInStream[i]);
      if (!fValid) {
        goto TOEND_LOOP;
      } // if
    } // for  

    // send the rest of the input buffer to COM
    fValid = FlushSend ();
    if (!fValid) {
      goto TOEND_LOOP;
    } // if

    // clears the receive buffer queue
    fValid = ClearCOMBuffers (TRUE, FALSE);
    if (!fValid) {
      return FALSE;
    } // if  

    // send data stop byte for the end of the input stream
    pabBuf[0] = CMDST_DATASTOP;  // '%'
    fValid = SendDirect (pabBuf, 1);
    if (!fValid) {
      goto TOEND_LOOP;
    } // if

    // wait for one acknowledge command from the box
    k = 0;
    bOut = 0;
    do {
      // ignore all received 0 bytes
      fValid = RecvDirect (&bOut);
      k++;
    } while(!(fValid && bOut != 0) && k < 2);

  TOEND_LOOP:
    j++;
  } while(!(fValid && bOut == CMDST_ACKNOWLEDGE) && j < 2);

  // if no acknowledge command received ==>> ERROR
  if (!fValid || bOut != CMDST_ACKNOWLEDGE) {
    goto EXIT_CONTROL;
  } // if

  // send the acknowledge to the box
  fValid = SendDirect (&bOut, 1);
  if (!fValid) {
    goto EXIT_CONTROL;
  } // if

  //==================
  // RECEIVE-sequence
  //==================
  // read the output stream
  cLoop = cbOut / 3;
  cRest = cbOut % 3;
  memset (pabBuf, 0, 3);
  for (i = 0; i < cLoop; i++) { 
    nOut = 3;
    fValid = RecvBytes (pabBuf, &nOut);
    memcpy (pbOutStream + (3 * i), pabBuf, (size_t)nOut);
    if (!fValid) {
      goto EXIT_CONTROL;
    } // if

    // The box has not sent all, may be there is anything wrong
    if (fValid && nOut != 3) {
      return TRUE;
    } // if
  } // for  

  // read the tail of the output stream
  nOut = cRest;
  fValid = RecvBytes (pabBuf, &nOut);
  memcpy (pbOutStream + (3 * cLoop), pabBuf, (size_t)nOut);
  if (!fValid) {
    goto EXIT_CONTROL;
  } // if

  // The box has not sent all, may be there is anything wrong
  if (fValid && nOut != cRest) {
    return TRUE;
  } // if
  
  // wait for the termination byte
  bOut = 0;
  fValid = RecvDirect (&bOut);
  if (bOut != 0) {
    fValid = SendDirect (&bOut, 1);
    if (!fValid) {
      goto EXIT_CONTROL;
    } // if
  } // if
  // if no termination byte received ==>> exit with error
  if (!fValid || bOut != CMDST_TERMINATE) {
    goto EXIT_CONTROL;
  } // if

  return TRUE;      // all okay

EXIT_CONTROL:
  return FALSE;     // exit with error
} // ControlCOM()


/*****************************************************************************
DestroyDeviceListCOM
==============================================================================

This function destroys the whole list of detected WibuBox /ST's.

Parameters

Return

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static void DestroyDeviceListCOM (void)
{
  DEVOBJECT *pdevobj, *pdevobjNext;

  pdevobj = pdevobjlist;
  while (pdevobj != NULL) {
    ASSERT_VALUE(pdevobj);

    pdevobjNext = pdevobj->pNext;   // save pointer to next element
    BsrFreeMem (pdevobj);           // destroy current element
    pdevobj = pdevobjNext;
  } // while

  cdevicelist = 0;
  pdevobjlist = NULL;               // global list is now empty

} // DestroyDeviceListCOM()


/*****************************************************************************
UpdateDeviceListCOM
==============================================================================

This function cates a list of all connected boxes (/ST's).

Parameters
  plwst  points to the the low level status structure which device list is
         destroyed.

Return

Attributes
  This function will be only used by the main Srl-driver-functions.

*****************************************************************************/

static BOOL8 UpdateDeviceListCOM (void)
{
  UINT cDevice = 0;    // now no devices exist
  BOOL8 fValid = TRUE;  // any read operation was true
  BYTE abBuf[6], idCmd, bStatus = 0;
  DEVOBJECT *pdevobjNew = NULL, *pdevobjFirst = NULL;
  WKBSERIAL wkbsr;

  // cleans/deletes the list of detected WibuBox /ST's
  DestroyDeviceListCOM ();

  // set append position to begin of list
  pdevobjFirst = pdevobjlist; // pdevobjlist == NULL
  idCmd = BOXCMD_FSEARCH;     // first command is FSEARCH

  // check the access to COM
  fValid = AccessCOM (cstWkd.nSelPort, 0);
  if (!fValid) {
    return FALSE;             // no device found
  } // if

  // search at the chosen port until no more WibuBox has found
  do {
    fValid = ControlCOM (&idCmd, 1, abBuf, 6);

    // new WibuBox detected: copy serial number into buffer
    if (fValid) {
      bStatus = abBuf[5];
      cDevice++;     // one more WibuBox read

      // create new device entry
      pdevobjNew = (DEVOBJECT*) WkdAllocMem ((UINT)sizeof(DEVOBJECT));
      if (pdevobjNew == NULL) {
        // memory too small => return 0, exit loop
        DestroyDeviceListCOM ();        // cleanup built device table
        cDevice = 0;
        goto EXIT_UPDATE;
      } // if

      // fill the new element with the datas of the founded box
      wkbsr.ulSerial = GET_UINT32_LE(abBuf);
      wkbsr.usMask = abBuf[4];     // set mask value
      wkbsr.usReserved = 0;
      pdevobjNew->wkbsr = wkbsr;
      pdevobjNew->wkdevobj.hwkdev = (HWKDEV)NULL; // handle to device here not used
      pdevobjNew->pNext = NULL;

      // put the new element (BOX /ST) on the stack
      pdevobjNew->pNext = pdevobjFirst;
      pdevobjFirst = pdevobjNew;
      pdevobjNew = NULL;


      if (bStatus == 0) {
        idCmd = BOXCMD_NSEARCH;       // there is another WibuBox to search
      } // if
    } // if
  } while (fValid && bStatus == 0);

EXIT_UPDATE:
  // all boxes scanned: return with from function with <cbData>
  fValid = ReleaseCOM (cstWkd.nSelPort);

  // exit function
  if (fValid) {         // at least one device variant was read properly
    if (cDevice > 0) {
      cdevicelist = cDevice;          // set global variable
      pdevobjlist = pdevobjFirst;
      cstWkd.idError = WKB_NO_ERROR;    // any devices found => no error
      return TRUE;      // all okay
    } else {  // no devices found but no wrong device state: set special error  
      cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
      return FALSE;     // ERROR
    } // if
  } else {
    return FALSE;     // ERROR
  } // if

  return FALSE;
} // UpdateDeviceListCOM()


/*****************************************************************************

---- Serial (ST) Class Driver Export Functions -------------------------------

*****************************************************************************/

/*****************************************************************************
SrlReset
==============================================================================

This function resets port and its driver with the specified number is valid.

Parameters
  idPort ... contains the number of the port in the range 0..127.
  fuCtrl ... contains the reset control flags.
             WKB_RESET_SYSTEM .. (0x01) reserved for future use.
             WKB_RESET_ADDRESS   (0x02) resets the driver's profiling caching
             WKB_RESET_TIMING .. (0x04) resets the driver's timing caching
             WKB_RESET_BOXSCAN   (0x08) resets the driver's WibuBox scanning
             WKB_RESET_BOXLOCKS  (0x10) frees all select-locked WIBU-BOXes

Return
  The function returns a WibuKey error code which can be one of
  the following values:
    0  WKB_NO_ERROR
    3  WKB_ERROR_INVALID_PORT
    7  WKB_ERROR_NOT_READY

Attributes
  Class driver export function

Implementation
  This function realize the ASIC-command RESET.

*****************************************************************************/

static void SrlReset (INT idPort, UINT fuCtrl)
{
  // check, if the port available
  ASSERT (idPort >= prtgrpWkdSrl.nFirstPort
      && idPort <= prtgrpWkdSrl.nLastPort);

  // close access to the port
  if ((fuCtrl & WKB_RESET_ADDRESS) && (nPortSel != FREE_COM)) {
    ReleaseCOM (nPortSel);
    // reset the selected port
    nPortSel = FREE_COM;
  } // if

  // reset the baudrate
  gnBaudrate = 0;

  // destroy device table 
  if (fuCtrl & WKB_RESET_BOXSCAN) {
    DestroyDeviceListCOM ();
  } // if

  cBuf = 0;
} // SrlReset()


/*****************************************************************************
SrlSelect
==============================================================================

This function selects the selected port in <cstWkd.nPort> within the
class driver if this is possible.

Parameters
  nPort  contains the number of the port in the range 0..127.

Return
  The function returns TRUE if the port is available and can be selected.
  and returns FALSE if not.
  In the latter case, the last error code is set to WKB_ERROR_INVALID_PORT;

Attributes
  Class driver export function

*****************************************************************************/

static BOOL SrlSelect (INT nPort)
{
  BOOL8 fValid;

  // only one port available
  if (nPort < prtgrpWkdSrl.nFirstPort || nPort > prtgrpWkdSrl.nLastPort) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;  // bad port => return with error
    return FALSE;
  } // if

  if (gnBaudrate == 0) {
    // read the baudrate out of the registry
    gnBaudrate = WkdGetProfileInt(WKPRFTPC_DRIVER_KERNEL, "Baudrate");
    if (gnBaudrate == 0) {
      // set default baudrate = 19200
      gnBaudrate = BR_19200;
    } // if
    // transform the value of the baudrate into the divisor value
    switch (gnBaudrate) {
    #if _UNIX
      case BR_4800:
      case BR_9600:
      case BR_19200:
      case BR_38400:
      case BR_57600:
      case BR_115200:
        // right number ==>> nothing to do!!!
        break;
      default:
        // not a valid baudrate ==>> set default value
        gnBaudrate = BR_19200;
    #endif
    } // switch
  } // if

  // opens the chosen port
  fValid = AccessCOM (nPort, 0);
  if (!fValid) {
    // bad port => return with error
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;        
  } // if

  return TRUE;
} // SrlSelect()


/*****************************************************************************
SrlRelease
==============================================================================

This function terminates the last selection of the port.

Return
  The function returns the error code of the release operation. It is set to
  WKB_NO_ERROR (0) if no error occurred.

Attributes
  Class driver export function

*****************************************************************************/

static UINT SrlRelease (void)
{
  if (!(nPortSel == FREE_COM)) {
    BOOL8 fValid;
    fValid = ReleaseCOM (nPortSel);
    if (!fValid) {
      return WKB_ERROR_INVALID_PORT;  // an error occured
    } // if
  } // if 

  return WKB_NO_ERROR;  // all okay
} // SrlRelease()


/*****************************************************************************
SrlGetDeviceList
==============================================================================

This function returns a pointer to a list of all available serial numbers
at the last selected port.

Parameters
  ppdevobj  points to a variable which receives the first element of a list
            of found WibuBox devices.

Return
  The function returns the number of found devices or 0 if no device was
  found. In the latter case, the error code in <cstWkd.idError> is set to
  WKB_ERROR_NOT_CONNECTED.

Attributes
  Class driver export function

Implementation
  This function realize the ASIC-commands FSEARCH and NSEARCH.

*****************************************************************************/

static UINT SrlGetDeviceList (DEVOBJECT **ppdevobj)
{
  cstWkd.idError = WKB_NO_ERROR;

  // detect the attached boxes
  UpdateDeviceListCOM ();

  *ppdevobj = pdevobjlist;    // set the ppointer of the first element
  if (cdevicelist == 0) {
    cstWkd.idError = WKB_ERROR_NOT_CONNECTED;   // no device found
  } // if

  ASSERT(cdevicelist != 0 || cstWkd.idError != WKB_NO_ERROR);

  return cdevicelist;       // return number of found devices
} // SrlGetDeviceList()


/*****************************************************************************
SrlGetDeviceFromSerial
==============================================================================

This function searches the specified serial number in the list of read serial
numbers at the last selected port and returns a pointer to the corresponding
entry.

Parameters
  pwkbsr  points to the serial number of the device. It is assumed that the
          serial number is valid. An exception is setting this pointer to
          NULL, which is used for modifying a serial number of WIBU-BOXes
          with version 1.

Return
  The function returns a reference to the WibuKey device object or NULL if
  the device is not found.

Attributes
  Class driver export function

*****************************************************************************/

static DEVOBJECT * SrlGetDeviceFromSerial (const WKBSERIAL *pwkbsr)
{
  if (pwkbsr == NULL) {
    // no serial number set
    // * used to modify a serial number of WibuBox version 1
    // * set us mask value to 0
    devobj.wkbsr.usMask = 0;
  } else if (pwkbsr->usMask == 0 || pwkbsr->usMask > 255) {
    // only byte storable value possible: return error
    cstWkd.idError = WKB_ERROR_INVALID_PARAMETER;
    return NULL;
  } else {
    // set serial number into device object
    devobj.wkbsr = *pwkbsr;
  } // if

  // return address of this device object
  return &devobj;
} // SrlGetDeviceFromSerial()


/*****************************************************************************
SrlFindDevice
==============================================================================

This function searches a WibuBox/ST at the last selected port with an entry
which matches the specified Firm and User Code and WibuBox version.

Parameters
  pope  points to the open entry structure. This is partially initialized:
        * <flCtrl> defines all flags except WKB_OE_MASTER which is reset.
        * <hwkbsys> contains the handle of the subsystem.
        * <lPort> contains the port number.
        * <pszPort> is undefined.
        * <ulFirmCode> contains the required Firm Code
        * <ulUserCode> contains the required Firm Code
        These entry must not be changed by this function except
        the WKB_OE_MASTER flag.

Return
  The function returns TRUE if the WibuBox was found and FALSE if not. If
  the WibuBox entry was found, its serial number and entry index is stored
  into the structure.

Attributes
  Class driver export function

Implementation
  This function realize the ASIC-commnad FIND.

*****************************************************************************/

static BOOL SrlFindDevice (OPENENTRY *pope)
{
  BOOL8 fValid;
  UINT uVersion = 0;
  BYTE abAnswer[8], abBuf[7];

  // port is determined
  pope->lPort = cstWkd.nSelPort;
  // access the right port
  fValid = AccessCOM (cstWkd.nSelPort, 0);
  if (!fValid) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  }

  // set the input array for the ASIC-command FIND
  memset (abBuf, 0, sizeof(abBuf));
  // store parameters
  abBuf[0] = BOXCMD_FIND;
  abBuf[1] = LBYTE32(pope->ulFirmCode);
  abBuf[2] = MBYTE32(pope->ulFirmCode);
  abBuf[3] = NBYTE32(pope->ulFirmCode);
  abBuf[4] = LBYTE32(pope->ulUserCode);
  abBuf[5] = MBYTE32(pope->ulUserCode);
  abBuf[6] = NBYTE32(pope->ulUserCode);

  // send the ASIC-command FIND and wait for the answer
  fValid = ControlCOM (abBuf, 7, abAnswer, 8);
  if (!fValid) {
    cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
    goto EXIT_FIND;
  }

  // execution successfully and no error code returned
  // store the output parameters back
  if (fValid && abAnswer[0] == 0) {
    // no version check - all is version 4
    pope->wkbsr.ulSerial = GET_UINT32_LE((abAnswer + 1));
    pope->wkbsr.usMask = abAnswer[5];
    pope->iEntry = abAnswer[6];
    uVersion = abAnswer[7];
    // opened via master entry: set flag
    if (!(uVersion & 0x10)) { 
      pope->flCtrl |= WKB_OE_MASTER;
    } // if
    // calculate version value (0xXF = 2, 0xXE = 3, 0xXD = 4)
    pope->usBoxVersion = 0x11 - (uVersion & 0x0F);

    // WIBU-BOX found => return without error
    cstWkd.idError = WKB_NO_ERROR;

    // check if version of WibuBox is OK
    if (!WkdCheckBoxVersion((ULONG*)&(pope->flCtrl), pope->usBoxVersion, &pope->wkbsr)) {
      // invalid version of the found WibuBox => return with error
      cstWkd.idError = WKB_ERROR_INVALID_VERSION;
      goto EXIT_FIND;
    } // if
  } else {
    // no WibuBox found => return with error
    cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
    goto EXIT_FIND;
  } // if

EXIT_FIND:
  // close the port
  fValid = ReleaseCOM (cstWkd.nSelPort);

  if (cstWkd.idError == WKB_NO_ERROR) {
    // no error occured ==>> return without error
    return TRUE;
  } // if
  // an error occured ==>> return with error
  return FALSE;
} // SrlFindDevice()


/*****************************************************************************
SrlRequest
==============================================================================

This function executes a box command and sends the neccessary parameters
to the box. After sending this function waits for the answer of the box
and returns the values.

Parameters
  pwkdevreq  points to the request block which contains the request header
             and an optional request byte structure.
  pabDest .. points to a memory block which receives the read data.
  preqopt .. points to the request option structure or is NULL if no such
             structure is required. This structure contains or receives
             additional members which are request type specific. Please note
             that the BOXCMD_PROGRAM and BOXCMD_CRYPT commands always assume
             a non-NULL valid pointer.

Return
  The function returns TRUE if it was successful and FALSE if not.

Attributes
  Class driver export function

Implementation
  The function allows only the ASIC-Commands SELECT1, SELECT2, CRYPT,
  LIST and PROGRAM. The commands FIND, RESET, FSEARCH and NSEARCH will
  be implement in their own functions in this file.

*****************************************************************************/

static BOOL SrlRequest (WKDEV_REQUEST *pwkdevreq, BYTE *pabDest,
                                const REQOPTIONS *preqopt)
{
  INT nInLen = 0, nOutLen = 0;
  BOOL8 fValid = TRUE;
  BYTE *pabInBuf = NULL, *pabOutBuf = NULL;

  ASSERT(pwkdevreq != NULL && pabDest != NULL);
  // reset error code        
  cstWkd.idError = WKB_NO_ERROR;

  // verify value of entry
  if (pwkdevreq->usIndex > 9) {
    cstWkd.idError = WKB_ERROR_INVALID_ENTRY;
    return FALSE;
  } // if
  // allocate enough space for the input array
  pabInBuf = (BYTE*) WkdAllocMem (7 + pwkdevreq->cbRequest);
  if (pabInBuf == NULL) {
    goto RET_FALSE; // ERROR
  }

  // open the selected port
  fValid = AccessCOM (cstWkd.nSelPort, 0);
  if (!fValid) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    goto RET_NOT_OK;
  }

  // set the complete array, which will be send to COM (first 7 bytes)
  pabInBuf[0] = (BYTE) pwkdevreq->usCmd;                // ASIC-command
  SET_UINT32_LE(pabInBuf + 1, devobj.wkbsr.ulSerial); // = byte 1..4
  pabInBuf[5] = (BYTE) devobj.wkbsr.usMask;             // mask byte
  pabInBuf[6] = (BYTE) pwkdevreq->usIndex;              // index of entry 1..9

  // saving the output pointer
  pabOutBuf = pabDest;

  ASSERT(pwkdevreq->usCmd == BOXCMD_CRYPT
      || pwkdevreq->usCmd == BOXCMD_LIST
      || pwkdevreq->usCmd == BOXCMD_PROGRAM
      || pwkdevreq->usCmd == BOXCMD_SELECT1
      || pwkdevreq->usCmd == BOXCMD_SELECT2);

  // >>> SPECIAL CASE 1 in LIST <<<
  // the output buffer is too small ==>> use local variable
  if (pwkdevreq->usCmd == BOXCMD_LIST && preqopt != NULL) {
    // WkdGetOpenBoxEntry is calling this case
    ASSERT(pwkdevreq->cbAnswer >= 1 + 2 * 6);   // space for 2 entries !
    // set the variable back
    pabOutBuf = NULL;
    pabOutBuf = (BYTE*) WkdAllocMem (61);
    if (pabOutBuf == NULL) {
      // destroy input buffer
      // memory too small ==>> ERROR
      goto RET_NOT_OK;
    } // if
    pwkdevreq->cbRequest = 0;
    pwkdevreq->cbAnswer = 61;
  } // if

  memcpy (pabInBuf + 7, pwkdevreq->abRequest, (size_t)pwkdevreq->cbRequest);
  // length of input array
  nInLen = 7 + pwkdevreq->cbRequest;
  nOutLen = pwkdevreq->cbAnswer;

  // send the array and wait for the output
  fValid = ControlCOM (pabInBuf, nInLen, pabOutBuf, nOutLen);
  if (!fValid) {
    BsrFreeMem (pabInBuf);
    cstWkd.idError = WKB_ERROR_KRNL_COMMUNICATION;
    goto RET_NOT_OK;
  }

  // ===============================
  // analyse the command execution
  // ===============================
  // >>> SPECIAL CASE in LIST<<<
  // this is the special case from CMDBOX_LIST (see above)
  // get only one box entry from the box list
  if (fValid && pwkdevreq->usCmd == BOXCMD_LIST && preqopt != NULL) {
    INT iEntry = preqopt->List.iEntry;

    // transfer entry specific part to specified destination buffer
    pabDest[0] = pabOutBuf[0];
    // store status code (number of version)
    memcpy(pabDest + 1 + 0, pabOutBuf + 6 * iEntry + 1, 6);
    // store the main entry
    if (iEntry < 5) {
      // transfer data of added entry
      memcpy(pabDest + 1 + 6, pabOutBuf + (iEntry + 5) * 6 + 1, 6);
    } // if
    // destroy allocated temporary output memory
    if (pabOutBuf != NULL){
      BsrFreeMem(pabOutBuf);
    }//if
  } // if

  // close the selected port
  fValid = ReleaseCOM (cstWkd.nSelPort);
  if (!fValid) {
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    //!Ch:mn return FALSE;
    goto RET_NOT_OK;
  }
  // destroy allocated temporary input memory
  if (pabInBuf != NULL){
    BsrFreeMem (pabInBuf);
  }//if

  return TRUE;
RET_NOT_OK:
  // destroy input buffer if exists
  if (pabInBuf != NULL){
    BsrFreeMem (pabInBuf);
  }//if
RET_FALSE:
  return FALSE;
} // SrlRequest()


/*****************************************************************************
SrlPortData
==============================================================================

This function reads or updates the port control data structure.

Parameters
  pwkbprt  points to a WKBPORT structure which contains the source data
           (fSet = FALSE) receives the destination data (fSet = TRUE).
  fSet     contains the flag "set data instead of get data".

Return
  TRUE    if the function  was successful
  FALSE   if not (never)

Attributes
  Class driver export function

*****************************************************************************/

static BOOL SrlPortData (WKBPORT *pwkbprt, BOOL fSet)
{
  ASSERT(fSet == TRUE || fSet == FALSE);
  ASSERT_VALUE(pwkbprt);
  ASSERT(cstWkd.nSelPort >= N_MINPORT && cstWkd.nSelPort <= N_MAXPORT);

  // the variable wkbprtSrl is never used by this driver
  if (fSet) {
    // save data into local variable
    wkbprtSrl = *pwkbprt;

    // set the given values for the WibuBox/ST driver (kesrl.c)
    pusComAddr[cstWkd.nSelPort - N_MINPORT] = (USHORT) wkbprtSrl.ulAddress;
  } else {                     
    USHORT fsCtrl;
    // return last stored structure (initial all data are 0)
    SetComAddr (cstWkd.nSelPort, &fsCtrl);
    *pwkbprt = wkbprtSrl;
  } // if

  return TRUE;
} // SrlPortData()


/*****************************************************************************
SrlCleanUp
==============================================================================

This function cleans up the class driver before the complete driver is
terminated.

Parameters
  nPort  contains the port which is cleaned up.

Attributes
  Class driver export function

*****************************************************************************/

static void SrlCleanUp (INT idPort)
{
  // deletes the internal parameters of WIBU-BOXes /ST and close access to COM
  ReleaseCOM (idPort);      // close the selected port
  DestroyDeviceListCOM ();     // destroy the list of detected boxes
  cBuf = 0;
} // SrlCleanUp()

// ============================================
// Interface structure for ST driver interface
// ============================================

CLASSDRIVER clsdrvWkdSrl = {
  SrlReset,
  SrlSelect,
  SrlRelease,
  SrlGetDeviceList,
  SrlGetDeviceFromSerial,
  SrlFindDevice,
  SrlRequest,
  SrlPortData,
  SrlCleanUp,
  31                  // the number of bytes allowed to crypt in one stream
}; // clsdrvWkdSrl

PORTGROUP prtgrpWkdSrl = {
  N_MINPORT,          // = 16
  N_MAXPORT,          // = 19
  "SRL",
  "C1|C2|C3|C4",      // control flags of the WibuBox/ST
  &clsdrvWkdSrl,
  NULL                // no low level driver
}; // prtgrpWkdSrl

#endif // DRVKRNL

// =================================
// end of kesrl.c
// =================================
