/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
kesys.c                                      Kernel Specific System Adaptation
==============================================================================

This C source file implements operating system specificfunctions which are
called from the specific low level drivers for the different supported
interface and functions to management the handling in the low level part of
the driver.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu), Marcus Balb (marb), Stephan Sptitz (StS)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
93Mar11 1.33 mabu: version 1.33 released
93Mar21 1.35 mabu: adaption to Windows
93Apr13 1.35 mabu: compatible with BSRLIB
93Jun19 1.35 mabu: adapation to Win32
93Jun21 1.35 mabu: error in Timing calculation of _WkbGetSysTime() removed
93Jul22 1.35 mabu: adapation to DOS
93Aug02 1.35 mabu: _WkbGetEnvAddr() can read 32 bit values
93Aug28 2.00 mabu: adaptation to WibuKey version 2
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94May28 2.00 mabu: _Wkb... functions renamed to _Wkd...
94Sep14 2.00 mabu: BIOS RAM addressing under Windows realized via __0040H
94Oct17 2.00 marb: adaptation to Netware NLM
94Nov11 2.00 mabu: version 2.0 shipped
94Dec30 2.01 mabu: bug in environment reading under Borland DPMI removed
95Feb06 2.02 mabu: using of ENABLE/DISABLED_INTERRUPT macros
95Feb27 2.05 mabu: parsing of decimal numbers added
95Oct13 2.08 mabu: VxD registry addressing implemented
95Oct19 2.08 mabu: system T value changed to 50 microseconds
95Oct31 2.08 mabu: absolute timing mechanism implemented
96Jan30 2.10 mabu: renamed from SYS.C
96Jun25 2.13 mabu: ignoring of system port addresses in range 0..15
96Jun25 2.13 mabu: proper port acquiring functions implemented
96Oct15 2.15 mabu: "AquireWait" delay specification (driver setting) added
97Mar09 2.31 mabu: separate machine dependencies to new module KEMACH.C
97May09 2.31 mabu: bug in WkdMapPort() removed (returns always FALSE)
97May25 2.31 mabu: full Plug & Play driver implemented in VXD
98Mar05 2.41 mabu: different implementation for NT and WDM kernel drivers
98Sep29 2.51 sts:  adaptation to Borland DPMI
06Jan25 5.10 ha:   CodeMerge
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

/*****************************************************************************

---- System Port Address Management ------------------------------------------

*****************************************************************************/

/*****************************************************************************
W k d Q u e r y S y s t e m P o r t s
==============================================================================

This function replaces all default entries in the specified port specification
table to system specified ports or sets them to invalid. After returning from
this functions, the table does not contain any default entry.

The simple function variant for DOS, DOS extenders or the 16-Bit-Windows
drivers reads the system specified ports via the specified pointer from an
array of USHORT values.

Parameters
  aptsDest .. points to the array port specifications which default addresses
              are replaced.
  cDestPort   contains the number of port specifications in the table at
              <aptsDest>.

The Windows NT kernel driver variant uses following parameters:
  idDevice .. contains the type of system port. This can be one of the
              following values:
                SerialController (17) .. for COM ports
                ParallelController (20)  for LPT ports

All other drivers uses other parameters:
  pusSysAddr  points to the USHORT address table of the system.
  cSysPort .. contains the number of entries address table of the system. This
              value may be less than <cDestPort> which results in invalid
              entries which index exceed this paramter.

*****************************************************************************/

void WkdQuerySystemPorts(PORTSPEC *aptsDest, INT cDestPort,
    USHORT *pusSysAddr, INT cSysPort)
{
  INT i = 0;
  PORTSPEC *ppts = NULL;
  INT cchLength;
  char pszBuf[256];
  USHORT wklpt_adr[3];

  cchLength = WkdGetProfileValue(WKPRFTPC_DRIVER_KERNEL, WKVAR_LPT, pszBuf, 256);
  
  // bad hack, to determine WKLPT="*,/D,*"
  wklpt_adr[0] = 0;
  wklpt_adr[1] = 0;
  wklpt_adr[1] = 0;
  
  if ( (NULL != pszBuf) && (pszBuf[0] != '\0') ) {
    if ( (strcmp(pszBuf, "*,*,*") == 0) ||
         (strcmp(pszBuf, "*,/D,*") == 0) ||
         (strcmp(pszBuf, "*,/D,/D") == 0) ||
         (strcmp(pszBuf, "*,*,/D") == 0)
        ) {
      // 1st lpt allowed - 0x378
      wklpt_adr[0] = 1;
    }
    if ( (strcmp(pszBuf, "*,*,*") == 0) ||
         (strcmp(pszBuf, "*,*,/D") == 0) ||
         (strcmp(pszBuf, "/D,*,*") == 0) ||
         (strcmp(pszBuf, "/D,*,/D") == 0)
        ) {
      // 2nd lpt allowed - 0x278
      wklpt_adr[1] = 1;
    }
    if ( (strcmp(pszBuf, "*,*,*") == 0) ||
         (strcmp(pszBuf, "*,/D,*") == 0) ||
         (strcmp(pszBuf, "/D,*,*") == 0) ||
         (strcmp(pszBuf, "/D,/D,*") == 0)
        ) {
      // 3th lpt allowed - 0x3bc
      wklpt_adr[2] = 1;
    }
  }

  for (i = 0, ppts = aptsDest; i < cDestPort; i++, ppts++) {
    // address already defined => continue
    if ((INT)ppts->pta != -1) {
      continue;
    } // if
    // convert default address to corresponding system port or set to invalid
    if ( (i < cSysPort) && (pusSysAddr[i] > 15) ) {
      // address in system port list and valid address (out of 0..15):
      // store this address
        // bad hack for Linux to determine LPT-Port to be used 
        if ( (pusSysAddr[i] == 0x378) ||  (pusSysAddr[i] == 0x278) || (pusSysAddr[i] == 0x3BC) ) {
          // LPT-Port 
          if ( (i >= 0 ) && (i < 3) ) {
            ppts->pta = (wklpt_adr[i] == 1) ? pusSysAddr[i] : -1;
            // ppts->pta = pusSysAddr[i];
          } else {
            ppts->pta = -1;
          }
        } else {
          // fixme: check if this is valid 
          ppts->pta = pusSysAddr[i];
        }
    } else {
      // set this port to invalid
      ppts->pta = 0;
    } // if
  } // for
} // WkdQuerySystemPorts()


#if ABSTIME

/*****************************************************************************
W k d S t a l l E x e c u t i o n
==============================================================================

This function implements a delay routine. The exceution of the processor is
stopped for the specified number of microseconds.

Parameters
 idDelay  specifies the factor of the delay. This must be one of the
          STALL_... constants.

Comments
  This function does not disable interrupts or the task switching. Thus, the
  execution may be much more longer delayed than the specified time specifies.
  This valus is more a minimum value.

*****************************************************************************/

void WkdStallExecution(UINT idDelay)
{
  #error function WkdStallExecution is not implemented!
} // WkdStallExecution()

#else

/*****************************************************************************
W k d D e l a y T i m e
==============================================================================

This function waits for the specified T value factor cycles. During each
cycle, the specified physical port is read blindly to synchronize the delay
of the loop.

Parameters
  ptadrSync  contains the physical port number which corresponds with the
             time synchronisation port.
  uCycles .. contains the number of cycles within the loop. The waiting loop
             is executed once more than this value specifies. Thus <uCycles>
             may be 0.

*****************************************************************************/

void WkdDelayTime(PORTADDR ptaSync, UINT uCycles)
{
  volatile PORTADDR pta = ptaSync;
  volatile UINT c = uCycles;
  // one step more
  c++;
  do {
    // dummy access
#if UNIX
    pta = ptaSync;
#else
    READ_PORT_BYTE(pta);
#endif
  } while (--c != 0);
} // WkdDelayTime()

#endif // #else #if ABSTIME


// ===================================
// end of kesys.c
// ===================================
