/*****************************************************************************************
D r v C                                             Portable WibuKey Driver, Written in C
==========================================================================================
keusb.c                                    Middle-Level Driver for USB Devices
==========================================================================================

This module implements the specific middle level driver parts for the USB
interface of the WibuBox/U via the plug & play driver WIBUKEY2.SYS.

This module is only implemented in WIBUKEY.VXD and WIBUKEY.SYS.

==========================================================================================

The portable WibuKey driver source is used under the following systems:
 - WKDOS2(d).LIB - for internal use to link to DOS programs.
 - WKWIN.DLL - 16 Bit calling driver on 32 Bit Windows systems.
 - WKWIN32.DLL - 32 Bit Windows calling driver.
 - WIBUKEY.SYS - NT based kernel driver.
 - WIBUKEY.VXD - W9x based kernel driver.
 - WkMAC.LIB - MacOS 9 driver (Shared Lib).
 - WkMacLibX.Framework - MacOS X driver.
 
All sources are written in ANSI C, except the using of C++ style
single line comments.

==========================================================================================

Copyright  1998-2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu\wk\dev\runtime\drv\c
Author ... Marcellus Buchheit (mabu), Alexander Schmitt (as)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
98Mar16 2.41 mabu: initial version
99Jan07 2.52 as:   adaption to W2K
99Aug10 2.52 as:   any bugs fixed
00Sep07 3.00 as:   MacOS adaptation
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************************/

#include "proj.h"

#if USE_USB





// specifies if the USB low level driver is already accessed
static int gfIsUsbAccessed = FALSE;




// pointer to current list of detected USB WibuBox/U devices
static DEVOBJECT *pdevobjList;
// number of read devices
// * initialized with 0: no devices assumed
static UINT cDeviceLastScan = 0;
// last read list identification
// * initialized with 0: force rereading
static UINT gidListLastScan = 0;
// flag "profile is read"
static INT nIsProfileRead = -1;
// WKBPORT structure as dummy for saving & returning data
static WKBPORT wkbprtUsb;


/*****************************************************************************
CheckUSB
==============================================================================

This functiuon checks if the USB Port is already accessed or not.

Return
  The function returns TRUE if the USB driver is accessed and FALSE if not.

*****************************************************************************/

static BOOL8 CheckUSB ()
{
  // access to device driver on MacOS already exists => port is valid
  return gfIsUsbAccessed;
} // CheckUSB()


/*****************************************************************************
AccessUSB
==============================================================================

This function tries to make an access to the WibuBox/U USB Plug & Play driver
and stores a successful reference in a handle.
If an access to USB driver already exists, the function returns really fast.

Return
  The function returns TRUE if the USB driver is accessed and FALSE if not.

*****************************************************************************/

static BOOL8 AccessUSB ()
{

  // check if already an access to the USB driver succeeded
  if (CheckUSB()) {
    // USB device driver already accessed ==>> return
    return TRUE;
  } // if

  // call to the function AccessWkUSBDriver
  gfIsUsbAccessed = AccessWkUSBDriver ();
  return gfIsUsbAccessed;
} // AccessUSB()


/*****************************************************************************
ReleaseUSB
==============================================================================

This function frees the access to the USB driver. This function is normally
only called if the driver is terminated.

*****************************************************************************/

static void ReleaseUSB ()
{
  // call to the function ReleaseWkUSBDriver
  ReleaseWkUSBDriver ();
  gfIsUsbAccessed = FALSE;
} // ReleaseUSB()


/*****************************************************************************
ControlUSB
==============================================================================

This function sends a sequence (as it is described in the Adaptation Guide)
to wibukey2.sys which send it to the WibuBox/U. After sending this function
waits for the answer of the WibuBox/U an gives the result back.

Parameters
  idCtrlCode .. contains the IOCTL-Code of the current request.
  pvInput ..... contains the input sequence.
  cbInput ..... contains the length of the input sequence.
  pvOutput .... contains the output sequence.
  cbOutput .... contains the length of the output sequence.

Return
  The function returns TRUE if the USB driver is accessed and FALSE if not.

*****************************************************************************/

static BOOL8 ControlUSB (UINT idCtrlCode, void *pvInput,
    UINT cbInput, void *pvOutput, UINT cbOutput)
{
  // port access check
  if (!AccessUSB()) {
    // no access to wibukey2.sys => return with error
    cstWkd.idError = WKB_ERROR_KRNL_COMMUNICATION;
    return FALSE;
  } // if

  // call to the function ControlWkUSBDriver
  BOOL8 fValid = ControlWkUSBDriver (idCtrlCode, pvInput, cbInput, pvOutput, cbOutput);
  return fValid;

} // ControlUSB()


/*****************************************************************************
CheckVersionUSB
==============================================================================

This function checks if the used plug & play driver has a proper version and
if the version of this calling driver is accepted as client by the WibuBox/U
plug and play driver.

Return
  The function returns TRUE if the version is ok and FALSE if not. In the
  latter case, the error code is set to 

Implementation
  If any version check error occurres, an error message string is set into
  the diagnosis of the driver.

*****************************************************************************/

static BOOL8 CheckVersionUSB ()
{
  BSRMODVERSION bsrmver;

  // For MacOS there is no version checking neccassry because the
  // whole PnP management is implemented in this module and not in a
  // seperated one as on Windows systems (wibukey2.sys).
  if (!ControlUSB(IOCTL_WK_GET_VERSION, NULL, 0,
      &bsrmver, (UINT)sizeof(bsrmver))) {
    return FALSE;
  } // if

  // check if plug & play accepts this driver as client
  if (!ControlUSB(IOCTL_WK_GET_CALLVERSION, NULL, 0,
      &bsrmver, (UINT)sizeof(bsrmver))) {
    return FALSE;
  } // if
  
  // return without error
  return TRUE;
} // CheckVersionUSB()


/*****************************************************************************
DestroyDeviceListUSB
==============================================================================

This function destroys the loaded device list and frees the access to the
USB driver.

*****************************************************************************/

static void DestroyDeviceListUSB ()
{
  DEVOBJECT *pdevobj;
  DEVOBJECT *pdevobjNext;
  for (pdevobj = pdevobjList; pdevobj != NULL; pdevobj = pdevobjNext) {
    ASSERT_VALUE(pdevobj);
    // save pointer to next element
    pdevobjNext = pdevobj->pNext;
    // destroy current element
    BsrFreeMem(pdevobj);
  } // for
  // set root element to NULL (list now empty)
  pdevobjList = NULL;
  // next read must rescan list
  cDeviceLastScan = 0;
  gidListLastScan = 0;
} // DestroyDeviceListUSB()


/*****************************************************************************
UpdateDeviceListUSB
==============================================================================

This function destroys the loaded device list and frees the access to the
USB driver.

Return
  The function returns TRUE if the update was not required or valid and
  FALSE it is failed. In the latter case, no device is read and an error
  code is set in <cstWkd.idError>.

*****************************************************************************/

static BOOL8 UpdateDeviceListUSB ()
{
  size_t cbData = 0;
  WKDEVOBJ_LIST *pwkdevobjl = NULL;
  UINT cDevice = cDeviceLastScan;
  BOOL8 fValid = FALSE;

#if MAC_MACHO
  // call to the function AccessWkUSBDriver
  gfIsUsbAccessed = AccessWkUSBDriver ();
  if (!gfIsUsbAccessed) {
    return FALSE;
  } // if
#endif  
  for (;;) {
    // read new table until all table is large enough and updated
    for (;;) {
      // check if new table: support for last number of devices
      cbData = FIELDOFFSET(WKDEVOBJ_LIST, awkdevobj) +
          sizeof(WKDEVOBJ) * cDevice;
      pwkdevobjl = WkdAllocMem(cbData);

      memset(pwkdevobjl, 0, cbData);
      if (pwkdevobjl == NULL) {
        // memory too small => return with error
        ASSERT(!fValid);
        goto EXITFN;
      } // if
      pwkdevobjl->cDevice = cDevice;
      pwkdevobjl->idList = gidListLastScan;
      // get newest list of devices from Plug & Play driver
      // * send initialized header of this list
      if (!ControlUSB(IOCTL_WK_GET_DEVICE_LIST, pwkdevobjl,
          (UINT)FIELDOFFSET(WKDEVOBJ_LIST, awkdevobj), pwkdevobjl, cbData)) {
        // any error during read => return with error
        cstWkd.idError = WKB_ERROR_NOT_READY;
        goto EXITFN;
      } // if
      if (pwkdevobjl->idList == gidListLastScan) {
        // no change in read list => nothing to do, return without error
        fValid = TRUE;
        goto EXITFN;
      } // if
      // set new number of devices for create next buffer
      if (pwkdevobjl->cDevice <= cDevice) {
        // new list read and enough memory => exit loop
        // * set new valid list identification
        gidListLastScan = pwkdevobjl->idList;
        // set new size of table
        cDevice = pwkdevobjl->cDevice;
        goto EXIT_GETLIST;
      } else {
        // new devices found => create new table & read again
        // * set required size of table
        cDevice = pwkdevobjl->cDevice;
        // * destroy table which is too small
        BsrFreeMem(pwkdevobjl);
      } // if
      // read list again with new number of devices
    } // for
  EXIT_GETLIST:
    if (cDevice == 0) {
      // no device read but no error: set special error code
      DestroyDeviceListUSB ();
      fValid = TRUE;
      goto EXITFN;
    } // if
    {// first update scan:
     // * remove all devices from list which are not in new list
      DEVOBJECT **ppdevobj;
      UINT iDeviceStart = 0;
      for (ppdevobj = &pdevobjList; *ppdevobj != NULL;) {
        WKDEVOBJ wkdevobj;
        UINT iDevice = 0;
        // cache device specification which is searched
        wkdevobj = (*ppdevobj)->wkdevobj;
        for (iDevice = iDeviceStart; iDevice < cDevice; iDevice++) {
          if (!memcmp(&wkdevobj, &pwkdevobjl->awkdevobj[iDevice],
              sizeof(wkdevobj))) {
            // device found => exit loop
            break;
          } // if
        } // for
        if (iDevice == cDevice) {
          // device not found: remove from list
          *ppdevobj = (*ppdevobj)->pNext;
          // do not move search position because current element destroyed
        } else {
          // both lists are ordered: set new start position for next search
          iDeviceStart = iDevice + 1;
          ASSERT(iDeviceStart <= cDevice);
          // search at next element
          ppdevobj = &(*ppdevobj)->pNext;
        } // if
      } // for
    } // block
    {// second update scan:
      // * insert/append all new devices into list
      UINT iDevice = 0;
      DEVOBJECT **ppdevobjInsert = &pdevobjList;
      for (iDevice = 0; iDevice < cDevice; iDevice++) {
        // search device from plug&play driver in local driver device list
        DEVOBJECT **ppdevobjSearch;
        WKDEVOBJ wkdevobj;
        wkdevobj = pwkdevobjl->awkdevobj[iDevice];
        for (ppdevobjSearch = ppdevobjInsert; *ppdevobjSearch != NULL;
            ppdevobjSearch = &(*ppdevobjSearch)->pNext) {
          // device specification which is searched
          if (!memcmp(&(*ppdevobjSearch)->wkdevobj, &wkdevobj,
            sizeof(wkdevobj))) {
            // device found => exit loop
            break;
          } // if
        } // for
        if (*ppdevobjSearch != NULL) {
          // device found: set as new insert-after position for next check
          // * (both lists are ordered with old devices)
          ppdevobjInsert = &(*ppdevobjSearch)->pNext;
        } else {
          // create new device object and set at insert position
          DEVOBJECT *pdevobjNew = WkdAllocMem((UINT)sizeof(DEVOBJECT));
          if (pdevobjNew == NULL) {
            // memory too small => return with error
            ASSERT(!fValid);
            goto EXITFN;
          } // if
          // set device parameters into list object
          pdevobjNew->wkdevobj = wkdevobj;
          // set serial number mask byte to 0 to force reading serial number
          pdevobjNew->wkbsr.usMask = 0;
          // insert into list
          pdevobjNew->pNext = *ppdevobjInsert;
          *ppdevobjInsert = pdevobjNew;
          // this is also the new insert position
          ppdevobjInsert = &pdevobjNew->pNext;
        } // if
      } // for
    } // block
    // store last number of found devices
    cDeviceLastScan = cDevice;
    // free buffer for reading devices
    BsrFreeMem(pwkdevobjl);
    pwkdevobjl = NULL;
    {// read now serial numbers of all new devices
      DEVOBJECT *pdevobj;
      for (pdevobj = pdevobjList; pdevobj != NULL; pdevobj = pdevobj->pNext) {
        if (pdevobj->wkbsr.usMask == 0) {
          BYTE abAnswer[6];
          BOOL8 fValid = FALSE;
          WKDEV_REQUEST wkdevreq;
          ASSERT_VALUE(pdevobj);
          memset(abAnswer, 0x0, sizeof(abAnswer));
          wkdevreq.hwkdev = pdevobj->wkdevobj.hwkdev;
          wkdevreq.idList = gidListLastScan;
          wkdevreq.usCmd = BOXCMD_FSEARCH;
          wkdevreq.cbRequest = 0;
          wkdevreq.cbAnswer = sizeof(abAnswer);
          wkdevreq.idList = gidListLastScan;
          fValid = ControlUSB(IOCTL_WK_EXECUTE, &wkdevreq,
              (UINT)FIELDOFFSET(WKDEV_REQUEST, abRequest), abAnswer, wkdevreq.cbAnswer);
          if (!fValid) {
            // any error => start update operation again
            // * device list change during list update
            goto CONT_GET_DEVICE_LIST;
          } // if
          // initialize WKBSERIAL structure to 0
          memset(&pdevobj->wkbsr, 0, sizeof(pdevobj->wkbsr));
          // set now read data into structure
          pdevobj->wkbsr.ulSerial = GET_UINT32_LE((abAnswer + 0));
          pdevobj->wkbsr.usMask = abAnswer[4];
          // no further WibuBox may follow
          ASSERT(abAnswer[5] == 1);
        } // if
      } // for
    } // block
    // new list read & new serial numbers determined
    // * exit loop without error
    fValid = TRUE;
    goto EXITFN;
CONT_GET_DEVICE_LIST:;
  } // for

EXITFN:
  if (pwkdevobjl != NULL) {
    BsrFreeMem(pwkdevobjl);
  } // if
  if (!fValid) {
    // any error code must be set
    ASSERT(cstWkd.idError != WKB_NO_ERROR);
    // any error => destroy existing device list
    DestroyDeviceListUSB();
  } // if
  // return status
  return fValid;
} // UpdateDeviceListUSB()


/*****************************************************************************
ExecuteRequest
==============================================================================

This function executes an initialized request.

Parameters

Attributes
  Class driver export function

Implementation
  The USB driver does nothing in the reset phase.

*****************************************************************************/

static BOOL8 ExecuteRequest (WKDEV_REQUEST *pwkdevreq, BYTE *pabDest)
{
  UINT cbSend = 0;
  int i = 0;
  cbSend = FIELDOFFSET(WKDEV_REQUEST, abRequest) + pwkdevreq->cbRequest;
  for (; i < 5; i++) {
    BOOL8 fValid;
    fValid = ControlUSB(IOCTL_WK_EXECUTE, pwkdevreq, cbSend,
        pabDest, pwkdevreq->cbAnswer);
    if (fValid) {
      // operation was succesfully => return without error
      return TRUE;
    } // if
    // execution failed: update device table
    fValid = UpdateDeviceListUSB();
    if (!fValid) {
      // device table could not be updated => return with error
      ASSERT(cstWkd.idError != WKB_NO_ERROR);
      return FALSE;
    } // if
    {// check if accessed device still part of device table
      HWKDEV hwkdevRequest;
      DEVOBJECT *pdevobj = NULL;
      hwkdevRequest = pwkdevreq->hwkdev;
      for (pdevobj = pdevobjList; pdevobj != NULL; pdevobj = pdevobj->pNext) {
        // The reference is allways static because there must not be done a
        // rereding of the device. In this cse the operationg system is handling
        // the whole plug & play.
        if (hwkdevRequest == pdevobj->wkdevobj.hwkdev) {
          // device found: is still valid
          break;
        } // if
      } // for
      if (pdevobj == NULL) {
        // device not found => return with error
        cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
        return FALSE;
      } // if
    } // block
    // set new list identification into request datagram
    pwkdevreq->idList = gidListLastScan;
  } // for
  // cycle, exit only by return
  cstWkd.idError = WKB_ERROR_NOT_READY;
  return FALSE;
} // ExecuteRequest()


/*****************************************************************************

---- USB Class Driver Export Functions ---------------------------------------

*****************************************************************************/

/*****************************************************************************
UsbReset
==============================================================================

This function resets port and its driver with the specified number is valid.

Parameters
  idPort  contains the number of the port in the range 0..127.
  fuCtrl  contains the reset control flags.

Return
  The function returns a WibuKey error code which can be one of
  the following values:
    0  WKB_NO_ERROR
    3  WKB_ERROR_INVALID_PORT
    7  WKB_ERROR_NOT_READY

Attributes
  Class driver export function

Implementation
  The USB driver does nothing in the reset phase.

*****************************************************************************/

static void UsbReset (INT idPort, UINT fuCtrl)
{
  // only one port available
  if (idPort != prtgrpWkdUsb.nFirstPort) {
    // bad port => return with error
    return;
  } // if
  if (fuCtrl & WKB_RESET_BOXSCAN) {
    // destroy device table and close access to plug & play driver
    DestroyDeviceListUSB();
  } // if
  if (fuCtrl & WKB_RESET_ADDRESS) {
    // force rereading of WKUSB profile at next call
    nIsProfileRead = -1;
  } // if
} // UsbReset()


/*****************************************************************************
UsbSelect
==============================================================================

This function selects the selected port in <cstWkd.nPort> within the
class driver if this is possible.

Parameters
  nPort  contains the number of the port in the range 0..127.

Return
  The function returns TRUE if the port is available and can be selected.
  and returns FALSE if not.
  In the latter case, the last error code is set to WKB_ERROR_INVALID_PORT;

Attributes
  Class driver export function

Implementation
  The USB driver returns independent of the received port number a successful
  state if the USB plug & play driver WIBUKEY2.SYS driver can be opened and
  FALSE if not.

*****************************************************************************/

static BOOL UsbSelect (INT nPort)
{
  BOOL8 fValid = FALSE;

  // only one port available
  if (nPort != prtgrpWkdUsb.nFirstPort) {
    // bad port => return with error
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if

#if LINUX
  // For MacOS no Environment check is needed, because all Mac's of
  // the new generation has only USB support and no other one. WK also
  // supports only USB for Mac's.
  
  // read the registry key in the WibuKey environment
  if (nIsProfileRead < 0) {
    // profile not determined: read it
    // * preliminary: assume that valid
    nIsProfileRead = 1;
    if (WkdCheckEnvVar("WKPORT", "USB") < 0) {
      // WKPORT exists and USB no part of it => disable
      nIsProfileRead = 0;
    } else {
      // WKPORT does not invalidate USB: check variable WKUSB
      INT cch;
      char szBuffer[10];
      cch = WkdGetProfileValue(WKPRFTPC_DRIVER_KERNEL, "WkUsb", szBuffer, 10);
      if (cch != 0) {
        // return is not 0 ==>> environemt WkUsb set ==>> read it
        cch = WkdGetProfileInt (WKPRFTPC_DRIVER_KERNEL, "WkUsb");
        if (cch == 0) {
          // 0 as argument => disable USB access
          nIsProfileRead = 0;
        } // if
      } // if
    } // if
  } // if
  if (nIsProfileRead == 0) {
    // USB port disabled => return with error
    // * destroy device list if any exists, close PnP driver access if any
    DestroyDeviceListUSB();
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if
#endif

  // check if already an access to the USB driver succeeded
  fValid = CheckUSB ();
  if (fValid) {
    // USB device driver already accessed ==>> return
    return TRUE;
  } // if
  
  // access to the USB driver
  fValid = AccessUSB();
  if (!fValid) {
    // no access of USB device driver => return with error code
    cstWkd.idError = WKB_ERROR_INVALID_PORT;
    return FALSE;
  } // if
  // check version of plug & play driver
  if (!CheckVersionUSB()) {
    // bad version of plug & play driver => return with error
    // * destroy device list if any exists, close PnP driver access if any
    DestroyDeviceListUSB ();
    ReleaseUSB ();
    cstWkd.idError = WKB_ERROR_KRNL_COMMUNICATION;
    return FALSE;
  } // if
  // PnP driver newly accessed: return without error
  return TRUE;
} // UsbSelect()


/*****************************************************************************
UsbRelease
==============================================================================

This function terminates the last selection of the port.

Return
  The function returns the error code of the release operation. It is set to
  WKB_NO_ERROR (0) if no error occurred.

Attributes
  Class driver export function

Implementation
  The function does nothing and returns always WKB_NO_ERROR.

*****************************************************************************/

static UINT UsbRelease (void)
{
  // close the access to wibukey2.sys
  //!I:as ReleaseUSB ();
  // The handle to wibukey2.sys will never be close because of
  // the better performance. For this case in the procedure 
  // ControlUSB any special checkings have to be done. Without
  // these checkings there could be some problems with the access
  // to wibukey2.sys because the handle is not valid any more.
  return WKB_NO_ERROR;
} // UsbRelease()


/*****************************************************************************
UsbGetDeviceList
==============================================================================

This function returns a pointer to a list of all available serial numbers
at the last selected port.

Parameters
  ppdevobj  points to a variable which receives the first element of a list
            of found WibuBox devices.

Return
  The function returns the number of found devices or 0 if no device was
  found. In the latter case, the error code in <cstWkd.idError> is set to
  WKB_ERROR_NOT_CONNECTED.

Attributes
  Class driver export function

*****************************************************************************/

static UINT UsbGetDeviceList (DEVOBJECT **ppdevobj)
{
  cstWkd.idError = WKB_NO_ERROR;
  // update device list if required
  UpdateDeviceListUSB ();
  // set pointer of first element of created list (NULL if none)
  *ppdevobj = pdevobjList;
  if (cDeviceLastScan == 0) {
    // no device found: set special error if no other set
    if (cstWkd.idError == WKB_NO_ERROR) {
      cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
    } // if
  } // if
  // return number of found devices
  ASSERT( (cDeviceLastScan != 0) || (cstWkd.idError != WKB_NO_ERROR) );
  return cDeviceLastScan;
} // UsbGetDeviceList()


/*****************************************************************************
UsbGetDeviceFromSerial
==============================================================================

This function searches the specified serial number in the list of read serial
numbers at the last selected port and returns a pointer to the corresponding
entry.

Parameters
  pwkbsr  points to the serial number of the device. It is assumed that the
          serial number is valid. An exception is setting this pointer to
          NULL, which is used for modifying a serial number of WibuBoxes
          with version 1.

Return
  The function returns a reference to the WibuKey device object or NULL if
  the device is not found.

Attributes
  Class driver export function

Implementation
  The support of invalid serial numbers is not supported because a WibuBox/U
  has always version 4 or higher.

*****************************************************************************/

static DEVOBJECT * UsbGetDeviceFromSerial (const WKBSERIAL *pwkbsr)
{
  if (pwkbsr != NULL) {
    DEVOBJECT *pdevobj = NULL;
    ASSERT_VALUE(pwkbsr);
    cstWkd.idError = WKB_NO_ERROR;

    // update device list if required
    UpdateDeviceListUSB ();
    if (pwkbsr->usMask > 200 ) {
      ASSERT(0);
    } // if
  
    for (pdevobj = pdevobjList; pdevobj != NULL; pdevobj = pdevobj->pNext) {
      // search the rigth serial number of the WibuBox/U
      if (   (pdevobj->wkbsr.usMask == pwkbsr->usMask)
          && (pdevobj->wkbsr.ulSerial == pwkbsr->ulSerial) ) {
        // serial number found => return this address
        return pdevobj;
      } // if
    } // for
  } // if
  // serial number not found: set error code, return NULL
  cstWkd.idError = WKB_ERROR_SERIAL_NOT_FOUND;
  return NULL;
} // UsbGetDeviceFromSerial()


/*****************************************************************************
UsbFindDevice
==============================================================================

This function searches a WibuBox/U at the last selected port with an entry
which matches the specified Firm and User Code and WibuBox version.

Parameters
  pope  points to the open entry structure. This is partially initialized:
        * <flCtrl> defines all flags except WKB_OE_MASTER which is reset.
        * <hwkbsys> contains the handle of the subsystem.
        * <lPort> contains the port number.
        * <pszPort> is undefined.
        * <ulFirmCode> contains the required Firm Code
        * <ulUserCode> contains the required Firm Code
        These entry must not be changed by this function except
        the WKB_OE_MASTER flag.

Return
  The function returns TRUE if the WibuBox was found and FALSE if not. If
  the WibuBox entry was found, its serial number and entry index is stored
  into the structure.

Attributes
  Middle level driver export function

*****************************************************************************/

static BOOL UsbFindDevice (OPENENTRY *pope)
{
  struct {
    WKDEV_REQUEST wkdevreq;
    BYTE ab[6];
  } s;
  DEVOBJECT *pdevobj = NULL;

  // update device list if required
  UpdateDeviceListUSB ();
  if (cDeviceLastScan != 0) {
    // device list read => search entry sequentially
    memset(&s, 0, sizeof(s));
    // store parameters
    s.wkdevreq.usCmd = BOXCMD_FIND;
    s.wkdevreq.cbRequest = 6;
    s.wkdevreq.cbAnswer = 8;
    s.wkdevreq.idList = gidListLastScan;
    // store argument
    s.wkdevreq.abRequest[0] = LBYTE32( pope->ulFirmCode );
    s.wkdevreq.abRequest[1] = MBYTE32( pope->ulFirmCode );
    s.wkdevreq.abRequest[2] = NBYTE32( pope->ulFirmCode );
    s.wkdevreq.abRequest[3] = LBYTE32( pope->ulUserCode );
    s.wkdevreq.abRequest[4] = MBYTE32( pope->ulUserCode );
    s.wkdevreq.abRequest[5] = NBYTE32( pope->ulUserCode );
    // search now first found WibuBox/U in list of devices
    for (pdevobj = pdevobjList; pdevobj != NULL; pdevobj = pdevobj->pNext) {
      WKDEVOBJ *pwkdevobj = &pdevobj->wkdevobj;
      BOOL8 fValid;
      BYTE abAnswer[8];
      memset(abAnswer, 0x0, sizeof(abAnswer));
      // set handle of this device
      s.wkdevreq.hwkdev = pwkdevobj->hwkdev;
      fValid = ExecuteRequest(&s.wkdevreq, abAnswer);
      if (fValid && abAnswer[0] == 0) {
        // execution successfully and no error code returned
        // * store back parameters
        UINT uVersion = 0;
        pope->wkbsr.ulSerial = GET_UINT32_LE((UINT*)(abAnswer + 1));
        pope->wkbsr.usMask = (USHORT)abAnswer[5];
        pope->iEntry = (USHORT)abAnswer[6];
        uVersion = (UINT)abAnswer[7];
        if (!(uVersion & 0x10)) {
          // opened via master entry: set flag
          ULONG temp = GET_UINT32_LE(&pope->flCtrl);
          temp  |= WKB_OE_MASTER;
          SET_UINT32_LE(&pope->flCtrl, temp);
        } // if
        // calculate version value (0xXF = 2, 0xXE = 3, 0x=D = 4)
        pope->usBoxVersion = (USHORT)(0x11 - (uVersion & 0x0F));
        // check if version of WibuBox is OK
        if (WkdCheckBoxVersion((ULONG*)&(pope->flCtrl), (USHORT)pope->usBoxVersion, 
              &pope->wkbsr)) {
          // WibuBox found => return without error
          cstWkd.idError = WKB_NO_ERROR;
          return TRUE;
        } else { 
          // algo not supported, not a valid version of the wibubox
          cstWkd.idError = WKB_ERROR_INVALID_VERSION;
          return FALSE;
        } // if(
      } // if
      // WibuBox does not match: search next
    } // for
  } // if

  // no WibuBox found => return with error
  cstWkd.idError = WKB_ERROR_ENTRY_NOT_FOUND;
  return FALSE;
} // UsbFindDevice()


/*****************************************************************************
UsbRequest
==============================================================================

This function returns a pointer to a list of all available serial numbers
at the specified port.

Parameters
  pwkdevreq  points to the request block which contains the request header
             and an optional request byte structure.
  pabDest .. points to a memory block which receives the read data.
  preqopt .. points to the request option structure or is NULL if no such
             structure is required. This structure contains or receives
             additional members which are request type specific. Please note
             that the BOXCMD_PROGRAM and BOXCMD_CRYPT commands always assume
             a non-NULL valid pointer.

Return
  The function returns TRUE if it was successful and FALSE if not.

Attributes
  Class driver export function

*****************************************************************************/

static BOOL UsbRequest (WKDEV_REQUEST *pwkdevreq, BYTE *pabDest,
    REQOPTIONS *preqopt)
{
  BOOL8 fValid = FALSE;
  BYTE *pabBuf = NULL;
  BYTE *pbDest = pabDest;

  // reset error code
  cstWkd.idError = WKB_NO_ERROR;
  if (pwkdevreq->usCmd == BOXCMD_LIST && preqopt != NULL) {
    // WibuBox list command and options specified: read only one index
    ASSERT(pwkdevreq->cbAnswer >= 1 + 2 * 6);
    // * read into separate buffer which is created locally
    pabBuf = WkdAllocMem(61);
    if (pabBuf == NULL) {
      // memory too small => return with error
      return FALSE;
    } // if
    // read into this buffer 61 bytes
    pwkdevreq->cbAnswer = 61;
    pbDest = pabBuf;
  } // if
  // set current list identification into request datagram
  pwkdevreq->idList = gidListLastScan;
  // execute WibuBox/U request, store data into <pbDest>
  fValid = ExecuteRequest (pwkdevreq, pbDest);
  // analyse execution
  if (fValid && pabBuf != NULL) {
    // transfer entry specific part to specified destination buffer
    INT iEntry = preqopt->List.iEntry;
    // store version byte
    pabDest[0] = pabBuf[0];
    memcpy(pabDest + 1 + 0, pabBuf + 6 * iEntry + 1, (size_t)6);
    if (iEntry < 5) {
      // transfer data of added entry
      memcpy(pabDest + 1 + 6, pabBuf + (iEntry + 5) * 6 + 1, (size_t)6);
    } // if
  } // if
  if (fValid && pwkdevreq->usCmd == BOXCMD_PROGRAM && preqopt != NULL) {
    ASSERT_VALUE(preqopt);
    if (preqopt->Program.fSetSerialNumber) {
      // serial number changed: invalidate accessed device
      // * this forces rereading of the device table with new serial number
      fValid = ControlUSB(IOCTL_WK_INVALIDATE_DEVICE, pwkdevreq,
          (UINT)FIELDOFFSET(WKDEV_REQUEST, abRequest), NULL, 0);
    } // if
  } // if
  // destroy allocated block
  if (pabBuf != NULL) {
    BsrFreeMem (pabBuf);
  } // if
  // return determined state
  return (BOOL)fValid;
} // UsbRequest()


/*****************************************************************************
UsbPortData
==============================================================================

This function reads or updates the port control data structure.

Parameters
  pwkbprt  points to a WKBPORT structure which contains the source data
           (fSet = FALSE) receives the destinatino data (fSet = TRUE).
  fSet ... contains the flag "set data instead of get data".

Return
  The function returns TRUE if it was successful and FALSE if not. In the
  latter case, <cstWkd.idError> contains an error code.

Attributes
  Class driver export function

Implementation
  The USB port stores and returns the WKPORT structure without doing anything
  within the driver itself.

*****************************************************************************/

static BOOL UsbPortData (WKBPORT *pwkbprt, BOOL fSet)
{
  ASSERT(fSet == TRUE || fSet == FALSE);
  ASSERT_VALUE(pwkbprt);
  if (fSet) {
    // save data into local variable
    // * this variable is not used by the driver
    wkbprtUsb = *pwkbprt;
  } else {
    // return last stored structure (initial all data are 0)
    *pwkbprt = wkbprtUsb;
  } // if
  return TRUE;
} // UsbPortData()


/*****************************************************************************
UsbCleanUp
==============================================================================

This function cleans up the class driver before the complete driver is
terminated.

Parameters
  nPort  contains the port which is cleaned up.

Attributes
  Class driver export function

Implementation
  The USB driver frees all allocated memory; especially the dynamically
  created list.

*****************************************************************************/

static void UsbCleanUp (INT nPort)
{
  // destroy device list and close access to plug & play driver
  DestroyDeviceListUSB ();
  ReleaseUSB ();
} // UsbCleanUp()


// ============================================
// Interface structure for USB driver interface
// ============================================

CLASSDRIVER clsdrvWkdUsb = {
  UsbReset,
  UsbSelect,
  UsbRelease,
  UsbGetDeviceList,
  UsbGetDeviceFromSerial,
  UsbFindDevice,
  UsbRequest,
  UsbPortData,
  UsbCleanUp,
  31
}; // clsdrvWkdUsb

PORTGROUP prtgrpWkdUsb = {
  48,
  48,
  "USB",
  "C1|C2|C3|C4",
  &clsdrvWkdUsb,
  NULL
}; // prtgrpWkdUsb

#endif // USE_USB

// =================================
// wibu\wk\dev\runtime\drv\c: end of keusb.c
// =================================
