/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
mncommon.c:                                     Middle Management common parts
==============================================================================

This C source file implements functions which are driver internally called
from several parts of the middle management.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1994 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu), Heiko Abraham (ha)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
96Jan29 2.10 mabu: initial version, built from CTRL.C and NTDISP.C
96Jul11 2.11 mabu: bug in memory management (leaks by handles) removed
96Oct24 2.20 mabu: automatic freeing of opened WIBU-BOX entries when terminate
96Oct24 2.20 mabu: WKB_OPEN_GLOBAL management added
97Jun18 2.31 mabu: keeping management added
98Mar05 2.41 mabu: bug in reducing handle table removed
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

sem_t* gThreadSemaphoreID = NULL;

// universal driver status variable
INT gidWkdLastError = WKB_NO_ERROR;

// global driver status
CALLSTATUS cstWkd = {
  0,     // fuSysType
  0,     // idError
  0,     // flCtrl
  (ULONG)-1, // flCtrlDefMask: all bits permitted
  0,     // flCtrlDefSet
  0,     // flGlobalInternal
  -1,    // nSelPort: no port selected
  -1,    // iSelPortGroup; no port group selected
  NULL,  // pwkprtSel: no port group selected
  NULL,  // pclsdrvSel; no port class selected
  FALSE, // fAccessed: no port accessed
#if ASYNCMODE
  NULL,  // pasacThis
#endif
  NULL,  // hwkbsys
  NULL,  // psys
  0,     // bLevel
  FALSE, // fUnicodeCal
#if ASYNCMODE
  FALSE,  // fAsyncCall
#endif
  NULL,  // pope
#if NTKRNL || WINVXD
  -1,    // idProcess
#endif
};


#if defined(__APPLE__) || defined(__linux__)
  pthread_mutex_t api_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

/*****************************************************************************

---- Memory Management -------------------------------------------------------

*****************************************************************************/

/*****************************************************************************
WkdAllocMem
==============================================================================

This function calls the based WklAllocMem() function and returns the address
of the allocated memory. Additionally, the function sets the error code
WKB_ERROR_MEMORY_TOO_SMALL if the requested memory could not be allocated.

Parameters
  cb  contains the length of the requested memory block. This value must not
      be 0, to guarantee a proper handling with different memory management
      schemes.

Globals
  cstWkd*

Return
  The function returns the address of the allocated memory or NULL if the
  allocation was not possible.

*****************************************************************************/

void * WkdAllocMem (size_t cb)
{
  void *pv;
  ASSERT(cb != 0);
  pv = BsrAllocMem(cb);
  if (pv) {
    return pv;
  } // if
  // memory too small => set error code
  cstWkd.idError = WKB_ERROR_MEMORY_TOO_SMALL;
  return NULL;
} // WkdAllocMem()

#if USE_KEEPMEM

/*****************************************************************************

---- Keeping memory management -----------------------------------------------

*****************************************************************************/

/*****************************************************************************
KEEPMEM                                 Keep Memory Block Management Structure
==============================================================================

This structure handles a single keep-memory block. A keep-memory block is
stored in a separate connected list to permit the checking and freeing of all
keep-memory blocks independent of other allocated memory.

Keep-memory objects are addressed by type specific handles. Each handle
value is individual created and is unique. Thus, it is impossible that a
discarded keep memory object can be used as a zombie.

Members
-------
pNext ...  points to the next item in the list of linked KEEPMEM elements.
hkm ...... contains the handle which specifies this keep memory object.
ulTimeOut  contains the timeout tick after this the keep-memory object is
           discarded.
abData ... contains as stub declarations the contents data.

*****************************************************************************/

typedef struct _KEEPMEM {
  struct _KEEPMEM *pNext;
  HKEEPMEM hkm;
  ULONG ulTimeOut;
  BYTE abData[1];
} KEEPMEM; // prefix "kpm"

// pointer to the root of allocated keep-memory objects
static KEEPMEM *pkpmRoot;

// counter to create unique keep-memory handles
static UINT idKeepMem = 0;

/*****************************************************************************
WkdAllocKeepMem
==============================================================================

This function allocates a new keep memory object with the specified number of
bytes, inserts it into the list of keep-memory blocks and sets the time out
value.

Parameters
  cbData ... contains the number of bytes in the keep memory area. For 16
             bits, the maximum size is 65500 bytes.
  ulTimeOut  contains the time out value in milliseconds. A typical value is
             10000 for 10 seconds.
  phDest ... points to a HKEEPMEM variable which receives the handle value.


Return
  The function returns NULL if memory is too small. In this case
  <cstWkd.idError> contains an error code. Otherwise the pointer to the
  contents area of the created block is returned.

Comments
  The created memory block is not initialized.

*****************************************************************************/

void *WkdAllocKeepMem (UINT cbData, ULONG ulTimeOut, HKEEPMEM *phDest)
{
  KEEPMEM *pkpmNew;
  pkpmNew = WkdAllocMem(cbData + FIELDOFFSET(KEEPMEM, abData));
  if (pkpmNew == NULL) {
    // block not allocated: return with error
    return NULL;
  } // if
  // create & store handle value
  if (idKeepMem == 0) {
    // handle of 0 never used => get next
    idKeepMem++;
  } // if
  pkpmNew->hkm = (HKEEPMEM)(unsigned long)idKeepMem;
  // create next object, overrun to 0 handled by next call
  idKeepMem++;
  // store back created handle value
  *phDest = pkpmNew->hkm;
  // insert block at begin of keep-memory list
  pkpmNew->pNext = pkpmRoot;
  pkpmRoot = pkpmNew;
  pkpmNew->ulTimeOut = GetTickCount() + ulTimeOut;
  // return pointer to contents area
  return pkpmNew->abData;
} // WkdAllocKeepMem()


/*****************************************************************************
WkdGetKeepMem
==============================================================================

This function searchs for a keep-memory block with the specified handle. If
the block is found, the time-out value is checked. It the block is out of
time, is automatically discarded. Otherwise a pointer to the information
context of the block is returned.

Parameters
  hkm  contains the the handle of the keep memory which should be deleted. If
       this handle is NULL, nothing happens and NULL is returned.

Return
  The function returns NULL if the keep-memory object is not found or is
  out of time. Otherwise a pointer to the information context of the block is
  returned.

*****************************************************************************/

void *WkdGetKeepMem (HKEEPMEM hkm)
{
  KEEPMEM **ppkpmFind;
  if (hkm == NULL) {
    // empty handle => nothing to do
    return NULL;
  } // if
  ppkpmFind = &pkpmRoot;
  // search keep memory object in list
  while (*ppkpmFind != NULL) {
    KEEPMEM *pkpmThis = *ppkpmFind;
    if (pkpmThis->hkm == hkm) {
      // block found: check the time-out value
      if ((INT)(GetTickCount() - pkpmThis->ulTimeOut) < 0) {
        // block found and in time: return pointer to information
        return pkpmThis->abData;
      } // if
      // remove block from keep-memory list
      *ppkpmFind = pkpmThis->pNext;
      BsrFreeMem(pkpmThis);
      // return with state "not found"
      return NULL;
    } // if
    // check next item
    ppkpmFind = &pkpmThis->pNext;
  } // while
  // not found => return with state "not found"
  return NULL;
} // WkdGetKeepMem()


/*****************************************************************************
WkdDiscardKeepMem
==============================================================================

This function discards the specified keep-memory block independent of its
stored time out value. Normally this function is called when the stored block
is no longer required.

Parameters
  hkm  contains the the handle of the keep memory which should be deleted.
       If this handle is NULL, nothing happens and FALSE is returned.

Return
  The function returns TRUE if the discard operation is executed and FALSE
  if the keep-memory block which the specified handle was not found.

*****************************************************************************/
BOOL WkdDiscardKeepMem (HKEEPMEM hkm)
{
  KEEPMEM **ppkpmFind;
  if (hkm == NULL) {
    // empty handle => nothing to do
    return FALSE;
  } // if
  ppkpmFind = &pkpmRoot;
  // search keep memory object in list
  while (*ppkpmFind != NULL) {
    KEEPMEM *pkpmThis = *ppkpmFind;
    if (pkpmThis->hkm == hkm) {
      //// block found: discard it
      // remove block from keep-memory list
      *ppkpmFind = pkpmThis->pNext;
      BsrFreeMem(pkpmThis);
      // return with state "discarding executed"
      return TRUE;
    } // if
    // check next item
    ppkpmFind = &pkpmThis->pNext;
  } // while
  // not found => return with FALSE (no discarding executed)
  return FALSE;
} // WkdDiscardKeepMem()


/*****************************************************************************
WkdDiscardAllKeepMem
==============================================================================

This function discards all keep-memory blocks which are allocated,
independent of their time out value. Normally this function is called when
the driver is terminated or when the available memory runs out.

*****************************************************************************/

void WkdDiscardAllKeepMem (void)
{
  KEEPMEM *pkpmThis;
  pkpmThis = pkpmRoot;
  while (pkpmThis != NULL) {
    KEEPMEM *pkpmNext;
    pkpmNext = pkpmThis->pNext;
    BsrFreeMem(pkpmThis);
    pkpmThis = pkpmNext;
  } // while
  // list now emptyx
  pkpmRoot = NULL;
} // WkdDiscardAllKeepMem()

#endif // #if USE_KEEPMEM

/*****************************************************************************

---- Handle based memory management ------------------------------------------

*****************************************************************************/

// number of reserved handles
#define HNDRES 16

/*****************************************************************************
HNDPT                                             Handle Table Entry Structure
==============================================================================

This structure specifies a single entry in the handle array which allocates
and manages all handles which are managed via the driver.

Members
-------
uType
    contains the type of the handle and is one of HTYP_... values
    (HTYP_NULL, HTYP_SYS, HTYP_ENTRY, HTYP_RMTENTRY, HTYP_MEMBLOCK).
    HTYP_SHMLINUX only for Linux.
ulData
    contains 32 bit data which doesn't represent a valid address.
pvData
    contains a pointer to the data which relates with the handle.
idProcess
    contains the identification of the process/file which has created the
    handle.

*****************************************************************************/

typedef struct {
  // prefix "hpt"
  UINT uType;
  union {
    ULONG ulData;
    void *pvData;
  } D;
} HNDPT;

    // number handles in <pahpt> array, 0 if no array
	static UINT cNumberOfHandles = 0;

	// number of next handle after filled table area
	static UINT uNextHandle = 0;

	// pointer to array of handles
	static HNDPT *pahpt = NULL;


#if defined(__APPLE__) || defined(__linux__)

#if defined(__linux__)
  #include <linux/version.h>
#else 
  #define KERNEL_VERSION(a,b,c) a##b##c
  #define LINUX_VERSION_CODE 240
#endif


/*****************************************************************************
WkdEnterCS
==============================================================================

This function enters/synchronize the access to a critical section.

Parameter
---------
pidSemaphore
    contains the ID if the semaphore.
pszSemahore
    contains the name of the semaphore.
    
Return
------
This funtion return TRUE, if the access can be done,
FALSE if an error occured.

*****************************************************************************/
#if defined(__APPLE__) || ( defined(__linux__) && (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,1) ))
#else
static int fdFutex = 0;
#endif 

BOOL WkdEnterCS(sem_t** pidSemaphore, const char* pszSemahore)
{
#if defined(__APPLE__) || ( defined(__linux__) && (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,1) ))
  // create a named semaphore using a string version of our key to identify it.
  if (*pidSemaphore == NULL) {
    *pidSemaphore = sem_open(pszSemahore, O_CREAT, GLOBAL_PERMISSIONS, 1);
    if ( *pidSemaphore == SEM_FAILED) {
      // Semaphore cannot be created - sem_open(O_CREAT) failed 
      perror("sem_open() failed");
      return FALSE;
    } // if
    sem_init(*pidSemaphore, TRUE, 0);
  } else {
    ASSERT(*pidSemaphore != NULL);
    if((sem_wait(*pidSemaphore) == -1)) {
      // exit only if a normal error occures
      if (errno != EAGAIN) {
        // normally return with error
        perror("sem_wait() failed");
        return FALSE;
      } // if
    } // if
  } // if
  return TRUE;
#else 
  // oh, bad, sem_open not implemented 
  // simulate it by using file-lock 
  struct stat buf;
  int fd = 0;
  char szFutexName[32];
  if ( 0 == stat( "/dev/shm/", &buf) ) {
    // write it to share memory 
    snprintf(szFutexName, sizeof(szFutexName), "/dev/shm/%s", pszSemahore);
    szFutexName[sizeof(szFutexName)-1] = 0;
  } else {
    // bad - write it to harddrive/flash
    snprintf(szFutexName, sizeof(szFutexName), "/tmp/%s", pszSemahore);
    szFutexName[sizeof(szFutexName)-1] = 0;
  }
  if ( 0 != stat( szFutexName, &buf) ) {
    // create it 
    fd = open(szFutexName, O_CREAT|O_SYNC|O_WRONLY);
    if ( fd >= 0 ) {
      close(fd);
      fd = 0;
    }
  }
  fdFutex = open(szFutexName, O_SYNC|O_WRONLY);
  if ( fdFutex >= 0 ) {
    if ( 0 == flock(fdFutex, LOCK_EX)) {
      return TRUE;
    }
  }
  perror("open(szFutexName, O_SYNC)");
  fdFutex = -1;
  return FALSE;
#endif
} // WkdEnterCS()


/*****************************************************************************
WkdLeaveCS
==============================================================================

This function exits the synchronisation of a critical section.

Parameter
---------
pidSemaphore
    contains the ID if the semaphore.

*****************************************************************************/

void WkdLeaveCS(sem_t** pidSemaphore)
{
#if defined(__APPLE__) || ( defined(__linux__) && (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,1) ))
  // release semaphore
  if (*pidSemaphore != NULL) {
    sem_post(*pidSemaphore);
  } // if
#else 
  if ( fdFutex >= 0 ) {
    flock(fdFutex, LOCK_UN);
    close(fdFutex);
  } else {
    perror("close(fdFutex=-1);");
  }
#endif
} // WkdLeaveCS()

#endif 

/*****************************************************************************
WkdCreateHandle
==============================================================================

This function searchs a new entry in the handle table, sets it to an empty
entry, allocates a memory block with the specified length, fills it with 0
bytes and stores its address in the table entry. Normally the address of the
created block is returned. If the specified length is 0, no additional memory
block is allocated and the address of the entry in the handle table is
returned.
If the handle could not be created, the error WKB_ERROR_NO_MORE_HANDLES is
set.

Parameters
  uType   contains the type of handle which is stored in the table entry.
  cbSize  contains the needed size of memory block in bytes.
  phDest  points to a HANDLE variable which receives the handle value.

Return
  If the handle is allocated, at <phDest> the handle index is stored. If
  <cSize> is not NULL, the address of the allocated memory is returned, other-
  wise the address of the handle in table. If the handle is not allocated,
  <phDest> receives NULL (0) and the function returns NULL.

Comments
  The returned address of the table handle is only valid until a further
  WkdCreateHandle() or WkdDestroyHandle() function call occurs.

*****************************************************************************/

void * WkdCreateHandle (UINT uType, UINT cbSize, HANDLE *phDest)
{
  HNDPT *phpt;
  void *pvBuf = NULL;
  if (cbSize != 0) {
    // allocate memory block with destination data
    pvBuf = BsrAllocMem(cbSize);
    if (pvBuf == NULL) {
      // memory too small
      goto FNERROR;
    } // if
    // initialize with 0 bytes
    memset(pvBuf, 0, cbSize);
  } // if
  if (uNextHandle < cNumberOfHandles) {
    // get quick free entry at end of table
    phpt = pahpt + uNextHandle;
    uNextHandle++;
  } else {
    // search free entry from start of table
    HNDPT *phptEnd = pahpt + cNumberOfHandles;
    phpt = pahpt;
    while (phpt != phptEnd) {
      if (phpt->uType == HTYP_NULL) {
        break;
      } // if
      phpt++;
    } // while
    if (phpt == phptEnd) {
      // no free entry: increase table by 16 entries
      UINT cNewHandle;
      HNDPT *pahptNew;
      if (cNumberOfHandles >= 10000) {
        // no more than 10000 entries
        goto FNERROR;
      } // if
      cNewHandle = cNumberOfHandles + 16;
      // <cNewHandle> must be always a multiple of 16
      ASSERT(cNewHandle % 16 == 0);
      pahptNew = BsrAllocMem(cNewHandle * sizeof(HNDPT));
      if (!pahptNew) {
        // memory too small
        goto FNERROR;
      } // if
      if (cNumberOfHandles != 0) {
        // copy data from previous block, free this block
        memcpy(pahptNew, pahpt, cNumberOfHandles * sizeof(HNDPT));
        BsrFreeMem(pahpt);
      } // if
      // set new dimension of table
      pahpt = pahptNew;
      cNumberOfHandles = cNewHandle;
      // get new free position (always garantueed because new allocated)
      phpt = pahpt + uNextHandle;
      // index of new position
      uNextHandle++;
      // fill new created memory area
      memset(phpt, 0, sizeof(HNDPT) * 16);
    } // if
  } // if
  // initialize table entry
  phpt->uType = uType;
  // calculate handle value: index in table + HNDRES (= 16)
  *phDest = (HANDLE)(phpt - pahpt + HNDRES);
  if (cbSize == 0) {
    // no additional memory needed => return address of table entry
    return phpt;
  } // if
  // set address of created block into handle table & return this address
  phpt->D.pvData = pvBuf;
  return pvBuf;
FNERROR:
  // no allocation: set error code
  cstWkd.idError = WKB_ERROR_NO_MORE_HANDLES;
  // free allocated memory
  if (pvBuf != NULL) {
    BsrFreeMem(pvBuf);
  } // if
  // no handle created
  *phDest = NULL;
  return NULL;
} // WkdCreateHandle()


/*****************************************************************************
WkdGetHandlePtr
==============================================================================

This function determines the address of the block which corresponds with the
specified handle.

Parameters
  h  contains the handle which data address block is specified.

Return
  The address of the handle data is returned.

Comments
  This function is only permitted for handles which have a separately allocated
  memory block.

*****************************************************************************/

void * WkdGetHandlePtr (HANDLE h)
{
  HNDPT *phpt;
  UINT uPos = (UINT)(unsigned long)(h - HNDRES);
  if (uPos >= uNextHandle) {
    // illegal handle => return
    return NULL;
  } // if
  phpt = pahpt + uPos;
  if (phpt->uType == HTYP_NULL) {
    return NULL;
  } // if
  // validate type of handle
  ASSERT(phpt->uType == HTYP_SYS ||
         phpt->uType == HTYP_ENTRY || phpt->uType == HTYP_RMTENTRY ||
         phpt->uType == HTYP_MEMBLOCK);
  // return address of allocated data block
  return phpt->D.pvData;
} // WkdGetHandlePtr()


/*****************************************************************************
WkdGetHandleType
==============================================================================

This function returns the type of the specified handle.

Parameters
  h  contains the handle which data address block is specified.

Return
  The type of the handle is returned.

*****************************************************************************/

UINT WkdGetHandleType (HANDLE h)
{
  UINT uPos;
  if(h == NULL) {
    return (UINT)HTYP_NULL;
  }
  uPos = (UINT)(unsigned long)(h - HNDRES);
  if (uPos >= uNextHandle) {
    // illegal handle => return
    return (UINT)HTYP_NULL;
  } // if
  return pahpt[uPos].uType;
} // WkdGetHandleType()


/*****************************************************************************
WkdGetHandleData
==============================================================================

This function determines the direct data in the handle table value which
corresponds with the specified handle.

Parameters
  h  contains the handle which data address block is specified.

Return
  The value of the handle data is returned.

Comments
  This function is only permitted for handles which have no separately
  allocated memory block.

*****************************************************************************/

ULONG WkdGetHandleData (HANDLE h)
{
  UINT uPos = (UINT)(unsigned long)(h - HNDRES);
  ASSERT(uPos < uNextHandle);
  if (uPos >= uNextHandle) {
    // illegal handle => return
    return 0;
  } // if
  return pahpt[uPos].D.ulData;
} // WkdGetHandleData()


/*****************************************************************************
WkdSetHandleData
==============================================================================

This function sets the direct data in the handle table value which
corresponds with the specified handle.

Parameters
  h  contains the handle which data address block is specified.

Return
  The value of the handle data is returned.

Comments
  This function is only permitted for handles which have no separately
  allocated memory block.

*****************************************************************************/

void WkdSetHandleData (HANDLE h, ULONG ul)
{
  UINT uPos = (UINT)(unsigned long)(h - HNDRES);
  ASSERT(uPos < uNextHandle);
  if (uPos >= uNextHandle) {
    // illegal handle => return
    return;
  } // if
  pahpt[uPos].D.ulData = ul;
} // WkdSetHandleData()


/*****************************************************************************
WkdDestroyHandle
==============================================================================

This function frees the memory which is allocated in corresponce with the
specified handle.

Parameters
  h  contains the handle which memory is freed.

Comments
  If the destroying of the handle results in 32 or more free handles at the
  end of the list, the size of the handle array is reduced to the next length
  which is a multiple of 16.

*****************************************************************************/

void WkdDestroyHandle (HANDLE h)
{
  HNDPT *phpt;
  UINT uPos = (UINT)(unsigned long)(h - HNDRES);
  ASSERT(uPos < uNextHandle);
  if (uPos >= uNextHandle) {
    // illegal handle => return
    return;
  } // if
  phpt = pahpt + uPos;
  if (phpt->uType == HTYP_NULL) {
    return;
  } // if
  switch (phpt->uType) {
    case HTYP_NULL:
      // entry already empty => return
      return;
    case HTYP_SYS:
    case HTYP_ENTRY:
    case HTYP_RMTENTRY:
    case HTYP_MEMBLOCK:
      if (phpt->D.pvData != NULL) {
        // free allocated memory
        BsrFreeMem(phpt->D.pvData);
      } // if
      break;
  } // switch
  // free handle
  phpt->uType = HTYP_NULL;
  ASSERT(uPos < uNextHandle);
  if (uPos + 1 != uNextHandle) {
    // entry not at end of list => return immediately
    return;
  } // if
  {// determine new end of list
    UINT uLastHandle = uPos;
    while (uLastHandle != 0) {
      phpt--;
      if (phpt->uType != HTYP_NULL) {
        break;
      } // if
      uLastHandle--;
    } // while
    // save new end of table
    uNextHandle = uLastHandle;
  } // block
  if (uNextHandle + 32 <= cNumberOfHandles) {
    // 32 or more entries free at end of list: reduce size of list
    // * always round up new size to next multiple of 16
    UINT cNewHandle = (uNextHandle + 15) & ~15;
    if (cNewHandle == 0) {
      // table is empty: delete it
      BsrFreeMem(pahpt);
      // set new size of alloatable handles
      cNumberOfHandles = 0;
    } else {
      // reduce allocated memory if possible
      HNDPT *pahptNew = BsrAllocMem(cNumberOfHandles * sizeof(HNDPT));
      if (pahptNew) {
        // copy data from previous block, free this block
        memcpy(pahptNew, pahpt, cNumberOfHandles * sizeof(HNDPT));
        BsrFreeMem(pahpt);
        pahpt = pahptNew;
        // set new size of alloatable handles
        cNumberOfHandles = cNewHandle;
      } // if
    } // if
  } // if
  // <cNumberOfHandles> must be always a multiple of 16
  ASSERT(cHandle % 16 == 0);
} // WkdDestroyHandle()


/*****************************************************************************
WkdCleanUp
==============================================================================

This function cleans up all memory which was allocated dynamically in the
heap. It should be called before the driver is terminated.

Comments
  This function is currently only implemented for the Novell Netware driver.

*****************************************************************************/

void WkdCleanUp()
{
  WkdResetPortEnumeration();
  // select next port (no preparation for access)
  while (WkdSelectNextPort(FALSE)) {
    // next port exist: reset it
    ASSERT_VALUE(cstWkd.pprtgrpSel);
    ASSERT_VALUE(cstWkd.pprtgrpSel->pclsdrv);
    ASSERT_VALUE(cstWkd.pprtgrpSel->pclsdrv->pfClDrvCleanUp);
    // call clean up routine of class driver
    (cstWkd.pprtgrpSel->pclsdrv->pfClDrvCleanUp)(cstWkd.nSelPort);
  } // while
  //!mabu:P not proper implemented: release only never-freed memory!
  if (cNumberOfHandles != 0) {
    BsrFreeMem(pahpt);
    cNumberOfHandles = 0;
  } // if
} // WkdCleanUp()


/*****************************************************************************

---- Reading Driver Setting (Profiling) Management functions -----------------

*****************************************************************************/

/*****************************************************************************
WkdCompString
==============================================================================

This function compares the specified pattern string with the string in the
environment at the specified position. It is assumed that the string to be
searched is specified in upper case. The string in the environment can be in
lowercase or uppercase. Special characters (bit 7=1) are not permitted.

Parameters
 pszEnvir .. points to the current position in the environment.
 pszPattern  points to the string to be searched in the environment.

Return
 If the string <pszPattern> is not found NULL is returned. Otherwise the
 function returns the position in the environment after the found string. The
 calling routine must check if this character is a separator (Null, Space,
 Comma etc).

*****************************************************************************/

char * WkdCompString (char *pszEnvir, char *pszPattern)
{
  for (;;) {
    INT ch;
    if (*pszPattern == 0) {
      // end of pattern string => pattern found, return next position
      return pszEnvir;
    } // if
    // next character, convert to upper case
    ch = *pszEnvir++;
    ch = WkdToUpper(ch);
    if (ch != *pszPattern) {
      // character doesn't match => not found, return NULL
      return NULL;
    } // if
    pszPattern++;
  } // for cycle, exit by return
} // WkdCompString()


/*****************************************************************************
WkdSkipSpace
==============================================================================

The next space characters at the specified parsing location are skipped
and the pointer is set to the next non-space-character is returned.
Following characters are space characters:
  Horizontal Tab (ASCII 9)
  Space (ASCII 32)

Parameters
  ppszParse  points to a variable which contains a pointer to the current
             parsing position. This pointer is set to the next valid
             character.

Return
  The parsing position of the next non-space character is returned.

*****************************************************************************/

void WkdSkipSpace (char **ppszParse)
{
  while (**ppszParse == ' ' || **ppszParse == '\t') {
    (*ppszParse)++;
  } // while
} // WkdSkipSpace()


/*****************************************************************************
WkdSkipSep
==============================================================================

This function skips the next command separator. Leading or trailing spaces
are also skipped.

Parameters
  ppszParse  points to a variable which contains a pointer to the current
             parsing position. This pointer is set to the next valid
             character.

Return
  The parsing position of the next non-space character is returned.

*****************************************************************************/

void WkdSkipSep (char **ppszParse)
{
  WkdSkipSpace(ppszParse);
  // skip comma if avaiable
  if (**ppszParse == ',') {
    (*ppszParse)++;
  } // if
  WkdSkipSpace(ppszParse);
} // WkdSkipSep()


/******************************************************************************
WkdReadHexNumber
===============================================================================

This function reads and converts an ULONG hexadecimal number at the specified
parsing position.

Parameters
  ppszNum  points to a variable which contains a pointer to the current
           parsing position. This pointer is set to the first character
           after the hexadecimal number.

Return
  The read value is returned as a 32 bit value.

Comments
  This function doesn't detect any overflow. An invalid number (contains no
  digits) is returned as 0.

Comments
 In a hexadecimal number, upper- and lowercase letters are ignored. The
 function doesn't check an overflow. All values are modulo (MAX_ULONG+1).

******************************************************************************/

ULONG WkdReadHexNumber (char* * ppszNum)
{
  char *pszParse = *ppszNum;
  ULONG ulSum = 0;
  for (;;) {
    UINT chDigit = *pszParse;
    UINT uDigit;
    chDigit = WkdToUpper(chDigit);
    if ( (chDigit < '0') || 
         (chDigit > '9') && (chDigit < 'A') || 
         (chDigit > 'F') ) {
      // end of hexadecimal number
      // * exit loop
      break;
    } // if
    uDigit = chDigit <= '9'? chDigit - '0' : chDigit - 'A' + 10;
    // calculate value, next position, number is valid
    ulSum = 16 * ulSum + uDigit;
    pszParse++;
  } // for cycle, exit by break
  // store back new parsing position, return number value
  *ppszNum = pszParse;
  return ulSum;
} // WkdReadHexNumber()


/******************************************************************************
WkdReadDecimalNumber
===============================================================================

This function reads and converts an ULONG decimal number at the specified
parsing position.

Parameters
  ppszNum  points to a variable which contains a pointer to the current
           parsing position. This pointer is set to the first character
           after the decimal number.

Return
  The read value is returned as a 32 bit value.

Comments
  This function doesn't detect any overflow. An invalid number (contains no
  digits) is returned as 0.

Comments
  In a hexadecimal number, upper- and lowercase letters are ignored. The
  function doesn't check an overflow. All values are modulo (MAX_ULONG+1).

******************************************************************************/

ULONG WkdReadDecimalNumber (char* * ppszNum)
{
  char *pszParse = *ppszNum;
  ULONG ulSum = 0;
  for (;;) {
    UINT chDigit = *pszParse;
    UINT uDigit;
    chDigit = WkdToUpper(chDigit);
    if (chDigit < '0' || chDigit > '9') {
      // end of decimal number
      // * exit loop
      break;
    } // if
    uDigit = chDigit - '0';
    // calculate value, next position, number is valid
    ulSum = 10 * ulSum + uDigit;
    pszParse++;
  } // for cycle, exit by break
  // store back new parsing position, return number value
  *ppszNum = pszParse;
  return ulSum;
} // WkdReadDecimalNumber()


/*****************************************************************************
WkdGetProfileInt
==============================================================================

This function searches in the driver's profiling data the specified variable,
converts their text into an integer variable and returns this value. If the
variable is not found or the syntax is invalid, 0 is returned.

Parameters
  idTopic   contains one of the WKPRFTPC_ constants which specify the fixed
            WIBU-KEY topic in WIBUKEY.INI res. the registry.
  pszName   points to the name of the profile variable.

Return
  The function returns the integer value of the variable.

*****************************************************************************/

INT WkdGetProfileInt (enum EWkProfileTopic idTopic, const char *pszName)
{
  char szBuf[100];
  char *pszParse = NULL;
  INT cch = 0;
  INT iRet = 0;

  cch = WkdGetProfileValue (idTopic, pszName, szBuf, sizeof(szBuf));
  if (cch == 0) {
    // not found => return 0
    return 0;
  } // if
  // variable valid: interpret contents
  pszParse = szBuf;
  WkdSkipSpace(&pszParse);
  // read (decimal) value, 0 if any error
  iRet = (INT)WkdReadDecimalNumber(&pszParse);
  return iRet;
} // WkdGetProfileInt()


/*****************************************************************************
WkdCheckEnvVar
==============================================================================

This function searches the specified type name in the argument of the
specified environment variable.

Parameters
  pszVarName   points to the variable name string. The string must be specified
               in uppercase and terminated by a 0-byte.
  pszTypeName  points to the type name string. The string must be specified in
               uppercase and terminated by a 0-byte.

Return
  The function returns one of the following values:
    -1  variable found, specified port type is no part of it.
     0  variable not found or no environment found.
    +1  variable found, specified port type is part of it.

Attributes
  module entry

*****************************************************************************/

INT WkdCheckEnvVar (char *pszVarName, char *pszTypeName)
{
  char *pszBuf;
  char *pszParse;
  UINT cchLength;
  INT idRet = 0;
  pszBuf = BsrAllocMem(256 * sizeof(char));
  if (pszBuf == NULL) {
    // memory too small => return
    return 0;
  } // if
  // search specified environment variable (system specific)
  cchLength = WkdGetProfileValue(WKPRFTPC_DRIVER_KERNEL, pszVarName, pszBuf, 256);
  if (cchLength == 0 || cchLength > 256) {
    // variable not found in environment => return 0
    goto RETURN;
  } // if
  // set search position
  pszParse = pszBuf;
  for (;;) {
    char *pszNext;
    // search specified type name in environment string
    // * skip leading spaces
    WkdSkipSpace(&pszParse);
    // check next string;
    pszNext = WkdCompString(pszParse, pszTypeName);
    if (pszNext != NULL) {
      // string matched: check if valid end of name
      // * skip trailing spaces
      WkdSkipSpace(&pszNext);
      if (*pszNext == ',' || *pszNext == 0) {
        // end of name (comma follows or end of string) => name matched
        idRet = 1;
        break;
      } // if
    } // if
    // search next comma to check next entry
    while (*pszParse != 0 && *pszParse != ',')
      pszParse++;
    if (*pszParse == 0) {
      // end of string, name not found
      idRet = -1;
      break;
    } // if
    // to character after comma
    pszParse++;
  } // for cycle, exit by break
RETURN:
  ASSERT(pszBuf != NULL);
  BsrFreeMem(pszBuf);
  return idRet;
} // WkdCheckEnvVar()

/*****************************************************************************
WkdFindEnumName
==============================================================================

This function checks if the specified string is a part of the second string.
The second string contains several names which are separated with the
vertical bar ("|") character. The matching is independent of the case of
letters.

Parameters
  pszName .... points to the name which is searched.
  pszNameList  points to the list with the names.

Return
  The address of the found name is returned. If this address is NULL, the
  name was not found.

*****************************************************************************/

const char *WkdFindEnumName (const char *pszName, const char *pszNameList)
{
  INT cchName = (INT)strlen(pszName);
  for (;;) {
    if (!_memicmp(pszName, pszNameList, cchName) &&
        (pszNameList[cchName] == '|' || pszNameList[cchName] == 0)
      ) {
      // name found: return position
      return pszNameList;
    } // if
    // search separator
    while (*pszNameList != 0 && *pszNameList != '|') {
      pszNameList++;
    } // while
    // end of string => return NULL
    if (*pszNameList == 0) {
      return NULL;
    } // if
    // set search position after separator
    pszNameList++;
  } // for
} // WkdFindEnumName()





/*****************************************************************************
WkdIsBoxSelected
==============================================================================

This function checks if the specified WIBU-BOX is selected by another open
operation. The checking is done via the serial number.

Parameters
  pwkbsr  points to the serial number of the WIBU-BOX which selection state
          is checked.

Return
  The function returns TRUE if the entry is read and FALSE if the entry is
  empty.

Comments
  This function assumes that different WIBU-BOX types (for example /P and /SP)
  have different serial numbers.

*****************************************************************************/

BOOL WkdIsBoxSelected (WKBSERIAL const *pwkbsr)
{
  HNDPT *phptEnd = pahpt + cNumberOfHandles;
  HNDPT *phpt = pahpt;
  while (phpt != phptEnd) {
    if (phpt->uType == HTYP_ENTRY) {
      // get pointer to open entry
      OPENENTRY *pope = phpt->D.pvData;
      // analyse if this is selected and matches current selected WIBU-BOX
      if (pope->flCtrl & WKB_OE_SELECTED &&
          !memcmp(&pope->wkbsr, pwkbsr, FIELDOFFSET(WKBSERIAL, usReserved))
        ) {
        // open entry selected and matches current WIBU-BOX => is selected
        return TRUE;
      } // if
    } // if
    phpt++;
  } // while
  // current WIBU-BOX is not selected by opened entry
  return FALSE;
} // WkdIsBoxSelected()


/*****************************************************************************
WkdFreeLockedBoxes
==============================================================================

This function resets the locking flag of all opened WIBU-BOX entries.

*****************************************************************************/

void WkdFreeLockedBoxes (void)
{
  HNDPT *phptEnd = pahpt + cNumberOfHandles;
  HNDPT *phpt = pahpt;
  while (phpt != phptEnd) {
    if (phpt->uType == HTYP_ENTRY) {
      // get pointer to open entry
      OPENENTRY *pope = phpt->D.pvData;
      // reset selection flag of this entry
      pope->flCtrl &= ~WKB_OE_SELECTED;
    } // if
    phpt++;
  } // while
} // WkdFreeLockedBoxes()


/*****************************************************************************
W k d I s E n t r y G l o b a l
==============================================================================

This function checks if the specified WIBU-BOX entry is used by another
process with the (exclusive) WKB_OPEN_GLOBAL mode. The checking is done via
the serial number and the used entry.

Parameters
  pwkbsr ...... points to the serial number of the WIBU-BOX which global
                state is checked.
  nEntry ...... contains the index of the entry (0..9) of the entry which is
                checked.
  hwkbsysOwner  contains the handle of the subsystem which wants to open the
                specified WIBU-BOX entry. This this subsystem possesses the
                global entry, this open is permitted.

Return
  The function returns TRUE if the entry is used by another process with the
  WKB_OPEN_GLOBAL mode and FALSE if not.

Comments
  This function assumes that different WIBU-BOX types (for example /P and /SP)
  have different serial numbers.

*****************************************************************************/
/*
BOOL WkdIsEntryGlobal(WKBSERIAL const *pwkbsr, INT nEntry,
  HWKBSYSTEM hwkbsysOwner)
{
  HNDPT *phptEnd = pahpt + cNumberOfHandles;
  HNDPT *phpt = pahpt;
  while (phpt != phptEnd) {
    if (phpt->uType == HTYP_ENTRY) {
      // get pointer to open entry
      OPENENTRY *pope = phpt->D.pvData;
      // analyse if this is selected and matches current selected WIBU-BOX
      if (pope->flCtrl & WKB_OE_GLOBAL &&
          (INT)pope->iEntry == nEntry &&
          !memcmp(&pope->wkbsr, pwkbsr, FIELDOFFSET(WKBSERIAL, usReserved)) &&
          (pope->hwkbsys == HWKB_LOCAL || pope->hwkbsys != hwkbsysOwner)
        ) {
        // open entry globally declared and matches current WIBU-BOX entry
        return TRUE;
      } // if
    } // if
    phpt++;
  } // while
  // current WIBU-BOX entry is not globally opened by another entry
  return FALSE;
} // WkdIsEntryGlobal()
*/


/*****************************************************************************
WkdIsEntryConflict
==============================================================================

This function checks if the specified entry conflicts with another opened
entry. If so, the entry cannot be opened and the function returns TRUE.

Two entries have a conflict:
* If one or both entries have set the WKB_OE_GLOBAL flag and the other entry
  has not set the WKB_OE_NONETLIMIT flag.
or
* If one or both entries have set the WKB_OE_EXCLUSIVE flag and the two
  entries are opened with the same Firm Code and User Code.

Note: If both the WKB_OE_GLOBAL and the WKB_OE_NONETLIMIT flag is set, the
      WKB_OE_GLOBAL overwrites the WKB_OE_NONETLIMIT flag to check the
      conflict.

Parameters
  popeConflict  points to a new entry which is created by the current
                open operation and which is fully initialized. This function
                checks if it conflicts with any other entry which is
                exclusively opened.

Return
  The function returns TRUE if the entry is used by another process with the
  WKB_OPEN_GLOBAL mode and FALSE if not.

Comments
  This function assumes that different WIBU-BOX types (for example /P and /SP)
  have different serial numbers because port numbers are not compared.

Implementation
  This function uses the entry handle table implementation in mncommon.c and
  must be stored in this module. It is only called from the clopen.c module.

*****************************************************************************/

BOOL WkdIsEntryConflict (const OPENENTRY *popeConflict)
{
  HNDPT *phptEnd, *phpt;
  phptEnd = pahpt + cNumberOfHandles;
  phpt = pahpt;
  // pre-check of new entry which may conflict
  if ((popeConflict->flCtrl & WKB_OPEN_NONETLIMIT)
      && !(popeConflict->flCtrl & WKB_OPEN_GLOBAL)
      && !(popeConflict->flCtrl & WKB_OPEN_EXCLUSIVE)
    ) {
    // new entry with unlimited user quantity but not opened exclusively:
    // * return (no conflict possible)
    return FALSE;
  } // if
  while (phpt != phptEnd) {
    if (phpt->uType == HTYP_ENTRY) {
      // get pointer to open entry
      const OPENENTRY *popeThis = GetCommonMem(phpt->D.pvData);
      if (popeThis == popeConflict) {
        // block is itself: this is trivial and no checking candidate
        goto NEXT_ENTRY;
      } // if
      if (popeThis->iEntry != popeConflict->iEntry) {
        // entry indices not identical => is another entry
        goto NEXT_ENTRY;
      } // if
      if (memcmp(&popeThis->wkbsr, &popeConflict->wkbsr,
         FIELDOFFSET(WKBSERIAL, usReserved)) != 0
        ) {
        // serial number not identical => is another entry
        goto NEXT_ENTRY;
      } // if
      if (popeThis->hwkbsys != HWKB_LOCAL &&
          popeThis->hwkbsys == popeConflict->hwkbsys
        ) {
        // entry opened via same true subsystem: permitted within this system
        // * required to accept a WibuKey server as a subsystem shell
        // * the uniqueness of the entry use is handled by the server itself
        // * also exclusive access is permitted within a subsystem shell
        goto NEXT_ENTRY;
      } // if
     if ((popeThis->flCtrl & WKB_OE_GLOBAL) &&
          (!(popeConflict->flCtrl & WKB_OE_NONETLIMIT) || (popeConflict->flCtrl & WKB_OE_GLOBAL))
        ) {
        // entry is globally opened and new entry has NONETLIMIT not set
        // * this is the first conflict => return TRUE
        cstWkd.idError = WKB_ERROR_ENTRY_ALREADY_USED;
        return TRUE;
      } // if
      if ((!(popeThis->flCtrl & WKB_OE_NONETLIMIT) || (popeConflict->flCtrl & WKB_OE_GLOBAL))
          && (popeConflict->flCtrl & WKB_OE_GLOBAL)
        ) {
        // entry has no NONETLIMIT flag set and new entry has GLOBAL flag
        // * this is the second conflict => return TRUE
        cstWkd.idError = WKB_ERROR_ENTRY_ALREADY_USED;
        return TRUE;
      } // if
      if (popeThis->flCtrl & WKB_OE_EXCLUSIVE &&
          popeConflict->flCtrl & WKB_OE_EXCLUSIVE &&
          popeConflict->ulFirmCode == popeThis->ulFirmCode &&
          popeConflict->ulUserCode == popeThis->ulUserCode
        ) {
        // * both entries have EXCLUSIVE flag and contain
        //   same Firm Code and User Code:
        // * this is an exclusive conflict => return TRUE
        cstWkd.idError = WKB_ERROR_EXCLUSIVE_MODE_CONFLICT;
        return TRUE;
      } // if
      // no conflict: check next entry
    } // if
  NEXT_ENTRY:
    phpt++;
  } // while
  // current WIBU-BOX entry is not globally opened by another entry
  return FALSE;
} // WkdIsEntryConflict()


/*****************************************************************************
WkdCloseAllOfSubsystem
==============================================================================

This function closes all WIBU-BOX entries which were opened with the
specified subsystem handle.

*****************************************************************************/

void WkdCloseAllOfSubsystem (HANDLE hwkbsys)
{
  HNDPT *phptEnd = pahpt + cNumberOfHandles;
  HNDPT *phpt = pahpt;
  UINT uHandle = HNDRES;
  for (; phpt != phptEnd; phpt++, uHandle++) {
    if (phpt->uType == HTYP_ENTRY) {
      // get pointer to open entry
      OPENENTRY *pope = phpt->D.pvData;
      // analyse if this is selected and matches current selected WIBU-BOX
      if (pope->hwkbsys == hwkbsys) {
        WkdClose((HANDLE)(unsigned long)uHandle);
      } // if
    } // if
  } // for
} // WkdCloseAllOfSubsystem()






// ============================================
// wibu\wk\dev\runtime\drv\c: end of mncommon.c
// ============================================
