/*****************************************************************************
S r c D r v                                Portable Source Code WibuKey Driver
==============================================================================
mnsys.c                                            Low-Level System Adaptation
==============================================================================

This module implements operating system specificfunctions which are called
from the specific low level drivers for the different supported interface and
functions to management the handling in the general section of the driver.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1992 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu), Marcus Balb (marb)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
93Mar11 1.33 mabu: version 1.33 released
93Mar21 1.35 mabu: adaption to Windows
93Apr13 1.35 mabu: compatible with BSRLIB
93Jun19 1.35 mabu: adapation to Win32
93Jun21 1.35 mabu: error in Timing calculation of _WkbGetSysTime() removed
93Jul22 1.35 mabu: adapation to DOS
93Aug02 1.35 mabu: _WkbGetEnvAddr() can read 32 bit values
93Aug28 2.00 mabu: adaptation to WibuKey version 2
94Apr26 2.00 mabu: adaptation to Windows NT kernel driver
94May28 2.00 mabu: _Wkb... functions renamed to _Wkd...
94Sep14 2.00 mabu: BIOS RAM addressing under Windows realized via __0040H
94Oct17 2.00 marb: adaptation to Netware NLM
94Nov11 2.00 mabu: version 2.0 shipped
94Dec30 2.01 mabu: bug in environment reading under Borland DPMI removed
95Feb06 2.02 mabu: using of ENABLE/DISABLED_INTERRUPT macros
95Feb27 2.05 mabu: parsing of decimal numbers added
95Oct13 2.08 mabu: VxD registry addressing implemented
95Oct19 2.08 mabu: system T value changed to 50 microseconds
95Oct31 2.08 mabu: absolute timing mechanism implemented
96Jan30 2.10 mabu: renamed from SYS.C
96Apr09 2.11 mabu: toupper() -> WkdToUpper to avoid including MS-C ctype array
97Apr05 2.31 mabu: WkdGetRegistryValue() implemented
00Okt05 3.00 mn:   GetLocalTime() and GetTickCount() fixed for Linux
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

#if UNIX
#include <sys/time.h>    // gettimeofday()
#include <unistd.h>      // gettimeofday()
#include <time.h>        // time(), localtime(), asctime(), clock()
#include <errno.h>       // errno
/*****************************************************************************

---- Compatibility to Microsoft C extensions ---------------------------------

*****************************************************************************/

/*****************************************************************************
_ m e m i c m p
==============================================================================

This function compares two memory areas without a difference of upper and
lower case characters. Except of this special character compare, the return
value is identical with the ANSI C function memcmp().

*****************************************************************************/

int _memicmp(const void *pv1, const void *pv2, size_t cb)
{
  UINT i;
  const BYTE *pb1 = (BYTE*)pv1;
  const BYTE *pb2 = (BYTE*)pv2;
  for (i = 0; i < cb; i++) {
    BYTE b1 = WkdToUpper(*pb1);
    BYTE b2 = WkdToUpper(*pb2);
    if (b1 < b2) {
      return -1;
    } // if
    if (b1 > b2) {
      return +1;
    } // if
    // continue in loop
    pb1++;
    pb2++;
  } // for
  // the strings are identical
  return 0;
} // _memicmp()

#endif // #if UNIX

/*****************************************************************************
GetLocalTime
==============================================================================

This function reads the current system time relative to the local time zone.

Parameters
  pst  points to the structure which receives the current system time.

Comments
* These function sets the milliseconds always to 0. The weekday may
  be set but is not required.

*****************************************************************************/

void GetLocalTime (SYSTEMTIME *pst)
{
  SYSTEMTIME stRead;

  #define WKB_DATEOFFSET 1900

    struct tm       *p_tm;        
    struct timeval  p_timeval;        
    struct timezone *p_timezone = NULL;

    gettimeofday(&p_timeval, p_timezone);

    p_tm = localtime((time_t*) &(p_timeval.tv_sec));
                                                                                 
    stRead.wYear         = p_tm->tm_year + WKB_DATEOFFSET;
    stRead.wMonth        = p_tm->tm_mon + 1;
    stRead.wDay          = p_tm->tm_mday;
    stRead.wDayOfWeek    = p_tm->tm_wday;
    stRead.wHour         = p_tm->tm_hour;
    stRead.wMinute       = p_tm->tm_min;
    stRead.wSecond       = p_tm->tm_sec;
    stRead.wMilliseconds = p_timeval.tv_usec;

  *pst = stRead;
} // GetLocalTime()


/*****************************************************************************
GetTickCount
==============================================================================

This function emulates the GetTickCount() function of Windows.

Return
  This function returns the milliseconds the program runs

Implementation
* The delta value is only valid for a distance of one day or less. Midnight
  overrun is determined properly, but a difference more than one day is
  detected only as a difference of one day.
* The implementation is portable but really slow. Whenever possible optimize
  this function reading a continued millisecond value which is returend by
  the operating system (a granularity of 500 ms is acceptable).

*****************************************************************************/

ULONG GetTickCount (void)
{
  static unsigned long long refTick = 0;
  unsigned long long  ulLastTick;
  struct timeval t1;

  gettimeofday(&t1, NULL);
  ulLastTick = ((unsigned long long )t1.tv_sec * (unsigned long long)1000) + 
    ((unsigned long long )t1.tv_usec / (unsigned long long)1000);
  
  if (0 == refTick) {
    refTick = ulLastTick;
  }
  ulLastTick -= refTick;

  return (ULONG)ulLastTick;
} // GetTickCount()


/*****************************************************************************

---- Basic Read Function for Driver Setting (Profiling) ----------------------

*****************************************************************************/

/*****************************************************************************
WkdGetProfileValue
==============================================================================

Parameters
  pszName   points to the string in the \Parameters registry subkey or in the
            [Drivers] section of the WIBUKEY.INI file.
  pszValue  points to a buffer which gets the contents of the profile.
  cchValue  contains the length of the buffer which gets the contents of the
            profile.

Return
  The function returns 0 if the profile was not found or the length of the
  written string without the trailing 0.

*****************************************************************************/

UINT WkdGetProfileValue (enum EWkProfileTopic idTopic, const char *pszName,
    char *pszValue, UINT cchValue)
{
  // for the Linux the INI-path is hardcoded
  int nRet = 0;
  INT id = -1;
  INT iEnt =  0;
  INT i = 0;
  char *pszCurSection = NULL;

  switch (idTopic) {
    case WKPRFTPC_GENERAL:
      pszCurSection = "General";
      break;
    case WKPRFTPC_DRIVER_CLIENT:
      pszCurSection = "Driver";
      break;
    case WKPRFTPC_DRIVER_KERNEL:
      pszCurSection = "Driver";
      break;
    case WKPRFTPC_WKLAN:
      pszCurSection = "WkLAN";
      break;
    case WKPRFTPC_CLIENT_USER:
      pszCurSection = "User";
      break;
    case WKPRFTPC_CLIENT_WKLAN:
      pszCurSection = "WkLAN Client";
      break;
    case WKPRFTPC_CLIENT_WKNET:
      pszCurSection = "WkNet Clients";
      break;
    default:
      ASSERT(0);
  } // switch

  memset(pszValue, 0x0, sizeof(pszValue) );
  nRet = BsrGetPrivateProfileString(pszCurSection, pszName, "",
    pszValue, cchValue, "/etc/wibu/WIBU-KEY/WIBUKEY.INI");
    
  id = -1;
  iEnt = SizeofEnvVarArray();
  while(i < iEnt){
    if (!(strncmp(pszName, gaevWK[i].pszName, (size_t)strlen(gaevWK[i].pszName)))){
      id = i;
      break;
    }//if
    i++;
  }//while
  if ( id == -1) {
    if (!(strncmp(pszName, "IpPort", sizeof(pszName) ))){
      id = ENV_IPPORT;
    } else if (!(strncmp(pszName, "ConnectTimeout", sizeof(pszName) ))) {
      id = ENV_TIMEOUT;
    } else if (!(strncmp(pszName, "SubSystem", sizeof(pszName) ))) {
      id = ENV_SYSTEM;
    } else if (!(strncmp(pszName, "Baudrate", sizeof(pszName) ))){
      id = ENV_BAUDRATE;
    } else if (!(strncmp(pszName, "SERVER", sizeof(pszName) - 1))){
      id = ENV_SERVERNAME;
    } 
  }

  if (id != -1){
    // Set gaevWK[] apropriately
    if (pszValue == NULL){
      // ENV not found: Set default
      SetEnvVarArrayEntry(id, NULL);
    } else {
      // ENV found:
      if (  (gaevWK[id].pszValue == NULL) || 
            strncmp(gaevWK[id].pszValue, pszValue, (size_t) strlen(pszValue) ) ){
        SetEnvVarArrayEntry(id, pszValue);
      }//if
    }//if (!pszProfile)
  }else{
    SetEnvVarArrayEntry(id, NULL);
  }// if(id != -1)

  return nRet;
} // WkdGetProfileValue()


// ===================================
// end of mnsys.c
// ===================================
