/*****************************************************************************
D r v C                                 Portable WibuKey Driver, Written in C
==============================================================================
usblinux.c                           Shared-Object Driver for Wibu-USB-Devices
==============================================================================

This module implements the specific low level driver parts for the USB
interface of the WibuBox/U & /RU for Linux/Un!x systems.

==============================================================================

The portable WibuKey driver source is used under the following systems:
 - WKDOS2(d).LIB       - for internal use to link to DOS programs.
 - WKWIN.DLL           - 16 Bit calling driver on 32 Bit Windows systems.
 - WKWIN32.DLL         - 32 Bit Windows calling driver.
 - WIBUKEY.SYS         - NT based kernel driver.
 - WIBUKEY.VXD         - W9x based kernel driver.
 - WkMAC.LIB           - MacOS 9 driver (Shared Lib).
 - WkMacLibX.Framework - MacOS X driver.
 - LibWkLin.so         - Linux and other Un!xes (Shared Object).
 
All sources are written in ANSI C, except the using of C++ style
single line comments.

==============================================================================

Copyright  2000 - 2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. libwklin.so
Author ... Markus Neumann (mn), Heiko Abraham (ha)
System ... Linux or Un!x
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
01Aug07 0.1  mn:   initial version
04Sep09 4.11 ha:   bugfix for timeout - protected ext mem readout
04Sep09 4.11 ha:   Changes for thread safe error code 
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************/

#include "proj.h"

#if defined(__linux__)
// ==============================
// Linux specific implementations
// ==============================
#include "usblinux.h"

#ifdef __cplusplus
extern "C" {
#endif


sem_t* gUsbSemaphoreID = NULL;
#define USB_BOX_LOCK(a)
#define USB_BOX_UNLOCK

  
/*****************************************************************************
wibu_usb_close
==============================================================================

Todo

*****************************************************************************/
  
void wibu_usb_close(struct usbdevice *dev)
{
  if (dev == NULL) {
    return;
  }
  if(dev->fIsStillOpen) {
    close(dev->fd);
  }
  dev->fIsStillOpen = FALSE;
  free(dev);
}// wibu_usb_close()


/*****************************************************************************
wibu_usb_open
==============================================================================

Todo

*****************************************************************************/

struct usbdevice *wibu_usb_open(BOXLIST *pbox, BOOL fLeaveOpen)
{
    UINT bn = (*(pbox->ppDev))->uBusNum;
    UINT dn = (*(pbox->ppDev))->uDevNum;
#if !HAVE_GCC_3X
    INT pid = (*(pbox->ppDev))->pDev->desc.idProduct[0] |
             ((*(pbox->ppDev))->pDev->desc.idProduct[1] << 8);
#else
    UInt16 pid = (UInt16)(*(pbox->ppDev))->pDev->desc.idProduct;
#endif //!HAVE_GCC_3X
    return wibu_usb_open_bynumber(pbox, bn, dn, pid, fLeaveOpen);
} // wibu_usb_open()


/*****************************************************************************
wibu_usb_open
==============================================================================

This is called by usb_open()
open USBdevice at "DEVFS/{bus#}/{device#}", read usb_device_descriptor at once,
ret OPENED usbdevice!

*****************************************************************************/

struct usbdevice *wibu_usb_open_bynumber(BOXLIST *pbox,
                                         unsigned int busnum,
                                         unsigned int devnum,
                                         int productid,
                                         BOOL fLeaveOpen)
{
    struct usbdevice *dev = (*(pbox->ppDev))->pDev;
    struct usb_device_descriptor *pdesc = &(dev->desc);
    UInt16 vid = 0, pid = 0;
    char devsfile[256], dummy[2];
    int ret = 0, fd = 0;
    int vendorid = WK_VENDOR_ID;
    FILE *os = NULL;

    dummy[0] = '\0';
    dummy[1] = '\0';

    os = stderr;

    ASSERT(dev!= NULL);

    errno = 0;
    snprintf(devsfile, sizeof(devsfile), "%s/%03u/%03u", 
        WkhUsbGetDeviceMountPoint(), busnum, devnum);
    if( ((dev->fIsStillOpen) && (dev->fd)) ) {
      if (lseek(dev->fd, 0, SEEK_SET) == (off_t)-1){
        if (errno != EBADF){
          close(dev->fd);
          dev->fIsStillOpen = FALSE;
          pbox->fBoxAlreadyInList = FALSE;
          return NULL;
        }
      } else {
        pbox->fBoxAlreadyInList = TRUE;
        goto EXIT_IF_LEAVE_OPEN;
      }
    }

    errno = 0;

    snprintf(devsfile, sizeof(devsfile), "%s/%03u/%03u", WkhUsbGetDeviceMountPoint(), busnum, devnum);
    if ((fd = open(devsfile, O_RDWR)) == -1) {
      return NULL;
    }

    errno = 0;
    if ((ret = (INT) read(fd, pdesc, sizeof(*pdesc))) != sizeof(struct usb_device_descriptor)) {
        if (ret > 0)
            errno = EIO;
        close(fd);
        dev->fIsStillOpen = FALSE;
        pbox->fBoxAlreadyInList = FALSE;
        return NULL;
    }

#if !HAVE_GCC_3X
    vid = pdesc->idVendor[0] | (pdesc->idVendor[1] << 8);
    pid = pdesc->idProduct[0] | (pdesc->idProduct[1] << 8);
#else
    vid = (UInt16)pdesc->idVendor;
    pid = (UInt16)pdesc->idProduct;
#endif //!HAVE_GCC_3X

    if ((vid != vendorid && vendorid == -1) ||
        (pid != productid && productid == -1)) {
        errno = -ENOENT;
        close(fd);
        dev->fIsStillOpen = FALSE;
        pbox->fBoxAlreadyInList = FALSE;
        return NULL;
    }

    dev->fd = fd;

EXIT_IF_LEAVE_OPEN:
    if (fLeaveOpen) {
      dev->fIsStillOpen = TRUE;
      return dev;
    } else {
      dev->fIsStillOpen = FALSE;
      close(dev->fd);
      return dev;
    }
} // wibu_usb_open_bynumber()


/*****************************************************************************
wibu_control_msg
==============================================================================

Send a usb control command to wibubox by using ioctl. If a timeout is detected,
we set timeout a little bit higher and retransmit the data.

Parameters
  dev   the usb device to use
  ctrl  the command data to send

Return
  Return 0 (zero) or a non negativ value, if successfull. On some USB-COntrollers
  we return the number of transfered bytes.
  On error, -1 is returned. (See errno for more details).

*****************************************************************************/

int wibu_control_msg (struct usbdevice *dev, IOUSBDevRequest *ctrl)
{
  int i = 0;
  int err = 0;

#if defined(__i386__)
  assert( sizeof(IOUSBDevRequest) == 16 );
#elif defined(__x86_64__)
  assert( sizeof(IOUSBDevRequest) == 24 );
#endif
  
  USB_BOX_LOCK();

  if ( ctrl->timeout < 300 ) {
    // make sure, timeout is not to small
    ctrl->timeout += 230;
  } // if
  errno = 0;

  i = ioctl(dev->fd, USBDEVFS_CONTROL, ctrl);
  err = errno;

  if (-1 == i ) {
    printf("ioctl() - %s\n", strerror(err));
  }
  if ( err == ETIMEDOUT ){
    // we have a timeout, now increase timeout and send again
    ctrl->timeout += 500;

    i = ioctl(dev->fd, USBDEVFS_CONTROL, ctrl);
    err = errno;
  }

  USB_BOX_UNLOCK;
  
  if (i < 0) {
    // error by receive data (ioctl)
    // typical, we have receive an usb-error, perhaps a usb-stall, cast error message
#if !HAVE_GCC_3X
  if ( (ctrl->request == 0x06) && (ctrl->value == 0x03)  ) {
#else
    if ( (ctrl->bRequest == 0x06) && (ctrl->wValue == 0x03)  ) {
#endif 
      // dirty hack 
      // for old boxes (WKB_VERSION, that not understand command 06-03 and will send a "usb-stall"
      return 0;
    } 
    return -1;
  }
  return i;
} // wibu_control_msg()


/*****************************************************************************
wibu_reset2
==============================================================================

Send a reset command to usb-wibubox. 

Parameters
  dev      the usb device to use
  timeout  the limit for request/answer-time

Return
  Return 0 (zero) or a non negativ value, if successfull. On some USB-Controllers
  we return the number of transfered bytes.
  On error, -1 is returned. (See errno for more details).

*****************************************************************************/

int wibu_reset2(struct usbdevice *dev, UINT timeout)
{
  IOUSBDevRequest ctrl;
  int nRet = 0;
  ctrl = (struct usbdevfs_ctrltransfer){ WIBU_WRITE, WIBU_RESET, 0, 0, 0, timeout, NULL };
  nRet = wibu_control_msg(dev, &ctrl);
  return nRet;
}//wibu_reset2()


/*****************************************************************************
WkhUsbUseSysFs
==============================================================================

Check if this version of linuix kernel use SYSFS or DEVFS. 

Return
  Return '0' if use PROCFS  (/proc/bus/usb/devices)
  Return '1' if use SYSFS   (/sys/bus/usb/ *)
  Return '2' if no usb acccess is possible 

*****************************************************************************/
int WkhUsbUseSysFs() 
{
  static int nUseSysFs = 0;
  int nRet = 0;

  if ( nUseSysFs == 0) {
    char szBuf[64];
    struct stat my_stat;
    const char *use_devfs = "/proc/bus/usb";
    const char *use_sysfs = "/dev/bus/usb";
    
    // RedHat 8,9, Fedora Core 1- 4, Debian 3.0, Debian 3.1, Ubuntu 5.10, 6.06, 6.10
    // SuSE 8.x, 9.0, 9,2, 9.3 
    // use /proc/bus/usb/devices
    //
    // SuSE 9.1 
    // use /proc/bus/usb/devices_please-use-sysfs-instead
    
    // SuSE 10.1, Ubuntu 6.06, Ubuntu 6.10 
    // use /sys/bus/usb
    
    snprintf(szBuf, sizeof(szBuf), "%s/devices", use_devfs);
    if ( stat( szBuf, &my_stat ) == 0 ) {
      // ok, we have a kernel-2.4.x like /proc-System 
      nUseSysFs = 2;
    } else {
      snprintf(szBuf, sizeof(szBuf), "%s", use_sysfs); 
      if ( stat( szBuf, &my_stat ) == 0 ) {
        // ok, have SYSFS
        nUseSysFs = 1;
      } else {
        snprintf(szBuf, sizeof(szBuf), "%s/devices_please-use-sysfs-instead", use_devfs); 
        if ( stat( szBuf, &my_stat ) == 0 ) {
          // ok, make SuSE 9.1 Workaround 
          nUseSysFs = 2;
        }
      }
    }
  }

  if ( nUseSysFs == 1 ) {
    // use SYSFS
    nRet = 1;
  } else if ( nUseSysFs == 2 ) {
    // use DEVFS, or unknown
    nRet = 0;
  } else {
    // an error occured
    nRet = 2;
  }
  return nRet;
} // WkhUsbUseSysFs


/*****************************************************************************
WkhUsbGetDeviceMountPoint
==============================================================================

This function return the current used IO communication mount point for an 
bulk transfer, like /proc/bus/usb/%03i/%03i.

Return
  Return "/proc/bus/usb" or "/sys/bus/usb" depends on the system

*****************************************************************************/

char *WkhUsbGetDeviceMountPoint()
{
  int nRet = WkhUsbUseSysFs();
  if ( nRet == 1 ) {
    // we uses newer dev interface
    return "/dev/bus/usb";
  } else {
    // we use older proc interface
    return "/proc/bus/usb";
  } 
} // WkhUsbGetDeviceMountPoint()

#ifdef __cplusplus
}
#endif

#endif //LINUX
