/*****************************************************************************************
D r v C                                             Portable WibuKey Driver, Written in C
==========================================================================================
usbmgr.c                                                  Low-Level Driver for USB Devices
==========================================================================================

This module implements the specific low level driver parts for the USB
interface of the WibuBox/U & /RU for MacOS systems. On Windows systems 
this part of the driver is seperated in an own driver (the wibuke2.sys).

==========================================================================================

The portable WibuKey driver source is used under the following systems:
 - WKDOS2(d).LIB - for internal use to link to DOS programs.
 - WKWIN.DLL - 16 Bit calling driver on 32 Bit Windows systems.
 - WKWIN32.DLL - 32 Bit Windows calling driver.
 - WIBUKEY.SYS - NT based kernel driver.
 - WIBUKEY.VXD - W9x based kernel driver.
 - WkMAC.LIB - MacOS 9 driver (Shared Lib).
 - WkMacLibX.Framework - MacOS X driver.
 
All sources are written in ANSI C, except the using of C++ style
single line comments.

==========================================================================================

Copyright  2000-2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu\wk\dev\runtime\drv\c
Author ... Alexander Schmitt (as), Heiko Abraham (ha)
System ... MacOS 9/X, Linux
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
00Dec21 3.00 as:   initial version
01Mar08 3.01 as:   adapted to pre release of Mac OS X
01Nov08 3.10 as:   redesign Mac OS X part ==>> port is not permanent locked
08Oct02 4.30 ha:   add usb support for arm/linux v4l
10Jun24 4.31 ha:   build update

*****************************************************************************************/

#include "config.h"

#if USE_USB
  #if defined(__linux__)
    #include "proj.h"
    #include "usblinux.h"
    #include <sys/types.h>
    #include <dirent.h> 
    #include <sys/param.h>
    #include <unistd.h>
  
    #define TIMEOUT_MS_RECV  180000 // = 180 ms (timeout for select())
    // timeout after terminating a USB request if USB is not answering
  
    #define USBRAWTIMEOUT    180
  
    #ifndef MAXPATHLEN
      #ifdef PATH_MAX
        #define MAXPATHLEN PATH_MAX
      #else
        #define MAXPATHLEN 4096
      #endif
    #endif  
  #endif 

  #if defined(__APPLE__)
    #include <unistd.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <string.h>
    #include <sysexits.h>
    
    #include <sys/errno.h> 
    #include <sys/fcntl.h>
    #include <sys/stat.h>   
    
    #include <semaphore.h>
  
    #if defined(ULONG)
      #undef ULONG
    #endif
    
    #include <mach/mach.h>
    #include <mach/mach_error.h>
    
    #include <CoreFoundation/CoreFoundation.h>
    #include <CoreFoundation/CFURL.h>
    #include <CoreFoundation/CFPlugIn.h>
  
    #include <IOKit/IOKitLib.h>
    
    
    #include <IOKit/IOCFPlugIn.h>
    #include <IOKit/IOKitKeys.h>
    #include <IOKit/IOMessage.h>
    #include <IOKit/usb/IOUSBLib.h>

  #endif // defined(__APPLE__)

    
#include "proj.h"
    
#if defined(__APPLE__)
  // internal structure of the list for all connected WibuBox/U & /RU
    typedef struct _BOXLIST {
      // prefix "boxlist"
      struct _BOXLIST *pNext;
      struct _BOXLIST *pPrev;
      WKBSERIAL wkbsr;
      IOUSBDeviceInterface **ppDev;
      BOOL8 fBoxAlreadyInList;
    } BOXLIST;
  
    // semaphore to synchronize the access to the Box/U & /RU
    sem_t* gUsbSemaphoreID = NULL;

    // global parameters for the notification handling
    IONotificationPortRef gNotifyPort = 0;
    io_iterator_t gAddedIter = 0;
    io_iterator_t gRemovedIter = 0;
#endif

// list of all connected boxes
static BOXLIST *gpboxlist = NULL;
// real number of connected boxes
static UINT gcBoxList = 0;
// force rescanning of the internal list
static BOOL8 gfForceReading = TRUE;
// last read list identification
// * initialized with 0: force rereading of the real list
static UINT gidListLastScanReal = 1;


// for synchronisation between the IOCTL Task and the PnP Task
static BOOL8 gfPnPIsActive = FALSE;


#if defined(__APPLE__)

/*****************************************************************************
EnterUsbCS
==============================================================================

This function intialize a semaphore for synchronization of the USB port
access.

Return:
-------
The function returns TRUE if the intialisation was successfully, FALSE
if not.

*****************************************************************************/

static BOOL8 EnterUsbCS ()
{
  // see if we can update intenal box list
  if (gUsbSemaphoreID != NULL) {
    if((sem_trywait(gUsbSemaphoreID) == -1)) {
      // exit only if a normal error occures
      if (errno != EAGAIN) {
        return FALSE;
      } // if
    } // if
  } // if
  return TRUE;
} // EnterUsbCS()


/*****************************************************************************
ExitUsbCS
==============================================================================

This function cleans up a semaphore for synchronization of the USB port
access.

Return:
-------
The function returns TRUE if the clean up was successfully, FALSE if not.

*****************************************************************************/

static void ExitUsbCS ()
{
  // release semaphore
  if (gUsbSemaphoreID != NULL) {
    sem_post(gUsbSemaphoreID);
  } // if
} // ExitUsbCS()
#endif // defined(__APPLE__)


/*****************************************************************************
QuerySerialNumber
==============================================================================

This function sends a FSEARCH command to the box and receives the serial
number.

//!N:mn WKBSERIAL: added *pDev
//!N:mn IOUSBDeviceInterface: f()-pointer to the BoxCalls added *pDev

*****************************************************************************/

static BOOL8 QuerySerialNumber (IOUSBDeviceInterface **ppDev, WKBSERIAL *pwkbser)
{
  IOUSBDevRequest ioRequest; // LINUX: equal with: struct usbdevfs_ctrltransfer
  UInt8 abBuffer[10];
  IOReturn kr; // LINUX: equal with: INT

  memset(abBuffer, 0x0, sizeof(abBuffer));
#if defined(__linux__)
  #if 0
  USHORT length = 6;
  UINT timeout = USBRAWTIMEOUT; 
  USHORT value = 0;
  ioRequest = (IOUSBDevRequest)
      { WIBU_READ, WIBU_FSEARCH, value, 0, length, timeout, abBuffer };
  #else 
  memset(&ioRequest, 0x0, sizeof(ioRequest));
  ioRequest.bRequestType = WIBU_READ;
  ioRequest.bRequest = 0x04;
  ioRequest.wValue = 0;
  ioRequest.wIndex = 0;
  ioRequest.wLength = 6;
  ioRequest.timeout = USBRAWTIMEOUT; // msec
  ioRequest.data = abBuffer;
  #endif 
#else
  ioRequest.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
  ioRequest.bRequest = 0x04;  // comand for FSEARCH
  ioRequest.wValue = 0;
  ioRequest.wIndex = 0;
  ioRequest.wLength = 6;      // the length of the answer sequence
  ioRequest.pData = abBuffer; // the answer buffer
#endif //LINUX

  kr = (*ppDev)->DeviceRequest(ppDev, &ioRequest);

#if defined(__linux__)
  if (kr >= 0) {
#else
  if (kr != 0) {
    // QuerySerialNumber(failed)
  }
  if (!kr) {
#endif 
    // set the serial number in the right format
    pwkbser->ulSerial = (abBuffer[3] << 24) | (abBuffer[2] << 16) | (abBuffer[1] << 8) | abBuffer[0];
    pwkbser->usMask = (unsigned short) abBuffer[4];
    pwkbser->usReserved = 0;
    return TRUE;
  } // if
 
  return FALSE;
} // QuerySerialNumber()


/*****************************************************************************
BoxDeviceRequest
==============================================================================

This function calls the USBRequest operation and waits for its termination.

Parameters
  dev            contains the device reference of the accessed device
  BMRequestType  contains the type of the USB request
  BRequest       contains the request command
  wValue
  wIndex         contains the WIndex of the request
  pvBuffer
  cbBuffer

Return
  The function returns a WibuKey error code which can be
  one of the following values:

*****************************************************************************/

static BOOL8 BoxDeviceRequest (IOUSBDeviceInterface **ppDev,
    UInt8 BMRequestType, UInt8 BRequest, UInt16 wValue, UInt16 wIndex, 
    void *pvBuffer, unsigned int cbBuffer)
{
  IOReturn kr;
  IOUSBDevRequest ioRequest;
#if defined(__linux__)
  UInt32 timeout = USBRAWTIMEOUT;
  ASSERT(BMRequestType == WIBU_WRITE || BMRequestType == WIBU_READ);
  ASSERT(BRequest != WIBU_SELECT_1 || BRequest != WIBU_NSEARCH);
#else
  ASSERT( (BMRequestType == (USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice))) ||
          (BMRequestType == (USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice))) );
#endif //LINUX
  if (cbBuffer == 0) {
    // the in- or output buffer is 0 ==>> nothing to do
    return TRUE;
  } // if

  // set the parameters for the communication with the device
#if defined(__linux__)
  ioRequest = (IOUSBDevRequest)  /* (struct usbdevfs_ctrltransfer) */
    { BMRequestType, BRequest, wValue, wIndex, cbBuffer, timeout, pvBuffer };
#else
  ioRequest.bmRequestType = BMRequestType;
  ioRequest.bRequest = BRequest; // WibuBox ASIC comand
  ioRequest.wValue = wValue;
  ioRequest.wIndex = wIndex;
  ioRequest.wLength = cbBuffer; // the length of the answer sequence
  ioRequest.pData = pvBuffer; // the answer buffer
#endif //LINUX

  // send the request to the device
  kr = (*ppDev)->DeviceRequest(ppDev, &ioRequest);

  // everything all right ==>> return with TRUE
#if defined(__linux__)
  return (kr >= 0);
#else
  if (kr != 0 ) {
      // request failed 
  }
  return (kr == 0);
#endif //LINUX
} // BoxDeviceRequest()


/*****************************************************************************
MarkBoxListEmpty
==============================================================================

This function reads mark all elements of the global box list to FALSE.

*****************************************************************************/

static void MarkBoxListEmpty()
{
  BOXLIST *pbox = NULL;

  // mark all elements of the list to FALSE
  if (gpboxlist != NULL) {
    // EnumConnectedBoxes
    for (pbox = gpboxlist; pbox != NULL; pbox = pbox->pNext) {
      pbox->fBoxAlreadyInList = FALSE;
    } // for
  }//if
} // MarkBoxListEmpty()

#if defined(__linux__)


/*****************************************************************************
TestBoxInList
==============================================================================

This function check, if an given USB-entry is already stored inside the 
global box list.

*****************************************************************************/

static BOOL TestBoxInList(INT nBusNo, INT nDeviceNo, 
  UInt16 nProductID, UInt16 nVendorID)
{
  BOXLIST *pBox = NULL;
  BOOL fBoxIsInList = FALSE;
  char szBuffer[64];

  for (pBox = gpboxlist; pBox != NULL; pBox = pBox->pNext) {
    
    if ( pBox->fBoxAlreadyInList == FALSE ) {
      // check if we can access current pBox
      snprintf(szBuffer, sizeof(szBuffer), "%s/%03u/%03u", 
        WkhUsbGetDeviceMountPoint(),
        (*(pBox->ppDev))->uBusNum,(*(pBox->ppDev))->uDevNum);
      if (access(szBuffer, W_OK | R_OK) != 0) {
        // access failed 
        pBox->fBoxAlreadyInList = FALSE;
      } // if (access)
    } // if
  
    // Check if found USB-device is same as in List
    // * devnum may have changed (new hotplug DevNum).
  #if !HAVE_GCC_3X
    if ( (nBusNo     == (*(pBox->ppDev))->uBusNum) &&
         (nDeviceNo  == (*(pBox->ppDev))->uDevNum) &&
        ((nProductID == (*(pBox->ppDev))->pDev->desc.idProduct[0])|
                       ((*(pBox->ppDev))->pDev->desc.idProduct[1]<<8)) )
  #else
    if ( (nBusNo     == (*(pBox->ppDev))->uBusNum) &&
         (nDeviceNo  == (*(pBox->ppDev))->uDevNum) &&
         (nProductID == (UInt16)(*(pBox->ppDev))->pDev->desc.idProduct) )
  #endif //!HAVE_GCC_3X
    {
      // Box found in List, so stop scanning here
      pBox->fBoxAlreadyInList = TRUE;
      fBoxIsInList = TRUE;
      break;
    } // if
  } // for

  return fBoxIsInList;
} // TestBoxInList()


static BOOL AddBoxToList(INT nBusNo, INT nDeviceNo, 
  UInt16 nProductID, UInt16 nVendorID)
{
 // Doing a BoxList-Walk !
              // We have 2 Lists: global 1st element = gpboxlist and  local = pbox.
              IOUSBDeviceInterface *pIoUsbDevIf;
              struct usbdevice         *pUsbDev;
              BOXLIST                 *pboxlist; // local list to add boxes
              // INT       idVendor = WK_VENDOR_ID; // default


              // Chaining into List: first to pboxlist, then add gpboxlist to the end
              // * Allocate lots of Space to write to:
              // * allocate memory for the new elements of the list.
              pboxlist = (BOXLIST*) malloc (sizeof(BOXLIST));
              if (pboxlist == NULL) {
                return FALSE;
              }// if
              // preinititialize:
              pboxlist->pNext = NULL;
              pboxlist->pPrev = NULL;

              // allocate memory for the new element (ppDev) of the list
              pIoUsbDevIf = (IOUSBDeviceInterface*) malloc (sizeof(IOUSBDeviceInterface));
              if (pIoUsbDevIf == NULL) {
                return FALSE;
              }// if
              // preinititialize:
              setUSBDevIf(pIoUsbDevIf);

              // Allocate memory for the new element (desc) of the list
              // * this struct is prefilled here with product, idVendor
              // * (ned to access the box); but its overwritten later by
              // * wibu_proc_usb_devices_open() at once !!
              pUsbDev = (struct usbdevice*) malloc (sizeof(struct usbdevice));
              if (pUsbDev == NULL) {
                return FALSE;
              }// if
              // preinititialize:
              pUsbDev->fd = 0;
              pUsbDev->fIsStillOpen = 0;

              // Chaining allocated elements into Structs, pointer bending stuff -
              // * faked (*this)-Self-Pointer is used to handle **ppDev of BOXLIST.
              pIoUsbDevIf->pvThis = (void*) pIoUsbDevIf;
              pboxlist->ppDev = /*(IOUSBDeviceInterface*)*/(void*)(pIoUsbDevIf->pvThis);

              pIoUsbDevIf->pDev = pUsbDev;

              // Now filling box with values
              // set "box is in list"
              pboxlist->fBoxAlreadyInList = TRUE;

              // Set the access values for USB box in '/proc/bus/usb/'<busnum>'/'<devnum>
              pIoUsbDevIf->uBusNum = nBusNo;
              pIoUsbDevIf->uDevNum = nDeviceNo;

              // the default value for vendor = 'WIBU-SYSTEMS AG'
            #if !HAVE_GCC_3X
              pIoUsbDevIf->pDev->desc.idVendor[0] = (UInt8) nVendorID;
              pIoUsbDevIf->pDev->desc.idVendor[1] = (UInt8) (nVendorID>>8);

              pIoUsbDevIf->pDev->desc.idProduct[0] = (UInt8) nProductID;
              pIoUsbDevIf->pDev->desc.idProduct[1] = (UInt8) (nProductID>>8);
            #else
              pIoUsbDevIf->pDev->desc.idVendor = (UInt16) nVendorID;

              pIoUsbDevIf->pDev->desc.idProduct = (UInt16) nProductID;
            #endif //!HAVE_GCC_3X

              // Setting serial number of the box will happen LATER
              // * The complete global BOXLIST is attached to the end of
              // * local (new detected) box.
              pboxlist->pNext = gpboxlist;
              if (gpboxlist != NULL) {
                // the previous box of the new second element
                // * is now the new detected box
                gpboxlist->pPrev = pboxlist;
              } // if

              // global boxlist is the new detected box (as first element)
              gpboxlist = pboxlist;

              // increment global WibuBox counter
              gcBoxList++;

              // force rereading of the internal box list
              gfForceReading = TRUE;// means "gpboxlist has changed"
              gidListLastScanReal++;
  return TRUE;
}

/*****************************************************************************
ParseProcFs
==============================================================================

This function parse /proc/bus/usb/devices for all connected WibBoxes.
For each BOX in the global list, this function will check if the 
BOX is longer accessible (HOTplug). If the Box is new connected, then 
this function add a new entry into the global box list.

Return 
  Return '0' if successful
  Return '1' if open failed
  Return '2' if access error 
  Return '3' if no changes - all data are cached 

*****************************************************************************/

static int ParseProcFs(BOXLIST *pvDummy)
{
  struct pollfd pfd;
  char devsfile[64];
  struct stat my_stat;
  INT ret = 0;
  char buf[16384]; //16k
  char *start = NULL, *end = NULL, *lineend = NULL, *cp = NULL;
  
  INT devnum = -1;
  INT busnum = -1;
  UInt16 vendor = (UInt16)-1;
  UInt16 product = (UInt16)-1;
  INT vendorid = WK_VENDOR_ID;
  INT productidU = WK_PROD_ID_BOXU;
  INT productidRU = WK_PROD_ID_BOXRU;
  INT productidS = WK_PROD_ID_BOXS;
  static time_t gLastModTime = 0;
  
  memset(devsfile, 0x0, sizeof(devsfile) );
  memset(buf, 0x0, sizeof(buf) );
  
  { // bugfix for SuSE 9.1 - we have no /proc/bus/usb/devices 
    if ( stat( "/proc/bus/usb/devices", &my_stat ) == 0 ) {
      // ok, we have a kernel-2.4.x like /proc-System 
      snprintf(devsfile, sizeof(devsfile), "%s", "/proc/bus/usb/devices");
    } else {
      // ok, check if we have a suse-kernel-2.6 like sysfs-system 
      snprintf(devsfile, sizeof(devsfile), "%s", 
        "/proc/bus/usb/devices_please-use-sysfs-instead"); 
    }
  } // bugfix for SuSE 9.1
  
  if ( devsfile[0] != '\0' ) {
    // resolve links (i.e. it can be a link to '.usb_fs/device')
    char szResolvBuf[MAXPATHLEN];
    memset(szResolvBuf, 0x0, sizeof(szResolvBuf) );
    if ( NULL != realpath(devsfile, szResolvBuf) ) {
      snprintf(devsfile, sizeof(devsfile), "%s", szResolvBuf);
    } // if
  } // if 
  if ( stat(devsfile, &my_stat) == 0 ) {
    if (my_stat.st_ctime == gLastModTime) {
      // no changes on ProcFS - use caching
      return 3;
    } // if
    gLastModTime = my_stat.st_mtime;
  } // if
  MarkBoxListEmpty();

  if ((pfd.fd =  open(devsfile, O_RDONLY | O_EXCL)) == -1) { 
    // error 
    pvDummy = NULL;
    return 1;
  } //if

  // parsing proc/bus/usb/devices
  for (;;) {
    //#################################
    // POLLING: nonblocking read, return immediately; pause; do again.
    //P:mn See select(2): synchronous I/O-multiplexing
    fd_set rfds;
    struct timeval tv;
    int fValid;

    // Now parsing the '/proc/bus/usb/devices'
    // reset all fdset-pointer-bits
    FD_ZERO(&rfds);
    // add pDev to a set of readfds
    FD_SET(pfd.fd, &rfds);//FD_SET(0, &rfds);

    // timeout
    tv.tv_sec = 0L;
    tv.tv_usec = TIMEOUT_MS_RECV;

    // SVR4 and BSD not POSIX.1
    // non-blocking check if read() is possible
    fValid = select(pfd.fd + 1, &rfds, NULL, NULL, &tv);
    // value of tv is UNDEFINED now!

    if (fValid > 0){
      // success
      // only read if pDev is ready for reading
      if ( FD_ISSET(pfd.fd, &rfds) ){

      // reset Filepointer to start
      errno = 0;
      if (lseek(pfd.fd, 0, SEEK_SET) == (off_t)-1){
        ret = -1;
        goto RET_INSERT_2;
      }//if

      ret = read(pfd.fd, buf, sizeof(buf)-1);
      if (ret == -1){
        goto RET_INSERT_2;
      }//if
  
      }// if
    } else if (fValid == 0){
      // time is out
    } 
  
    //###################################################
    end = buf + ret;
    *end = 0;
    // start of opened '/proc/bus/usb/devices'
    start = buf;
    ret = 0;
    // further Lines in '/proc/bus/usb/devices'
    while (start < end) {
      // step the lines
      lineend = strchr(start, '\n');
      if (!lineend){
        break;
      }//if
      *lineend = 0;
      // checking line contents:
      switch (start[0]) {
      // T = Topology (etc.) line
      case 'T':
        if ((cp = strstr(start, "Dev#="))) {
          devnum = strtoul(cp + 5, NULL, 0);
        } else
          devnum = -1;
        if ((cp = strstr(start, "Bus="))) {
          busnum = strtoul(cp + 4, NULL, 0);
        } else {
          busnum = -1;
        }//if
        break;//'T'
      /* P = Product ID line */
      case 'P':
          if ((cp = strstr(start, "Vendor="))) {
            vendor = (UInt16)strtoul(cp + 7, NULL, 16);
          } else {
            vendor = -1;//not found
          }//if
          if ((cp = strstr(start, "ProdID="))) {
            product = (UInt16)strtoul(cp + 7, NULL, 16);
          } else {
            product = -1;//not found
          }//if

          if ( vendor == (UInt16)vendorid && 
              (product == (UInt16)productidU || product == (UInt16)productidRU
              || product == (UInt16)productidS) 
            ) {

            // Create global BOXLIST-element and chain the first 3 values
            // * (busnum, devnum, product) of the Box into global BOXLIST-element.
            // * There is no Check if box existst in list yet!
            // BOXLIST                 *pBox;
            char        testdevsfile[256]; //DeviceString for testing access
            // char     testolddevsfile[256]; //OLD-DeviceString (hotplugged off)
            BOOL8    fBoxIsInList = FALSE; // Default: Box NOT in List

            // Box found - Check if newly found USB-device is writeable
            errno = 0;
            snprintf(testdevsfile, sizeof(devsfile),
                     "%s/%03u/%03u", WkhUsbGetDeviceMountPoint(), busnum, devnum);
            if (access(testdevsfile, W_OK) == -1){
              switch(errno){
                case ENOMEM:       /* 12 Out of memory                       */
                case EACCES:       /* 13 Permission denied                   */
                case EFAULT:       /* 14 Bad address                         */
                case ENOTDIR:      /* 20 Not a directory                     */
                case EINVAL:       /* 22 Invalid argument                    */
                case ENAMETOOLONG: /* 36 File name too long                  */
                case ENOENT:       /*  2 No such file or directory           */
                case ELOOP:        /* 40 Too many symbolic links encountered */
                  cstWkd.idError = WKB_ERROR_INVALID_PORT;
                  return 2;
                  // break;
                default:
                  return 2;
              }//switch
            }//if (access)

            // test if (busnum, devnum, product) is still in global List!
            fBoxIsInList = TestBoxInList(busnum, devnum, product, vendor);

            // Only reel New Boxes need a new Listelement:
            if (!fBoxIsInList) {
              AddBoxToList(busnum, devnum, product, vendor);
            }

            ret++; // increment Box-Counter

           ; /*nothing */
          } else{
            //default:
            // Found None WIBU-Device: do nothing in this case
          }//if
          break;
      }//switch (start[0])
      start = lineend + 1;
    }//while (start < end)

    // All Lines in '/proc/bus/usb/devices' are now parsed:
    if (gpboxlist == NULL){
      goto RET_INSERT_2;
    } //if

RET_INSERT_2:
    if (ret <= 0) {
      close(pfd.fd);
      pvDummy = NULL;
      return 1;
    } //if
    if (ret > 0){ 
      // success
      break;
    }//if
  }//for(;;)
  close(pfd.fd);
  // END of parsing proc/devices

  return 0;
} // ParseProcFs()



/*****************************************************************************
ParseSysFs
==============================================================================

This function parse /sys/bus/usb for connected boxes and insert the bus 
topology into the global box list gpboxlist

Return
  0        if successful
  1        if insert failed 
  2        if parse failed, exit with error, no box connected

*****************************************************************************/

static int ParseSysFs()
{
  struct stat my_stat;
  struct dirent *pDirent;
  const char* szSysDir= "/sys/bus/usb/devices";
  char szBuffer[128];
  char szReadBuffer[64];
  DIR *dirp = NULL;
  int fd = 0;
  int cbLen = 0, k= 0;
  INT devnum = -1;
  INT busnum = -1;
  UInt16 vendor = (UInt16)-1;
  UInt16 product = (UInt16)-1;

  memset(szBuffer, 0x0, sizeof(szBuffer));
  
  if ( stat( szSysDir, &my_stat ) != 0 ) {
    return 1;
  } // if 

  MarkBoxListEmpty();

  dirp = opendir(szSysDir);
  if (dirp != 0) {
    while ((pDirent = readdir(dirp)) != NULL) {
      vendor = (UInt16)-1;
      vendor = (UInt16)-1;  
      devnum = -1;
      busnum = -1;
    
      // exclude '.' and '..' dirs
      if ( (strcmp(pDirent->d_name, "..") == 0) 
        || (strcmp(pDirent->d_name, ".") == 0) ) {
          continue;
      } // if
    
      // get VendorID
      snprintf(szBuffer, sizeof(szBuffer), "%s/%s/idVendor", szSysDir, pDirent->d_name);
      if ( (fd = open(szBuffer, O_RDONLY | O_EXCL) ) == -1) {
        // no VendorID-File 
        continue;
      } else {
        cbLen = read(fd, szReadBuffer, sizeof(szReadBuffer));
        close(fd);
        if ( memcmp(szReadBuffer, "064f", 4) != 0 ) {
          // invalid VendorID
          continue;
        } // if
        vendor = WK_VENDOR_ID;
      } // if 
      
      // get ProductID 
      snprintf(szBuffer, sizeof(szBuffer), "%s/%s/idProduct", szSysDir, pDirent->d_name);
      if ( (fd = open(szBuffer, O_RDONLY | O_EXCL) ) == -1) {
        // no ProductID 
        continue;
      } else {
        memset(szReadBuffer, 0x0, sizeof(szReadBuffer));
        cbLen = read(fd, szReadBuffer, sizeof(szReadBuffer));
        close(fd);
        if ( 1 == sscanf(szReadBuffer, "%x", &k) ) {
          product = k;
        } else {
          // invalid ProductID 
          continue;
        } // if
        if ( ! ((k == WK_PROD_ID_BOXU) || (k == WK_PROD_ID_BOXRU) ) 
          || (k == WK_PROD_ID_BOXS)
          ) {
          // invalid ProductID 
          continue;
        } // if
        
      } // if 
      
      // get Bus Device Number
      snprintf(szBuffer, sizeof(szBuffer), "%s/%s/devnum", szSysDir, pDirent->d_name);
      if ( (fd = open(szBuffer, O_RDONLY | O_EXCL) ) == -1) {
        // no ProductID 
        continue;
      } else {
        memset(szReadBuffer, 0x0, sizeof(szReadBuffer));
        cbLen = read(fd, szReadBuffer, sizeof(szReadBuffer));
        close(fd);
        if ( 1 == sscanf(szReadBuffer, "%i", &k) ) {
          devnum = k;
        } else {
          // invalid ProductID 
          continue;
        } // if
      } // if
    
      // get busnum 
      if ( 2 == sscanf(pDirent->d_name, "%i-%s", &k, szReadBuffer) ) {
        busnum = k;
      } // if 
      
      // check for readable (busnum,devnum) 
      snprintf(szBuffer, sizeof(szBuffer), "%s/%03i/%03i" , WkhUsbGetDeviceMountPoint(), busnum, devnum);
      if ( 0 != access(szBuffer, W_OK | R_OK) ) {
        continue;
      } // if
    
      { // block 
        // make new global box entry 
        if ( FALSE == TestBoxInList(busnum, devnum, product, vendor) ) {
          // missing in list, add it 
          AddBoxToList(busnum, devnum, product, vendor);
        } // if 
          
      } // block 
      
    } // while 
    closedir(dirp);
  } // if

  return 0;
} // ParseSysFs()

/*****************************************************************************
EnumConnectedBoxes                                                     (LINUX)
==============================================================================

This function reads all connected boxes out of the IORegistry in an
internal double linked list.

Parameters
  holdSem   TRUE  if Semaphore[0] should be hold  at return.
  FALSE if Semaphore[0] should be left at return.

*****************************************************************************/

static BOOL8 EnumConnectedBoxes (BOOL8 holdSem)
{
  char devsfile[64];
  int nRet = 0;
  BOXLIST *pvDummy = NULL;
  BOXLIST *pbox = NULL;

  // parse for connected boxes 
  if ( 0 == WkhUsbUseSysFs() ) {
    nRet = ParseProcFs(pvDummy);
  } else {
    nRet = ParseSysFs();
  } // if 

  switch (nRet) {
    case 1:
      goto RET_INSERT;
    case 2:
      goto EXIT_ERROR;
    case 3:
      // already up-to date
      gfPnPIsActive = FALSE;
      goto RET_CHECK;
      // return TRUE;
    default:
      // nothing 
      ;
  } // switch 
  
  // Walk again through global list:
  // * for each prescanned Box, now get usb_device_descriptor
  for (pbox = gpboxlist; pbox != NULL; pbox = pbox->pNext) {
    if(pbox->fBoxAlreadyInList){
      // read busnum,devnum as from parsing above

      // open / try to open if fd exists:
      // Close the Boxes?
      errno = 0;

      // because pbox contents: busnum, devnum, productid !
      (*(pbox->ppDev))->pDev = wibu_usb_open(pbox, TRUE);

      if (errno){
        switch(errno){
        // 13  Permission denied :: open():
        case EACCES:
          snprintf(devsfile, sizeof(devsfile),
            "%s/%03u/%03u", WkhUsbGetDeviceMountPoint(), (*(pbox->ppDev))->uBusNum,
            (*(pbox->ppDev))->uDevNum);//256
          cstWkd.idError = WKB_ERROR_INVALID_PORT;//!N:mn
          goto EXIT_ERROR;
          break;
        case EIO:
          // BSRMSGM("Read error: Not enough bytes: \"%s\"\n", strerror(errno) );
          goto EXIT_ERROR;
          break;
        case -ENOENT:
          // vendor not WIBU or prodID != lookupID(in pbox)
          // Just Destroy Listelement
          break;
        default:
          // Any unknown Error, no box found
          goto EXIT_ERROR;
        } //switch
      } //if (errno)
    } //if(pbox->fBoxAlreadyInList)
  } // for each prescanned Box

  pvDummy = gpboxlist;

RET_INSERT:

  if (pvDummy == NULL) {
      // BSRMSGM("Neither WibuBox/U nor WibuBox/RU detected!\n");
      cstWkd.idError = WKB_ERROR_NOT_CONNECTED;
      goto EXIT_ERROR;
  } else{
    if (pbox == NULL){
      // BSRMSGM("USB \nNo more Boxes in List !\n");
    }else if (((*(pbox->ppDev))->pDev) == NULL){
      // BSRMSGM("USB \nNo more Boxes connected !\n");
    }//if
  }//else

RET_CHECK:
  // Now in the list all valid boxes marked as "fBoxAlreadyInList=TRUE"
  // * all Invalid boxes could be deleted!
  for (pbox = gpboxlist; pbox != NULL; /*pbox = pbox->pNext*/) {
    // For each (valid) Box, now get the Serial-Number into each Box' usbdevice*
    if(pbox->fBoxAlreadyInList){
      if (!QuerySerialNumber (&(*(pbox->ppDev))->pvThis, &(pbox->wkbsr))){
        // If ERROR:  asume Box isn't valid, so delete it.
        // * Force "redo" scanning, with same element (due to delete Box from List):
        pbox->fBoxAlreadyInList = FALSE;
        continue;
      }//if(!QuerySerialNumber)
    } else {
      // For each (Invalid) Box: delete all elements in the list,
      // *  which are not any more connected to the port.
      BOXLIST *pboxlistH = NULL;

      IOUSBDeviceInterface **ppDev = NULL;
      IOReturn kRet;

      // get the reference to the device
      //!P:mn ppDev = pbox->ppDev;
      ppDev = (void*)(*(pbox->ppDev))->pvThis;

      //!C:mn NO QuerySerialNumber() here, not accessible (See above)

      //This is from case IOCTL_WK_EXECUTE: in ControlWkUSBDriver()
      // free the reference to the box and free the usbdevice*
      kRet = (*ppDev)->USBDeviceClose(ppDev);
      // free the IOUSBDeviceInterface**
      kRet = (*ppDev)->Release(ppDev);

      // set a new pre- & successor
      if (pbox->pNext != NULL) {
        pbox->pNext->pPrev = pbox->pPrev;
      } // if
      if (pbox->pPrev != NULL) {
        pbox->pPrev->pNext = pbox->pNext;
      } else {
        // pbox->pPrev != NULL ==>> pbox == gpboxlist
        // ==>> pbox is the first element in the list
        gpboxlist = pbox->pNext;
      } // if

      // set the next element to check
      pboxlistH = pbox;
      pbox = pbox->pNext;
      // free the allocated memory (BOXLIST*)
      free (pboxlistH);

      // decrease WibuBox counter
      gcBoxList--;

      // force rereading of the internal box list
      gfForceReading = TRUE;// means "gpboxlist has changed"
      gidListLastScanReal++;
    }//else //if(pBox->fBoxAlreadyInList)

    // check next device in the list
    if (pbox != NULL){
      pbox = pbox->pNext;
    } else {
      break;// spare the way up
    }
  } // for

  gfPnPIsActive = FALSE;
  // No error occured ==>> exit with TRUE
  return TRUE;

EXIT_ERROR:
  // An error occured ==>> exit with FALSE
  gfPnPIsActive = FALSE;
  return FALSE;
} // EnumConnectedBoxes()

#endif // defined(__linux__)


#if defined(__APPLE__)
// fix for ASIC 1804 - invalid configuration after suspend 
static BOOL8 SetupDeviceConfiguration(IOUSBDeviceInterface187 **ppDev)
{

  UInt8 configNum = 0;
  kern_return_t kRet;
  kRet = (*ppDev)->GetConfiguration(ppDev, &configNum);
  
  if (kRet) {
    // this case should never happens
    usleep(1000 * 1000);
  
    return FALSE;
  }
  return TRUE;
} // SetupDeviceConfiguration


/*****************************************************************************
EnumConnectedBoxes
==============================================================================

This function reads all connected boxes out of the IORegistry in an
internal double linked list.

*****************************************************************************/

static BOOL8 EnumConnectedBoxes ()
{
  kern_return_t kRet;
  mach_port_t nMasterPort;
  CFMutableDictionaryRef nUsbMatch;
  SInt32 nUsbVendor = WK_VENDOR_ID;
  SInt32 nUsbDeviceSubClass = 0;
  io_iterator_t nIter;
  io_service_t nUsbDevice;
  BOXLIST *pbox = NULL;
  BOOL8 fRetValue = FALSE;
  int k = 0;

  // see if we can update intenal box list
  EnterUsbCS();

  // initiate the comminication with the IOKit
  kRet = IOMasterPort (MACH_PORT_NULL, &nMasterPort);
  if (kRet || !nMasterPort) {
    goto EXIT_ERROR;
  } // if

  for (k=0;k<4;k++) {

    // connect to the USB service
    nUsbMatch = IOServiceMatching (kIOUSBDeviceClassName);
    if (!nUsbMatch) {
      mach_port_deallocate(mach_task_self(), nMasterPort);
      goto EXIT_ERROR;
    } // if
  
    { // Add our vendor and product IDs to the matching criteria
      CFNumberRef num1;
      CFNumberRef num2;
      num1 = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &nUsbVendor);
      CFDictionarySetValue (nUsbMatch, CFSTR(kUSBVendorName), num1 ); 
      num2 = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &nUsbDeviceSubClass);
      CFDictionarySetValue (nUsbMatch, CFSTR(kUSBDeviceSubClass), num2 ); 
      CFRelease(num1);
      CFRelease(num2);
    } // block 
  
    // create an iterator over all matching IOService nubs
    kRet = IOServiceGetMatchingServices(nMasterPort, nUsbMatch, &nIter);
    if (kRet) {
      CFRelease(nUsbMatch);
      mach_port_deallocate(mach_task_self(), nMasterPort);
      goto EXIT_ERROR;
    }
  
    MarkBoxListEmpty();
  
    // iterate over all found devices and interfaces
    while ( nUsbDevice = IOIteratorNext(nIter) ) {
      IOCFPlugInInterface **ppIoDev = NULL;
      IOUSBDeviceInterface **ppDev = NULL;
      HRESULT nRes;
      SInt32 nScore;
      UInt16 idVendor = 0;
      UInt16 idProduct = 0;
      
      kRet = IOCreatePlugInInterfaceForService (nUsbDevice,
          kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &ppIoDev, &nScore);
      // done with the device object now that I have the plugin
      IOObjectRelease(nUsbDevice);
      if (kRet || !ppIoDev) {
        // ==>> goto next detected device
        continue;
      } // if
              
      // i have the device plugin. I need the device interface
      nRes = (*ppIoDev)->QueryInterface (ppIoDev, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID), (LPVOID)&ppDev);
      // done with this
      (*ppIoDev)->Release(ppIoDev);
      if (nRes || !ppDev) {
        // ==>> goto next detected device
        continue;
      } // if
  
      // technically should check these kRet values
      kRet = (*ppDev)->GetDeviceVendor (ppDev, &idVendor);
      kRet = (*ppDev)->GetDeviceProduct (ppDev, &idProduct);
      if (idVendor != WK_VENDOR_ID) {
        // We should never get here because the matching criteria we specified above
        // will return just those devices with our vendor and product IDs
        (*ppDev)->Release(ppDev);
        // ==>> goto next detected device
        continue;
      } // if
  
      { // open USB device
        int i = 0;
        for (i = 0; i < 5; i++) {
          // need to open the interface in order to change its state
          kRet = (*ppDev)->USBDeviceOpen(ppDev);
          if (kRet == kIOReturnExclusiveAccess) {
            sleep(1);  // 1sec
            continue;
          } else {
            // The device cannot be opend if it is already opened!!!
            // ==>> goto next detected device
            break;
          } // if
        } // for
        if (kRet) {
          // The device cannot be opend if it is already opened!!!
          // ==>> This means that the device is already connected and inserted in the internal list.
          (*ppDev)->Release(ppDev);
          // ==>> goto next detected device
          continue;
        } // if
      } // block
      
      // Now, this device is not yet in the internal list ==>> insert it now!!!
      switch (idProduct) {
        case WK_PROD_ID_BOXU:
        case WK_PROD_ID_BOXS:
        case WK_PROD_ID_BOXRU: {
          BOOL8 fValid, fBoxFound;
          BOXLIST *pboxlist = NULL;
          WKBSERIAL wkbser;
  
          // get the serial number of the box
          fValid = QuerySerialNumber (ppDev, &wkbser);
          if (!fValid) {
          
            // bugfix ASIC1804
            // try to receive GetConfiguration to trigger a reinnit of IOUSBLib 
            // sometimes a re-enumeration of the device is required
            if ( FALSE == SetupDeviceConfiguration(ppDev) ) {
              // failed - wait and reenumerate complete new over IOService 
              if (nIter) {
                IOObjectRelease(nIter);
                nIter = 0;
              } // if
              break;
            } else {
              // of, was success, try again 
              k = 4;
              goto END_LOOP;
            }
          } else {
            // of, was success, try again 
            k = 4;
          } // if
  
          // check if box is in alredy in list
          fBoxFound = FALSE;
          pbox = gpboxlist;
          while (pbox != NULL) {
            if (!pbox->fBoxAlreadyInList) {
              // check if the box in the list and the current box are equal 
              if (pbox->wkbsr.usMask == wkbser.usMask && pbox->wkbsr.ulSerial == wkbser.ulSerial) {
                // box found ==>> exit loop
                pbox->fBoxAlreadyInList = TRUE;
                fBoxFound = TRUE;
                break;
              } // if
              // box not equal ==>> check next device in the list
              pbox = pbox->pNext;
            } else {
              // box already marked ==>> check next device in the list
              pbox = pbox->pNext;
            } // if
          } // while
          if (!fBoxFound) {
            // box not found in the list ==>> insert the box in the list
            // allocate memory for the new element of the list
            pboxlist = (BOXLIST*) malloc (sizeof(BOXLIST));
            if (pboxlist == NULL) {
              goto END_LOOP;
            } // if
            
            // box is in list
            pboxlist->fBoxAlreadyInList = TRUE;
            // set the pointer to the device into the list
            pboxlist->ppDev = ppDev;
            // set serial number of the box
            pboxlist->wkbsr = wkbser;
            // previous box is allways NULL, because the cuurent box is set to the beginning
            pboxlist->pPrev = NULL;
            // successor is the whole list
            pboxlist->pNext = gpboxlist;
            if (gpboxlist != NULL) {
              // the previous box of the new second element is now the new detected box
              gpboxlist->pPrev = pboxlist;
            } // if
            // new first box is the new detected box
            gpboxlist = pboxlist;
            // increase WibuBox counter
            gcBoxList++;
  
            // force rereading of the internal box list
            gfForceReading = TRUE;
            gidListLastScanReal++;
          } // if
        } // case
        break;
        default:
          // do nothing in this case
          ;
      } // switch
  
    END_LOOP:
      // close access to the device
      kRet = (*ppDev)->USBDeviceClose(ppDev);
      // close the access to the driver
      kRet = (*ppDev)->Release(ppDev);
    } // while

  } // for

  // delete all boxes out of the list which are not any longer connected to the computer
  pbox = gpboxlist;
  while (pbox != NULL) {
    BOXLIST *pboxH = pbox->pNext;
    if (!pbox->fBoxAlreadyInList) {
      BOXLIST *pboxlistH;
      // set a new pre- & successor
      if (pbox->pNext != NULL) {
        pbox->pNext->pPrev = pbox->pPrev;
      } // if
      if (pbox->pPrev != NULL) {
        pbox->pPrev->pNext = pbox->pNext;
      } else {
        // pbox->pPrev != NULL ==>> pbox == gpboxlist
        // ==>> pbox is the first element in the list
        gpboxlist = pbox->pNext;
      } // if

      // set the next element to check
      pboxlistH = pbox;
      pbox = pbox->pNext;
      // free the allocated memory
      free (pboxlistH);     
       // decrease WibuBox counter
      gcBoxList--;

      // force rereading of the internal box list
      gfForceReading = TRUE;
      gidListLastScanReal++;
    } // if
    pbox = pboxH;
  } // while

  // release the device and intrerface iterator
  if (nIter) {
    IOObjectRelease(nIter);
    nIter = 0;
  } // if

  if (nMasterPort) {
    mach_port_deallocate(mach_task_self(), nMasterPort);
    nMasterPort = 0;
  } // if

  fRetValue = TRUE;
EXIT_ERROR:
  // release semaphore
  ExitUsbCS();
  
  // if an error occured ==>> exit with FALSE
  // if no error occured ==>> exit with TRUE
  return fRetValue;
} // EnumConnectedBoxes()

#endif // defined(__APPLE__)

#if defined(__APPLE__)
/*****************************************************************************
SignalHandler
==============================================================================

This function is called when application is interrupted.

*****************************************************************************/

void SignalHandler (int sigraised)
{
  // Clean up here
  IONotificationPortDestroy(gNotifyPort);
  if (gAddedIter) {
    IOObjectRelease(gAddedIter);
    gAddedIter = 0;
  } // if
  if (gRemovedIter) {
    IOObjectRelease(gRemovedIter);
    gRemovedIter = 0;
  } // if
  exit(0);
} // SignalHandler()



/*****************************************************************************
RawDeviceRefreshCallback
==============================================================================

This callback is called when a device is plugged in or out.

*****************************************************************************/

void RawDeviceRefreshCallback (void *refCon, io_iterator_t nIterator)
{
  BOOL8 fChanged = FALSE;
  io_service_t nObject;
  while (nObject = IOIteratorNext(nIterator)) {
    IOObjectRelease(nObject);
    fChanged = TRUE;
  } // while
  if (fChanged) {
    // build up the internal list of all connected Box/U & /RU
    if (!EnumConnectedBoxes()) {
      return;
    } // if
  } // if
} // RawDeviceRefreshCallback()


/*****************************************************************************
GetRef
==============================================================================

This function checks if the given WibuBox is still in the internal list
and is so connected to the computer.

*****************************************************************************/

static IOUSBDeviceInterface **GetRef (WKBSERIAL *pwkbser)
{
  IOUSBDeviceInterface **ppDevGlobal = NULL;
  kern_return_t kRet;
  mach_port_t nMasterPort;
  CFMutableDictionaryRef devUsbMatch;
  SInt32 nUsbVendor = WK_VENDOR_ID;
  SInt32 nUsbDeviceSubClass = 0;
  io_iterator_t nIter;
  io_service_t devUsbDevice;

  // initiate the comminication with the IOKit
  kRet = IOMasterPort (MACH_PORT_NULL, &nMasterPort);
  if (kRet || !nMasterPort) {
    goto EXIT_ERROR;
  } // if

  // connect to the USB service
  devUsbMatch = IOServiceMatching (kIOUSBDeviceClassName);
  if (!devUsbMatch) {
    mach_port_deallocate(mach_task_self(), nMasterPort);
    goto EXIT_ERROR;
  } // if

  { // Add our vendor and product IDs to the matching criteria
    CFNumberRef num1;
    CFNumberRef num2;
    num1 = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &nUsbVendor);
    CFDictionarySetValue (devUsbMatch, CFSTR(kUSBVendorName), num1 ); 
    num2 = CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &nUsbDeviceSubClass);
    CFDictionarySetValue (devUsbMatch, CFSTR(kUSBDeviceSubClass), num2 ); 
    CFRelease(num1);
    CFRelease(num2);
  } // block

  // create an iterator over all matching IOService nubs
  kRet = IOServiceGetMatchingServices(nMasterPort, devUsbMatch, &nIter);
  if (kRet) {
    CFRelease(devUsbMatch);
    mach_port_deallocate(mach_task_self(), nMasterPort);
    goto EXIT_ERROR;
  }

  // iterate over all found devices and interfaces
  while ( devUsbDevice = IOIteratorNext(nIter) ) {
    IOCFPlugInInterface **ppIoDev = NULL;
    IOUSBDeviceInterface **ppDev = NULL;
    HRESULT nRes;
    SInt32 nScore;
    UInt16 idVendor = 0;
    UInt16 idProduct = 0;
    BOOL8 fValid = 0;
    WKBSERIAL wkbserCur;
    memset(&wkbserCur, 0x0, sizeof(WKBSERIAL) );
    
    kRet = IOCreatePlugInInterfaceForService (devUsbDevice,
        kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, 
        (IOCFPlugInInterface**)&ppIoDev, &nScore);
    // done with the device object now that I have the plugin
    IOObjectRelease(devUsbDevice);
    if (kRet || !ppIoDev) {
      // ==>> goto next detected device
      continue;
    } // if
            
    // i have the device plugin. I need the device interface
    nRes = (*ppIoDev)->QueryInterface (ppIoDev, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID), (LPVOID)&ppDev);
    // done with this
    (*ppIoDev)->Release(ppIoDev);
    if (nRes || !ppDev) {
      // ==>> goto next detected device
      continue;
    } // if

    // technically should check these kRet values
    kRet = (*ppDev)->GetDeviceVendor (ppDev, &idVendor);
    kRet = (*ppDev)->GetDeviceProduct (ppDev, &idProduct);
    if (idVendor != WK_VENDOR_ID) {
      // We should never get here because the matching criteria we specified above
      // will return just those devices with our vendor and product IDs
      (*ppDev)->Release(ppDev);
      // ==>> goto next detected device
      continue;
    } // if

    { // open USB device
      int i = 0;
      for (i = 0; i < 5; i++) {
        // need to open the interface in order to change its state
        kRet = (*ppDev)->USBDeviceOpen(ppDev);
        if (kRet == kIOReturnExclusiveAccess) {
          sleep(1);
          continue;
        } else {
          // The device cannot be opend if it is already opened!!!
          // ==>> goto next detected device
          break;
        } // if
      } // for
      if (kRet) {
        // The device cannot be opend if it is already opened!!!
        // ==>> This means that the device is already connected and inserted in the internal list.
        (*ppDev)->Release(ppDev);
        // ==>> goto next detected device
        continue;
      } // if
    } // block

    // get the serial number of the box
    fValid = QuerySerialNumber (ppDev, &wkbserCur);
    if (fValid) {
      if (pwkbser->usMask == wkbserCur.usMask && pwkbser->ulSerial == wkbserCur.ulSerial) {
        ppDevGlobal = ppDev;
        // box found ==>> exit loop (device reference is still opened!!!)
        break;
      } // if
    } // if

    // close access to the device
    kRet = (*ppDev)->USBDeviceClose(ppDev);
    // close the access to the driver
    kRet = (*ppDev)->Release(ppDev);
  } // while

  // release the device and intrerface iterator
  if (nIter) {
    IOObjectRelease(nIter);
    nIter = 0;
  } // if

  if (nMasterPort) {
    mach_port_deallocate(mach_task_self(), nMasterPort);
    nMasterPort = 0;
  } // if

  // no error occured ==>> exit with TRUE
  return ppDevGlobal;
EXIT_ERROR:
  // an error occured ==>> exit with FALSE
  return NULL;
} // GetRef()
#endif // #if !LINUX

#if defined(__linux__)

/*****************************************************************************
IsStillAValidBox
==============================================================================

This function checks if the given WibuBox is still in the internal list
and is so connected to the computer.

*****************************************************************************/

static int IsStillAValidBox (IOUSBDeviceInterface **ppDev)
{
  BOXLIST *pboxlist = (BOXLIST*)gpboxlist;

  while (pboxlist != NULL) {
    if (pboxlist->ppDev == ppDev) {
      // the address is still valid ==>> exit with TRUE
      return TRUE; // Both Pointers in both Lists are equal: OK
    } else {
      // check next element
      pboxlist = pboxlist->pNext;
    } // if
  } // while

  // whole list checked and no right address found
  // ==>> box not connected any more
  return FALSE;
} // IsStillAValidBox()

#endif 

#if defined(__APPLE__)

/*****************************************************************************
IsStillAValidBox
==============================================================================

This function checks if the given WibuBox is still in the internal list
and is so connected to the computer.

*****************************************************************************/

static IOUSBDeviceInterface **IsStillAValidBox (WKBSERIAL *pwkbser)
{
  BOXLIST *pboxlist = gpboxlist;
  while (pboxlist != NULL) {
    if (&(pboxlist->wkbsr) == pwkbser) {
      // the address is still valid ==>> give device reference back
      return GetRef(pwkbser);
    } else {
      // check next element
      pboxlist = pboxlist->pNext;
    } // if
  } // while

  // whole list checked and no rigth address found
  // ==>> box not connected any more
  return NULL;
} // IsStillAValidBox()

#endif // #if LINUX

#define GLOBAL_PERMISSIONS   0666
/*****************************************************************************
AccessWkUSBDriver
==============================================================================

This function tries to make an access to the WibuBox/U USB Plug & Play driver
and stores a successful reference in a handle.
If an access to USB driver already exists, the function returns really fast.

Return
  The function returns TRUE if the USB driver is accessed and FALSE if not.

*****************************************************************************/

BOOL8 AccessWkUSBDriver ()
{

#if USE_USB
 
#if defined(__APPLE__)
  BOOL8 fValid = TRUE;
  static BOOL8 fIsInit = FALSE;
  
  // create a named semaphore using a string version of our key to identify it.
  if (gUsbSemaphoreID == NULL) {
    gUsbSemaphoreID = sem_open("WKXX_USB", O_CREAT, GLOBAL_PERMISSIONS, 1);
    if ( gUsbSemaphoreID == SEM_FAILED) {
      return FALSE;
    } // if
    sem_init(gUsbSemaphoreID, TRUE, 0);
  } else {
    // wait for released semaphore
    EnterUsbCS();
  } // if
#endif //
  
#if defined(__linux__)

  EnumConnectedBoxes (FALSE);
  if ( 2 == WkhUsbUseSysFs() ) {
    return FALSE;
  } else {
    // we have DEVFS or SYSFS
    return TRUE;
  }
#else
  // initialize internal device list
  if (!fIsInit) {
    // read all conected boxes out of the IORegistry
    if (fValid) {
      fIsInit = TRUE;
    } // if
  } // if
  return fValid;
#endif //
#endif //USE_USB
} // AccessWkUSBDriver()


/*****************************************************************************
ReleaseWkUSBDriver
==============================================================================

This function frees the access to the USB driver. This function is normally
only called if the driver is terminated.

*****************************************************************************/

BOOL8 ReleaseWkUSBDriver ()
{
#if defined(__APPLE__)
  // close our semaphore
  if (gUsbSemaphoreID != NULL) {
    sem_close(gUsbSemaphoreID);
    gUsbSemaphoreID = NULL;
  } // if
#endif
  // Return allways with TRUE!!!
  return TRUE;
} // ReleaseWkUSBDriver()


/*****************************************************************************
ControlWkUSBDriver
==============================================================================

This function sends a sequence (as it is described in the Adaptation Guide)
to wibukey2.sys which send it to the WibuBox/U. After sending this function
waits for the answer of the WibuBox/U an gives the result back.

Parameters
  idCtrlCode .. contains the IOCTL-Code of the current request.
  pvInput ..... contains the input sequence.
  cbInput ..... contains the length of the input sequence.
  pvOutput .... contains the output sequence.
  cbOutput .... contains the length of the output sequence.

Return
  The function returns TRUE if the USB driver is accessed and FALSE if not.

*****************************************************************************/

BOOL8 ControlWkUSBDriver (UINT idCtrlCode, void *pvInput,
    UINT cbInput, void *pvOutput, UINT cbOutput)
{
  BOOL8 fValid = FALSE;

  // if this flag is set another task is blocked this procedure
  if (gfPnPIsActive) {
    // wait until other task set the flag back
    goto RET_FALSE;
  } // if

  // For MacOS the work of the wibukey2.sys on Windows systems must be simulated!
  switch (idCtrlCode) {
    case IOCTL_WK_EXECUTE: {
#if defined(__APPLE__)
      // set the parameters for the request
      WKDEV_REQUEST *pwkdevreq = (WKDEV_REQUEST*) pvInput;

      UInt8 BMRequestType;
      IOUSBDeviceInterface **ppDev;
      WKBSERIAL *pwkbsr = NULL;
      kern_return_t kRet;

      // see if we can update intenal box list
      EnterUsbCS();
    
      // set the reference of the connected WibuBox
      pwkbsr = (WKBSERIAL*) pwkdevreq->hwkdev;
      ppDev = IsStillAValidBox (pwkbsr);
      if (ppDev == NULL) {
        // box not found any more in the internal list ==>> exit with error
        goto RET_FALSE;
      } // if

      BMRequestType = USBmakebmRequestType(kUSBOut, kUSBVendor, kUSBDevice);
      fValid = BoxDeviceRequest (ppDev, BMRequestType, (UInt8) pwkdevreq->usCmd, 
                 (UInt16) pwkdevreq->usIndex, (UInt16)0, (void*) pwkdevreq->abRequest,
                 (UINT) pwkdevreq->cbRequest);
      if (!fValid) {
        // an error occured ==>> break request and exit
        goto RET_FALSE;
      } // if
      
      // request data from the device
      BMRequestType = USBmakebmRequestType(kUSBIn, kUSBVendor, kUSBDevice);
      fValid = BoxDeviceRequest (ppDev, BMRequestType, (UInt8) pwkdevreq->usCmd,
          (UInt16) pwkdevreq->usIndex, (UInt16)0, (void*)pvOutput, (UINT)cbOutput);    
      // free the reference to the box
      // close access to the device
      kRet = (*ppDev)->USBDeviceClose(ppDev);
      // close the access to the driver
      kRet = (*ppDev)->Release(ppDev);
      // release semaphore
      ExitUsbCS();
      if (!fValid) {
        // an error occured ==>> break request and exit
        if ((pwkdevreq->usCmd == BOXCMD_LIST) && (pwkdevreq->usIndex == 0x03)) {
          fValid = TRUE;
        } else {
          // special error handling because box behavior is not by spec!!!
          goto RET_FALSE;
        }
      } // if

#elif defined(__linux__)

      UInt8 BMRequestType;
      IOUSBDeviceInterface **ppDev;
      // set the parameters for the request
      WKDEV_REQUEST *pwkdevreq = (WKDEV_REQUEST*) pvInput;

      // set the reference of the connected WibuBox
      errno = 0;
      ppDev = (IOUSBDeviceInterface**) pwkdevreq->hwkdev;
      fValid = IsStillAValidBox (ppDev);
      if (!fValid) {
        // box not found any more in the internal list ==>> exit with error
        goto RET_FALSE;//return FALSE;
      } // if

      errno = 0;
      if ((pwkdevreq->usCmd != WIBU_FSEARCH) && (pwkdevreq->usCmd != WIBU_LIST)){
        BMRequestType = WIBU_WRITE;//WIBU_WRITE = 64 = 0x40
        fValid = BoxDeviceRequest (ppDev, BMRequestType, (UInt8) pwkdevreq->usCmd,
                 (UInt16) pwkdevreq->usIndex, (UInt16)0, (void*) pwkdevreq->abRequest,
                 (UINT) pwkdevreq->cbRequest);
        if (!fValid) {
          //75: [EOVERFLOW] Value too large for defined data type
          // an error occured ==>> break request and exit
          goto RET_FALSE;//return FALSE;
        } // if
      
      } // if ((pwkdevreq->usCmd != WIBU_FSEARCH) && (pwkdevreq->usCmd != WIBU_LIST)){
      // request data from the device
      errno = 0;
      if (pwkdevreq->usCmd != WIBU_RESET) {
        BMRequestType = WIBU_READ;// WIBU_READ = 192 = 0xc0
        fValid = BoxDeviceRequest (ppDev, BMRequestType, (UInt8) pwkdevreq->usCmd,
                 (UInt16) pwkdevreq->usIndex, (UInt16)0,(void*) pvOutput, (UINT)cbOutput);
        if (!fValid) {
          // 110: [ETIMEDOUT] Connection timed out
          //  32: [EPIPE] Broken pipe
          //  75: [EOVERFLOW] Value too large for defined data type

          // an error occured ==>> break request and exit
          goto RET_FALSE;//return FALSE;
        } // if
      } // if if (pwkdevreq->usCmd != WIBU_RESET)

#endif // MAC_MACHO || LINUX

      //all is o.k.
      break;
    //exitpoint for all 'return FALSE' of this case
   goto RET_FALSE;
    } // case
    case IOCTL_WK_GET_DEVICE_LIST: {
      WKDEVOBJ_LIST *pwkdevobjl = (WKDEVOBJ_LIST*) pvOutput;

#if USE_USB
      // check all connected boxes if already available.
  #if LINUX
      EnumConnectedBoxes(TRUE);
  #else
      EnumConnectedBoxes();
  #endif

      if (gfForceReading || pwkdevobjl->idList != gidListLastScanReal) {
        // box list is updated ==>> store new ID of the valid list
        pwkdevobjl->idList = gidListLastScanReal;
        if (pwkdevobjl->cDevice < gcBoxList) {
          // check if enough memory is available
          // ==>> if not enough place for all devices in destination
          // ==>> return only header of list with number of devices
          pwkdevobjl->cDevice = gcBoxList;
        } else {
          // transfer current device table into destination buffer
          unsigned int iDev;
          BOXLIST *pboxlist = gpboxlist;
          ASSERT(cbOutput >= FIELDOFFSET(WKDEVOBJ_LIST, awkdevobj) + sizeof(WKDEVOBJ) * gcBoxList);
          for (iDev = 0; iDev < gcBoxList; iDev++, pboxlist = pboxlist->pNext) {
            // set address of driver's device as handle
          #if LINUX
            pwkdevobjl->awkdevobj[iDev].hwkdev = (HWKDEV) pboxlist->ppDev;
          #else
            pwkdevobjl->awkdevobj[iDev].hwkdev = (HWKDEV) &(pboxlist->wkbsr);
          #endif
            pwkdevobjl->awkdevobj[iDev].idDevice = 0;
          } // for
          // list is read
          gfForceReading = FALSE;
        } // if
      } // if

#endif //MAC_MACHO || LINUX
      pwkdevobjl->cDevice = gcBoxList;
      break;
    } // case
    case IOCTL_WK_INVALIDATE_DEVICE: {
      // force rereading of the internal box list
      gfForceReading = TRUE;
      gidListLastScanReal++;
      break;
    } // case
    case IOCTL_WK_GET_VERSION: 
    case IOCTL_WK_GET_CALLVERSION: {
      // give version of this driver back
      BSRMODVERSION bsrver = {DRV_VERSION, DRV_SUBVERSION, DRV_VER_BUILD, DRV_VER_COUNT};
      if (cbOutput >= sizeof(BSRMODVERSION)) {
        memcpy (pvOutput, &bsrver, sizeof(BSRMODVERSION));
      } // if
      break;
    } // case
    default:
      ASSERT(0);
  } // switch

  // all okay ==>> return without error
  return TRUE;
RET_FALSE:
  // error case
  return FALSE;
} // ControlWkUSBDriver()

#endif // USE_USB

// ==========================================
// wibu\wk\dev\runtime\drv\c: end of usbmgr.c
// ==========================================
