/*****************************************************************************
S r c D r v                               Portable Source Code WibuKey Driver
==============================================================================
wkstd.h                                         Public Declarations of WKLIB.H
==============================================================================

This header file contains all declarations which are used by several
driver source modules and normally declared in BSRLIB.H or WKLIB.H. This
header file is used only for the source driver export.

==============================================================================

The public WibuKey source driver (SRCDRV) bases on the WIBU-SYSTEMS
internally used portable driver which is used for WKWIN32.DLL, WKWIN.DLL,
WIBUKEY.SYS, WIBUKEY.VXD and WKDOS.EXE.

The SRCDRV driver variant supports the full WibuKey API functionality except
indirect encryption, Expiration Dates, WkLAN, diagnosis event recording and
some minor features.

The SRCDRV can be used for several DOS extender programs or in special
hardware, system or processor technology environments.

All sources are written in ANSI C, except the using of C++ style single line
comments.

==============================================================================

Copyright  1998-2010 by WIBU-SYSTEMS AG
           Rueppurrer Strasse 52, D-76137 Karlsruhe, Germany/Europe
           Website: http://www.wibu.com, Email: info@wibu.com
           Phone: +49-721-93172-0, Fax: +49-721-93172-22
           All rights reserved.

Version .. 4.31 of 2010-Jun-24
Project .. wibu/wk/dev/drv/src
Author ... Marcellus Buchheit (mabu)
System ... All Posix OS
Tools .... GNU/gcc 3.3.x, gcc 4.0.x

Change Report
98Jun09 2.50 mabu: separated from PROJ.H
10Jun24 4.31 ha:   build update

*****************************************************************************/


#ifndef __WKSTD_H__
#define __WKSTD_H__

#define BSRAPI

#if defined(_DEBUG)
  #define BSRDEBUG 1
#else
  #define BSRDEBUG 0
#endif

#if !defined(LONG)
  #define LONG int
#endif
#define BOOLEAN BYTE
typedef char BOOL8;
#define UINT32 UINT
#define UINT16 USHORT

#define MAKEULONG(l,h) ((ULONG)(USHORT)(l)|(ULONG)(USHORT)(h)<<16)
#define FIELDOFFSET(type, field) ((int)(&((type *)1)->field)-1)
#define DIMOF(X)  (sizeof(X) / sizeof(X[0]))
#if defined(_DEBUG)
  #define ASSERT(x) assert(x)
#else
  #define ASSERT(x)
#endif
#define ASSERT_VALUE(p) ASSERT((p) != NULL)
#define ASSERT_ARRAY(p, n) ASSERT((n) == 0 || (p) != NULL)
#define SET_BIT(d, b)    ((d) |= (1 << (b)))
#define CLEAR_BIT(d, b)  ((d) &= ~(1 << (b)))
#define IS_BIT(d, b)     ((d) & (1 << (b)))
#define LBYTE32(ul) ((BYTE)( (ULONG)(ul) & 0xFF))
#define MBYTE32(ul) ((BYTE)(((ULONG)(ul) >> 8) & 0xFF))
#define NBYTE32(ul) ((BYTE)(((ULONG)(ul) >> 16) & 0xFF))
#define HBYTE32(ul) ((BYTE)(((ULONG)(ul) >> 24) & 0xFF))

// macros to handle little and big endian
#if !defined(BSR_BIG_ENDIAN)
  // Intel x86 processor: (standard) low endian and no alignment limitations
  #define GET_USHORT_VALUE(p)    (*(UINT16*)(p))
  #define GET_UINT32_LE(p)     (*(UINT32*)(p))
  #define SET_USHORT_VALUE(p, v) (*(UINT16*)(p) = (UINT16)(v))
  #define SET_UINT32_LE(p, v)  (*(UINT32*)(p) = (UINT32)(v))
#else
  // RISC processor or Macintosh: critical alignment, assume low endian in data
  #define GET_USHORT_VALUE(p) MAKEUSHORT(((BYTE*)(p))[0], ((BYTE*)(p))[1])
  #define GET_UINT32_LE(p)  MAKEULONG(\
                              MAKEUSHORT(((BYTE*)(p))[0], ((BYTE*)(p))[1]),\
                              MAKEUSHORT(((BYTE*)(p))[2], ((BYTE*)(p))[3]))
  
  #define SET_USHORT_VALUE(p, v) ((((BYTE*)(p))[0]) = LOBYTE(v),\
                                  (((BYTE*)(p))[1]) = HIBYTE(v))
  #define SET_UINT32_LE(p, v)  ((((BYTE*)(p))[0]) = LBYTE32(v),\
                                  (((BYTE*)(p))[1]) = MBYTE32(v),\
                                  (((BYTE*)(p))[2]) = NBYTE32(v),\
                                  (((BYTE*)(p))[3]) = HBYTE32(v))
#endif // #if LITTLE_ENDIAN




// macros to set to little endian at same address
#define SET_USHORT_VALUE_AT(p) {\
  USHORT us;\
  SET_USHORT_VALUE(&us, *(p));\
  *(p) = us;}

#define SET_UINT32_LE_INPLACE(p) {\
  ULONG ul;\
  SET_UINT32_LE(&ul, *(p));\
  *(p) = ul;}

#define MAX_PATH 260


typedef struct {
  USHORT wYear;
  USHORT wMonth;
  USHORT wDayOfWeek;
  USHORT wDay;
  USHORT wHour;
  USHORT wMinute;
  USHORT wSecond;
  USHORT wMilliseconds;
} SYSTEMTIME; // prefix "st"


#define FILE_SHARE_READ    0x00000001
#define FILE_SHARE_WRITE   0x00000002
#define GENERIC_READ       0x80000000
#define GENERIC_WRITE      0x40000000
#define OPEN_EXISTING      3
#define INVALID_HANDLE_VALUE ((HANDLE)-1)

// declaration of standard Win32 API functions
void GetLocalTime(SYSTEMTIME *psystm);
ULONG GetTickCount(void);

// declaration of basic services and resources library
UINT BsrGetDateCount(INT nYear, INT nMonth, INT nDay);

int BSRAPI BsrGetPrivateProfileString(const char* pszSection,
  const char* pszEntry, const char* pszDefault, char* pszDestBuf,
  unsigned int cbDestBuf, const char* pszFileName);

// declaration of standard WibuKey library functions
#define BsrAllocMem(cb) malloc((size_t) cb )
void BsrFreeMem(void *pPointer);

// handle of controlled device
DECLARE_HANDLE(HWKDEV);

enum EWkProfileTopic {
  WKPRFTPC_GENERAL,
  WKPRFTPC_DRIVER_CLIENT,
  WKPRFTPC_DRIVER_KERNEL,
  WKPRFTPC_DRVMGR_SETTING,
  WKPRFTPC_DRVMGR_CONTEXT,
  WKPRFTPC_DRVMGR_INSTALL,
  WKPRFTPC_DRVMGR_UPDATE,
  WKPRFTPC_WKLAN,
  WKPRFTPC_CLIENT_USER,
  WKPRFTPC_CLIENT_WKLAN,
  WKPRFTPC_CLIENT_WKNET,
  WKPRFTPC_SERVER_SETTING,
  WKPRFTPC_SERVER_RECENT,
  WKPRFTPC_SERVER_WKLAN,
  WKPRFTPC_SERVER_WKNET_FILE,
  WKPRFTPC_SERVER_CLUSTER,
  WKPRFTPC_SERVER_HLM_FILES
};

/*****************************************************************************
W K D E V O B J                   WibuKey Inter-Driver Device Object Structure
==============================================================================

This structure manages a single WibuKey device object which is transferred
between different drivers, especially functional and physical device drivers.

Members
-------
hwkdev
    contains the handle of the device. This handle is distribued and used
    by the physical device driver. Other driver may use this handle as unique
    device identification.
idDevice
    specifies the unique device identification which avoids that a reused
    device handle refers the wrong device.

Comments
--------
The <hwkdev> element is always unique to a specific device at a fixed moment.
But when a device is removed and another device is insert, both may use the
same handle. That's why the <idDevice> member is additionally used: It is
increased by 1 for each newly created device and is unique for a specific
device during the complete execution phase of the device driver.

*****************************************************************************/

typedef struct {
  HWKDEV hwkdev;
  UINT32 idDevice;
} WKDEVOBJ; // prefix "wkdevobj"


typedef struct {
  // prefix "wkdevobjl"
  UINT32 cDevice;
  UINT32 idList;
  WKDEVOBJ awkdevobj[1]; // variable length
} WKDEVOBJ_LIST;

/*****************************************************************************
W K D E V _ R E Q U E S T      WibuKey Inter-Driver Command Request Structure
==============================================================================

This structure is used to execute a command a lower-level driver from a
higher-level driver. A typical transfer is done from a functional to a
physical device driver.

Members
-------
hwkdev
    contains the handle of the addressed device. This handle must be returned
    from the accessed driver.
idList
    contains a unique identification of the list which returned the handle
    which is specified in <hwkdev>. The request command can be executed only
    if this identification is not obsolete.
usCmd
    contains the command code of the request operation. This is one of the
    BOXCMD_... constants.
usIndex
    is currently not used and must be set to 0.
cbRequest
    contains the number of bytes in the <abRequest> member.
cbAnswer
    contains the number of bytes in the allocated answer data block.
abRequest
    contains the <usCmd> specific request data block of the command.

*****************************************************************************/

typedef struct {
  HWKDEV hwkdev;
  UINT32 idList;
  UINT16 usCmd;
  UINT16 usIndex;
  UINT16 cbRequest;
  UINT16 cbAnswer;
  BYTE abRequest[1]; // variable length
} WKDEV_REQUEST; // prefix "wkdevreq"


typedef struct {
  BYTE cchName;
  BYTE bReserved;
  USHORT usPortIndex;
  USHORT offsBoxList;
  USHORT ausName[1];
} WKNET_PORT; // prefix "wkntp"

// prototype, normally defined in WKLIB.H
INT WKAPIENTRY _WkbProgramLow(INT nPort, WKBSERIAL const *pwkbsr,
  const BYTE *pabData, UINT cbData);

// ========================================
// Remote datagram request master structure
// ========================================

typedef struct {
  BYTE bMagic;
  BYTE bFlags1;
  UINT16 usLength;
  union {
    UINT16 _cbMaxDGAnswer;
    UINT16 _idAnswerDGError;
  } dgv;
  BYTE bDatagramId;
  BYTE bFlags2;
} WKDGHEAD; // prefix "dgh"

// =====================================
// Overloaded request datagram structure
// =====================================
typedef int HANDLE32;
typedef int INT32;



// constants for <bFlags1>
enum {
  WKDG_ANSWER    = 0x80,
  WKDG_ENCRYPTED = 0x40
};

#define WKDG_MAGIC 0xB2


#define wsprintf sprintf

#endif // __WKSTD_H__

// ===================================
// end of wkstd.h
// ===================================
